/*******************************************************************************
+
+  AGD 1.2
+
+
+  AGDModuleOption.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         AGDModuleOption.h
//
//    author: Carsten Gutwenger
//    date:   september 1998
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_AGD_MODULE_OPTION_H
#define _AGD_AGD_MODULE_OPTION_H


#include <AGD/AGDModule.h>
#include <AGD/PreCond.h>
#include <AGD/PostCond.h>


/*{\Manpage {AGDModuleOption} {M} {Module Options} {opt} }*/

template <class M>
class __exportAGD AGDModuleOption {

/*{\Mdefinition
	The class |\Mname| is the base class for module options, which are used
	for implementing exchangeable modules. An instance of type |\Mname|
	consists of a module |module| of type |M|, a guaranteed precondition
	$|pre|_1,\ldots,|pre|_n$, and a required postcondition $|post|_0,\ldots,|post|_m$,
	where $n$ is the number of input and $m$ is the number of output parameters
	of |module|. If |module| is called with an input that satisfies the
	guaranteed precondition, then the output must satisfy the required postcondition.
	More precise, for the postcondition of |module| must hold
	\[ |post|(i)(|pre|_1,\ldots,|pre|_n) \supseteq |post|_i \mbox{ for all } 0\leq i\leq m\]
	We represent $|pre|_1,\ldots,|pre|_n$ by data type |PreCond| with $n$ parameters,
	and $|post|_0,\ldots,|post|_m$ by data type |PostCond| with $m$ parameters.
}*/

public:

	/*{\Mcreation opt 5.5}*/

	AGDModuleOption() {
		_module = nil;
	}
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	virtual ~AGDModuleOption() {
		delete _module;
	}


	/*{\Moperations}*/

	void init (const PreCond& PRE, const PostCond& POST, const M& module);
	/*{\Mop initializes |\Mvar| with guaranteed precondition |PRE|, required
	postcondition |POST|, and initial module |module|.}*/

	bool set (const M& module);
	/*{\Mop sets the module option to a copy of |module| and returns |true| if |module|
	is applicable, returns |false| otherwise. We use the dynamic copy method [[clone]]
	of |AGDModule| for allocating the copy.}*/

	bool check (const M& module) const;
	/*{\Mop returns |true| if |module| is applicable, |false| otherwise.}*/

	const M &get() const;

	M &get();
	/*{\Mop returns a reference to the module contained in |\Mvar|,}*/

protected:

	M *clone(const M& m) {
		return (M *) m.clone();
	}

	PreCond  _PRE;
	PostCond _POST;
	M *_module;
};


template <class M>
void AGDModuleOption<M>::init (const PreCond& PRE, const PostCond& POST, const M& module) {
	if (_module != nil)
		leda::error_handler (1, "AGDModuleOption: reinitialization not possible!");

	_PRE = PRE; _POST = POST;

	if (!check(module))
		leda::error_handler (1, "AGDModuleOption: initial module not possible!");

	_module = clone(module);
}

template <class M>
bool AGDModuleOption<M>::set (const M& module) {
	if (_module == nil)
		leda::error_handler (1, "AGDModuleOption: module not initialized!");

	if (check(module)) {
		delete _module;
		_module = clone(module);
		return true;
	} else
		return false;
}

template <class M>
bool AGDModuleOption<M>::check (const M& module) const
{
	return module.applicable (_PRE, _POST);
}

template <class M>
const M &AGDModuleOption<M>::get() const
{
#if !defined(LEDA_CHECKING_OFF)
	if (_module == nil)
		leda::error_handler (1, "AGDModuleOption::get: module not initialized!");
#endif
	return *_module;
}

template <class M>
M &AGDModuleOption<M>::get()
{
#if !defined(LEDA_CHECKING_OFF)
	if (_module == nil)
		leda::error_handler (1, "AGDModuleOption::get: module not initialized!");
#endif
	return *_module;
}


/*{\Mtext
\bigskip
\headerline{Declaring Module Options}

The macro {\bf AGD\_DECLARE\_MODULE}(|type|,|name|) should be used to declare
a module option of type |type| with name |name| in a class declaration. This
macro expands to

{\tt protected:\\
\hspace*{1cm}AGDModuleOption<{\it type}> {\it name};\\
public:\\
\hspace*{1cm}bool set\_{\it name} (const {\it type} \&M) $\{$\\
\hspace*{2cm}{\it name}.set(M);\\
\hspace*{1cm}$\}$
}

It also defines a member function {\tt set\_\it name}(), which allows
the user of |M| to set the module option. This function sets the
module stored in |name| to |M| and returns |true| if |M| is applicable for |name|.
Otherwise, |false| is returned.
}*/

#define AGD_DECLARE_MODULE(type,name) \
protected: \
	AGDModuleOption<type> name; \
public: \
	bool set_##name (const type &M) { \
		return name.set(M); \
	} \


/*{\Mexample
We create a sample layout algorithm with an optional augmentation
module. The augmentation module gets a simple, planar graph and
must augment it to a simple, planar, biconnected graph. The
{\tt do\_call()} method shows how to use the module, and the
{\tt main()} function shows, how to set the module option to another
augmenter.

\small\tt
\begin{verbatim}
class SampleLayout : public LayoutModuleImpl<GraphCopyConcept>
{
  AGD_DECLARE_MODULE(AugmentationModule,augmenter);
    ...
public:
  SampleLayout();
  bool do_call(GraphCopyConcept& C);
};

SampleLayout::SampleLayout()
{
    ...
  PreCond  PRE;
  PostCond POST;
  PRE [1] << key::planar << key::simple;
  POST[1] << key::planar << key::biconnected << key::simple;
  augmenter.init (PRE,POST,PlanAug());
    ...
}

SampleLayout::do_call(GraphCopyConcept& C)
{
    ...
  augmenter.get().call(C.get_graph(),L);  // make graph biconnected
    ...
}

int main()
{
  GraphWin GW;
  GW.open();

  GraphWinInterface GWI(GW);
  SampleLayout L;

  L.set_augmenter (OptPlanAug());  // use an optimal augmenter
  L.call(GW.get_graph(),GWI);

    ...
  return 0;
}
\end{verbatim}\normalsize\rm
}*/


#endif
