/*******************************************************************************
+
+  AGD 1.2
+
+
+  CoffmanGrahamRanking.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                       CoffmanGrahamRanking.h
//
//
//    author: Carsten Gutwenger
//    date:   decembre 1998
//-------------------------------------------------------------------


#ifndef _AGD_COFFMAN_GRAHAM_RANKING_H
#define _AGD_COFFMAN_GRAHAM_RANKING_H


#include <AGD/RankAssignment.h>
#include <AGD/AGDModuleOption.h>
#include <AGD/SubgraphModule.h>
#include <LEDA/core/tuple.h>


/*{\Manpage {CoffmanGrahamRanking} {} {Ranking with Predefined Width} {R} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{1.5}{1.5}{5.5}{RankAssignment}
\arhook{7}{2}{8}
\hdclass{0.8}{0}{7.7}{CoffmanGrahamRanking}
\end{hierarchy}
}*/

class __exportAGD CoffmanGrahamRanking : public RankAssignment {

/*{\Mdefinition
The |\Mname| module computes the following ranking of a graph $G=(V,E)$:
First, a feedback set $R\subset E$ is computed, that is, the graph
$G'$ resulting from $G$ by reversing all the edges in $R$ is acyclic.
In this step, self-loops are simply ignored. For this purpose, we use
a subgraph module that computes a maximal acyclic subgraph (that also
implies a feedback set $R$). Then, we compute a ranking
with width at most $W$, in which all edges in $G'$ run downwards.
The width of a ranking is the maximal number of vertices on
the same layer. Notice that we do not consider dummy vertices
at this stage, which are created later when using a ranking
with |SugiyamaLayout|.

The algorithm~\cite{CG72} is actually from the theory of
multiprocessor scheduling. The relation to the rank assignment
problem is shown, e.g., in~\cite{DETT99}. The aim of the
algorithm is to ensure that the height of the ranking (the
number of layers) is kept small. The following performance
guarantee is shown in~\cite{LS77}: Let $h_{min}$ be the
minimum height of a ranking of width $W$. Then the height $h$
of the ranking computed by |\Mname| satisfies
\[ h\leq\left(2-\frac{2}{W}\right) h_{min} \]
For $W=2$ the height of the ranking is optimal, i.e., $h=h_{min}$.

\horizline
\commoninfo{Coffman-Graham Ranking}{Coffman-Graham Ranking}
{}{C. Gutwenger}{Decembre 1998}{1.0}

\begin{preposttab}
precondition & $\{\, |directed| \,\}$ \\
postcondition(|PRE|) & $\emptyset$
\end{preposttab}

\headerline{Exchangeable Modules}

Instances of type |\Mname| provide the following module options:
\begin{options}
\item {\bf subgraph}\\
an |SubgraphModule| option used for computing a maximal acyclic subgraph.

\begin{tabular}{ll}
guaranteed precondition: & $\{\emph{ directed }\}$\\
required postcondition: & $\{\emph{ directed, maximal\_acyclic }\}$\\
initial module: & |LEDAMakeAcyclic|
\end{tabular}
\end{options}

\headerline{Optional Parameters}

Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |int| {\bf width} = 3\\
the maximal number $W$ of vertices on the same layer.
\end{options}\horizline
}*/

public:

	/*{\Mcreation R 5.5}*/

	CoffmanGrahamRanking();
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname|.}*/

	~CoffmanGrahamRanking() { }


	/*{\Moperations}*/

	leda::string name () const {
		return "Coffman-Graham Ranking";
	}

	leda::string impl_author () const {
		return "C. Gutwenger";
	}

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "Decembre 1998";
	}

	AGDModule *clone() const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|RankAssignment|).
	}*/

	/* inherited
	bool check (const leda::graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	void call (const leda::graph& G, leda::node_array<int>& rank);
	/*{\Mop}*/

	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	/*
	bool set_subgraph (const SubgraphModule& M);
	*/
	/*{\Mop}*/

	AGD_DECLARE_MODULE (SubgraphModule, subgraph)

	int width() const {
		return W;
	}
	/*{\Mop}*/

	void width (int w) {
		W = w;
	}
	/*{\Mop}*/

	/*{\Mtext \stdend}*/

private:

	class _int_set {
		int *A, l, p;
	public:
		_int_set() : A(nil), l(0), p(0) { }
		_int_set(int len) : A(nil), l(len), p(len) {
			if (len > 0)
				A = new int[l];
		}
		~_int_set() { delete[] A; }

		void init(int len) {
			delete A;
			if ((l = len) == 0)
				A = nil;
			else
				A = new int[l];
			p = len;
		}

		int length() const {
			return l;
		}

		int operator[](int i) const {
			return A[i];
		}

		void insert(int x) {
			A[--p] = x;
		}

		bool ready() const {
			return (p == 0);
		}
	};

	void insert (leda::node u, leda::list<leda::two_tuple<leda::node,int> > &ready_nodes);
	void insert (leda::node u, leda::list<leda::node> &ready, const leda::node_array<int> &pi);

	int W;
	leda::node_array<_int_set> S;

};

/*{\Mimplementation
In a preprocessing step, we make $G$ acyclic using the module option
|subgraph|. Then, we delete all transitive edges with |ReducedDigraph|.
This takes time $\BigO(n(n+m))$. The actual Coffman--Graham algorithm
(applied to the reduced digraph) takes time $\BigO(n^2)$.
}*/


#endif
