/*******************************************************************************
+
+  AGD 1.2
+
+
+  EdgeInsertionModule.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        EdgeInsertionModule.h
//
//
//    author: Carsten Gutwenger
//    date:   august 1997
//-------------------------------------------------------------------


#ifndef _AGD_EDGE_INSERT_MODULE_H
#define _AGD_EDGE_INSERT_MODULE_H


#include <AGD/AGDModule.h>
#include <AGD/PlaneGraphCopy.h>


/*{\Manpage {EdgeInsertionModule} {} {Edge Insertion Modules} {I} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5.5}{AGDModule}
\arhook{5.5}{2}{6.5}
\hdclass{0}{0}{7}{EdgeInsertionModule}
\end{hierarchy}
}*/

class __exportAGD EdgeInsertionModule : public AGDModule {

/*{\Mdefinition
The class |\Mname| is the base class for edge insertion modules. An
edge insertion module gets as input a planarized graph |PG|, which is
a copy of a graph $G$, and a list of edges
$e_1,\ldots,e_k\in G$. It inserts $e_1,\ldots,e_k$
into |PG| by replacing edge crossings with dummy nodes.

Edge insertion modules are used for computing a planarization with
a planar subgraph. The edges that are not in the planar subgraph
are reinserted by means of an edge insertion module
(see Section~\ref{Planarization with Planar Subgraph}).

\headerline{Input and Output Parameters}

\begin{tabular}{lcc}
input parameters: & planar map & |in_planar_map=1| \\
 & edge list & |in_edge_list=2| \\
output parameter: & planarized graph & |out_planarized_graph=1|
\end{tabular}

The integer constants |in_planar_map|, |in_edge_list| and
|out_planarized_graph| denote the
positions in the input and output parameter lists. The values can be used to
reference the desired parameter when setting the pre- and postcondition.  

\headerline{Implementation of Edge Insertion Modules}

An implementation of an edge insertion module must override the protected
method do\_call(|PG|, |list<edge>|\& |e_insert|),
which gets as input a planarized graph |PG|, and a
list of edges $|e_insert|=e_1,\ldots,e_k$, which must be inserted.

The planarized graph |PG| is a copy of $G$, and
$e_1,\ldots,e_k\in G$. For each edge $e_i$, the implementation
inserts a chain of edges starting at $|copy|(|source|(e_i))$ and ending at
$|copy|(|target|(e_i))$ that leads over dummy nodes. The dummy nodes are
new nodes resulting from splitting an edge, which will represent edge
crossings in a drawing of the graph.

The class |PlaneGraphCopy| stores these chains and provides methods
(split, new\_edge\_chain) for reinserting edges.
}*/

public:

	enum RemoveReinsertType { none, inserted, all, mostCrossed };

	// the input and output parameters of a planarization module
	static const int in_planar_map;
	static const int in_edge_list;
	static const int out_planarized_graph;

	/*{\Minitialization I 4.5}*/

	EdgeInsertionModule () : AGDModule(2,1) { }

	/*{\Mcreate initializes an edge insertion module.}*/

	virtual ~EdgeInsertionModule () { }


	/*{\Moperations 1.5 3.5}*/

	/*{\Mtext
	\headerline{Standard Interface} }*/

	bool check (const PlaneGraphCopy& G, leda::list<leda::edge>& e_insert, AgdKey &p) const;
	/*{\Mop returns |true| if $G$ satisfies the precondition of
	|\Mvar|. Otherwise, |false| is returned and |p| contains a property
	that is not satisfied.}*/

	void call (PlaneGraphCopy& PG, leda::list<leda::edge>& e_insert);
	/*{\Mop calls the edge insertion module for plane graph copy $PG$.}*/

protected:

	/*{\Mtext
	\bigskip
	\headerline{Protected Overridables to Implement Functionality} }*/

	virtual void do_call (PlaneGraphCopy& G, leda::list<leda::edge>& e_insert)
	/*{\Moppure implements the edge insertion module for plane graph copy $PG$,
	and list of edges |e_insert|.}*/
	= 0;

};


inline void EdgeInsertionModule::call (PlaneGraphCopy& G, leda::list<leda::edge>& e_insert)
{
	do_call (G, e_insert);
}

inline bool EdgeInsertionModule::check (const PlaneGraphCopy& G,
	leda::list<leda::edge>& /*e_insert*/, AgdKey &p) const

{
	return check_graph (in_planar_map, G, p);
}


#endif
