/*******************************************************************************
+
+  AGD 1.2
+
+
+  FPPLayout.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         FPPLayout.h
//
//    author: Carsten Gutwenger
//    date:   octobre 1997
//-------------------------------------------------------------------


#ifndef _AGD_FPP_LAYOUT_H
#define _AGD_FPP_LAYOUT_H


#include <AGD/GridLayoutModuleImpl.h>
#include <AGD/PlaneGraphCopyGridConcept.h>


/*{\Manpage {FPPLayout} {} {The de Fraysseix, Pach, Pollack Algorithm}
{L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{9.5}{5.5}
\hdclass{0}{4.5}{5}{AGDModule}
\arhook{5}{5}{6}
\hdclass{1.5}{3}{5}{LayoutModule}
\arhook{6.5}{3.5}{7.5}
\hdclass{2}{1.5}{6}{GridLayoutModule}
\arhook{8}{2}{9}
\hdclass{4.5}{0}{5}{FPPLayout}
\end{hierarchy}
\hfill\agdeps{fpp.epsi}{3.8cm}
}*/

class __exportAGD FPPLayout : public GridLayoutModuleImpl<PlaneGraphCopyGridConcept> {

/*{\Mdefinition
The class |\Mname| represents the layout algorithm by de Fraysseix, Pach,
Pollack~\cite{dpp-hdpgg-90}.

This algorithm draws a planar graph |G| straight-line
without crossings. |G| must not contain self-loops or multiple edges.
The grid layout size is $(2n-4) \times (n-2)$ for a graph with
$n$ nodes ($n \geq 3$).

The algorithm runs in three phases. In the first phase, the graph is
augmented by adding new artificial edges to get a triangulated plane
graph. Then, a so-called shelling order (also called canonical ordering)
for triangulated planar graphs
is computed. In the third phase the vertices are placed incrementally
according to the shelling order.

\horizline
\commoninfo{de Fraysseix et al.}{de Fraysseix, Pach, Pollack}
{H. de Fraysseix, J. Pach, R. Pollack}{C. Gutwenger}{February 1997}{1.0}

\begin{preposttab}
precondition & $\{\, \emph{planar, simple, no self loops} \,\}$ \\
postcondition(|PRE|) & $\{\, \emph{straight-line, no crossings} \,\}$
\end{preposttab}

\horizline
}*/

public:

	/*{\Mcreation}*/

	FPPLayout ();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	virtual ~FPPLayout() { }


	/*{\Moperations 1.5 2.5}*/

	leda::string name () const {
		return "de Fraysseix et al.";
	}

	leda::string long_name () const {
		return "de Fraysseix, Pach, Pollack";
	}

	leda::string author () const {
		return "H. de Fraysseix, J. Pach, R. Pollack";
	}

	leda::string impl_author () const {
		return "C. Gutwenger";
	}

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "february 1997";
	}

	AGDModule *clone () const;


	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base classes (|LayoutModule|, |GridLayoutModule|).}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	virtual void call (const graph& G, face f, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, face f, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	IRect bounding_box ();
	*/
	/*{\Mop}*/

	/* inherited
	void set_mapper (const GridCoordinateMapper& gcm);
	*/
	/*{\Mop}*/
	
	/*{\Mtext \stdend}*/

protected:
	void get_node_grid_size (PlaneGraphCopyGridConcept &, const LayoutInterface &) { }
	void get_node_grid_size (PlaneGraphCopyGridConcept &C, const GridLayout &) {
		C.set_zero_size();
	}

	void do_call (PlaneGraphCopyGridConcept& C);

private:
	void compute_order (const leda::graph &G,
		leda::node_array<int>&  num,
		leda::node_array<leda::edge>& e_wp,
		leda::node_array<leda::edge>& e_wq,
		leda::edge e_12, leda::edge e_2n, leda::edge e_n1);

	void compute_coordinates (PlaneGraphCopyGridConcept &C,
		leda::node_array<int> &num,
		leda::node_array<leda::edge> &e_wp,
		leda::node_array<leda::edge> &e_wq);

};

/*{\Mimplementation
The computation of the layout takes time $\BigO(n)$,
where $n$ is the number of nodes of the input graph.}*/


#endif
