/*******************************************************************************
+
+  AGD 1.2
+
+
+  GraphChecker.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         GraphChecker.h
//
//    author: Carsten Gutwenger
//    date:   novembre 1998
//-------------------------------------------------------------------


#ifndef _AGD_GRAPH_CHECKER_H
#define _AGD_GRAPH_CHECKER_H

#include <AGD/AGD.h>
#include <LEDA/core/list.h>


typedef AgdKey (*check_graph_handler)(const leda::graph &G, KeySet &props,
	KeySet &satisfied, bool check_all);

// avoid a warning with g++
inline istream &operator>> (istream &is, check_graph_handler &) { return is; }


/*{\Manpage {GraphChecker} {} {Graph Checker} {C} }*/

class __exportAGD GraphChecker {

/*{\Mdefinition
The class |\Mname| maintains a list of check-graph handler functions
|Checkers|, which are used in order to test if a graph satsifies a given set of
properties, or to compute the set of properties a given graph satisfies.
The type |check_graph_handler| is a pointer to a check-graph handler
function. Such a function, e.g. f, has to be declared as
\begin{quote}
|AgdKey| f(const |graph|\& G, |KeySet|\& |props|, |KeySet|\&
|satisfied|, |bool| |check_all|);
\end{quote}
It has to test if input graph |G| satisfies the properties in |props|.
Each property that has been checked must be removed from |props|.
If |check_all| is |true|, each satisfied property is added to |satsified|.
If |check_all| is |false|, f returns an unsatisfied property or a |nil| key
(cf. Sect.~\ref{Keys}) if it cannot find an unsatisfied property.

|\Mname| allows to add and remove functions from |Checkers|. New handler
functions are added at the beginning of |Checkers|, i.e., a newly
added handler is called before any other handler. Initially,
|Checkers| contains a default handler function, which can test for the
following properties:

\begin{tabular}{llllll}
|directed|  & |undirected|  & |planar|        & |simple|    & |no_self_loops| & |no_reversals| \\
|connected| & |biconnected| & |triconnected|  & |bipartite| & |four_graph|    & |acyclic| \\
|forest|    & |tree|        & |single_source| & |single_sink| \\
\end{tabular}
}*/

	leda::list<check_graph_handler> checkers;

public:

	/*{\Mcreation}*/

	GraphChecker();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	~GraphChecker() { }

	/*{\Moperations}*/

	bool check(const leda::graph &G, KeySet &props) const;
	/*{\Mop returns |true| iff |G| satisfies all properties in |props|.}*/

	bool check(const leda::graph &G, KeySet &props, AgdKey &not_satisfied) const;
	/*{\Mop returns |true| if |G| satisfies all properties in |props|,
	otherwise |false| is returned and |not_satsified| is a property in
	|props| that is not satisfied by |G|.}*/

	KeySet get_properties(const leda::graph &G, KeySet &props) const;
	/*{\Mop returns the set of all properties in |props| which are
	satisfied by |G|.}*/

	leda::list_item add_checker(check_graph_handler f);
	/*{\Mop appends a new handler function |f| at the beginning of |Checkers|.
	Returns the corresponding list item.}*/

	void rem_checker(leda::list_item it);
	/*{\Mop removes the check-handler function at position |it|
	in |Checkers|. \precond |it| is a list item in |Checkers|.}*/
	
private:

	static AgdKey default_checker(const leda::graph &G, KeySet &props,
		KeySet &satisfied, bool check_all);
};


#endif
