/*******************************************************************************
+
+  AGD 1.2
+
+
+  GridCoordinateMapper.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                   GridCoordinateMapper.h
//
//    author: Carsten Gutwenger
//    date:   september 1998
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_GRID_COORDINATE_MAPPER_H
#define _AGD_GRID_COORDINATE_MAPPER_H


#include <AGD/AGDModule.h>
#include <AGD/IPolyline.h>
#include <AGD/IRect.h>
#include <AGD/DPolyline.h>
#include <AGD/CGGridLayoutInfo.h>


class LayoutInterface;
class CGLayoutInterface;


/*{\Manpage {GridCoordinateMapper} {} {Grid Coordinate Mapper} {M} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0.5}{1.5}{5.5}{AGDModule}
\arhook{6}{2}{7}
\hdclass{0}{0}{7.5}{GridCoordinateMapper}
\end{hierarchy}
}*/

class __exportAGD GridCoordinateMapper : public AGDModule {
/*{\Mdefinition
The class |\Mname| defines a standard interface for grid coordinate
mappers and provides a generic method for their implementation.

Grid coordinate mappers are used to transform a grid layout of a graph $G$ into a
real layout ({\sl call()} function). Furthermore, they compute the initial
grid size of nodes in the grid layout ({\sl get\_node\_grid\_size()} function).

\headerline{Implementation of Grid Coordinate Mappers}

We refer to the input grid layout with $|column|(v)$, $|row|(v)$ for the position of
node $v$, $|ip|_1(e),\ldots,|ip|_{k_e}(e)$ for the polyline of edge $e$,
where $|ip|_i(e) = (|col|_i(e),|row|_i(e))$, with $|grid_width|(v)$, $|grid_height|(v)$
for the grid width and height of node $v$, and with |xmin|, |xmax|, |ymin|,
|ymax| for the bounding box of the grid layout. The width and height of
node $v$ before the transformation are denoted by |width(v)| and |height(v)|.

The output is given by $x(v)$, $y(v)$ for the position of node $v$,
|line(e)|, |source_anchor(e)| and |target_anchor(e)| for the polyline of
edge $e$, and $|width|'(v)$, $|height|'(v)$ for the width and height of node
$v$ after the transformation. This notation will also be used in the
documentation of implementations.

The input grid layout is passed to the grid coordinate mapper as a
grid layout info object. An implementation of a grid coordinate mapper
accesses the grid layout by using access functions provided by |\Mname|.
It always works on the graph for which the final layout is produced
independent of the actual type of the grid layout info object.

A grid coordinate mapper is implemented by overiding virtual methods.
The computation of initial node sizes ({\sl get\_node\_grid\_size()} function)
is implemented by overriding {\sl do\_get\_node\_grid\_size()}. When this method
is called, grid width and height of each node are set to $0$. For setting
grid width or height serve the two functions {\sl grid\_width}$(v,w)$ and
{\sl grid\_height}$(v,h)$. The default of {\sl do\_get\_node\_grid\_size}()
is the empty function.

The call() function does the transformation of grid coordinates into real
coordinates. In most cases, the mapping is a function
$\varphi : \nat^2 \rightarrow \real^2$ with
\[ \varphi(|column|,|row|) = \left( \varphi_x(|column|),\varphi_y(|row|)
\right). \]
In this case, the mapping of x-coordinates is independent of the
y-coordinates and vice versa. |\Mname| provides a basic scheme for
implementing the call() function. This scheme consists of several virtual
methods, which an implementation overrides if the default is not
appropriate for the particular mapping. The mapping function $\varphi$
must be defined by implementing {\sl map\_x}$(|column|)$ and
{\sl map\_y}$(|row|)$.
\begin{enumerate}
\item {\bfseries Initialization}

The {\sl init}$(A)$ method is called. Default is the empty function.

\item {\bfseries Mapping of node coordinates}

The {\sl map\_nodes}$(A)$ method is called. The default implementation computes
the positions of the nodes by calling {\sl map\_node}$(v,A)$ for each node $v$ in $G$.

The default of {\sl map\_node}$(v,A)$ returns the center of the rectangular
region of $v$. More exactly, if $v$ is at grid position $(x,y)$ and
has grid width $w$ and grid height $h$, it returns
\[ \frac{1}{2} \left( \varphi(x,y) + \varphi(x+w,y+h)\right). \]

\item {\bfseries Mapping of edge coordinates}

The {\sl map\_edges}$(A)$ method is called. The default implementation
determines for each edge $e$ the grid points |gsa| and |gta| which are the source (resp.
target) grid anchor point of $e$, and the remaining list of grid bend points |ipl|.
Then, it calls {\sl map\_edge}$(e,|ipl|,|gsa|,|gta|,|dpl|,|s_a|,|t_a|,A)$
in order to compute the polyline |dpl| and the source and target anchors |s_a|
and |t_a| for |e|.

The default of {\sl map\_edge}$(e,|ipl|,|gsa|,|gta|,|dpl|,|s_a|,|t_a|,A)$
computes the edge anchors by transforming
the coordinates of $\varphi(|gsa|)$ and $\varphi(|gta|)$ into $[-1..1]\times[-1,1]$
coordinates which are relativ to the source (resp.\ target) node.
The grid bend points in |ipl| are simply transformed using $\varphi$ and assigned to |dpl|.

\item {\bfseries Mapping of cluster regions}

If |\Mvar| is called for a cluster graph, the {\sl map\_clusters}$(A)$ method is
called. The default implementation uses the {\sl map\_x} and {\sl map\_y}
methods to transform the corner points of the rectangular cluster regions.

\item {\bfseries Cleaning-up}

Finally, the {\sl cleanup}$(A)$ method is called. Default is the empty function.
\end{enumerate}
By default, the sizes of nodes remain unchanged. However, it is usually
necessary to enlarge nodes in certain cases. An implementation must
take care for this (e.g., in {\sl init}()). It is also important to know
that the node grid sizes computed by {\sl get\_node\_grid\_size}() are not necessarily
the input grid sizes of the call() function. They are rather a proposal,
but the input of call() may differ in any way.
}*/

public:

	/*{\Minitialization M 5.5}*/

	GridCoordinateMapper() : AGDModule(1,1) { }
	/*{\Mcreate initializes a grid coordinate mapper.}*/

	virtual ~GridCoordinateMapper() { }


	/*{\Moperations}*/

	/*{\Mtext       
	\headerline{a) Standard Interface}}*/

	void get_node_grid_size(GridLayoutInfo &gli, const LayoutInterface& A);
	/*{\Mop assign for each node in $G$ a grid width and grid height via |gli| .
	\precond The grid layout corresponding to |gli| and the layout interface $A$
	are for the same graph.}*/


	void call (GridLayoutInfo &gli, LayoutInterface& A);
	/*{\Mop calls |\Mvar| for grid layout info |gli|. Assigns the
	transformed layout to |A|. \precond The grid layout corresponding to |gli|
	and the layout interface $A$ are for the same graph.}*/

	void call (CGGridLayoutInfo &gli, CGLayoutInterface& A);
	/*{\Mop calls |\Mvar| for cluster graph grid layout info |gli|. Assigns the
	transformed layout to |A|. \precond The grid layout corresponding to |gli|
	and the layout interface $A$ are for the same graph.}*/

protected:

	/*{\Mtext       
	\bigskip
	\headerline{b) Protected Access Functions}}*/

	const leda::graph& get_graph() const {
		return *_cgp;
	}
	/*{\Mop returns a reference to the input graph $G$.}*/

	int column (leda::node v) const {
		return _glip->x(v);
	}
	/*{\Mop returns the column of $v$ in the grid layout.}*/

	int row (leda::node v) const {
		return _glip->y(v);
	}
	/*{\Mop returns the row of $v$ in the grid layout.}*/

	IPolyline bends (leda::edge e) {
		return _glip->bends(e);
	}
	/*{\Mop returns the list of bend points of $e$ in the grid
	layout.}*/

	int grid_width (leda::node v) const {
		return _glip->grid_width(v);
	}
	/*{\Mop returns the width of $v$ in the grid layout.}*/

	int grid_height (leda::node v) const {
		return _glip->grid_height(v);
	}
	/*{\Mop returns the height of $v$ in the grid layout.}*/

	bool in_rect(const IPoint &ip, leda::node v) const;
	/*{\Mop returns true iff $|ip|\in|rect|(v)$.}*/

	IRect get_gbbox() {
		return _glip->get_gbbox();
	}
	/*{\Mop returns a tight bounding box of the grid layout.}*/

	DPoint map(const IPoint &ip) {
		return DPoint(map_x(ip.xcoord()),map_y(ip.ycoord()));
	}
	/*{\Mop returns the point $(\mbox{map}(x),\mbox{map}(y))$, where
	$|ip|=(x,y)$.}*/

	void get_anchor(DPoint &anchor, leda::node v, const DPoint &abs_anchor,
		const LayoutInterface &A);
	/*{\Mop returns in |anchor| the relativ anchor point corresponding
	to the absolute point |abs_anchor| for edge end point $v$.}*/

	/*{\Mtext
	The following two functions may only be called for implementing
	{\sl do\_get\_node\_grid\_size}().}*/

	void grid_width (leda::node v, int w) const {
		_glip->grid_width(v,w);
	}
	/*{\Mop sets the width of $v$ in the grid layout to $w$.}*/

	void grid_height (leda::node v, int h) const {
		_glip->grid_height(v,h);
	}
	/*{\Mop sets the height of $v$ in the grid layout to $h$.}*/

	/*{\Mtext       
	\bigskip
	\headerline{c) Protected Overridables} }*/

	virtual void do_get_node_grid_size(const LayoutInterface &A);
	/*{\Mop implements get\_node\_grid\_size. Initially, the width
	and height of each node is set to $0$. Default is the empty function.}*/

	virtual void init(LayoutInterface &A);
	/*{\Mop called for initialization.}*/

	virtual void map_nodes(LayoutInterface &A);
	/*{\Mop called for assigning node coordinates.}*/

	virtual void map_edges(LayoutInterface &A);
	/*{\Mop called for assigning edge polylines.}*/

	void map_clusters(CGLayoutInterface &A);
	/*{\Mop called for assigning cluster regions.}*/

	virtual void cleanup(LayoutInterface &A);
	/*{\Mop called for cleaning up.}*/

	virtual DPoint map_node (leda::node v, LayoutInterface &A);
	/*{\Mop returns the position of $v$.}*/

	virtual void map_edge (leda::edge e, IPolyline &ipl, const IPoint &gsa,
	const IPoint &gta, DPolyline& dpl, DPoint& s_a, DPoint& t_a, LayoutInterface &A);
	/*{\Mop returns the polyline for edge $e$ in |dpl| (bend points),
	|s_a| (source anchor point) and |t_a| (target anchor point). Input
	are the grid points |gsa| and |gta|, which are the anchor points at
	the source (resp. target) node, and the bend points |ipl|.}*/

	virtual double map_x (int column) const
	/*{\Moppure returns $\varphi_x(|column|)$.}*/
	= 0;

	virtual double map_y (int row) const
	/*{\Moppure returns $\varphi_y(|row|)$.}*/
	= 0;


private:

	const leda::graph   *_cgp;
	const ClusterGraph *_ccgp;
	GridLayoutInfo     *_glip;
	CGGridLayoutInfo   *_cgglip;
};


inline void GridCoordinateMapper::do_get_node_grid_size(
	const LayoutInterface &) { }
inline void GridCoordinateMapper::init(LayoutInterface &) { }
inline void GridCoordinateMapper::cleanup(LayoutInterface &) { }

#endif
