/*******************************************************************************
+
+  AGD 1.2
+
+
+  GridLayoutModule.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



//-------------------------------------------------------------------
//
//                       GridLayoutModule.h
//
//    author: Carsten Gutwenger
//    date:   january 1999
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_GRID_LAYOUT_H
#define _AGD_GRID_LAYOUT_H


#include <AGD/LayoutModule.h>
#include <AGD/GridCoordinateMapper.h>
#include <AGD/AGDModuleOption.h>
#include <AGD/GridLayout.h>


/*{\Manpage {GridLayoutModule} {} {Grid Layout Algorithms} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5}{AGDModule}
\arhook{5}{3.5}{6}
\hdclass{1.5}{1.5}{5}{LayoutModule}
\arhook{6.5}{2}{7.5}
\hdclass{2}{0}{6}{GridLayoutModule}
\end{hierarchy}
}*/

class __exportAGD GridLayoutModule : public LayoutModule {

/*{\Mdefinition
The class |\Mname| extends the standard interface of layout
algorithms by specialized call functions for grid layout algorithms.

Grid layout algorithms can be implemented by directly deriving from
|\Mname| and overriding all pure virtual methods of the standard interface
of |LayoutModule| and of the extended standard interface of |\Mname|.
More convenient is to derive from |GridLayoutModuleImpl| with a suitable
implementation concept class. This is explained in detail in
Chapter~\ref{Chapter Implementation of Layout Algorithms}, where we
also give a comprehensive example (see Sect.~\ref{Implementation of Grid Layout Algorithms}).
In the latter case, only one method (\texttt{do\_call()}) has to be overridden.

\headerline{Exchangeable Modules}
Each grid layout algorithm has a module option for its grid coordinate
mapper. Deriving classes must also copy the |mapper| option when
implementing the clone method (see Sect.~\ref{Implementation of Grid Layout Algorithms}
for an example).

\begin{options}
\item {\bf mapper}\\
a |GridCoordinateMapper| option used for translating grid coordinates
into real coordinates.

\begin{tabular}{ll}
guaranteed precondition: & $\emptyset$\\
required postcondition: & $\emptyset$\\
initial module: & |MaxAllMapper|
\end{tabular}
\end{options}
}*/

public:

	/*{\Minitialization L 4.5}*/

	GridLayoutModule ();
	/*{\Mcreate initializes a grid layout module. Sets the
	grid coordinate mapper to |MaxAllMapper|,}*/

	GridLayoutModule (const PreCond &P, const PostCond &Q,
		const GridCoordinateMapper &M);
	/*{\Mcreate initializes a grid layout module. Sets the
	grid coordinate mapper to $M$.}*/

	~GridLayoutModule() { }

	/*{\Moperations 1.5 3.5}*/

	/*{\Mtext       
	\headerline{Extended Standard Interface} }*/

	virtual void call_grid (const leda::graph& G, GridLayout &gl) = 0;
	/*{\Moppure calls |\Mvar| for graph $G$. Returns the grid layout in |gl|.}*/


	/*{\Mtext\smallskip
	The following call interfaces are not supported by all layout algorithms.
	The general postcondition (|\Mvar|.postcondition(0)) declares, which
	call interfaces are supported.}*/

	virtual void call_grid (const leda::graph& G, leda::face f, GridLayout &gl) = 0;
	/*{\Moppure calls |\Mvar| for graph $G$ and external face $f$. Returns
	the grid layout in |gl|. \precond $|key::planar_map_callable|\in |post|(0)$,
	$G$ represents a planar map, and $f$ is |nil| or a face in $G$.}*/

	virtual void call_grid (PlaneGraphCopy& PG, leda::face f, GridLayout &gl) = 0;
	/*{\Moppure calls |\Mvar| for planarized graph $PG$ and external face $f$.
	Returns the grid layout |gl|. \precond $|key::PlaneGraphCopy_callable|\in|post|(0)$,
	$f$ is |nil| or a face in $PG$.}*/


	/*{\Mtext\medskip}*/

	IRect bounding_box () const {
		return _grid_bbox;
	}
	/*{\Mop returns the bounding box of the computed grid layout.
	\precond {\sl call\_grid} has been called before.}*/


	AGD_DECLARE_MODULE(GridCoordinateMapper,mapper)

protected:

	IRect _grid_bbox;
};



#endif
