/*******************************************************************************
+
+  AGD 1.2
+
+
+  Hierarchy.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        Hierarchy.h
//
//
//    author: Carsten Gutwenger
//            Michael Krger
//    date:   february 1998
//-------------------------------------------------------------------


#ifndef _AGD_HIERARCHY_H
#define _AGD_HIERARCHY_H


#include <AGD/Layer.h>
#include <AGD/GraphCopy.h>
#include <LEDA/core/array.h>



/*{\Manpage {Hierarchy} {} {Node Hierarchies} {H} }*/

class __exportAGD Hierarchy {

/*{\Mdefinition
An instance |\Mvar| of type |\Mname| is an ordered collection
$L_0,\ldots,L_{n-1}$ of layers. The nodes on each layer are
numbered $0,1,\ldots$

We call the index of the layer to which a node $v$ of |GC|
belongs the rank of $v$. |\Mvar| always represents a proper
hierarchy, i.e, ${\rm rank}|(v)=|{\rm rank}(w)+1$ holds
for every edge $(v,w)$ of |GC|.

Node hierarchies are used by Sugiyama's layout algorithm, which
performs iteratively a two layer crossing minimization between
two neighbored layers, the fixed and the changeable layer.
The crossing minimization proceeds alternatingly
downward and upward, and reorders the nodes on the changeable
layer depending on the positions of the nodes on the fixed
layer.

The class |\Mname| provides a data type
|Direction|, which can be |downward| or |upward|, and maintains
the current direction of the crossing minimization and the
current changeable layer. If $L_i$ is the changeable layer, then
$L_{i-1}$ is the fixed layer if the direction is |downward|
(and $i\geq1$), and $L_{i+1}$ if the direction  is |upward|
(and $i\leq n-1$).
}*/

	friend class Layer;

public:

	/*{\Mtypes}*/ 

	enum Direction { downward, upward };
	/*{\Menum the direction of the crossing minimization.}*/


	/*{\Mcreation}*/

	Hierarchy (GraphCopy& GC, const leda::node_array<int>& R);
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|
	initialized to a proper hierarchy of |GC|. Places node v on layer
	$R[{\rm original}(v)]$. Changes |GC| by reversing and splitting
	edges.}*/

	~Hierarchy();


	/*{\Moperations}*/

	int number_of_layers() const {
		return _layer.size();
	}
	/*{\Mop returns the number of layers of |\Mvar|.}*/

	int length (int i) const {
		return _layer[i]->length();
	}
	/*{\Mop returns the number of nodes on layer $L_i$.}*/

	int rank (leda::node v) const {
		return _rank[v];
	}
	/*{\Mop returns the index of the layer to which $v$ belongs.}*/

	int pos (leda::node v) const {
		return _pos[v];
	}
	/*{\Mop returns the position of $v$ on its layer.}*/

	const Layer& operator[] (int i) const {
		return *(_layer[i]);
	}

	Layer& operator[] (int i) {
		return *(_layer[i]);
	}
	/*{\Marrop returns a reference to layer $L_i$.
	\precond $0\leq i\leq n-1$}*/

	Direction direction() const {
		return _direction;
	}
	/*{\Mop returns the current direction of crossing minimization.}*/

	void direction (Direction dir) {
		_direction = dir;
	}
	/*{\Mop sets the current direction of crossing minimization to |dir|.}*/

	void set_changeable_layer (int i) {
		_changeable = i;
	}
	/*{\Mop sets the current changeable layer to $L_i$.}*/

	bool is_dummy (leda::node v) const {
		return (gc_p->original(v) == nil);
	}
	/*{\Mop returns true iff $v$ is a dummy node. \precond $v$ is
	a node of |GC|.}*/

	bool is_long_edge_dummy (leda::node v) const {
		return (gc_p->original(v) == nil && gc_p->outdeg(v) >= 1);
	}

	int calc_crossings (int i);
	/*{\Mop calculates the number of crossings between layer $L_i$
	and layer $L_{i+1}$. \precond $0\leq i\leq n-1$}*/

	int calc_crossings();
	/*{\Mop calculates the total number of crossings.}*/

	void store_pos (leda::node_map<int>& old_pos);
	/*{\Mop stores the current positions of nodes in |old_pos|.}*/

	void restore_pos (const leda::node_map<int>& new_pos);
	/*{\Mop sets the positions of all nodes $v$ to $|new_pos|[v]$.
	\precond $\{|new_pos|[v]\mid v\in L_i\} = \{0,\ldots,{\rm length}(i)-1\}$
	holds for $0\leq i\leq n-1$.}*/

	void shuffle();
	/*{\Mop permutes the nodes on each layer randomly.}*/

private:
	GraphCopy *gc_p;

	leda::node_map<int> _rank;	// the layer of node v
	leda::node_map<int> _pos;		// the position of node v on its layer
	leda::array<Layer *>  _layer;	// the layers of the hierarchy

	Direction _direction;	// current direction of crossing minimization
	int _changeable;

	// adjacent nodes on upper or lower layer, resp.
	leda::node_array<leda::list<leda::node> > _upper_adj_nodes, _lower_adj_nodes;

	// adjacent edges to upper or lower layer, resp.
	leda::node_array<leda::list<leda::edge> > _upper_adj_edges, _lower_adj_edges;

	// used by calculation of crossings
	leda::node_array<leda::list_item>   high_occ;
	leda::node_array<leda::list<leda::node> > sorted_adj_nodes;
};


#endif
