/*******************************************************************************
+
+  AGD 1.2
+
+
+  Keymapper.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         Keymapper.h
//
//
//    author: Carsten Gutwenger
//    date:   january 1999
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_KEYMAPPER_H
#define _AGD_KEYMAPPER_H

#include <AGD/AgdKey.h>
#include <LEDA/core/h_array.h>
#include <LEDA/core/list.h>


LEDA_BEGIN_NAMESPACE
int Hash (const leda::string& s);
LEDA_END_NAMESPACE


/*{\Manpage {Keymapper} {} {Key Mappers} {M} }*/

class __exportAGD Keymapper {

/*{\Mdefinition
An instance |\Mvar| of type |\Mname| is an injective mapping from strings
to keys. We use |dom(\Mvar)| to denote the set of used strings, and |\Mvar(s)|
to denote the value of |\Mvar| for string s. Each key has a unique index
between $0$ and $n-1$, where $n$ is the number of keys in |\Mvar|.
}*/

private:
	leda::h_array<leda::string,AgdKey> stringToKey;
	leda::list<AgdKey> _keys;


public:

	/*{\Mcreation}*/

	Keymapper ();
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|.
	Sets |dom(\Mvar)| to the empty set.}*/

	~Keymapper ();


	class const_iterator;
	friend class const_iterator;

	
	/*{\Moperations}*/

	int number_of_keys() const {
		return _keys.length();
	}
	/*{\Mop returns the number of defined keys.}*/

	AgdKey ask_key (const leda::string& s) const {
		return stringToKey [s];
	}
	/*{\Mop returns |\Mvar(s)| if $|s|\in |dom(\Mvar)|$, |nil|
	otherwise.}*/

	AgdKey get_key (const leda::string& s);
	/*{\Mop returns |\Mvar(s)|. Creates a new key and adds |s| to
	|dom(\Mvar)| if $|s|\notin |dom(\Mvar)|$.}*/

	leda::string get_string (AgdKey k) const {
		return k.info();
	}
	/*{\Mop returns the string |s| with $|\Mvar(s)|=k$.}*/

	const_iterator begin() const;
	/*{\Mop returns an iterator pointing to first element in |\Mvar|.}*/

	const_iterator end() const;
	/*{\Mop returns an iterator pointing to one-past-last element in |\Mvar|.}*/

	const_iterator rbegin() const;
	/*{\Mop returns an iterator pointing to last element in |\Mvar|.}*/

	const_iterator rend() const;
	/*{\Mop returns an iterator pointing to one-before-first element in |\Mvar|.}*/
};



/*{\Manpage {Keymapper::const_iterator} {} {Keymapper Iterators} {it} }*/

class __exportAGD Keymapper::const_iterator {

/*{\Mdefinition
Instances of type |\Mvar| iterate over the elements of a keymapper. They
behave similar like const-pointers to |AgdKey|. A keymapper $K$
provides functions to retrieve iterators pointing to the first or
last element of $K$. We say an iterator of keymapper |K| is valid iff it
points to an element of |K|. Derefferencing is only allowed for
valid iterators.
}*/

public:

	/*{\Mcreation}*/

	const_iterator() : _mapper(nil), _it(nil) { }
	/*{\Mcreate creates an invalid iterator |\Mvar|.}*/

	const_iterator(const Keymapper &mapper) : _mapper((Keymapper*)&mapper), _it(nil) { }
	/*{\Mcreate creates an iterator |\Mvar| pointing to the first element of |mapper|.}*/

	const_iterator(const Keymapper &mapper, leda::list_item it) :
		_mapper((Keymapper*)&mapper), _it(it) { }

	const_iterator(const const_iterator &iter) : _mapper(iter._mapper), _it(iter._it) { }
	/*{\Mcreate creates an iterator identical to |iter|.}*/


	/*{\Moperations}*/

	bool valid() const { return (_it != nil); }
	/*{\Mop returns |true| iff |\Mvar| is valid, i.e., points to an
	element.}*/

	bool operator==(const const_iterator& iter) const {
		return (_it == iter._it);
	}
	/*{\Mbinop returns |true| iff |\Mvar| and |iter| point to the
	same element.}*/

	bool operator!=(const const_iterator& iter) const {
		return (_it != iter._it);
	}
	/*{\Mbinop returns |true| iff |\Mvar| and |iter| point to
	different elements.}*/

	const_iterator &operator=(const const_iterator &iter) {
		_mapper = iter._mapper;
		_it = iter._it;
		return *this;
	}
	/*{\Mbinop assigns |iter| to |\Mvar|.}*/

	const AgdKey &operator*() { return _mapper->_keys[_it]; }
	/*{\Munop returns the element to which |\Mvar| points.}*/

	const AgdKey *operator->() { return &_mapper->_keys[_it]; }
	/*{\Munop returns the address of the element to which |\Mvar| points.}*/

	const_iterator &operator++() {
		_it = (_it != nil) ? _mapper->_keys.succ(_it) : _mapper->_keys.first();
		return *this;
	}
	/*{\Munop increments and returns |\Mvar|.}*/

	const_iterator operator++(int) {
		const_iterator t = *this;
		_it = (_it != nil) ? _mapper->_keys.succ(_it) : _mapper->_keys.first();
		return t;
	}
	/*{\Munop increments |\Mvar| and returns its previous value.}*/

	const_iterator &operator--() {
		_it = (_it != nil) ? _mapper->_keys.pred(_it) : _mapper->_keys.last();
		return *this;
	}
	/*{\Munop decrements and returns |\Mvar|.}*/

	const_iterator operator--(int) {
		const_iterator t = *this;
		_it = (_it != nil) ? _mapper->_keys.pred(_it) : _mapper->_keys.last();
		return t;
	}
	/*{\Munop decrements |\Mvar| and returns its previous value.}*/

protected:
	Keymapper *_mapper;
	leda::list_item _it;
};


inline Keymapper::const_iterator Keymapper::begin() const
{
	return const_iterator(*this,_keys.first());
}

inline Keymapper::const_iterator Keymapper::end() const
{
	return const_iterator(*this,nil);
}

inline Keymapper::const_iterator Keymapper::rbegin() const
{
	return const_iterator(*this,_keys.last());
}

inline Keymapper::const_iterator Keymapper::rend() const
{
	return const_iterator(*this,nil);
}



#endif
