/*******************************************************************************
+
+  AGD 1.2
+
+
+  LeveledGraph.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



//-------------------------------------------------------------------
//
//                         LeveledGraph.h
//
//    author: Carsten Gutwenger
//    date:   july 1998
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_LEVELED_GRAPH_H
#define _AGD_LEVELED_GRAPH_H


#include <LEDA/graph/graph.h>
#include <LEDA/core/map.h>
#include <AGD/basic.h>




/*{\Manpage {LeveledGraph} {} {Leveled Graphs} {LG} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5}{graph}
\arhook{5}{2}{6}
\hdclass{1.5}{0}{5}{LeveledGraph}
\end{hierarchy}
}*/

class __exportAGD LeveledGraph : public leda::graph {

typedef leda::graph LG;

/*{\Mdefinition
An instance |\Mvar| of type |\Mname| is a graph $G=(V,E)$ with an injective
mapping $|level|:V\rightarrow {\rm Z\!\!Z}$ from the nodes of |\Mvar| into
the integer numbers. We call the set of nodes
$L_i := \{ v \mid |level|(v)=i\}$ the $i$th level of |\Mvar|. The number
of levels of |\Mvar| is
\[ \left\{ \begin{array}{cl}
\max \{i\mid L_i\neq\emptyset\}-\min\{i\mid L_i\neq\emptyset\}+1 & \mbox{ if }V\neq\emptyset\\
0 & \mbox{ otherwise}
\end{array} \right. \]
|\Mname| also maintains the current number of sources and sinks.
}*/

public:

	/*{\Mcreation}*/

	LeveledGraph();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname| initialized
	to an empty graph.}*/

	LeveledGraph(const LeveledGraph &LG);

	virtual ~LeveledGraph() { }

	LeveledGraph& operator=(const LeveledGraph &LG);

	/*{\Moperations}*/

	int number_of_levels() const {
		return (empty()) ? 0 : _max_level - _min_level + 1;
	}
	/*{\Mop returns the number of levels of |\Mvar|.}*/

	int min_level() const {
		return _min_level;
	}
	/*{\Mop returns $\min\{i\mid L_i\neq\emptyset\}$. \precond $V\neq\emptyset$}*/

	int max_level() const {
		return _max_level;
	}
	/*{\Mop returns $\max\{i\mid L_i\neq\emptyset\}$. \precond $V\neq\emptyset$}*/

	int level(leda::node v) const {
		return _level[v];
	}
	/*{\Mop returns the index $i$ of the level $L_i$ to which $v$ belongs.}*/

	void level(leda::node v, int i);
	/*{\Mop changes the level to which $v$ belongs to $L_i$.}*/

	const leda::list<leda::node> &nodes(int i) const {
		return _nodes[i];
	}
	/*{\Mop returns the list of nodes on level $L_i$.}*/

	int nodes(int i, leda::list<leda::node> &L) const {
		L = _nodes[i];
		return L.length();
	}
	/*{\Mop assigns the list of nodes on level $L_i$ to $L$ and returns
	the number of nodes on $L_i$.}*/
		
	int length(int i) const {
		return _nodes[i].length();
	}
	/*{\Mop returns the number of nodes on level $L_i$.}*/

	bool is_isolated (leda::node v) const {
		return (degree(v) == 0);
	}
	/*{\Mop returns true iff $v$ is a isolated node, i.e., $|degree(v)|=0$.}*/

	int number_of_sources() const {
		return _num_sources;
	}
	/*{\Mop returns the current number of sources (nodes with no
	incoming edges) in |\Mvar|.}*/

	int number_of_sinks() const {
		return _num_sinks;
	}
	/*{\Mop returns the current number of sinks (nodes with no
	outgoing edges) in |\Mvar|.}*/

	leda::node new_node (int i);
	/*{\Mop adds a new node on level $L_i$ to |\Mvar| and returns it.}*/

	leda::node new_node() {
		return LG::new_node();
	}
	/*{\Mop adds a new node on level $L_0$ to |\Mvar| and returns it.}*/

	void invert();

protected:

	void post_new_node_handler(leda::node v);
	void pre_del_node_handler(leda::node v);
	void pre_new_edge_handler(leda::node v, leda::node w);
	void post_del_edge_handler(leda::node v, leda::node w);
	void pre_restore_edge_handler(leda::edge e);
	void post_hide_edge_handler(leda::edge e);
	void pre_move_edge_handler(leda::edge e, leda::node v, leda::node w);

private:

	void _init(const LeveledGraph &LG);

	bool _set_level;
	int  _min_level, _max_level;
	int  _num_sources, _num_sinks;

	leda::node_map<int>       _level;
	leda::node_map<leda::list_item> _pos_in_nodes;
	leda::map<int,leda::list<leda::node> > _nodes;
};

/*{\Mimplementation
Most operations take time $\BigO(1)$ or expected time $\BigO(1)$, except for
invert, which takes time $\BigO(n+m+\ell)$, where $n$ is the
number of nodes, $m$ is the number of edges, and $\ell$ is the
number of levels in |\Mvar|.
}*/


#endif
