/*******************************************************************************
+
+  AGD 1.2
+
+
+  LongestPathRanking.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                       LongestPathRanking.h
//
//
//    author: Carsten Gutwenger
//    date:   octobre 1998
//-------------------------------------------------------------------


#ifndef _AGD_L_P_RANKING_H
#define _AGD_L_P_RANKING_H


#include <AGD/RankAssignment.h>
#include <AGD/AGDModuleOption.h>
#include <AGD/SubgraphModule.h>


/*{\Manpage {LongestPathRanking} {} {Ranking to Minimize Height} {R} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{1.5}{1.5}{5.5}{RankAssignment}
\arhook{7}{2}{8}
\hdclass{2}{0}{6.5}{LongestPathRanking}
\end{hierarchy}
}*/

class __exportAGD LongestPathRanking : public RankAssignment {

/*{\Mdefinition
The |\Mname| module computes the following ranking of a graph $G=(V,E)$:
First, a feedback set $R\subset E$ is computed, that is, the graph
$G'$ resulting from $G$ by reversing all the edges in $R$ is acyclic.
In this step, self-loops are simply ignored. For this purpose, we use
a subgraph module that computes a maximal acyclic subgraph (that also
implies a feedback set $R$). Then, we compute a ranking satisfying
\begin{itemize}
\item If $v,w$ are two nodes in $G'$ with $v$ is a source, $w$ is a sink,
and there is a path from $v$ to $w$ in $G'$, then exists also a
path $v=u_1,\ldots,u_p=w$ with $rank(u_{i+1})=1+rank(u_i)$ for
$1\leq i<p$.
\item $\min\{rank(v)\mid v\in V\} = 0$
\end{itemize}
Hence, the so--computed ranking of the acyclic graph $G'$ uses the
minimum number of layers, that is, the height of the layering is
minimal if we constrain each edge in $G'$ to run downward.
Optionally, all nodes with degree 0 can be placed on a separate
layer.

\horizline
\commoninfo{Longest-Path Ranking}{Longest-Path Ranking}
{}{C. Gutwenger}{May 1998}{1.0}

\begin{preposttab}
precondition & $\{\, |directed| \,\}$ \\
postcondition(|PRE|) & $\emptyset$
\end{preposttab}

\headerline{Exchangeable Modules}
Instances of type |\Mname| provide the following module options:
\begin{options}
\item {\bf subgraph}\\
an |SubgraphModule| option used for computing a maximal acyclic subgraph.

\begin{tabular}{ll}
guaranteed precondition: & $\{\emph{ directed }\}$\\
required postcondition: & $\{\emph{ directed, maximal\_acyclic }\}$\\
initial module: & |LEDAMakeAcyclic|
\end{tabular}
\end{options}

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |bool| {\bf separate\_deg0\_layer} = |true|\\
a boolean value determining if the nodes with degree $0$ are placed
on a separate layer, i.e., they get the highest rank.
\end{options}\horizline
}*/

public:

	/*{\Mcreation}*/

	LongestPathRanking(bool sdl = true);
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname|. Sets
	the option |separate_deg0_layer| to |sdl|.}*/

	~LongestPathRanking() { }


	/*{\Moperations}*/

	leda::string name () const {
		return "Longest-Path Ranking";
	}

	leda::string impl_author () const {
		return "C. Gutwenger";
	}

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "May 1998";
	}

	AGDModule *clone() const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|RankAssignment|).
	}*/

	/* inherited
	bool check (const leda::graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	void call (const leda::graph& G, leda::node_array<int>& rank);
	/*{\Mop}*/

	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	/*
	bool set_subgraph (const SubgraphModule& M);
	*/
	/*{\Mop}*/

	AGD_DECLARE_MODULE (SubgraphModule, subgraph)

	bool separate_deg0_layer() const {
		return _sep_deg0;
	}
	/*{\Mop}*/

	void separate_deg0_layer (bool sdl) {
		_sep_deg0 = sdl;
	}
	/*{\Mop}*/

	/*{\Mtext \stdend}*/


private:
	bool _sep_deg0;

	void dfs(leda::node v);
	void get_tmp_rank(leda::node v, leda::node_array<int>& rank);
	void dfs_add(leda::node v, leda::node_array<int>& rank);

	int offset, MAX_N;

	leda::node_array<bool> is_source, finished;
	leda::node_array<leda::list<leda::node> > adjacent;
	leda::node_array<int> ingoing;
};

/*{\Mimplementation
The computation of the ranking takes time $\BigO(\Labs{V}+\Labs{E})$.
}*/


#endif
