/*******************************************************************************
+
+  AGD 1.2
+
+
+  MaxAllMapper.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        MaxAllMapper.h
//
//    author: Carsten Gutwenger
//    date:   september 1998
//-------------------------------------------------------------------


#ifndef _AGD_MAX_ALL_MAPPER_H
#define _AGD_MAX_ALL_MAPPER_H


#include <AGD/GridCoordinateMapper.h>


/*{\Manpage {MaxAllMapper} {} {Mapping with Equal Sized Grid Width} {M} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{0}{1.5}{7.5}{GridCoordinateMapper}
\arhook{7.5}{2}{8.5}
\hdclass{3}{0}{6}{MaxAllMapper}
\end{hierarchy}
}*/

class __exportAGD MaxAllMapper : public GridCoordinateMapper {

/*{\Mdefinition
The MaxAll grid coordinate mapper provides a mapping, which guarantees
that the distance between two nodes is at least |sep|, and that the
slopes of edge segments are the same as in the grid layout. Moreover,
the computation of the mapping takes time linear in the size of $G$, the
total number of bends and the width and height of the grid layout.

The input grid sizes of nodes ({\sl get\_node\_gride\_size}() function) are
\begin{eqnarray*}
|grid_width|(v) & = & \max\left( 0,\left\lceil \frac{|width|(v)+|sep|}
{|width|}-1\right\rceil\right) \\
|grid_height|(v) & = & \max\left( 0,\left\lceil \frac{|height|(v)+|sep|}
{|width|}-1\right\rceil\right)
\end{eqnarray*}
Each row and each column have the same size
\[ w_{min} = \max_{v\in V} \left\{ \frac{|width|(v)+|sep|}{|grid_width|(v)+1},
\frac{|height|(v)+|sep|}{|grid_height|(v)+1}\right\}. \]
The nodes are eventually enlarged, such that a minimal overhang is guaranteed:
\begin{eqnarray*}
|width|'(v) & = & \max(|width|(v),2|min_overhang|+|grid_width|(v)*w_{min}) \\
|height|'(v) & = & \max(|height|(v),2|min_overhang|+|grid_height|(v)*w_{min})
\end{eqnarray*}
|\Mname| realizes the following mapping functions:
\begin{eqnarray*}
\varphi_x(|column|) & = & (|column|-|xmin|) * w_{min} + |origin.x|\\
\varphi_y(|row|) & = & (|row| - |ymin|) * w_{min} + |origin.y|
\end{eqnarray*}
The positions of nodes and the polylines of edges are given by the
default implementations of |GridCoordinateMapper| using the mapping
functions defined above.

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |double| {\bf width} = 40\\
determines the grid width used by {\sl get\_node\_grid\_size}()
in order to determine the grid size of nodes.
\item |double| {\bf min\_overhang} = 2\\
determines the minimal allowed overhang of each node.
\item |double| {\bf sep} = 20\\
determines the minimal distance (separation) between nodes.
\item |DPoint| {\bf origin} = $(0,0)$\\
determines the origin of the coordinate system, that is the point to which
the lower left grid point $(|xmin|,|ymin|)$ is mapped.
\end{options}\horizline
}*/

public:

	/*{\Mcreation}*/

	MaxAllMapper (double width = 40, double min_overhang = 2, double sep = 20,
		DPoint origin = DPoint(0,0)) :
		_w(width), _min_over(min_overhang), _sep(sep), _origin(origin) { }
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|, and sets
	the optional parameters |width|, |min_overhang|, |sep| and |origin|.}*/

	virtual ~MaxAllMapper() { }


	/*{\Moperations}*/

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|GridCoordinateMapper|).
	}*/

	/* inherited
	void get_node_grid_size(const leda::graph& G, GridLayout &gl, const LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void get_node_grid_size(const GraphCopy& GC, GridLayout &gl, const LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call (const leda::graph& G, const GridLayout &gl, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call (const GraphCopy& GC, const GridLayout &gl, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	double width() const {
		return _w;
	}
	/*{\Mop}*/

	void width (double x) {
		_w = x;
	}
	/*{\Mop}*/

	double min_overhang() const {
		return _min_over;
	}
	/*{\Mop}*/

	void min_overhang (double x) {
		_min_over = x;
	}
	/*{\Mop}*/

	double sep() const {
		return _sep;
	}
	/*{\Mop}*/

	void sep(double x) {
		_sep = x;
	}
	/*{\Mop}*/

	DPoint origin() const {
		return _origin;
	}
	/*{\Mop}*/

	void origin (DPoint p) {
		_origin = p;
	}
	/*{\Mop}*/

	/*{\Mtext \stdend}*/


	AGDModule *clone() const {
		return new MaxAllMapper (_w,_min_over,_sep,_origin);
	}


protected:
	void do_get_node_grid_size(const LayoutInterface &A);

	void init (LayoutInterface &A);

	double map_x (int column) const {
		return ax * column + bx;
	}

	double map_y (int row) const {
		return ay * row + by;
	}

	void increase_nodes(double w_min, LayoutInterface &A);

	double _w, _min_over, _sep;
	DPoint _origin;

	double ax, bx, ay, by;
};

/*{\Mimplementation
The get\_node\_size() function takes time $\BigO(n)$ and
the call() function takes time $\BigO(n+m+b)$, where $n$ is
the number of nodes and $m$ is the number of edges of $G$, and $b$ is the
total number of bend points.
}*/


#endif
