/*******************************************************************************
+
+  AGD 1.2
+
+
+  MaxEachMapper.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                       MaxEachMapper.h
//
//    author: Carsten Gutwenger
//    date:   january 1999
//-------------------------------------------------------------------


#ifndef _AGD_MAX_EACH_MAPPER_H
#define _AGD_MAX_EACH_MAPPER_H


#include <AGD/GridCoordinateMapper.h>
#include <LEDA/core/tuple.h>


/*{\Manpage {MaxEachMapper} {} {Mapping with Adjusted Grid Width} {M} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{0}{1.5}{7.5}{GridCoordinateMapper}
\arhook{7.5}{2}{8.5}
\hdclass{3}{0}{6}{MaxEachMapper}
\end{hierarchy}
}*/

class __exportAGD MaxEachMapper : public GridCoordinateMapper {

/*{\Mdefinition
|\Mname| provides a mapping, which tries to reduce the width and height
of the resulting layout, such that the distance between two nodes is
at least |sep_x| in x-direction and |sep_y| in y-direction.

The input grid sizes of nodes ({\sl get\_node\_gride\_size}() function) are
\begin{eqnarray*}
|grid_width|(v) & = & \max\left( 0,\left\lceil \frac{|width|(v)+|sep_x|}
{|width_x|}-1\right\rceil\right) \\
|grid_height|(v) & = & \max\left( 0,\left\lceil \frac{|height|(v)+|sep_y|}
{|width_y|}-1\right\rceil\right)
\end{eqnarray*}

Let |wc(x)| be the width of column |x| and |wr(y)| the width of row |y|.
|\Mname| computes a (not necessarily optimal) solution of the
two optimization problems
\begin{center}
\begin{tabular}{lrclp{5cm}}
\multicolumn{5}{l}{$\min \sum\limits_{x=xmin}^{xmax}wc(x)$\hspace{1.5cm}s.t.}\\[2.5ex]
\hspace*{1cm} & $wc(x)$ & $\geq$ & $2|min_over_x|+|sep_x|$ & for all $|xmin|\leq x\leq |xmax|$\\
& $\sum\limits_{i=xl}^{xr}wc(i)$ & $\geq$ & $|width|(v)+|sep_x|$ & for all $v\in V$, $|xl|=|column|(v)$, $|xr|=|xl|+|grid_width|(v)$
\end{tabular}
\end{center}
and
\begin{center}
\begin{tabular}{lrclp{5cm}}
\multicolumn{5}{l}{$\min \sum\limits_{y=ymin}^{ymax}wr(y)$\hspace{1.5cm}s.t.}\\[2.5ex]
\hspace*{1cm} & $wr(y)$ & $\geq$ & $2|min_over_y|+|sep_y|$ & for all $|ymin|\leq y\leq |ymax|$\\
& $\sum\limits_{i=yb}^{yt}wr(i)$ & $\geq$ & $|height|(v)+|sep_y|$ & for all $v\in V$, $|yb|=|row|(v)$, $|yt|=|yb|+|grid_height|(v)$
\end{tabular}
\end{center}
The width of node $v$ is eventually enlarged, such that it is at least
\[\left\{\begin{array}{cl}
2|min_over_x| & \mbox{if }|xl|=|xr|\\
2|min_over_x|+\sum\limits_{i=xl+1}^{xr-1}|wc|(i)+\frac{1}{2}\left(|wc|(|xl|)+|wc|(|xr|)\right) & \mbox{otherwise}
\end{array}\right.\]
where $|xl|=|column|(v)$ and $|xr|=|xl|+|grid_width|(v)$, and the height such that it is at least
\[\left\{\begin{array}{cl}
2|min_over_y| & \mbox{if }|yb|=|yt|\\
2|min_over_y|+\sum\limits_{i=yb+1}^{yt-1}|wr|(i)+\frac{1}{2}\left(|wr|(|yb|)+|wr|(|yt|)\right) & \mbox{otherwise}
\end{array}\right.\]
where $|yb|=|row|(v)$ and $|yt|=|yb|+|grid_height|(v)$.

The realized mapping is defined as
\begin{eqnarray*}
\varphi_x(i) & = & |origin.x| + \sum\limits_{j=xmin+1}^{i-1}|wc|(j) + \frac{1}{2}\left(|wc|(|xmin|)+|wc|(i)\right) \\
\varphi_y(i) & = & |origin.y| + \sum\limits_{j=ymin+1}^{i-1}|wr|(j) + \frac{1}{2}\left(|wr|(|ymin|)+|wr|(i)\right)
\end{eqnarray*}

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |double| {\bf width\_x} = 40\\
determines the grid width in x-direction used by {\sl get\_node\_grid\_size}()
in order to determine the grid size of nodes.
\item |double| {\bf width\_y} = 40\\
determines the grid width in y-direction used by {\sl get\_node\_grid\_size}().
\item |double| {\bf min\_overhang\_x} = 2\\
determines the minimal allowed overhang of each node in x-direction.
\item |double| {\bf min\_overhang\_y} = 2\\
determines the minimal allowed overhang of each node in y-direction.
\item |double| {\bf sep\_x} = 20\\
determines the minimal distance (separation) between nodes in x-direction.
\item |double| {\bf sep\_y} = 20\\
determines the minimal distance between nodes in y-direction.
\item |DPoint| {\bf origin} = $(0,0)$\\
determines the origin of the coordinate system, that is the point to which
the lower left grid point $(|xmin|,|ymin|)$ is mapped.
\end{options}\horizline
}*/

public:

	/*{\Mcreation}*/

	MaxEachMapper (double width_x = 40, double width_y = 40,
		double min_overhang_x = 2, double min_overhang_y = 2,
		double sep_x = 20, double sep_y = 20, DPoint origin = DPoint(0,0)) :
		_wx(width_x), _wy(width_y), _min_over_x(min_overhang_x), _min_over_y(min_overhang_y), 
		_sep_x(sep_x), _sep_y(sep_y), _origin(origin) { }
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|, and sets
	the optional parameters |width_x|, |width_y|, |min_overhang_x|,
	|min_overhang_y|, |sep_x|, |sep_y| and |origin|.}*/

	virtual ~MaxEachMapper() { }


	/*{\Moperations}*/

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|GridCoordinateMapper|).
	}*/

	/* inherited
	void get_node_grid_size(const leda::graph& G, GridLayout &gl, const LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void get_node_grid_size(const GraphCopy& GC, GridLayout &gl, const LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call (const leda::graph& G, const GridLayout &gl, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call (const GraphCopy& GC, const GridLayout &gl, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	double width_x() const {
		return _wx;
	}
	/*{\Mop}*/

	void width_x (double wx) {
		_wx = wx;
	}
	/*{\Mop}*/

	double width_y() const {
		return _wy;
	}
	/*{\Mop}*/

	void width_y (double wy) {
		_wy = wy;
	}
	/*{\Mop}*/

	double min_overhang_x() const {
		return _min_over_x;
	}
	/*{\Mop}*/

	void min_overhang_x (double x) {
		_min_over_x = x;
	}
	/*{\Mop}*/

	double min_overhang_y() const {
		return _min_over_y;
	}
	/*{\Mop}*/

	void min_overhang_y (double y) {
		_min_over_y = y;
	}
	/*{\Mop}*/

	double sep_x() const {
		return _sep_x;
	}
	/*{\Mop}*/

	void sep_x (double x) {
		_sep_x = x;
	}
	/*{\Mop}*/

	double sep_y() const {
		return _sep_y;
	}
	/*{\Mop}*/

	void sep_y (double y) {
		_sep_y = y;
	}
	/*{\Mop}*/

	DPoint origin() const {
		return _origin;
	}
	/*{\Mop}*/

	void origin (DPoint p) {
		_origin = p;
	}
	/*{\Mop}*/

	/*{\Mtext \stdend}*/


	AGDModule *clone() const {
		return new MaxEachMapper (_wx,_wy,_min_over_x,_min_over_y,
			_sep_x,_sep_y,_origin);
	}

	int get_grid_width(leda::node v) {
		return grid_width(v);
	}

	int get_grid_height(leda::node v) {
		return grid_height(v);
	}

protected:
	void do_get_node_grid_size(const LayoutInterface &A);

	void init (LayoutInterface &A);

	double map_x (int column) const {
		return xpos[column];
	}

	double map_y (int row) const {
		return ypos[row];
	}

	DPoint map_node (leda::node v, LayoutInterface &A);
	void cleanup(LayoutInterface &A);

	static int cmp(const leda::two_tuple<int,double> &a,
                       const leda::two_tuple<int,double> &b);

	void increase(leda::array<double> &w, int left, int right, double d);

	double _wx, _wy, _min_over_x, _min_over_y, _sep_x, _sep_y;
	DPoint _origin;

	leda::array<double> wc, wr, xpos, ypos;
};

/*{\Mimplementation
Let $n$ be the number of nodes and $m$ the number of edges in $G$,
$b$ the total number of bend points, $s$ the number of rows and columns
in the grid layout, and denote with $|gc|(v)$
the number of columns and with $|gr|(v)$ the number of rows covered by
$|rect|(v)$. Then, get\_node\_grid\_size() takes time
$\BigO(n)$ and call() takes time
\[ \BigO\left( s+n+m+b+
\sum_{v\in V} \left( 1+|gc|(v)\log |gc|(v)+|gr|(v)\log |gr|(v)\right) \right). \]
}*/


#endif
