/*******************************************************************************
+
+  AGD 1.2
+
+
+  MedianHeuristic.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                       MedianHeuristic.h
//
//
//    author: Carsten Gutwenger
//    date:   february 1998
//-------------------------------------------------------------------


#ifndef _AGD_MEDIAN_HEURISTIC_H
#define _AGD_MEDIAN_HEURISTIC_H


#include <AGD/TwoLayerCrossMin.h>
#include <LEDA/graph/graph.h>


/*{\Manpage {MedianHeuristic} {} {Median Heuristic} {M} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{0.9}{1.5}{6.1}{TwoLayerCrossMin}
\arhook{7}{2}{8}
\hdclass{2.5}{0}{6}{MedianHeuristic}
\end{hierarchy}
}*/

class __exportAGD MedianHeuristic : public TwoLayerCrossMin {

/*{\Mdefinition
The class |\Mname| implements the two layer crossing minimization heuristic
{\it Median}. It reorders the nodes on the changeable layer according
to the median weight~\cite{EW86}
\[ Median(v) = \left\{\begin{array}{cl}
\frac{1}{2} \left(P(w_{\lfloor\frac{n}{2}\rfloor})+P(w_{\lceil\frac{n}{2}\rceil})\right) & \mbox{ if }n\geq1\\
0.0 & \mbox{ otherwise}
\end{array}\right. \]
where $v$ is a node on the changeable layer, and $w_1,\ldots,w_n$
is the sequence of neighbors of $v$ on the fixed layer ordered
according to the position |P|.

\horizline
\commoninfo{Median}{Median (Eades, Wormald)}{P. Eades, N. Wormald}{C. Gutwenger}
{February 1998}{1.0}

\begin{preposttab}
precondition & $\emptyset$ \\
postcondition(|PRE|) & $\emptyset$
\end{preposttab}

\horizline
}*/

public:

	/*{\Mcreation}*/

	MedianHeuristic() { }
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname|.}*/

	~MedianHeuristic() { }


	/*{\Moperations}*/



	leda::string name () const {
		return "Median";
	}

	leda::string long_name () const {
		return "Median (Eades, Wormald)";
	}

	leda::string author () const {
		return "P. Eades, N. Wormald";
	}

	leda::string impl_author () const {
		return "C. Gutwenger";
	}

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "February 1998";
	}

	AGDModule *clone() const {
		return new MedianHeuristic();
	}

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|TwoLayerCrossMin|).
	}*/

	/* inherited
	bool check (const leda::graph& G, const Layer& L, AgdKey &p) const;
	*/
	/*{\Mop}*/

	void init (const leda::graph& G, const Hierarchy& H) {
		weight.init(G);
	}
	/*{\Mop}*/

	void call (const leda::graph& G, Layer& L);
	/*{\Mop}*/

	void cleanup () {
		weight.init();
	}
	/*{\Mop}*/

	/*{\Mtext \stdend}*/

private:

	leda::node_array<double> weight;
};


/*{\Mimplementation
We use a randomized quicksort for sorting the nodes on
the layer according to the median weight.
The init() function takes time $\BigO(n)$, and the call()
function takes expected time $\BigO(m_L+n_L\log n_L)$,
where $n$ is the number of nodes in the input graph, $n_L$
is the number of nodes on the changeable layer and $m_L$
is the number of edges between the fixed and the changeable layer.
}*/


#endif
