/*******************************************************************************
+
+  AGD 1.2
+
+
+  MixedModelLayout.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                      MixedModelLayout.h
//
//    author: Carsten Gutwenger
//    date:   september 1998
//-------------------------------------------------------------------


#ifndef _AGD_MIXED_MODEL25_LAYOUT_H
#define _AGD_MIXED_MODEL25_LAYOUT_H


#include <AGD/GridLayoutModuleImpl.h>
#include <AGD/PlaneGraphCopyGridConcept.h>
#include <AGD/AugmentationModule.h>
#include <AGD/AGDModuleOption.h>


/*{\Manpage {MixedModelLayout} {} {The Mixed-Model Algorithm} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{9.5}{5.5}
\hdclass{0}{4.5}{5}{AGDModule}
\arhook{5}{5}{6}
\hdclass{1.5}{3}{5}{LayoutModule}
\arhook{6.5}{3.5}{7.5}
\hdclass{2}{1.5}{6}{GridLayoutModule}
\arhook{8}{2}{9}
\hdclass{3.3}{0}{6.2}{MixedModelLayout}
\end{hierarchy}
\hfill\agdeps{mixed_model.epsi}{3.8cm}
}*/

class __exportAGD MixedModelLayout : public GridLayoutModuleImpl<PlaneGraphCopyGridConcept> {

/*{\Mdefinition
The class |\Mname| represents the layout algorithm Mixed-Model
by Gutwenger and Mutzel~\cite{GM98b}, which is based upon ideas
by Kant~\cite{Kan96}. In particularly, Kant's algorithm has
been changed concerning the placement phase and the vertex
boxes, and it has been generalized to work for connected planar graphs.

This algorithm draws a $d$-planar graph |G| on a grid such that every
edge has at most three bends and the minimum angle between two
edges is at least $\frac{2}{d}$ radians. |G| must not contain self-loops
or mulitple edges.
The grid size is at most $(2n-6) \times (\frac{3}{2}n-\frac{7}{2})$, the number of bends is
at most $5n-15$, and every edge has length $\BigO(n)$ if |G| has $n$ nodes.

The algorithm runs in several phases. In the preprocessing phase,
vertices with degree one are temporarily deleted and the graph is
being augmented to a biconnected planar graph using a planar biconnectivity
augmentation module. Then, a shelling order for biconnected plane
graphs~\cite{GM97} is computed. In the next step, boxes around each
vertex are defined in such a way that the incident edges appear
regularly along the box. Finally, the coordinates of the vertex boxes
are computed taking the degree-one vertices into account.

\horizline
\commoninfo{Mixed-Model}{Mixed-Model (Gutwenger/Mutzel)}{C. Gutwenger, P. Mutzel}{C. Gutwenger}
{September 1998}{2.5}

\begin{preposttab}
precondition & $\{\, \emph{planar, simple, no self loops} \,\} $ \\
postcondition(|PRE|) & $\{\, \emph{no crossings} \,\}$
\end{preposttab}

\headerline{Exchangeable Modules}
Instances of type |\Mname| provide the following module options:
\begin{options}
\item {\bf augmenter}\\
an |AugmentationModule| option used for planar biconnected augmentation.

\begin{tabular}{ll}
guaranteed precondition: & $\{\emph{ planar, simple, no self loops, directed }\}$\\
required postcondition: & $\{\emph{ planar, biconnected, simple, no self loops, directed }\}$\\
initial module: & |PlanAug|
\end{tabular}
\end{options}

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |double| {\bf base\_ratio} $= 0.33$\\
a double value determining, how many nodes are drawn on the base line.
If |base_ratio| is $r$ and the outerface is $F$, the number of nodes
on the base line is $\max\left(2,\min\left(\ell, r\Labs{F}\right)\right)$,
where $\ell = \max\{k \mid v_1,\ldots,v_k$ is a path on $F_{out}$ with
$(v_i,v_j)\notin E$ for all $j\neq i+1 \}$.
\end{options}
\horizline
}*/

public:

	/*{\Mcreation L 5}*/

	MixedModelLayout ();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	virtual ~MixedModelLayout() { }


	/*{\Moperations 1.5 2.5}*/

	leda::string name () const {
		return "Mixed-Model";
	}

	leda::string long_name () const {
		return "Mixed-Model (Gutwenger/Mutzel)";
	}

	leda::string author () const {
		return "C. Gutwenger, P. Mutzel";
	}

	leda::string impl_author () const {
		return "C. Gutwenger";
	}

	leda::string impl_version () const {
		return "2.5";
	}

	leda::string impl_date () const {
		return "june 1998";
	}


	AGDModule *clone () const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base classes (|LayoutModule|, |GridLayoutModule|).}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, face f, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, face f, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	IRect bounding_box ();
	*/
	/*{\Mop}*/

	/* inherited
	void set_mapper (const GridCoordinateMapper& gcm);
	*/
	/*{\Mop}*/
	
	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	/*
	bool set_augmenter (const AugmentationModule& M);
	*/
	/*{\Mop}*/

	double base_ratio() const {
		return the_base_ratio;
	}
	/*{\Mop}*/

	void base_ratio (double br) {
		the_base_ratio = br;
	}
	/*{\Mop}*/

	bool prefer_nodes() const {
		return the_prefer_nodes;
	}

	void prefer_nodes (bool pn) {
		the_prefer_nodes = pn;
	}

	/*{\Mtext \stdend}*/

	AGD_DECLARE_MODULE (AugmentationModule, augmenter)


protected:
	void get_node_grid_size (PlaneGraphCopyGridConcept &, const LayoutInterface &) { }
	void get_node_grid_size (PlaneGraphCopyGridConcept &C, const GridLayout &) {
		C.set_zero_size();
	}

	void do_call (PlaneGraphCopyGridConcept &C);


private:

	bool   the_prefer_nodes;
	double the_base_ratio;

};


/*{\Mimplementation
The computation of the layout takes time $\BigO(n)$,
where $n$ is the number of nodes of the input graph.}*/


#endif
