/*******************************************************************************
+
+  AGD 1.2
+
+
+  OptCompCoord.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        OptCompCoord.cpp
//
//
//    author: Carsten Gutwenger
//    date:   june 1999
//-------------------------------------------------------------------


#ifndef _AGD_O_COMP_COORD_H
#define _AGD_O_COMP_COORD_H

#include <AGD/HierarchyLayoutModule.h>
#include <AGD/DPoint.h>

class LayoutInterface;
class GraphCopy;
class Layer;



/*{\Manpage {OptCompCoord} {} {Optimal Layout of Hierarchies} {L} }*/

/*{\Mbaseclasses
\abacus
\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5}{AGDModule}
\arhook{5}{3.5}{6}
\hdclass{0}{1.5}{7.5}{HierarchyLayoutModule}
\arhook{7.5}{2}{8.5}
\hdclass{3}{0}{6}{OptCompCoord}
\end{hierarchy}
}*/

class __exportAGDopt OptCompCoord : public HierarchyLayoutModule {

/*{\Mdefinition
The class |\Mname| represents a layout algorithm for hierarchies
which assigns x-coordinates to nodes according to the optimal
solution of an optimization problem. The y-coordinates are
computed in a straight-forward way.

More precisely, the optimization problem is defined as follows
\begin{center}
\begin{tabular}{lrclp{4.5cm}}
\multicolumn{5}{l}{$\min \sum\limits_{(u,v)\in E}\Omega(u,v)\,|x(u)-x(v)|$\hspace{1.5cm}s.t.}\\[2.5ex]
\hspace*{0.5cm} & $x(v)-x(u)$ & $\geq$ & $\frac{1}{2}(|width|(v)+|width|(u)) + |node_distance|$ & where $v$ is successor of $u$ on the same layer\\
& $x(v)$ & $\geq$ & $|origin.y|$ & for all $v\in V$
\end{tabular}
\end{center}
where the hierarchy represents the graph $G=(V,E)$. The aim of the objective
function is to make edges run as vertically as possible. For each edge $e$,
$\Omega(e)$ defines the weight of $e$ in the objective function.
We distinguish three cases:
\[ \Omega(e) = \left\{
\begin{array}{cl}
|weight0| & \mbox{no endpoint of }e\mbox{ is a dummy vertex}\\
|weight1| & \mbox{exactly one endpoint of }e\mbox{ is a dummy vertex}\\
|weight2| & \mbox{both endpoints of }e\mbox{ are dummy vertices}
\end{array} \right.\]
The values of |weight0|, |weight1| and |weight2| are optional parameters.
The default settings are 1, 2 and 8, respectively, what gives edges
with to adjacent dummy vertices the heaviest weight because these edges
represent interior segments of long edges in the actual graph.

The y-coordinates of the layers $L_1,\ldots,L_n$ are defined recursively:
\begin{eqnarray*}
y(L_n) & = & |origin.y|\\
y(L_i) & = & y(L_{i+1}) + \frac{1}{2}(h_i+h_{i+1}) + |layer_distance|\quad\mbox{for }i<n
\end{eqnarray*}
where $h_i$ denotes $\max\{|height|(v)\mid v\in L_i\}$.

\horizline
\commoninfo{OptCompCoord}{Optimal Layout of Hierarchies}{}{C. Gutwenger}
{June 1999}{1.0}

\begin{preposttab}
precondition & $\emptyset$ \\
postcondition(|PRE|) & $\emptyset$
\end{preposttab}

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |double| {\bf node\_distance} $ = 3.0$\\
defines the minimal horizontal distance between two nodes on the
same layer.
\item |double| {\bf layer\_distance} $ = 3.0$\\
defines the minimal vertical distance between two nodes on
neighbored layers.
\item |DPoint| {\bf origin} $ = (0.0,0.0)$\\
defines the y-coordinate of the lowest layer and the smallest
possible x-coordinate of a vertex.
\item |double| {\bf weight0} $ = 1.0$\\
defines the weight of edges which are not adjacent to a
dummy vertex.
\item |double| {\bf weight1} $ = 2.0$\\
defines the weight of edges which are adjacent to exactly one
dummy vertex.
\item |double| {\bf weight2} $ = 8.0$\\
defines the weight of edges which are adjacent to two
dummy vertices.
\end{options}\horizline
}*/

public:

	/*{\Mcreation L 4.5}*/

	OptCompCoord() : nd(3), ld(3), w0(1.0), w1(2.0), w2(8.0), orig(DPoint()) { }
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	~OptCompCoord() { }


	/*{\Moperations 1.2 3}*/

	leda::string name() const {
		return "OptCompCoord";
	}

	leda::string long_name() const {
		return "Optimal Layout of Hierarchies";
	}

	leda::string impl_author() const {
		return "C. Gutwenger";
	}

	leda::string impl_date() const {
		return "June 1999";
	}

	leda::string impl_version() const {
		return "1.0";
	}


	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|HierarchyLayoutModule|).
	}*/

	/* inherited
	bool check (const GraphCopy& GC, const Hierarchy& H, AgdKey &p) const;
	*/
	/*{\Mop}*/

	void call(const GraphCopy& GC,const Hierarchy& H,LayoutInterface& A);
	/*{\Mop}*/

	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	double node_distance() const {
		return nd;
	}
	/*{\Mop}*/

	void node_distance(double x) {
		nd = x;
	}
	/*{\Mop}*/

	double layer_distance() const {
		return ld;
	}
	/*{\Mop}*/

	void layer_distance(double x) {
		ld = x;
	}
	/*{\Mop}*/

	DPoint origin() const {
		return orig;
	}
	/*{\Mop}*/

	void origin(DPoint p) {
		orig = p;
	}
	/*{\Mop}*/

	double weight0() const {
		return w0;
	}
	/*{\Mop}*/

	void weight0(double x) {
		w0 = x;
	}
	/*{\Mop}*/

	double weight1() const {
		return w1;
	}
	/*{\Mop}*/

	void weight1(double x) {
		w1 = x;
	}
	/*{\Mop}*/

	double weight2() const {
		return w2;
	}
	/*{\Mop}*/

	void weight2(double x) {
		w2 = x;
	}
	/*{\Mop}*/


	/*{\Mtext \stdend}*/

	AGDModule *clone() const;

private:
	double nd, ld, w0, w1, w2;
	DPoint orig;

	void set_y(const Layer &L, double y, LayoutInterface &li);
	double max_height(const Layer &L, const LayoutInterface &li);
};


/*{\Mimplementation
The optimization problem is refomulated as a linear program, which is
solved with an LP-solver.
}*/


#endif
