/*******************************************************************************
+
+  AGD 1.2
+
+
+  OptPlanAug.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         OptPlanAug.h
//
//    author: Sergej Fialko
//    date:   octobre 1997
//-------------------------------------------------------------------


#ifndef _AGD_PLAN_AUG_OPT_H
#define _AGD_PLAN_AUG_OPT_H

#include <AGD/PlanAug.h>


/*{\Manpage {OptPlanAug} {} {Optimal Planar Biconnected} {A} }*/

/*{\Mbaseclasses
\abacus
\begin{hierarchy}{8}{5.5}
\hdclass{0}{4.5}{5.5}{AGDModule}
\arhook{5.5}{5}{6.5}
\hdclass{0.3}{3}{6.7}{AugmentationModule}
\arhook{7}{3.5}{8}
\hdclass{3.5}{1.5}{5}{PlanAug}
\arhook{8.5}{2}{9.5}
\hdclass{5}{0}{5}{OptPlanAug}
\end{hierarchy}
}*/

class __exportAGDopt OptPlanAug : public PlanAug {

/*{\Mdefinition
The |\Mname| module augments a planar graph to a planar biconnected
graph by adding the minimum number of edges~\cite{Mut95,JM95,Fialko97}. 
Since this problem
is an NP-hard optimization problem, the algorithm may take exponential
time. But on average, instances with up to 300 vertices can be solved
to provable optimality within a few seconds. 

The algorithm is based on branch-and-cut using ABACUS~\cite{JT97}.
Branch-and-cut algorithms combine integer programming techniques
like polyhedral combinatorics with a branch-and-bound scheme.
The advantage of a branch-and-cut algorithm lies in improved
bounding. This can be achieved by investigating the mathematical
structure of the underlying polytope. For the planar augmentation problem
the mathematical investigation has been done in the
articles by Mutzel~\cite{Mut95}, J\"unger and Mutzel~\cite{JM95} and Fialko~\cite{Fialko97}.

\horizline
\commoninfo{Optimal Planar Biconnected}{Optimal Planar Biconnected}{P. Mutzel}{S. Fialko}
{October 1997}{}

\begin{preposttab}
precondition & $\{\, \emph{planar} \,\}$ \\
postcondition(|PRE|) & $\{\, \emph{planar, biconnected} \,\}
\cup \Big( \{\, \emph{simple, no self loops} \,\} \cap |PRE| \Big)$
\end{preposttab}

\headerline{Optional Parameters}
Instances of |\Mname| inherit two parameters from |PlanAug|, which
is used by |\Mname| in order to find an optimal solution without
branch-and-cut in advance. The following parameters vary this process:
\begin{options}
\item |bool| {\bf size\_strategy} = |false|\\
Setting of |size_strategy| to |false| gives statistical
better results, but differences for our test graphs were
minimal (less than 1\%).
\item |bool| {\bf use\_blabels} = |true|\\
Setting |use_blables| to |true| has better approximation
ratio (but not necessarily a better solution) and is a
little slower.
\end{options}\horizline
}*/

private:
  bool heurist_optim_;// = true, if approx. solution was optimal 
  bool only_bac; //for test purposes, call only branch and cut routine without heuristics


public:

	/*{\Mcreation}*/

  OptPlanAug ();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

  ~OptPlanAug () { }
  
	/*{\Moperations}*/


  leda::string name () const {
     return "Optimal Planar Biconnected";
  }

	leda::string author () const {
		return "P. Mutzel";
	}

  inline bool heurist_optim () { return heurist_optim_; }

  	AGDModule *clone() const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|AugmentationModule|).
	}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	bool call (graph& G);
	*/
	/*{\Mop}*/

	/* inherited
	bool call (graph& G, list<edge>& L);
	*/
	/*{\Mop}*/
	
	/* inherited
	int num_added_edges() const;
	*/
	/*{\Mop}*/

	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	/*
	bool size_strategy () const;
	*/
	/*{\Mop}*/

	/*
	void size_strategy (bool b);
	*/
	/*{\Mop}*/

	/*
	bool use_blabels () const;
	*/
	/*{\Mop}*/

	/*
	void use_blabels (bool b);
	*/
	/*{\Mop}*/

	/*{\Mtext \stdend}*/

  void set_bac_only () { only_bac = true; }
	
protected:
  void do_call (leda::graph& G, leda::list<leda::edge>& L);
};


#endif
