/*******************************************************************************
+
+  AGD 1.2
+
+
+  Ovid.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



//#line 75 "Lweb/Ovid.lw"

#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_OVID_H
#define _AGD_OVID_H

#include <AGD/GridLayoutModuleImpl.h>
#include <AGD/OrthoRepGridConcept.h>
#include <AGD/AGDModuleOption.h>

#include <AGD/CompactionModule.h>
#include <AGD/OrthogonalRepresentation.h>
//#include <AGD/Label.h>

/*{\Manpage {Ovid} {} {The Orthogonal Visibility Drawing Base Class} {O} }*/

/*{\Mbaseclasses
\begin{hierarchy}{9.5}{5.5}
\hdclass{0}{4.5}{5}{AGDModule} \arhook{5}{5}{6}
\hdclass{1.5}{3}{5}{LayoutModule} \arhook{6.5}{3.5}{7.5}
\hdclass{2}{1.5}{6}{GridLayoutModule} \arhook{8}{2}{9}
\hdclass{6.5}{0}{3.5}{Ovid}
\end{hierarchy}
}*/

class __exportAGD Ovid : public GridLayoutModuleImpl<OrthoRepGridConcept> {

/*{\Mdefinition
The class \Mname is an abstract interface class for implementations of algorithms within the topology--shape--metrics approach. These algorithms base on minimum cost flow computations in a special network $N$. The baseclass provides the construction of this network and can compute a minimum cost flow in $N$.

\horizline

\commoninfo{Ovid}{Ovid}{}{Gunnar W.~Klau}{November 1999}{1.2}

 \headerline{Pre- and Postcondition}
  
  \begin{tabular}{lcl}
  precondition & $=$ & \{\,$\emph{planar}$\,\} \\
  postcondition(|PRE|) & $=$ & $\{\,$\emph{no crossings}$\,\}$
  \end{tabular}\\

  \headerline{Exchangeable Modules}
  Instances of type |\Mname| provide the following module options:
  \begin{options}
  \item {\bf compactor}\\
  a |CompactionModule| option used for the compaction phase.
  
  \begin{tabular}{ll}
  guaranteed precondition: & \emph{no crossings, orthogonal}\\
  required postcondition: &  \emph{no crossings, orthogonal}\\
  initial module: & |FlowCompaction|
  \end{tabular}
  \end{options}

  \headerline{Optional Parameters}
  Instances of |\Mname| provide the following optional parameters:
  \begin{options}
  \item |bool| \textbf{traditional} = |false|\\
  When set to |true|, we construct the underlying network as described in \cite{t-eggmn-87}. Otherwise---and by default---we additionally maximize the number of $180^\circ$ angles as a second optimization goal.

  \item |bool| \textbf{compact\_further} = |true|\\
  If set to false, only an initial layout is computed. Setting the value to true (the default) causes the compaction module [[compactor]] to continue its work on the layout.

  \item |int| \textbf{output\_level} $= 0$\\
  The |output_level| determines the amount of output to |stdout| that
  is produced by the algorithm. The default value, zero, causes |\Mname| to
  run in silent mode, values $\geq 10$ cause the maximal amount of output.
  
  \item |bool| {\bf verify} = |true|\\
  If this flag is set, the checking routines are activated. Especially the orthogonal
  representation is verified, 
  certificating the correctness of the computation.

  %\horizline

%\item |int| {$\mathbf k$} $= 2$\\
%Determines the maximal set of possible angles, i.e. every angle $\alpha$ that 
%occurs in the representation holds $\alpha = \frac{i}{2^k}\cdot 360^\circ$ with $i \in \{-2^k, \ldots, 0, \ldots, 2^k\}$.
%\item |bool| {\bf allow\_bends} $ = true$\\
%Makes it possible to place bends on edges. 
%\item |bool| {\bf allow\_nonpos\_deg} $= true$\\
%For graphs $G$ with maximal vertex degree $\Delta(G) > 4$ this variable
%has to be set. It is possible to set it for 4--planar graphs, too. Then
%the number of bends will possibly decrease at the cost of losing the property
%that there is at most one edge per side.
%\item |bool| {\bf forbid\_small\_zero\_deg} $= \ldots$\\
%\marginpar{\fbox{\tiny NOT IMPLEMENTED.}} Causes zero degrees only to appear at high degree nodes.
%\item |bool| {\bf forbid\_cross\_zero\_deg} $= true$\\
%\marginpar{\fbox{\tiny NOT IMPLEMENTED.}} Forbids zero degrees at dummy vertices 
%standing for crossings in the drawing. Should always be true.
%\item |bool| {\bf allow\_neg\_deg} $= true$\\
%Extends the set of possible angle values by $\{-90^\circ, -180^\circ, -270^\circ, -360^\circ\}$.
%\item |bool| {\bf restrict\_to\_L\_shape} $= true$\\
%Only one $-90^\circ$ angle can appear at each vertex. Thus their form is
%restricted to L--shape.
\item |bool| {\bf containsCages} $= false$\\
Some algorithms replace high degree vertices by faces---internally called \emph{cages}---in order to proceed with a representative graph $G'$ with maximal degree
$\Delta(G') \leq 4$. In this case, the cages have to get rectangular shape which can be ensured by modifying the network.
\end{options} \horizline }*/
 
public:

  /*{\Mcreation}*/

  Ovid();
  /*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/
  
  virtual ~Ovid() {}
  
  /*{\Moperations 1.5 2.5}*/
  
  leda::string name() const { return leda::string("Orthogonal visibility representation"); }
  
  leda::string long_name() const { 
    return leda::string("Orthogonal visibility representation (fixed topology)"); 
  }
  
  leda::string author() const { return leda::string(""); }
  
  virtual leda::string impl_author() const { return leda::string("G.~W.~Klau"); }
  
  virtual leda::string impl_version() const { return leda::string("1.2"); }
  
  virtual leda::string impl_date() const { 
    return leda::string("November 1999"); 
  }

  void forbid_concave_bends(const leda::list<leda::edge>& L) {
     _noConcaveBends = L;
  }
  /*{\Mop The list $L$ contains those edges that must not have concave bends.}*/
  
  /*{\Moptions warnings=no}*/

  void verify(bool v) { _verify = v; } 
  /*{\Mop sets the verify status to |v|.}*/
  
  bool verify() const { return _verify; } 
  /*{\Mop returns the verify status.}*/

  void debug(bool d) { debug_ = d; }
  bool debug() const { return debug_; }

  void traditional(bool t) { _traditional = t; }
  /*{\Mop toggles the traditional mode. Setting it to [[false]] leads to drawings, in which 180$^\circ$ angles are preferred.}*/

  bool traditional() const { return _traditional; }
  /*{\Mop returns the traditional mode.}*/

  void progressive(bool p) { _progressive = p; }
  bool progressive() const { return _progressive; }
  // does not have any effect with the actual flow algorithm (this one returns in some sense a "small" flow between flows of equal cost. 

  void compact_layout(bool c) { _compactLayout = c; }
  /*{\Mop changes the compaction behaviour. If $c$ = [[false]], only an initial layout $L$ is computed, $c$ = [[true]] tells the compaction module to further proceed on $L$.}*/

  bool compact_layout() const { return _compactLayout; }
  /*{\Mop returns the compaction behaviour.}*/

  void output_level(int o) { _outputLevel = o; } 
  /*{\Mop sets the output level to |o|. 
    Reasonable values are $o \in [0 \ldots 10]$. 
    $o = 0$ causes the algorithm to run in silent mode, 
    a value $o \geq 10$ will create the greatest
    possible amount of output. }*/
  
  int output_level() const { return _outputLevel; } 
  /*{\Mop returns the output level}*/

  float compaction_time() const { return t_compact_; }
  /*{\Mop returns the time used for the compaction phase}*/
  
  bool variant() const { return _variant; } 
    
  void variant(bool d) { _variant = d; } 

  int number_of_internal_nodes() const { return n_internal_; }

  int number_of_internal_edges() const { return m_internal_; }

  int internal_edge_length() const { return l_internal_; }

  CompactionModule &get_compactor() { return compactor.get(); }

/*{\Mtext       
  \headerline{Standard Interface (Inherited Methods)}
  The detailed description of these methods can be found in the
  manual entries of the base classes (|LayoutModule|, |GridLayoutModule|).
  \stdstart{1.5cm}}*/

/* inherited
   bool check(const graph& G, string& message);
*/
/*{\Mop}*/

   void call (const leda::graph& G, LayoutInterface& A) { 
   GridLayoutModuleImpl<OrthoRepGridConcept>::call(G, A);
  }
 /*{\Mop}*/

   virtual void call (const leda::graph& G, leda::face f, LayoutInterface& A) {
   GridLayoutModuleImpl<OrthoRepGridConcept>::call(G,f, A);
}
/*{\Mop}*/

   virtual void call (PlaneGraphCopy& PG, leda::face f, LayoutInterface& A) {
GridLayoutModuleImpl<OrthoRepGridConcept>::call(PG,f,A);}

   virtual void call (OrthogonalRepresentation& H, leda::face f, LayoutInterface& A) {
GridLayoutModuleImpl<OrthoRepGridConcept>::call(H, f, A);}
/*{\Mop}*/

   void call_grid (const leda::graph& G, GridLayout &gl){
GridLayoutModuleImpl<OrthoRepGridConcept>::call_grid(G, gl);}
/*{\Mop}*/

   void call_grid (const leda::graph& G, leda::face f, GridLayout &gl){
GridLayoutModuleImpl<OrthoRepGridConcept>::call_grid(G, f, gl);
}
/*{\Mop}*/

   	void call_grid (PlaneGraphCopy& PG, leda::face f, GridLayout &gl) {
		GridLayoutModuleImpl<OrthoRepGridConcept>::call_grid(PG,f,gl);}

   void call_grid (OrthogonalRepresentation& H, leda::face f, GridLayout &gl){
GridLayoutModuleImpl<OrthoRepGridConcept>::call_grid(H, f, gl);
}

/*{\Mop}*/

/* inherited
   IRect bounding_box ();
*/
/*{\Mop}*/

/* inherited
   void set_mapper (const GridCoordinateMapper& gcm);
*/
/*{\Mop}*/

/*{\Mtext \stdend}*/

/*{\Mtext       
  \headerline{Extended Interface}
  All descendants of |Ovid| can also be called with an additional parameter:
  The array |max_b|, storing for each edge $e$ the maximal number of bends 
  allowed on $e$. The returned value of these functions is of type |bool|, it
  is true if the call was successful, and false if no feasible flow with the
  restrictions can be established.}*/

   /*{\Mtext \stdstart{1.5cm}}*/
   bool call(const leda::graph& G, const leda::edge_map<int> &max_b, LayoutInterface& A) {
     OrthoRepGridConcept C(G, A);
     get_node_grid_size(C, A);
     bool result = do_call(C, max_b);
     finish_call(C, A);
     return result;
   }
   /*{\Mop}*/
  
   bool call(const leda::graph& G, leda::face f, const leda::edge_map<int> &max_b, 
             LayoutInterface& A){
     OrthoRepGridConcept C(G, f, A);
     get_node_grid_size(C, A);
     bool result = do_call(C, max_b);
     finish_call(C, A);
     return result;
   }
   /*{\Mop}*/
  
   bool call(OrthogonalRepresentation &H, const leda::edge_map<int> &max_b,
             LayoutInterface& A) { 
     OrthoRepGridConcept C(H, A);
     get_node_grid_size(C, A);
     bool result = do_call(C, max_b);
     finish_call(C, A);
     return result;
   }
    /*{\Mop}*/

    bool call_grid(const leda::graph& G, const leda::edge_map<int> &max_b, 
                 GridLayout &gl) { 
      OrthoRepGridConcept C(G, gl);
      get_node_grid_size(C, gl);
      if (!do_call(C, max_b)) return false;
      finish_call_grid(C, gl);
      return true;
    }
    /*{\Mop}*/

    bool call_grid(const leda::graph& G, leda::face f, const leda::edge_map<int> &max_b, 
                 GridLayout &gl) {
      OrthoRepGridConcept C(G, f, gl);
      get_node_grid_size(C, gl);
      bool result = do_call(C, max_b);
      finish_call_grid(C, gl);
      return result;
    }
    /*{\Mop}*/
 
   bool call_grid(PlaneGraphCopy& PG, leda::face f, const leda::edge_map<int> &max_b, 
                 GridLayout &gl) {
     OrthoRepGridConcept C(PG, f, gl);
     get_node_grid_size(C, gl);
     bool result = do_call(C, max_b);
     finish_call_grid(C, gl);
     return result;
    }
    /*{\Mop}*/

    /*{\Mtext \stdend  }*/

protected:

  void get_node_grid_size (OrthoRepGridConcept &, const LayoutInterface &) { }
  void get_node_grid_size (OrthoRepGridConcept &C, const GridLayout &) {
    C.set_zero_size();
  }

  int _outputLevel;
  bool _verify, _variant;

  bool _allowBends, _allowNonPosDeg, _allowNegDeg,
    _forbidSmallZeroDeg, 
    _containsCages, _blowUp;
  
  int internal_edge_length_;
  float t_compact_;
  
  bool compute_optimal_shape(OrthogonalRepresentation& H, 
                             const leda::edge_map<int>& b);

  void INF(leda::string s) { 
    if (_outputLevel > 0) { cout << s; cout. flush(); }
  }

  void DONE() { if (_outputLevel > 0) cout << "done. " << endl; }

  void LINE() { 
    if (_outputLevel > 0) 
      cout << "----------------------------------------------------" << endl; 
  }

  void do_call(OrthoRepGridConcept& C);
  bool do_call(OrthoRepGridConcept& C, 
               const leda::edge_map<int>& max_b);

  virtual void preprocessing(OrthogonalRepresentation &H) = 0;
/*{\Moppure  }*/

  virtual void postprocessing(OrthogonalRepresentation &H, GridLayout &L) = 0;
/*{\Moppure  }*/



//#line 435 "Lweb/Ovid.lw"
private:

  enum n_type { high, low, inner, outer };
  enum a_type { angle, bend, cage_bend, neg_angle, cage_angle };
  
  inline int pow2(int n) { return 1 << n; }

  int k_;

bool _traditional, _progressive;
  bool _compactLayout, debug_;
  leda::list<leda::edge> _noConcaveBends;

  int n_internal_, m_internal_, l_internal_;

  AGD_DECLARE_MODULE(CompactionModule, compactor);

};


/*{\Mimplementation
Creation of the network and computation of a min cost flow takes time
$O((n \log n)^2)$.
}*/

#endif

