/*******************************************************************************
+
+  AGD 1.2
+
+
+  PlanarDrawLayout.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                     PlanarDrawLayout.h
//
//    author: Carsten Gutwenger
//    date:   august 1997
//-------------------------------------------------------------------


#ifndef _AGD_PLANAR_DRAW_LAYOUT_H
#define _AGD_PLANAR_DRAW_LAYOUT_H


#include <AGD/GridLayoutModuleImpl.h>
#include <AGD/PlaneGraphCopyGridConcept.h>
#include <AGD/AugmentationModule.h>
#include <AGD/AGDModuleOption.h>


class CanonicalOrder;


/*{\Manpage {PlanarDrawLayout} {} {The PlanarDraw Algorithm} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{9.5}{5.5}
\hdclass{0}{4.5}{5}{AGDModule}
\arhook{5}{5}{6}
\hdclass{1.5}{3}{5}{LayoutModule}
\arhook{6.5}{3.5}{7.5}
\hdclass{2}{1.5}{6}{GridLayoutModule}
\arhook{8}{2}{9}
\hdclass{3.5}{0}{6}{PlanarDrawLayout}
\end{hierarchy}
\hfill\agdeps{planar_draw.epsi}{3.8cm}
}*/

class __exportAGD PlanarDrawLayout : public GridLayoutModuleImpl<PlaneGraphCopyGridConcept> {

/*{\Mdefinition
The class |\Mname| represents the layout algorithm PlanarDraw
(Gutwenger and Mutzel~\cite{GM97}), which
is a generalization of the algorithm ConvexDraw (Chrobak and
Kant~\cite{CK97}).

This algorithm draws a planar graph |G| straight-line
without crossings. |G| must not contain self-loops or multiple edges.
The grid layout
size is at most $(n-1) \times (n-1)$ for a graph with $n$ nodes ($n \geq 3$).

The algorithm runs in three phases. In the first phase, the graph is
augmented by adding new artificial edges to get a biconnected plane
graph (using a planar biconnectivity augmentation module). Then, a new 
shelling order~\cite{GM97} (also called canonical ordering)
for biconnected planar graphs
is computed. In the third phase the vertices are placed incrementally
according to the shelling order.

\horizline
\commoninfo{PlanarDraw}{PlanarDraw (Gutwenger/Mutzel)}{C. Gutwenger, P. Mutzel}
{C. Gutwenger}{March 1997}{1.0}

\begin{preposttab}
precondition & $\{\, \emph{planar, simple, no self loops} \,\}$ \\
postcondition(|PRE|) & $\{\, \emph{straight-line, no crossings, convex} \,\}$
\end{preposttab}

\headerline{Exchangeable Modules}
Instances of type |\Mname| provide the following module options:
\begin{options}
\item {\bf augmenter}\\
an |AugmentationModule| option used for planar biconnected augmentation.

\begin{tabular}{ll}
guaranteed precondition: & $\{\emph{ planar, simple, no self loops, directed }\}$\\
required postcondition: & $\{\emph{ planar, biconnected, simple, no self loops }\}$\\
initial module: & |PlanAug|
\end{tabular}
\end{options}

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |bool| {\bf size\_opt} = |true|\\
a boolean value determining if the grid size of the
layout is optimized compared to the original Convex algorithm. This
option reduces the grid size especially for graphs with few edges.
If the graph is (almost) triangulated, the grid size will be (almost)
the same.
\item |bool| {\bf side\_opt} = |false|\\
a boolean value determining if it is allowed to have slopes other than
-1, 0, +1 on the outerface.
\item |bool| {\bf max\_outer} = |true|\\
a boolean value determining if the algorithm chooses
the face of |G| with maximal degree as outerface (|true|), or the
first face in the list of faces.
\item |double| {\bf base\_ratio} $= 0.33$\\
a double value determining, how many nodes are drawn on the base line.
If |base_ratio| is $r$ and the outerface is $F$, the number of nodes
on the base line is $\max\left(2,\min\left(\ell, r\Labs{F}\right)\right)$,
where $\ell = \max\{k \mid v_1,\ldots,v_k$ is a path on $F_{out}$ with
$(v_i,v_j)\notin E$ for all $j\neq i+1 \}$.
\end{options}
\horizline
}*/

public:

	/*{\Mcreation L 5}*/

	PlanarDrawLayout ();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	virtual ~PlanarDrawLayout() { }


	/*{\Moperations 1.5 2.5}*/

	leda::string name () const {
		return "PlanarDraw";
	}

	leda::string long_name () const {
		return "PlanarDraw (Gutwenger/Mutzel)";
	}

	leda::string author () const {
		return "C. Gutwenger, P. Mutzel";
	}

	leda::string impl_author () const {
		return "C. Gutwenger";
	}

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "March 1997";
	}


	AGDModule *clone () const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base classes (|LayoutModule|, |GridLayoutModule|).}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	virtual void call (const graph& G, face f, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, face f, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	IRect bounding_box ();
	*/
	/*{\Mop}*/

	/* inherited
	void set_mapper (const GridCoordinateMapper& gcm);
	*/
	/*{\Mop}*/
	
	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	/*
	bool set_augmenter (AugmentationModule& M);
	*/
	/*{\Mop}*/

	bool size_opt() const {
		return the_size_opt;
	}
	/*{\Mop}*/

	void size_opt (bool so) {
		the_size_opt = so;
	}
	/*{\Mop}*/

	bool side_opt() const {
		return the_side_opt;
	}
	/*{\Mop}*/

	void side_opt (bool so) {
		the_side_opt = so;
	}
	/*{\Mop}*/

	bool max_outer() const {
		return the_max_outer;
	}
	/*{\Mop}*/

	void max_outer (bool mo) {
		the_max_outer = mo;
	}
	/*{\Mop}*/

	double base_ratio() const {
		return the_base_ratio;
	}
	/*{\Mop}*/
	
	void base_ratio (double br) {
		the_base_ratio = br;
	}
	/*{\Mop}*/

	/*{\Mtext \stdend}*/

	AGD_DECLARE_MODULE (AugmentationModule, augmenter)

protected:
	void get_node_grid_size (PlaneGraphCopyGridConcept &, const LayoutInterface &) { }
	void get_node_grid_size (PlaneGraphCopyGridConcept &C, const GridLayout &) {
		C.set_zero_size();
	}

	void do_call (PlaneGraphCopyGridConcept &C);

private:
	bool the_size_opt;
	bool the_side_opt;
	bool the_max_outer;
	double the_base_ratio;

	void compute_coordinates (PlaneGraphCopyGridConcept &C,
		CanonicalOrder &c_o);
};


/*{\Mimplementation
The computation of the layout takes time $\BigO(n)$,
where $n$ is the number of nodes of the input graph.}*/


#endif
