/*******************************************************************************
+
+  AGD 1.2
+
+
+  PlaneGraphCopy.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         PlaneGraphCopy.h
//
//    author: Carsten Gutwenger
//    date:   february 2000
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_PLANARIZED_GRAPH_H
#define _AGD_PLANARIZED_GRAPH_H


#include <AGD/GraphCopy.h>

class StackDelNode;


/*{\Manpage {PlaneGraphCopy} {} {Plane Graph Copies} {PG} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5}{graph}
\arhook{5}{3.5}{6}
\hdclass{1.5}{1.5}{5}{GraphCopy}
\arhook{6.5}{2}{7.5}
\hdclass{2.5}{0}{5.5}{PlaneGraphCopy}
\end{hierarchy}
}*/

class __exportAGD PlaneGraphCopy : public GraphCopy {

/*{\Mdefinition
Plane graph copies represent copies of planar graphs or
planarized graphs as planar maps.
}*/

public:

	/*{\Mcreation PG 5}*/

	PlaneGraphCopy () : GraphCopy(), SS(0) { }
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname| initialized
	to an empty graph with undefined original graph.}*/
	
	PlaneGraphCopy(const leda::graph& G, bool preserveMap = true);
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname| initialized
	to a copy of |G|. If |preserveMap| is set to |false|, reversal edges are
	considered as multi-edges.}*/

	PlaneGraphCopy(const leda::graph& G, leda::list<leda::edge>& E, bool preserveMap = true);
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname| initialized
	to a copy of $G$ without the edges in $E$. \precond Each $e\in E$
	is an edge in $G$. If |preserveMap| is set to |false|, reversal edges are
	considered as multi-edges.}*/

	PlaneGraphCopy(const PlaneGraphCopy& PG);
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname| initialized
	to a copy of $|original|(PG)$.}*/

	PlaneGraphCopy (const PlaneGraphCopy& PG,
		leda::node_array<leda::node> &v_copy, leda::edge_array<leda::edge> &e_copy);
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname| initialized
	to a copy of $|original|(PG)$. Returns in |v_copy| the newly created
	nodes, i.e., $|v_copy|[v]$ is the node created for node $v\in|PG|$,
	and in |e_copy| the newly created edges, i.e., $|e_copy|[e]$ is the edge
	created for edge $e\in|PG|$.}*/

	~PlaneGraphCopy();


	/*{\Moperations 2 3.5}*/

	PlaneGraphCopy& operator= (const PlaneGraphCopy& PG);

	leda::face copy_of_face (leda::face f) const {
		return face_of(_chain[original().first_face_edge(f)].head());
	}
	/*{\Mop returns a face of |\Mvar| that corresponds to $F$.
	\precond $F$ is a face in $G$.}*/

	virtual leda::edge split (leda::edge e);
	/*{\Mop split edge $e=(v,w)$ and its reversal $r=(w,v)$ into
	edges $(v,u)$, $(u,w)$ and their reversals by introducing a new
	node $u$. Returns the edge $(u,w)$.}*/

	leda::edge unsplit (leda::node u);
	/*{\Mop unsplits a previously split edge $e=(v,w)$, where $u$
	has been created while splitting edge $e$. Removes $u$ and
	inserts edges $e'=(v,w)$ and $r=(w,v)$ according to the embedding.
	Returns edge $e'$.
	\precond $u$ has exactly two incoming and two outgoing edges,
	and has been created by a split operation.}*/

	edge unsplit_embedded (leda::node u);

	void new_edge_chain (leda::edge e, leda::edge from, leda::edge to,
		const leda::list<leda::edge>& crossings);
	/*{\Mop Inserts a chain of edges for edge e into |PG|. Let
	$|crossings|=e_1,\ldots,e_k$ for a $k\geq0$,
	$e_{k+1}=|to|$, and $e=(v,w)$.
	Then, we split each $e_i$ $(1\leq i\leq k)$ with a node $u_i$,
	and insert the
	edges $(|source|(|from|),u_1)$,$(u_i,u_{i+1})$,$(u_k,|source|(|to|))$ into
	|\Mvar|. The inserted edges are appended to |chain(e)|.
	\precond $e$ is an edge in $G$, all other edges are edges in |\Mvar|,
	$|source|(|from|)=|copy|(v)$,
	$|source|(|to|)=|copy|(w)$,
	$|face|(|from|)=|face|(e_1)$, and $|face|(|reversal|(e_i))=|face|(e_{i+1})$.}*/

	void new_edge_chain (leda::edge e, const leda::list<leda::edge>& crossings);
	/*{\Mop Inserts a chain of edges for edge e into |PG|. Let
	$|crossings|=e_1,\ldots,e_k$ for a $k\geq0$,
	$e_{k+1}=|to|$, and $e=(v,w)$.
	Then, we split each $e_i$ $(1\leq i\leq k)$ with a node $u_i$,
	and insert the
	edges $(|copy|(|source|(|e|)),u_1)$,$(u_i,u_{i+1})$,$(u_k,|copy|(|target|(|e|)))$ into
	|\Mvar|. The inserted edges are appended to |chain(e)| and a planar
	embedding of |\Mvar| is recomputed.
	\precond $e$ is an edge in $G$, all other edges are edges in |\Mvar|.}*/

	void del_edge_chain(leda::edge e);
	leda::list<leda::edge> del_edge_chain_embedded(leda::edge e);

	leda::face expand (leda::node v);
	/*{\Mop replaces node $v$ by a face $f$. Let $v_1,\ldots,v_k$ be the
	adjacent nodes of $v$. Then $f = w_1,\ldots,w_k$, where deg$(w_i)=3$
	and $w_i$ is connected to $v_i$. Returns $f$. \precond deg$(v)\geq 3$.}*/

	leda::node collapse (leda::face f, leda::node v_orig);
	/*{\Mop collapses face $f$ to a single node $v$. Sets the original
	node of $v$ to |v_orig|. Serves as undo for expand$(v)$. Returns $v$.
	\precond Each node on $f$ has degree $3$.}*/

	leda::face max_face() const;
	/*{\Mop returns a face of |\Mvar| with maximal degree.}*/

	virtual bool is_map() const {
		return true;
	}
	/*{\Mop returns |true|.}*/

	void remove_deg1_nodes(const leda::node_array<bool> &mark);
	void restore_deg1_nodes(leda::list<leda::node> &deg1s);

private:

	StackDelNode *SS;

	void extend_chain(leda::edge e1, leda::edge e2);
	leda::edge add_edge(leda::edge orig, leda::edge rev_orig, leda::edge e, leda::node w,
		int dir = leda::after);
	leda::edge add_edge(leda::edge orig, leda::edge rev_orig, leda::node v, leda::node w);


	void _init (leda::list<leda::edge>& E, bool preserveMap);
};



#endif
