/*******************************************************************************
+
+  AGD 1.2
+
+
+  RuleAdmin.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                           RuleAdmin.h
//
//
//    author: Carsten Gutwenger
//    date:   january 1999
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_RULE_ADMIN_H
#define _AGD_RULE_ADMIN_H


#include <AGD/Rule.h>
#include <AGD/Keymapper.h>
#include <LEDA/core/array.h>
#include <LEDA/core/map.h>


/*{\Manpage {RuleAdmin} {} {Rule Administrators} {A} }*/

class __exportAGD RuleAdmin {

/*{\Mdefinition
An instance |\Mvar| of type |\Mname| maintains a rule system
$R=\{r_1,\ldots,r_k\}$. The rules $r_i$ are of type |Rule|. Each
rule has a unique index betweeen $0$ and $k-1$. |\Mname|
allows to add rules to the system, and to compute the closure $P^*$
of a set of properties $P$ (a key set), which is defined as
\begin{eqnarray*}
P^0 & = & P \\
P^{i+1} & = & P^i\cup\left\{ p\mid\exists p_1,\ldots,p_k\in P^i:
p_1\wedge\ldots\wedge p_k\rightarrow p\in R\right\} \\
P^* & = & \bigcup_{i\geq0}P^i
\end{eqnarray*}
We say that a set of properties $P$ implies a set of properties $Q$
$(P\rightarrow Q)$ according to $R$ if and only if $P^*\supseteq Q$.
}*/

	leda::list<Rule *> _rules;
	const Keymapper *_mapper;
	//int index_count;
	leda::map<int,leda::list<Rule *> > on_left;

public:

	class const_iterator;
	friend class const_iterator;


	/*{\Mcreation}*/

	RuleAdmin(const Keymapper &mapper) : _mapper(&mapper) { }
	/*{\Mcreate creates an rule administrator |\Mvar| with empty rule
	system and associated keymapper |mapper|.}*/

	~RuleAdmin();


	/*{\Moperations}*/

	int number_of_rules() const { return _rules.length(); }
	/*{\Mop returns the number of rules in the rule system.}*/

	const_iterator begin() const;
	/*{\Mop returns an iterator pointing to first element in |\Mvar|.}*/

	const_iterator end() const;
	/*{\Mop returns an iterator pointing to one-past-last element in |\Mvar|.}*/

	const_iterator rbegin() const;
	/*{\Mop returns an iterator pointing to last element in |\Mvar|.}*/

	const_iterator rend() const;
	/*{\Mop returns an iterator pointing to one-before-first element in |\Mvar|.}*/


	bool implies(const KeySet &P, const KeySet &Q) const;
	/*{\Mop returns |true| iff $P\rightarrow Q$, i.e., iff $Q$ is contained in
	the closure of $P$ according to the rule system of |\Mvar|.}*/

	KeySet closure(const KeySet &P) const;
	/*{\Mop returns the closure $P^*$ of $P$ according to the rule system of |\Mvar|.}*/

	void add(const Rule &r);
	/*{\Mop adds the rule $r$ to the rule system maintained by |\Mvar|.}*/

private:
	// used by function closure
	void visit(AgdKey k, leda::array<bool> &mark, leda::array<int> &count) const;
};


/*{\Manpage {RuleAdmin::const_iterator} {} {Rule System Iterators} {it} }*/

class __exportAGD RuleAdmin::const_iterator {

/*{\Mdefinition
Instances of type |\Mvar| iterate over the elements of a rule system,
which is maintained by a rule administrator |admin|. They
behave similar to const-pointers to |Rule|. A rule administrator
provides functions to retrieve iterators pointing to the first or
last element of its rule system. We say an iterator of |admin| is valid
iff it points to an element of the rule system maintained by |admin|.
Derefferencing is only allowed for valid iterators.
}*/

public:

	/*{\Mcreation}*/

	const_iterator() : _admin(nil), _it(nil) { }
	/*{\Mcreate creates an invalid iterator |\Mvar|.}*/

	const_iterator(const RuleAdmin &admin) : _admin((RuleAdmin*)&admin), _it(nil) { }
	/*{\Mcreate creates an iterator |\Mvar| pointing to the first element of the
	rule system of |admin|.}*/

	const_iterator(const RuleAdmin &admin, leda::list_item it) :
		_admin((RuleAdmin*)&admin), _it(it) { }

	const_iterator(const const_iterator &iter) : _admin(iter._admin), _it(iter._it) { }
	/*{\Mcreate creates an iterator identical to |iter|.}*/


	/*{\Moperations}*/

	bool valid() const { return (_it != nil); }
	/*{\Mop returns |true| iff |\Mvar| is valid, i.e., points to an
	element.}*/

	bool operator==(const const_iterator& iter) const {
		return (_it == iter._it);
	}
	/*{\Mbinop returns |true| iff |\Mvar| and |iter| point to the
	same element.}*/

	bool operator!=(const const_iterator& iter) const {
		return (_it != iter._it);
	}
	/*{\Mbinop returns |true| iff |\Mvar| and |iter| point to
	different elements.}*/

	const_iterator &operator=(const const_iterator &iter) {
		_admin = iter._admin;
		_it = iter._it;
		return *this;
	}
	/*{\Mbinop assigns |iter| to |\Mvar|.}*/

	const Rule &operator*() { return *(_admin->_rules[_it]); }
	/*{\Munop returns the element to which |\Mvar| points.}*/

	const Rule *operator->() { return _admin->_rules[_it]; }
	/*{\Munop returns the address of the element to which |\Mvar| points.}*/

	const_iterator &operator++() {
		_it = (_it != nil) ? _admin->_rules.succ(_it) : _admin->_rules.first();
		return *this;
	}
	/*{\Munop increments and returns |\Mvar|.}*/

	const_iterator operator++(int) {
		const_iterator t = *this;
		_it = (_it != nil) ? _admin->_rules.succ(_it) : _admin->_rules.first();
		return t;
	}
	/*{\Munop increments |\Mvar| and returns its previous value.}*/

	const_iterator &operator--() {
		_it = (_it != nil) ? _admin->_rules.pred(_it) : _admin->_rules.last();
		return *this;
	}
	/*{\Munop decrements and returns |\Mvar|.}*/

	const_iterator operator--(int) {
		const_iterator t = *this;
		_it = (_it != nil) ? _admin->_rules.pred(_it) : _admin->_rules.last();
		return t;
	}
	/*{\Munop decrements |\Mvar| and returns its previous value.}*/

protected:
	RuleAdmin *_admin;
	leda::list_item _it;
};


inline RuleAdmin::const_iterator RuleAdmin::begin() const
{
	return const_iterator(*this,_rules.first());
}

inline RuleAdmin::const_iterator RuleAdmin::end() const
{
	return const_iterator(*this,nil);
}

inline RuleAdmin::const_iterator RuleAdmin::rbegin() const
{
	return const_iterator(*this,_rules.last());
}

inline RuleAdmin::const_iterator RuleAdmin::rend() const
{
	return const_iterator(*this,nil);
}


#endif
