/*******************************************************************************
+
+  AGD 1.2
+
+
+  SchnyderLayout.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                       SchnyderLayout.h
//
//    author: Dirk Ambras, AGD modifications David Alberts
//    date:   february 1997
//-------------------------------------------------------------------


#ifndef _AGD_SCHN_LAYOUT_H
#define _AGD_SCHN_LAYOUT_H


#include <AGD/GridLayoutModuleImpl.h>
#include <AGD/PlaneGraphCopyGridConcept.h>


/*{\Manpage {SchnyderLayout} {} {The Schnyder Algorithm} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{9.5}{5.5}
\hdclass{0}{4.5}{5}{AGDModule}
\arhook{5}{5}{6}
\hdclass{1.5}{3}{5}{LayoutModule}
\arhook{6.5}{3.5}{7.5}
\hdclass{2}{1.5}{6}{GridLayoutModule}
\arhook{8}{2}{9}
\hdclass{4.5}{0}{5}{SchnyderLayout}
\end{hierarchy}
\hfill\agdeps{schnyder.epsi}{3.8cm}
}*/

class __exportAGD SchnyderLayout : public GridLayoutModuleImpl<PlaneGraphCopyGridConcept>
{

/*{\Mdefinition
The class |\Mname| represents the layout algorithm by
Schnyder~\cite{s-epgg-90}.

This algorithm draws a planar graph |G| straight-line
without crossings. |G| must not contain self-loops or multiple edges.
The grid layout
size is $(n-2) \times (n-2)$ for a graph with $n$ nodes ($n \geq 3$).

The algorithm runs in three phases. In the first phase, the graph is
augmented by adding new artificial edges to get a triangulated plane
graph. Then, a partition of the set of interior edges in three trees 
(also called Schnyder trees) with special orientation properties is 
derived. In the third step, the actual coordinates are computed.


\horizline
\commoninfo{Schnyder}{Schnyder}{W. Schnyder}{D. Ambras, D. Alberts}
{February 1997}{}

\begin{preposttab}
precondition & $\{\, \emph{planar, simple, no self loops} \,\}$ \\
postcondition(|PRE|) & $\{\, \emph{straight-line, no crossings} \,\}$
\end{preposttab}

\horizline
}*/

  public:

	/*{\Mcreation}*/

    SchnyderLayout();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

    ~SchnyderLayout() { }

	/*{\Moperations}*/


    leda::string name () const { return "Schnyder"; }

    leda::string long_name () const { return "Schnyder"; }

    leda::string author () const { return "W. Schnyder"; }

	leda::string impl_author () const {
		return "D. Ambras, D. Alberts";
	}

  	leda::string impl_date () const {
		return "February 1997";
	}

	AGDModule *clone () const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base classes (|LayoutModule|, |GridLayoutModule|).}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	virtual void call (const graph& G, face f, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, face f, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	IRect bounding_box ();
	*/
	/*{\Mop}*/

	/* inherited
	void set_mapper (const GridCoordinateMapper& gcm);
	*/
	/*{\Mop}*/
	
	/*{\Mtext \stdend}*/

protected:
	void get_node_grid_size (PlaneGraphCopyGridConcept &, const LayoutInterface &) { }
	void get_node_grid_size (PlaneGraphCopyGridConcept &C, const GridLayout &) {
		C.set_zero_size();
	}

	void do_call (PlaneGraphCopyGridConcept &C);

  private:

    void Contract(leda::graph& G, leda::node a, leda::node b , leda::node c,
		leda::list<leda::node>& L);

    void Realizer(leda::graph& G, const leda::list<leda::node>& L, leda::node a,
		leda::node b, leda::node c, leda::GRAPH<leda::node, int>& T,
		leda::node_array<leda::node>& v_in_T);

    void Subtree_Sizes(leda::GRAPH<leda::node, int>& T, int i, leda::node r,
        leda::node_array<int>& size);

    void Prefix_Sum(leda::GRAPH<leda::node, int>& T, int i, leda::node r,
        const leda::node_array<int>& val, leda::node_array<int>& sum);

    int schnyder_embedding(leda::graph& G,leda::node_map<int>& xcoord,
        leda::node_map<int>& ycoord);
};

/*{\Mimplementation
The computation of the layout takes time $\BigO(n)$,
where $n$ is the number of nodes of the input graph.}*/


#endif
