/*******************************************************************************
+
+  AGD 1.2
+
+
+  SplitHeuristic.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         SplitHeuristic.h
//
//
//    author: Michael Krger
//    date:   march 1998
//-------------------------------------------------------------------


#ifndef _AGD_SPLIT_HEURISTIC_H
#define _AGD_SPLIT_HEURISTIC_H

#include <AGD/TwoLayerCrossMin.h>
#include <LEDA/core/array.h>


/*{\Manpage {SplitHeuristic} {} {Split Heuristic} {M} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{0.9}{1.5}{6.1}{TwoLayerCrossMin}
\arhook{7}{2}{8}
\hdclass{2.5}{0}{6}{SplitHeuristic}
\end{hierarchy}
}*/

class CrossingsMatrix;

class __exportAGD SplitHeuristic : public TwoLayerCrossMin {

/*{\Mdefinition
The module |\Mname| chooses a pivot vertex $v$ and places every
other vertex to the left or right of $v$ according to whether it 
would make fewer crossings. This step is applied recursively to 
order the left hand side and right hand side of $v$ (see Eades and Kelly~\cite{EK86}).

\horizline
\commoninfo{Split}{Split (Eades,Kelly)}{P. Eades, D. Kelly}{M. Kr\"uger, Th. Ziegler}
{March 1998}{1.0}

\begin{preposttab}
precondition & $\emptyset$ \\
postcondition(|PRE|) & $\emptyset$
\end{preposttab}

\horizline
}*/

public:

	/*{\Mcreation M 4.5}*/

	SplitHeuristic() { }
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname|.}*/

	~SplitHeuristic() { }


	/*{\Moperations}*/

	leda::string name () const {
		return "Split";
	}

	leda::string long_name () const {
		return "Split (Eades,Kelly)";
	}

	leda::string author () const {
		return "P. Eades, D. Kelly";
	}

	leda::string impl_author () const {
		return "M. Kr\201ger, Th. Ziegler";
	}

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "March 1998";
	}


	AGDModule *clone() const {
		return new SplitHeuristic();
	}

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|TwoLayerCrossMin|).
	}*/

	/* inherited
	bool check (const leda::graph& G, const Layer& L, AgdKey &p) const;
	*/
	/*{\Mop}*/

	void init (const leda::graph& G, const Hierarchy& H);
	/*{\Mop}*/

	void call (const leda::graph& G, Layer& L);
	/*{\Mop}*/

	void cleanup ();
	/*{\Mop}*/

	/*{\Mtext \stdend}*/

private:
	CrossingsMatrix *cp;
	leda::node *buffer;

	void _rec_call(Layer &, int low, int high);
};

/*{\Mimplementation
Let $n_L$ denote the number of nodes on the changeable layer
and $m_L$ the number of edges between the fixed and the changeable layer.
The algorithm computes in a preprocessing step a so-called
crossing matrix, which takes time $\BigO(n_L m_L)$.
The actual crossing reduction algorithm takes time
$\BigO(n_L \log n_L)$. Both steps are implemented in the
\texttt{call()}-method.
}*/


#endif
