/*******************************************************************************
+
+  AGD 1.2
+
+
+  StPlanarGridLayout.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                    StPlanarGridLayout.h
//
//    author: Carsten Gutwenger
//    date:   decembre 1999
//-------------------------------------------------------------------


#ifndef AGD_ST_PL_GRI_LAYOUT_H
#define AGD_ST_PL_GRI_LAYOUT_H

#include <AGD/GridLayoutModuleImpl.h>
#include <AGD/GraphCopyGridConcept.h>


/*{\Manpage {StPlanarGridLayout} {} {Dominance Layout} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{9.5}{5.5}
\hdclass{0}{4.5}{5}{AGDModule}
\arhook{5}{5}{6}
\hdclass{1.5}{3}{5}{LayoutModule}
\arhook{6.5}{3.5}{7.5}
\hdclass{2}{1.5}{6}{GridLayoutModule}
\arhook{8}{2}{9}
\hdclass{3.3}{0}{6.2}{StPlanarGridLayout}
\end{hierarchy}
\hfill\agdeps{st_planar_grid.epsi}{3.8cm}
}*/

class StPlanarGridLayout __exportAGD : public GridLayoutModuleImpl<GraphCopyGridConcept>
{
/*{\Mdefinition
The class |\Mname| represents the layout algorithm for st-planar
graphs by Di Battista, Tamassia, and Tollis \cite{BTT92}. If the
input graph $G$ contains no transitive edges, the algorithm computes
a planar dominance grid drawing of $G$, that is, a straight-line
embedding, such that, for any two vertices $u$ and $v$, there is
a directed path from $u$ to $v$ if and only if $x(u)\leq x(v)$
and $y(u)\leq y(v)$. Dominance drawings characterize the transitive
closure of a digraph by means of the geometric dominance relation
among the vertices \cite{BTT92}.

If $G$ contains transitive edges, each transitive edge is split
by introducing a new (dummy) vertex, and a dominance drawing for
the resulting graph is computed. The dummy vertices represent
bend points in the final drawing.

An st-planar graph is a directed, acyclic graph $G$ with exactly one
source $s$ and exactly one sink $t$, such that $G\cup(s,t)$ is planar.

The drawings constructed by |\Mname| have the following symmetry
properties, if |CompactionType| is |none| or |symmetryPreserving|.
Let $G$ be a reduced st-planar graph and $\Gamma$ the constructed
drawing.
\begin{enumerate}
\item Simply isomorphic components of $G$ have drawings in $\Gamma$
that are congruent up to a translation.

\item Axially isomorphic components of $G$ have drawings in $\Gamma$
that are congruent up to a translation and reflection.

\item Rotationally isomorphic components of $G$ have drawings in $\Gamma$
that are congruent up to a translation and a $180^\circ$ rotation.

\item The drawing of an axially symmetric component of $G$ is
symmetric with respect to the straight line that passes through its
source and sink.

\item Teh drawing of a rotationally symmetric component of $G$ is
symmetric with respect to a $180^\circ$ rotation around its centroid.
\end{enumerate}

If |CompactionType| is set to |minimumArea| and $G$ is a reduced
st-planar graph, then the dominance drawing constructed by |\Mname|
has minimum area among all dominance drawings of $G$ that places
vertices on grid points and reflect the same embedding of $G$.
Please refer to \cite{BTT92} for more details.

\horizline
\commoninfo{Dominance Layout}{Dominance Layout (Di Battista/Tamassia/Tollis)}{G. Di Battista, R. Tamassia, I. G. Tollis}
{C. Gutwenger}{Decembre 1999}{1.0}

\begin{preposttab}
precondition & $\{\, \emph{st-planar} \,\}$ \\
postcondition(|PRE|) & $\{\, \emph{no crossings} \,\}$
\end{preposttab}

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |CompactionType| {\bf compaction} = |symmetryPreserving|\\
selects the type of compaction used. Possible settings are
|none|, |symmetryPreserving|, and |minimumArea|.
\end{options}
\horizline
}*/

public:

	/*{\Mtypes}*/ 

	enum CompactionType { none, symmetryPreserving, minimumArea };
	/*{\Menum the types of compaction.}*/


	/*{\Mcreation}*/

	StPlanarGridLayout();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	~StPlanarGridLayout() { }



	/*{\Moperations 1.5 3}*/

	leda::string name () const {
		return "Dominance Layout";
	}

	leda::string long_name () const {
		return "Dominance Layout (Di Battista/Tamassia/Tollis)";
	}

	leda::string author () const {
		return "G. Di Battista, R. Tamassia, I. G. Tollis";
	}

	leda::string impl_author () const {
		return "C. Gutwenger";
	}

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "Decembre 1999";
	}


	void call_plane (GraphCopy &GC, leda::edge e_st, GridLayout &gl);
	/*{\Mop calls |\Mvar| for embedded graph copy |GC| with grid layout |gl|
	and st-edge |e_st|. Computes a grid layout for the original graph of |GC| and
	preserves the embedding defined by |GC|. \precond |GC| is a planar map with
	computed faces, $|e_st|\in|GC|$ and |gl| is a grid layout for
	the original graph of |GC|.}*/

	AGDModule *clone () const;


	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base classes (|LayoutModule|, |GridLayoutModule|).}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, GridLayout &gl);
	*/
	/*{\Mop}*/


	/* inherited
	IRect bounding_box ();
	*/
	/*{\Mop}*/

	/* inherited
	void set_mapper (const GridCoordinateMapper& gcm);
	*/
	/*{\Mop}*/
	
	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	CompactionType compaction() const {
		return _compaction;
	}
	/*{\Mop}*/

	void compaction(CompactionType c) {
		_compaction = c;
	}
	/*{\Mop}*/

	/*{\Mtext \stdend}*/


private:
	void get_node_grid_size (GraphCopyGridConcept &, const LayoutInterface &) { }
	void get_node_grid_size (GraphCopyGridConcept &C, const GridLayout &) {
		C.set_zero_size();
	}

	void do_call(GraphCopyGridConcept &C);

	void draw(GraphCopy &GC, GridLayout &gl, const leda::list<leda::edge> &R);

	void labelX(const leda::graph &G, leda::node v);
	void labelY(const leda::graph &G, leda::node v);
	void compact(const leda::graph &G, GridLayout &gl);
	void findTransitiveEdges(const GraphCopy &GC, leda::list<leda::edge> &T);

	CompactionType _compaction;

	leda::node s, t;
	leda::edge st;
	bool stMissing;
		
	int count;
	leda::node_map<leda::edge> firstout, lastout;
	leda::node_map<leda::edge> firstin,  lastin;

	leda::node_array<int> X, Y;
	leda::node *xNode, *yNode;
};


#endif
