/*******************************************************************************
+
+  AGD 1.2
+
+
+  StPlanarLayout.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        StPlanarLayout.h
//
//    author: Carsten Gutwenger
//    date:   decembre 1999
//-------------------------------------------------------------------


#ifndef AGD_ST_PL_LAYOUT_H
#define AGD_ST_PL_LAYOUT_H

#include <AGD/StPlanarGridLayout.h>
#include <AGD/LayoutModuleImpl.h>
#include <AGD/ConstLedaGraphConcept.h>


/*{\Manpage {StPlanarLayout} {} {Layout of st-Planar Graphs} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{6}
\hdclass{0}{5}{5}{AGDModule}
\arhook{5}{5.5}{6}
\hdclass{1.5}{3.5}{5}{LayoutModule}
\arhook{6.5}{4}{7.5}
\hdclass{2}{2}{6}{StPlanarLayout}
\end{hierarchy}
\hfill\agdeps{st_planar.epsi}{3.7cm}
}*/

class StPlanarLayout __exportAGD : public LayoutModuleImpl<ConstLedaGraphConcept>
{
/*{\Mdefinition
The class |\Mname| represents the layout algorithm for st-planar
graphs by Di Battista, Tamassia, and Tollis \cite{BTT92}. It computes
in a first step a dominance grid layout by calling |StPlanarGridLayout|,
and rotates the computed layout by an angle of $45^{\circ}$.

An st-planar graph is a directed, acyclic graph $G$ with exactly one
source $s$ and exactly one sink $t$, such that $G\cup(s,t)$ is planar.

The drawings conctructed by |\Mname| have the same symmetry properties
as the drawings constructed by |StPlanarGridLayout| with the same
setting of |CompactionType|.

\horizline
\commoninfo{st-planar Draw}{st-planar Draw (Di Battista/Tamassia/Tollis)}{G. Di Battista, R. Tamassia, I. G. Tollis}
{C. Gutwenger}{Decembre 1999}{1.0}

\begin{preposttab}
precondition & $\{\, \emph{st-planar} \,\}$ \\
postcondition(|PRE|) & $\{\, \emph{no crossings} \,\}$
\end{preposttab}

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |StPlanarGridLayout::CompactionType| {\bf compaction} = |symmetryPreserving|\\
selects the type of compaction used in the call of |StPlanarGridLayout|.
\item |double| {\bf separation} $= 20$\\
determines the minimal distance between node in the final layout.
\end{options}
\horizline
}*/

public:

	/*{\Mcreation}*/

	StPlanarLayout();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	~StPlanarLayout() { }


	/*{\Moperations 1.5 3}*/

	leda::string name () const {
		return "st-planar Draw";
	}

	leda::string long_name () const {
		return "st-planar Draw (Di Battista/Tamassia/Tollis)";
	}

	leda::string author () const {
		return "G. Di Battista, R. Tamassia, I. G. Tollis";
	}

	leda::string impl_author () const {
		return "C. Gutwenger";
	}

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "Decembre 1999";
	}


	void call_plane (GraphCopy &GC, leda::edge e_st, LayoutInterface &A);
	/*{\Mop calls |\Mvar| for embedded graph copy |GC| with layout interface $A$
	and st-edge |e_st|. Computes a layout for the original graph of |GC| and
	preserves the embedding defined by |GC|. \precond |GC| is a planar map with
	computed faces, $|e_st|\in|GC|$ and |A| is a layout interface for
	the original graph of |GC|.}*/

	AGDModule *clone () const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\medskip\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|LayoutModule|).
	}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/
	
	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/


	StPlanarGridLayout::CompactionType compaction() const {
		return _compaction;
	}
	/*{\Mop}*/

	void compaction(StPlanarGridLayout::CompactionType c) {
		_compaction = c;
	}
	/*{\Mop}*/

	double separation() const {
		return _sep;
	}
	/*{\Mop}*/

	void separation(double x) {
		_sep = x;
	}
	/*{\Mop}*/

	/*{\Mtext \stdend}*/


private:
	void do_call(ConstLedaGraphConcept &C);
	
	void draw(const leda::graph &G, GridLayout &gl, LayoutInterface &A);

	StPlanarGridLayout::CompactionType _compaction;
	double _sep;
};


#endif
