/*******************************************************************************
+
+  AGD 1.2
+
+
+  SubgraphChecker.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        SubgraphChecker.h
//
//    author: Carsten Gutwenger
//    date:   novembre 1998
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_PM_CHECKER_H
#define _AGD_PM_CHECKER_H

#include <AGD/AGD.h>
#include <LEDA/core/list.h>


typedef AgdKey (*check_subgraph_handler)(const leda::graph &G, const leda::list<leda::edge> &E,
	KeySet &props, KeySet &satisfied, bool check_all);

// avoid a warning with g++
inline istream &operator>> (istream &is, check_subgraph_handler &) { return is; }


/*{\Manpage {SubgraphChecker} {} {Subgraph Checker} {C} }*/

class __exportAGD SubgraphChecker {

/*{\Mdefinition
The class |\Mname| maintains a list of check-subgraph handler functions
|Checkers|, which are used in order to test if a subgraph satsifies a given set of
properties, or to compute the set of properties a given subgraph satisfies.
The type |check_subgraph_handler| is a pointer to a check-subgraph handler
function. Such a function, e.g. f, has to be declared as
\begin{quote}
|AgdKey| f(const |graph|\& G, const |list<edge>|\& E, |KeySet|\& |props|, |KeySet|\&
|satisfied|, |bool| |check_all|);
\end{quote}
It has to test if input subgraph $G\setminus E$ satisfies the properties in |props|.
Each property that has been checked must be removed from |props|.
If |check_all| is |true|, each satisfied property is added to |satsified|.
If |check_all| is |false|, f returns an unsatisfied property or a |nil| key
(cf. Sect.~\ref{Keys}) if it cannot find an unsatisfied property.

|\Mname| allows to add and remove functions from |Checkers|. New handler
functions are added at the beginning of |Checkers|, i.e., a newly
added handler is called before any other handler. Finally, a graph checker
(cf. Sect.~\ref{Graph Checker}) is called, which has been passed to
the constructor of |\Mname|.
}*/

	leda::list<check_subgraph_handler> checkers;

public:

	/*{\Mcreation}*/

	SubgraphChecker(GraphChecker &GC) : gcp(&GC) { }
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	~SubgraphChecker() { }

	/*{\Moperations}*/

	bool check(const leda::graph &G, const leda::list<leda::edge> &E, KeySet &props) const;
	/*{\Mop returns |true| iff |G| satisfies all properties in |props|.}*/

	bool check(const leda::graph &G, const leda::list<leda::edge> &E, KeySet &props, AgdKey &not_satisfied) const;
	/*{\Mop returns |true| if |G| satisfies all properties in |props|,
	otherwise |false| is returned and |not_satsified| is a property in
	|props| that is not satisfied by |G|.}*/

	KeySet get_properties(const leda::graph &G, const leda::list<leda::edge> &E, KeySet &props) const;
	/*{\Mop returns the set of all properties in |props| which are
	satisfied by |G|.}*/

	leda::list_item add_checker(check_subgraph_handler f);
	/*{\Mop appends a new handler function |f| at the beginning of |Checkers|.
	Returns the corresponding list item.}*/

	void rem_checker(leda::list_item it);
	/*{\Mop removes the check-handler function at position |it|
	in |Checkers|. \precond |it| is a list item in |Checkers|.}*/

private:
	GraphChecker *gcp;
};


#endif
