/*******************************************************************************
+
+  AGD 1.2
+
+
+  TreeLayout.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


/************************************************************************

            Filename    :  TreeLayout.h

            Version     :  01.1997

            Author      :  Sebastian Leipert

            Language    :  C++

************************************************************************/


#ifndef _AGD_TREE_LAYOUT_H
#define _AGD_TREE_LAYOUT_H

#include <AGD/impl/TreeLayout/tree.h>
#include <AGD/LayoutModuleImpl.h>
#include <AGD/ConstLedaGraphConcept.h>


/*{\Manpage {TreeLayout} {} {Walker's Algorithm} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{6}
\hdclass{0}{5}{5}{AGDModule}
\arhook{5}{5.5}{6}
\hdclass{1.5}{3.5}{5}{LayoutModule}
\arhook{6.5}{4}{7.5}
\hdclass{3}{2}{5}{TreeLayout}
\end{hierarchy}
\hfill\agdeps{tree.epsi}{3.8cm}
}*/

class __exportAGD TreeLayout : public LayoutModuleImpl<ConstLedaGraphConcept>
{
/*{\Mdefinition
The class |\Mname| represents the layout algorithm by Walker II
(Reingold and Tilford, Walker~\cite{rt-tdt-81,w-npagt-90,Lei96}).

This algorithm draws binary and general rooted trees. The graph $G=(V,E)$
has to be directed, connected, and has to have exactly one source. 
Furthermore the equality $\Labs{V}-1 = \Labs{E}$ must hold. 

The drawing is straight-line.
Nodes of the same level are placed on a horizontal line. 

The algorithm works recursively starting on the lowest level of the tree.
In each step, the subtrees of a tree vertex (that have been layed out
already) are placed as closely to each other as possible.
This guarantees to have small layout size. Moreover, isomorphic subtrees
are layed out symmetrically.

\horizline
\commoninfo{TreeLayout}{Layout algorithm for general trees}{E. Reingold, J. Tilford, J. Q. Walker II}
{S. Leipert}{May 1997}{0.1}

\begin{preposttab}
precondition & $\{\, \emph{simple, tree} \,\}$ \\
postcondition(|PRE|) & $\{\, \emph{straight-line, no crossings, upward planar} \,\}$
\end{preposttab}

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |double| {\bf sib\_separation} $= 4.0$\\
The spacing between two sibling nodes.
\item |double| {\bf subtree\_separation} $= 4.0$\\
The spacing between two adjacent subtrees.
\item |double| {\bf level\_separation} $= 4.0$\\
The spacing between the nodes of two adjacent levels.
\item |double| {\bf left\_margin} $= 0.0$\\
Additional spacing at the left side of the drawing.
\item |double| {\bf upper\_margin} $= 0.0$\\
Additional spacing at the upper side of the drawing.
\end{options}\horizline
}*/

public:

	/*{\Mcreation}*/

   TreeLayout();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

   virtual ~TreeLayout();


  /*{\Moperations}*/

   leda::string name() const;

   leda::string long_name() const;

   leda::string author() const;

   leda::string impl_author() const;

   leda::string impl_version() const;

   leda::string impl_date() const;


	AGDModule *clone () const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|LayoutModule|).
	}*/

	/* inherited
	virtual bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/
	
	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

   double sib_separation() const;
	/*{\Mop}*/

   void sib_separation(double s);
	/*{\Mop}*/

   double level_separation() const;
	/*{\Mop}*/

   void level_separation(double s);
	/*{\Mop}*/

   double subtree_separation() const;
	/*{\Mop}*/

   void subtree_separation(double s);
	/*{\Mop}*/

   double left_margin() const;
	/*{\Mop}*/

   void left_margin(double s);
	/*{\Mop}*/

   double upper_margin() const;
	/*{\Mop}*/

   void upper_margin(double s);
	/*{\Mop}*/

	/*{\Mtext \stdend}*/

protected:

   void do_call(ConstLedaGraphConcept &C);

private:

   bool Initialize(const leda::graph& G,LayoutInterface &a);
   
   Tree    _T;
   leda::node    _root;
   leda::node  *_arrayOfNodes;
};

/*{\Mimplementation
The running time of the algorithm is $\BigO(\Lvert V\Lvert)$.}*/

#endif


