/*******************************************************************************
+
+  AGD 1.2
+
+
+  TriconnectivityTest.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



/**********************************************************

                TriconnectivityTest.h

  Test multigraph G = (V,E) if triconnected in
  time O(n+m).

  reference: Hopcroft and Tarjan, "Dividing a Graph into
             Triconnected Components"
             SIAM J. Comput., Vol. 2, No. 3, Sept. 1973
			 pages 135-158

  author:    Carsten Gutwenger
  date:      Novembre 1998

**********************************************************/


#ifndef _AGD_TRIC_TEST_H
#define _AGD_TRIC_TEST_H


#include <AGD/AGDModule.h>
#include <LEDA/graph/graph.h>
#include <LEDA/core/array.h>


/*{\Manpage {TriconnectivityTest} {} {Testing Triconnectivity} {T} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5.5}{AGDModule}
\arhook{5.5}{2}{6.5}
\hdclass{1}{0}{6}{TriconnectivityTest}
\end{hierarchy}
}*/

class __exportAGD TriconnectivityTest : public AGDModule {

/*{\Mdefinition
The class |\Mname| represents the triconnectivity test algorithm
according to Hopcroft and Tarjan~\cite{HT73}. It tests if a graph $G$ is
triconnected and returns a witness (cut vertex, separation pair) if not.
There are no restrictions on the input graph $G$, i.e., $G$ may contain
multiple edges or self loops.

\horizline
\commoninfo{Triconnectivity Test}{Triconnectivity Test (Hopcroft/Tarjan)}{J. E. Hopcroft, R. E. Tarjan}
{C. Gutwenger}{Novembre 1998}{1.0}

\begin{preposttab}
precondition & $\emptyset$
\end{preposttab}

\horizline
}*/

public:

	/*{\Mcreation T 4.5}*/

	TriconnectivityTest() : AGDModule(1,0) { }
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	~TriconnectivityTest() { }


	/*{\Moperations 1.5 2.5}*/

	leda::string name () const {
		return "Triconnectivity Test";
	}

	leda::string long_name () const {
		return "Triconnectivity Test (Hopcroft/Tarjan)";
	}

	leda::string author () const {
		return "J. E. Hopcroft, R. E. Tarjan";
	}

	virtual leda::string impl_author () const {
		return "C. Gutwenger";
	}

	virtual leda::string impl_version () const {
		return "1.0";
	}

	virtual leda::string impl_date () const {
		return "Novembre 1998";
	}


	bool call (const leda::graph& G) {
		leda::node s1, s2;
		return call(G,s1,s2);
	}
	/*{\Mop calls |\Mvar| for graph $G$. Returns |true| iff
	$G$ is triconnected.}*/

	bool call (const leda::graph& G, leda::node& s1, leda::node& s2);
	/*{\Mop calls |\Mvar| for graph $G$. Returns |true| iff
	$G$ is triconnected. Returns in |s1| and |s2| nodes whose
	deletion disconnect the graph, i.e. $|s1| = |nil|$ and
	$|s2| = |nil| \Rightarrow$ $G$ is not connected, $|s1| \neq |nil|$ and
	$|s2| = |nil| \Rightarrow$ |s1| is a cut vertex, $|s1| \neq |nil|$ and
	$|s2| \neq |nil| \Rightarrow$ $(|s1|,|s2|)$ is a separation pair.}*/


	AGDModule *clone() const {
		return new TriconnectivityTest;
	}

private:

	int *TSTACK_h, *TSTACK_a, *TSTACK_b;
	int top;

	void TSTACK_push (int h, int a, int b) {
		TSTACK_h[++top] = h;
		TSTACK_a[top] = a;
		TSTACK_b[top] = b;
	}

	enum edge_type { unseen, tree, tree_rev, frond, frond_rev };

	leda::node src (leda::edge e) const {
		return (TYPE[e] == tree_rev || TYPE[e] == frond_rev) ?
			leda::target(e) : leda::source(e);
	}

	leda::node tgt (leda::edge e) const {
		return (TYPE[e] == tree_rev || TYPE[e] == frond_rev) ?
			leda::source(e) : leda::target(e);
	}

	int Phi (leda::edge e);

	void build_graph (const leda::graph &G);
	void DFS1 (const leda::graph& G, leda::node v, leda::node u, leda::node& s1);
	void build_acceptable_adj_struct (const leda::graph& G);
	void DFS2 (const leda::graph& G);
	void path_finder(const leda::graph& G, leda::node v);
	bool PathSearch (const leda::graph& G, leda::node v, leda::node& s1, leda::node& s2);

	leda::node_array<int>       NUMBER, LOWPT1, LOWPT2, ND;
	leda::array<leda::node>      NODEAT;
	leda::node_array<leda::node> FATHER;
	leda::edge_array<edge_type> TYPE;
	leda::node_array<leda::list<leda::edge> > A;
	leda::node_array<int>  NEWNUM, HIGHPT;
	leda::edge_array<bool> START;
	leda::node_array<int>  DEGREE;

	leda::list<leda::edge> edges;
	leda::edge_array<bool> is_edge;

	leda::node s;
	int num_count;
	bool new_path;
};


/*{\Mimplementation
Testing a graph $G=(V,E)$ for triconnectivity takes time
$\BigO(\Labs{V}+\Labs{E})$.}*/


#endif
