/*******************************************************************************
+
+  AGD 1.2
+
+
+  VisibilityRepresentation.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


// ---------------------------------------------------------------- //
// compute a visibility representation of the planar graph G        //
// using the algorithm by Rosenstiehl and Tarjan                    //
//                                                                  //
// David Alberts (1996), AGD Version (1997)                         //
// last change: september 1998                                      //
// ---------------------------------------------------------------- //

#ifndef _AGD_VISREP_H
#define _AGD_VISREP_H


#include <AGD/GridLayoutModuleImpl.h>
#include <AGD/GraphCopyGridConcept.h>
#include <AGD/AugmentationModule.h>
#include <AGD/AGDModuleOption.h>
#include <LEDA/core/map.h>


/*{\Manpage {VisibilityRepresentation} {} {Visibility Representation} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{11}{5.5}
\hdclass{0}{4.5}{5}{AGDModule}
\arhook{5}{5}{6}
\hdclass{1.5}{3}{5}{LayoutModule}
\arhook{6.5}{3.5}{7.5}
\hdclass{2}{1.5}{6}{GridLayoutModule}
\arhook{8}{2}{9}
\hdclass{2}{0}{7.5}{VisibilityRepresentation}
\end{hierarchy}
\hfill\agdeps{visibility.epsi}{3.8cm}
}*/

class __exportAGD VisibilityRepresentation : public GridLayoutModuleImpl<GraphCopyGridConcept>
{

/*{\Mdefinition
The class |\Mname| represents the visibility representation algorithm
by Rosenstiehl and Tarjan~\cite{rt-rplbo-86}.

This algorithm draws a planar graph |G| such that
each node is represented by a horizontal
segment and each edge is represented by a vertical segment, and incidence
between nodes and edges translates to intersection of horizontal and
vertical segments.

The grid size is at most $(2n-4) \times (n-1)$ for a graph with $n$ nodes.
|G| must not contain self-loops or multiple edges.

\horizline
\commoninfo{Visibility Representation}{Weak Visibility Representation}{P. Rosenstiehl, R. E. Tarjan}
{D. Alberts}{1996/1997}{}

\begin{preposttab}
precondition & $\{\, \emph{planar, simple, no self loops} \,\}$ \\
postcondition(|PRE|) & $\{\, \emph{straight-line, no crossings} \,\}$
\end{preposttab}

\headerline{Exchangeable Modules}
Instances of type |\Mname| provide the following module options:
\begin{options}
\item {\bf augmenter}\\
an |AugmentationModule| option used for planar biconnected augmentation.

\begin{tabular}{ll}
guaranteed precondition: & $\{\emph{ planar, simple, no self loops }\}$\\
required postcondition: & $\{\emph{ planar, biconnected, simple, no self loops }\}$\\
initial module: & |PlanAug|
\end{tabular}
\end{options}

\horizline
}*/

	AGD_DECLARE_MODULE (AugmentationModule, augmenter)

  public:

	/*{\Mcreation L 5}*/

    VisibilityRepresentation();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

    ~VisibilityRepresentation() { }

	/*{\Moperations}*/

    leda::string name() const {
		return "Visibility Representation";}

    leda::string long_name() const {
		return "Weak Visibility Representation";
	}

    leda::string author () const {
		return "P. Rosenstiehl, R. E. Tarjan";
	}

	leda::string impl_author () const {
		return "D. Alberts";
	}

	leda::string impl_date () const {
		return "1996/1997";
	}

	AGDModule *clone() const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base classes (|LayoutModule|, |GridLayoutModule|).}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	IRect bounding_box ();
	*/
	/*{\Mop}*/

	/* inherited
	void set_mapper (const GridCoordinateMapper& gcm);
	*/
	/*{\Mop}*/

	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	/*
	bool set_augmenter (AugmentationModule& M);
	*/
	/*{\Mop}*/

	/*{\Mtext \stdend}*/

  protected:

	void get_node_grid_size (GraphCopyGridConcept &, const LayoutInterface &) { }
	void get_node_grid_size (GraphCopyGridConcept &C, const GridLayout &) {
		C.set_zero_size();
	}

	void do_call (GraphCopyGridConcept &C);

  private:

    int compute_depth(leda::graph& G, leda::node s, leda::node_array<int>& depth,
                      int offset=0);

    int VISIBILITY_REPRESENTATION(GraphCopyGridConcept &C);

    leda::node_map<int> leftx;
    leda::node_map<int> rightx;
    leda::node_map<int> ypos;
    leda::edge_map<int> xpos;
    leda::edge_map<int> lowy;
    leda::edge_map<int> topy;
};

/*{\Mimplementation
The computation of the layout takes time $\BigO(n)$,
where $n$ is the number of nodes of the input graph.}*/


#endif
