/*******************************************************************************
+
+  AGD 1.2
+
+
+  ShortestPathInserter.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


#include <AGD/ShortestPathInserter.h>

using namespace leda;


//-------------------------------------------------------------------
// 
// Shortest Path Inserter
// 
//-------------------------------------------------------------------

AGDModule *ShortestPathInserter::clone() const
{
	ShortestPathInserter *M = new ShortestPathInserter;

	if (insert_method==0) { (*M).set_to_basic();}
	else if (insert_method==1) {(*M).set_to_shortest_first();}
	else if (insert_method==2) {(*M).set_to_rr();}
	else {(*M).set_to_err();}

	return M;
}

void ShortestPathInserter::do_call (PlaneGraphCopy& PG, list<edge>& e_insert)
{
	if (e_insert.length() == 0) return;

	if (insert_method==0) {call_basic(PG,e_insert);}
	else if (insert_method==1) { call_sf(PG,e_insert);}
	else if (insert_method==2) {call_rr(PG,e_insert);}
	else{call_err(PG,e_insert);	}
	
	// clean up
	delete pFaceNode;
	delete pPrimalEdge;
	delete pDual;
}

void ShortestPathInserter::call_basic (PlaneGraphCopy& PG, list<edge>& e_insert)
{
	pMark = new face_map<bool>(PG,false);
	construct_dual(PG);

	edge eOrig;
	forall(eOrig,e_insert) {
		list<edge> crossed;
		face fCommon = find_shortest_path(PG.copy(source(eOrig)),
			PG.copy(target(eOrig)), crossed);

		insertEdge(PG,eOrig,crossed,fCommon);
	}
}

void ShortestPathInserter::call_sf (PlaneGraphCopy& PG, list<edge>& e_insert)
{
	pMark = new face_map<bool>(PG,false);
	construct_dual(PG);

	const leda::graph& G=PG.original();
	edge_map<bool> already_inserted(G,false);

	int num_inserted=0;
	while (num_inserted<e_insert.size()) {
		num_inserted++;

		edge eInsert = 0;
		int length=-1;
		edge eOrig;
		forall(eOrig,e_insert) {
			if (!already_inserted[eOrig]) {
				list<edge> crossed;
				/*face fCommon = */ // unused (s.n.)
                                find_shortest_path(PG.copy(source(eOrig)),
					PG.copy(target(eOrig)), crossed);

				if (length==-1||length>crossed.size()) {
					length=crossed.size();
					eInsert=eOrig;
				}
			}
		}

		list<edge> crossed;
		face fCommon = find_shortest_path(PG.copy(source(eInsert)),
		PG.copy(target(eInsert)), crossed);

		insertEdge(PG,eInsert,crossed,fCommon);
		already_inserted[eInsert]=true;
	}
}

void ShortestPathInserter::call_rr (PlaneGraphCopy& PG, list<edge>& e_insert)
{
	pMark = new face_map<bool>(PG,false);
	construct_dual(PG);

	edge eOrig;
	forall(eOrig,e_insert) {
		list<edge> crossed;
		face fCommon = find_shortest_path(PG.copy(source(eOrig)),
			PG.copy(target(eOrig)), crossed);

		insertEdge(PG,eOrig,crossed,fCommon);
	}

	/* const graph &G = PG.original(); */ // unused (s.n.)

	list<edge> rr_edges;
	rr_edges = e_insert;

	bool improved;
	do {
		improved = false;
		forall(eOrig,rr_edges) {
			int chainLength = PG.chain(eOrig).length();
			if (chainLength == 1) continue;

			removeEdge(PG,eOrig);
			list<edge> crossed;
			face fCommon = find_shortest_path(PG.copy(source(eOrig)),
				PG.copy(target(eOrig)), crossed);

			insertEdge(PG,eOrig,crossed,fCommon);
			if (crossed.length()+1 < chainLength) improved = true;
		}
	} while (improved);
}

void ShortestPathInserter::call_err (PlaneGraphCopy& PG, list<edge>& e_insert)
{
	pMark = new face_map<bool>(PG,false);
	construct_dual(PG);

	edge eOrig;
	forall(eOrig,e_insert) {
		list<edge> crossed;
		face fCommon = find_shortest_path(PG.copy(source(eOrig)),
			PG.copy(target(eOrig)), crossed);

		insertEdge(PG,eOrig,crossed,fCommon);
	}

	const graph &G = PG.original();
	list<edge> rr_edges;
//	rr_edges = G.all_edges();  //bad order for removing and reinsert
//	permute(rr_edges);  //slightly better (?)

	rr_edges=e_insert; //good
	leda::edge_map<bool> insertdone(G,false);
	edge e;
	forall(e,e_insert) insertdone[e]=true;
	forall_edges(e,G) if (!insertdone[e]) rr_edges.append(e);

	bool improved;
	do {
		improved = false;
		forall(eOrig,rr_edges) {
			int chainLength = PG.chain(eOrig).length();
			if (chainLength == 1) continue;

			removeEdge(PG,eOrig);
			list<edge> crossed;
			face fCommon = find_shortest_path(PG.copy(source(eOrig)),
				PG.copy(target(eOrig)), crossed);

			insertEdge(PG,eOrig,crossed,fCommon);
			if (crossed.length()+1 < chainLength) improved = true;
		}
	} while (improved);
}

//-------------------------------------------------------------------
// construct augmented dual
//-------------------------------------------------------------------

void ShortestPathInserter::construct_dual(const graph &G)
{
	pDual = new graph;
	pPrimalEdge = new edge_map<edge>(*pDual,nil);
	pFaceNode = new face_map<node>(G);

	// constructs nodes (for faces in G)
	face f;
	forall_faces(f,G) {
		(*pFaceNode)[f] = pDual->new_node();    }

	// construct dual edges (for primal edges in exp)
	edge e;
	forall_edges(e,G) {
		(*pPrimalEdge)[pDual->new_edge((*pFaceNode)[G.face_of(e)],
			(*pFaceNode)[G.face_of(G.reversal(e))])] = e;
	}

	// augment dual by m_vS and m_vT
	m_vS = pDual->new_node();
	m_vT = pDual->new_node();
}


//-------------------------------------------------------------------
// find shortest path in dual from m_vS representing s to m_vT
// representing t; output this path in L by omitting first and
// last edge, and translating edges to PG
//-------------------------------------------------------------------

face ShortestPathInserter::find_shortest_path(node s, node t, list<edge> &L)
{
	// add augmented edges to dual (temporarily)
	face f;
	forall_adj_faces(f,s)
		pDual->new_edge(m_vS,(*pFaceNode)[f]);

	forall_adj_faces(f,t)
		pDual->new_edge((*pFaceNode)[f], m_vT);

	node_array<edge> spPred(*pDual,nil); // predecessor in shortest path tree
	list<edge> queue; // candidate edges

	// start with all edges leaving from m_vS
	edge e;
	forall_out_edges(e,m_vS)
		queue.append(e);

	node fCommon;

	while(true) {
		edge eCand = queue.pop_front(); // next candidate from front of queue
		node v = target(eCand);

		// hit an unvisited node ?
		if (spPred[v] == nil) {
			spPred[v] = eCand;

			// if it is m_vT, we have found the shortest path
			if (v == m_vT) {
				fCommon = source(eCand);
				// build path from shortest path tree
				while(v != m_vS) {
					edge eG = (*pPrimalEdge)[spPred[v]];
					if (eG != nil) // == nil for first and last edge
						L.push_front(eG);
					v = source(spPred[v]);
				}
				break;
			}

			// append next candidates to end of queue
			forall_out_edges(e,v)
				queue.append(e);
		}
	}

	// remove added edges from dual
	while ((e = pDual->first_adj_edge(m_vS)))
		pDual->del_edge(e);

	while ((e = pDual->first_in_edge(m_vT)))
		pDual->del_edge(e);

	if (L.empty()) {
		// find common face of s and t
		forall_adj_faces(f,t)
			if ((*pFaceNode)[f] == fCommon) return f;
	}
	return nil;
}


void ShortestPathInserter::insertEdge(PlaneGraphCopy &PG, edge eOrig,
	list<edge> &crossed, face fCommon)
{
	// updating PG
	edge e_first = 0; 
        edge e_last = 0;
        edge e;
	node src = PG.copy(source(eOrig));
	node tgt = PG.copy(target(eOrig));

	if (fCommon != nil) {
		forall_face_edges(e, fCommon) {
			if (PG.source(e) == src) e_first = e;
			if (PG.source(e) == tgt) e_last  = e;
		}
	}

	else {
		face f = PG.face_of(crossed.head());
		forall_face_edges(e,f) {
			if (PG.source(e) == src) e_first = e;
		}
  
		f = PG.face_of(PG.reversal(crossed.tail()));
		forall_face_edges(e,f) {
			if (PG.source(e) == tgt) e_last = e;
		}
	}

	// delete face nodes in dual
	if (fCommon) {
		pDual->del_node((*pFaceNode)[fCommon]);

	} else {
		forall(e,crossed)
			pDual->del_node((*pFaceNode)[PG.face_of(e)]);
		pDual->del_node((*pFaceNode)[PG.face_of(PG.reversal(crossed.tail()))]); // oder head?
	}

	// update PG
	PG.new_edge_chain (eOrig, e_first, e_last, crossed);

	// insert new face nodes into dual
	forall(e,PG.chain(eOrig)) {
		(*pFaceNode)[PG.face_of(e)] = pDual->new_node();
		(*pFaceNode)[PG.face_of(PG.reversal(e))] = pDual->new_node();
	}

	// insert new edges into dual
	forall(e,PG.chain(eOrig)) {
		face f = PG.face_of(e);
		edge ef;
		forall_face_edges(ef,f) {
			node v1 = (*pFaceNode)[f];
			node v2 = (*pFaceNode)[PG.face_of(PG.reversal(ef))];
			(*pPrimalEdge)[pDual->new_edge(v1,v2)] = ef;
			(*pPrimalEdge)[pDual->new_edge(v2,v1)] = PG.reversal(ef);
		}

		f = PG.face_of(PG.reversal(e));
		forall_face_edges(ef,f) {
			node v1 = (*pFaceNode)[f];
			node v2 = (*pFaceNode)[PG.face_of(PG.reversal(ef))];
			(*pPrimalEdge)[pDual->new_edge(v1,v2)] = ef;
			(*pPrimalEdge)[pDual->new_edge(v2,v1)] = PG.reversal(ef);
		}
	}
}



void ShortestPathInserter::removeEdge(PlaneGraphCopy &PG, edge eOrig)
{
	edge e;
	forall(e,PG.chain(eOrig)) {
		pDual->del_node((*pFaceNode)[PG.face_of(e)]);
		pDual->del_node((*pFaceNode)[PG.face_of(PG.reversal(e))]);
	}

	list<edge> updatedEdges = PG.del_edge_chain_embedded(eOrig);
	list<face> updatedFaces;

	forall(e,updatedEdges) {
		face f = PG.face_of(e);
		if ((*pMark)[f] == false) {
			(*pMark)[f] = true; updatedFaces.append(f);
		}
		f = PG.face_of(PG.reversal(e));
		if ((*pMark)[f] == false) {
			(*pMark)[f] = true; updatedFaces.append(f);
		}
	}

	face f;
	forall(f,updatedFaces) {
		(*pMark)[f] = false;
		(*pFaceNode)[f] = pDual->new_node();
	}

	forall(f,updatedFaces) {
		forall_face_edges(e,f) {
			(*pPrimalEdge)[pDual->new_edge((*pFaceNode)[f],
				(*pFaceNode)[PG.face_of(PG.reversal(e))])] = e;
		}
	}
}

