/*******************************************************************************
+
+  AGD 1.2
+
+
+  checkers.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                          checkers.cpp
//
//    author: Carsten Gutwenger
//    date:   novembre 1998
//-------------------------------------------------------------------


#include <AGD/GraphChecker.h>
#include <AGD/ClusterGraphChecker.h>
#include <AGD/SubgraphChecker.h>
#include <AGD/KeySet.h>

using namespace leda;


//-------------------------------------------------------------------
//                          GraphChecker
//-------------------------------------------------------------------

GraphChecker::GraphChecker()
{
	add_checker(&default_checker);
}


list_item GraphChecker::add_checker(check_graph_handler f)
{
	return checkers.push(f);
}


void GraphChecker::rem_checker(list_item it)
{
	checkers.del_item(it);
}


bool GraphChecker::check(const graph &G, KeySet &props) const
{
	AgdKey p;

	return check(G,props,p);
}


bool GraphChecker::check(const graph &G, KeySet &props, AgdKey &not_satisfied) const
{
	check_graph_handler f;
	KeySet satisfied;

	not_satisfied = AgdKey();

	forall(f,checkers) {
		not_satisfied = f(G,props,satisfied,false);
		if (not_satisfied.valid()) return false;
	}

	return props.empty();
}


KeySet GraphChecker::get_properties(const graph &G, KeySet &props) const
{
	check_graph_handler f;
	KeySet satisfied;

	forall(f,checkers) {
		f(G,props,satisfied,true);
	}

	return satisfied;
}


AgdKey GraphChecker::default_checker(const graph &G, KeySet &props,
	KeySet &satisfied, bool check_all)
{
	AgdKey k;

	// explicit call of AgdKey's constructor is workaround for SUN CC
	k = AgdKey(key::planar);
	if (props.member(k)) {
		props.del(k);
		if (AGD::is_planar(G)) satisfied.insert(k);
		else if (!check_all) return k;
	}

	k = key::simple;
	if (props.member(k)) {
		props.del(k);
		if (AGD::is_simple(G)) satisfied.insert(k);
		else if (!check_all) return k;
	}

	k = key::connected;
	if (props.member(k)) {
		props.del(k);
		if (AGD::is_connected(G)) satisfied.insert(k);
		else if (!check_all) return k;
	}

	k = key::biconnected;
	if (props.member(k)) {
		props.del(k);
		if (AGD::is_biconnected(G)) satisfied.insert(k);
		else if (!check_all) return k;
	}

	k = key::triconnected;
	if (props.member(k)) {
		props.del(k);
		if (AGD::is_triconnected(G)) satisfied.insert(k);
		else if (!check_all) return k;
	}

	k = key::four_graph;
	if (props.member(k)) {
		props.del(k);
		if (AGD::is_four_graph(G)) satisfied.insert(k);
		else if (!check_all) return k;
	}

	k = key::no_self_loops;
	if (props.member(k)) {
		props.del(k);
		if (AGD::has_no_self_loops(G)) satisfied.insert(k);
		else if (!check_all) return k;
	}

	k = key::directed;
	if (props.member(k)) {
		props.del(k);
		if (G.is_directed()) satisfied.insert(k);
		else if (!check_all) return k;
	}

	k = key::undirected;
	if (props.member(k)) {
		props.del(k);
		if (G.is_undirected()) satisfied.insert(k);
		else if (!check_all) return k;
	}

	k = key::no_reversals;
	if (props.member(k)) {
		props.del(k);
		if (AGD::has_no_reversals(G)) satisfied.insert(k);
		else if (!check_all) return k;
	}

	k = key::acyclic;
	if (props.member(k)) {
		props.del(k);
		if (AGD::is_acyclic(G)) satisfied.insert(k);
		else if (!check_all) return k;
	}

	k = key::st_graph;
	if (props.member(k)) {
		props.del(k);
		if (AGD::is_st_graph(G)) satisfied.insert(k);
		else if (!check_all) return k;
	}

	k = key::st_planar;
	if (props.member(k)) {
		props.del(k);
		if (AGD::is_st_planar(G)) satisfied.insert(k);
		else if (!check_all) return k;
	}

	k = key::forest;
	if (props.member(k)) {
		props.del(k);
		if (AGD::is_forest(G)) satisfied.insert(k);
		else if (!check_all) return k;
	}

	k = key::tree;
	if (props.member(k)) {
		props.del(k);
		if (AGD::is_tree(G)) satisfied.insert(k);
		else if (!check_all) return k;
	}

	k = key::single_source;
	if (props.member(k)) {
		props.del(k);
		if (AGD::has_single_source(G)) satisfied.insert(k);
		else if (!check_all) return k;
	}

	k = key::single_sink;
	if (props.member(k)) {
		props.del(k);
		if (AGD::has_single_sink(G)) satisfied.insert(k);
		else if (!check_all) return k;
	}

	k = key::bipartite;
	if (props.member(k)) {
		props.del(k);
		if (AGD::is_bipartite(G)) satisfied.insert(k);
		else if (!check_all) return k;
	}

	return AgdKey();
}


//-------------------------------------------------------------------
//                          ClusterGraphChecker
//-------------------------------------------------------------------

ClusterGraphChecker::ClusterGraphChecker(GraphChecker &GC) : gcp(&GC)
{
        add_checker(&default_checker);
}

list_item ClusterGraphChecker::add_checker(check_clustergraph_handler f)
{
	return checkers.push(f);
}

void ClusterGraphChecker::rem_checker(list_item it)
{
	checkers.del_item(it);
}

bool ClusterGraphChecker::check(const ClusterGraph &CG, KeySet &props) const
{
	AgdKey p;

	return check(CG,props,p);
}

bool ClusterGraphChecker::check(const ClusterGraph &CG, 
				KeySet &props, 
				AgdKey &not_satisfied) const
{
	check_clustergraph_handler f;
	KeySet satisfied;

	not_satisfied = AgdKey();

	forall(f,checkers) {
		not_satisfied = f(CG,props,satisfied,false);
		if (not_satisfied.valid()) return false;
	}

	//return props.empty();
	return gcp->check(CG,props,not_satisfied);
}

KeySet ClusterGraphChecker::get_properties(const ClusterGraph &CG, 
					   KeySet &props) const
{
	check_clustergraph_handler f;
	KeySet satisfied;

	forall(f,checkers) {
		f(CG,props,satisfied,true);
	}

	return satisfied + gcp->get_properties(CG,props);
}

AgdKey ClusterGraphChecker::default_checker(const ClusterGraph &CG, 
					    KeySet &props,
					    KeySet &satisfied, 
					    bool check_all)
{
    AgdKey k;

	k = key::c_planar;
	if (props.member(k)) {
	        props.del(k);
		if (AGD::is_cplanar(CG)) satisfied.insert(k);
		else if (!check_all) return k;
	}

	return AgdKey();
}

//-------------------------------------------------------------------
//                         SubgraphChecker
//-------------------------------------------------------------------

list_item SubgraphChecker::add_checker(check_subgraph_handler f)
{
	return checkers.push(f);
}


void SubgraphChecker::rem_checker(list_item it)
{
	checkers.del_item(it);
}


bool SubgraphChecker::check(const graph &G, const list<edge> &E, KeySet &props) const
{
	AgdKey p;

	return check(G,E,props,p);
}


bool SubgraphChecker::check(const graph &G, const list<edge> &E,
	KeySet &props, AgdKey &not_satisfied) const
{
	check_subgraph_handler f;
	KeySet satisfied;

	not_satisfied = AgdKey();

	forall(f,checkers) {
		not_satisfied = f(G,E,props,satisfied,false);
		if (not_satisfied.valid()) return false;
	}

	return gcp->check(G,props,not_satisfied);
}


KeySet SubgraphChecker::get_properties(const graph &G, const list<edge> &E,
	KeySet &props) const
{
	check_subgraph_handler f;
	KeySet satisfied;

	forall(f,checkers) {
		f(G,E,props,satisfied,true);
	}

	return satisfied + gcp->get_properties(G,props);
}
