/*******************************************************************************
+
+  AGD 1.2
+
+
+  graph.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                          graph.cpp
//
//    author: Carsten Gutwenger
//    date:   february 2000
//-------------------------------------------------------------------


#include <AGD/PlaneGraphCopy.h>
#include <AGD/LeveledGraph.h>

#include <LEDA/core/array.h>
#include <LEDA/core/stack.h>
#include <LEDA/internal/std/limits.h>

using namespace leda;


//-------------------------------------------------------------------
//                            GraphCopy
//-------------------------------------------------------------------

GraphCopy::GraphCopy(const GraphCopy& GC) : graph(GC)
{
	_init_GC (GC);
}

GraphCopy::GraphCopy(const GraphCopy& GC, node_array<node> &v_copy,
	edge_array<edge> &e_copy) : graph(GC)
{
	_init_GC (GC,v_copy,e_copy);
}

GraphCopy& GraphCopy::operator=(const GraphCopy& GC)
{
	graph::operator=(GC);
	_init_GC(GC);
	return *this;
}

void GraphCopy::_init_GC (const GraphCopy& GC)
{
	node_array<node> v_copy(GC);
	edge_array<edge> e_copy(GC);
	_init_GC(GC,v_copy,e_copy);
}

void GraphCopy::_init_GC (const GraphCopy& GC, node_array<node> &v_copy,
	edge_array<edge> &e_copy)
{
	node v, w;
	edge e, e1;

	g_orig = GC.g_orig;

	e_orig.init(*this,nil);
	v_orig.init(*this,nil);
	_in_chain.init(*this, nil);

	if (g_orig != nil) {
		_chain   .init(*g_orig);
		_copy    .init(*g_orig);
		//node_array<node> v_copy(GC);
		//edge_array<edge> e_copy(GC);
		v_copy.init(GC);
		e_copy.init(GC);

		map_to_copy (v_copy, e_copy, GC);

		forall_nodes (v, GC)
			v_orig[v_copy[v]] = GC.original(v);

		forall_edges (e, GC)
			e_orig[e_copy[e]] = GC.original(e);

		forall_nodes (v, *this) {
			if ((w = v_orig[v]) != nil)
				_copy[w] = v;
		}

		forall_edges (e, *g_orig)
			forall (e1, GC._chain[e]) {
				_in_chain[e_copy[e1]] = _chain[e].append(e_copy[e1]);
			}
	}
}

GraphCopy::GraphCopy(const graph& G) : graph(G)
{
	node v;
	edge e;

	g_orig = &G;
	e_orig.init(*this,nil);
	v_orig.init(*this,nil);
	_copy    .init(G);
	_chain   .init(G);
	_in_chain.init(*this, nil);


	edge_array<edge> e_copy(G);

	map_to_copy (_copy, e_copy, G);

	forall_nodes (v, G)
		v_orig[_copy[v]] = v;

	forall_edges (e, G)
		e_orig[e_copy[e]] = e;

	forall_edges (e, *this) {
		_in_chain[e] = _chain[original(e)].append(e);
	}

}

void GraphCopy::map_to_copy (node_array<node>& v_copy,
	edge_array<edge>& e_copy, const graph& G)
{
	node v, vc;
	edge e, ec;

	vc = first_node();
	forall_nodes (v, G) {
		v_copy[v] = vc;
		ec = _first_adjacent(vc);
		for (e = _first_adjacent(v); e != nil; e = _succ_adjacent(e,v)) {
			e_copy[e] = ec;
			ec = _succ_adjacent(ec,vc);
		}
		vc = succ_node(vc);
	}
}

edge GraphCopy::_first_adjacent (node v)
{
	edge e = First_Adj_Edge(v,0);

	if (e != nil && source(e) != v)
		e = _succ_adjacent(e,v);

	return e;
}

edge GraphCopy::_succ_adjacent (edge e, node v)
{
	do {
		e = Succ_Adj_Edge(e,v);
	} while (e != nil && source(e) != v);

	return e;
}

edge GraphCopy::split (edge e)
{
	edge e2;

	split_edge (e, e, e2);

	if (original(e) != nil) {
		_in_chain[e2] = _chain[original(e)].insert(e2,_in_chain[e],leda::after);
		e_orig[e2] = original(e);
	}

	return e2;
}


//-------------------------------------------------------------------
//                          PlaneGraphCopy
//-------------------------------------------------------------------

struct DelNode
{
	DelNode() { }
	DelNode(edge _eo, edge _eor, node _vo, edge _eref) :
		eo(_eo), eo_rev(_eor), vo(_vo), e_ref(_eref) { }

	edge eo, eo_rev;
	node vo;
	edge e_ref;

	friend ostream& operator<< (ostream& os, const DelNode&)
	{ return os; }
	friend istream& operator>> (istream& is, DelNode&)
	{ return is; }
};

class StackDelNode : public stack<DelNode> {
public:
	StackDelNode() { }
	~StackDelNode() { }
};

void PlaneGraphCopy::_init (list<edge>& E, bool preserveMap)
{
	make_directed();

	edge e;
	forall (e, E) {
		del_edge(chain(e).head());
		_chain[e].clear();
	}

	if (preserveMap) {
		if (Is_Planar_Map(*this)) {
			compute_faces();
		} else {
			list<edge> R;
			make_map(R);
			make_planar_map();
		}

	} else {
		list<edge> edges = all_edges();
		forall(e,edges)
			set_reversal(e,new_edge(target(e),source(e)));
		make_planar_map();
	}
}

PlaneGraphCopy::PlaneGraphCopy (const PlaneGraphCopy& PG,
	node_array<node> &v_copy, edge_array<edge> &e_copy) :
	GraphCopy(PG,v_copy,e_copy), SS(0)
{
	if (g_orig != nil) compute_faces();
}

PlaneGraphCopy::PlaneGraphCopy (const PlaneGraphCopy& PG) :
	GraphCopy(PG), SS(0)
{
	if (g_orig != nil) compute_faces();
}

PlaneGraphCopy::PlaneGraphCopy(const graph& G, bool preserveMap) :
	GraphCopy(G), SS(0)
{
	list<edge> empty;
	_init (empty, preserveMap);
}

PlaneGraphCopy::PlaneGraphCopy(const graph& G, list<leda::edge>& E,
	bool preserveMap) : GraphCopy (G), SS(0)
{
	_init (E, preserveMap);
}

PlaneGraphCopy::~PlaneGraphCopy()
{
	delete SS;
}


PlaneGraphCopy& PlaneGraphCopy::operator= (const PlaneGraphCopy& PG)
{
	GraphCopy::operator=(PG);
	if (g_orig != nil) compute_faces();
	return *this;
}

edge PlaneGraphCopy::split (edge e)
{
	edge r = reversal(e), e1 = split_map_edge(e);

	if ((e_orig[e1] = original(e)) != nil) {
		_in_chain[e1] = _chain[original(e)].insert(e1,_in_chain[e],leda::after);
	}

	if ((e_orig[reversal(e)] = original(r)) != nil) {
		_in_chain[reversal(e)] =
			_chain[original(r)].insert(reversal(e),_in_chain[r],leda::after);
	}

	return e1;
}

edge PlaneGraphCopy::unsplit (node u)
{
	if (degree(u) == 0) {
		del_node(u);
		return nil;
	}

	if (indeg(u) != 2)
		error_handler(1,"PlaneGraphCopy::unsplit: indeg(u) != 2");

	edge e = first_in_edge(u);
	edge r = last_in_edge (u);

	edge er = reversal(e);
	edge rr = reversal(r);

	if (original(e) != original(rr) || original(r) != original(er))
		error_handler(1,"PlaneGraphCopy::unsplit: edge was not split!");

	edge x = new_map_edge (e, r);
	edge xr = reversal(x);

	if ((e_orig[x] = original(e)) != nil) {
		list<edge> &L = _chain[original(e)];

		/*_in_chain[x] = _in_chain[e];
		L[_in_chain[e]] = x;*/
		L[_in_chain[x] = _in_chain[e]] = x;

		L.del_item(_in_chain[rr]);
	}
	
	if ((e_orig[xr] = original(r)) != nil) {
		list<edge> &L = _chain[original(r)];

		/*_in_chain[xr] = _in_chain[r];
		L[_in_chain[r]] = xr;*/
		L[_in_chain[xr] = _in_chain[r]] = xr;

		L.del_item(_in_chain[er]);
	}

	del_node(u);

	return x;
}
	
edge PlaneGraphCopy::unsplit_embedded (node u)
{
	if (degree(u) == 0) {
		del_node(u);
		return nil;
	}

	if (indeg(u) != 2)
		error_handler(1,"PlaneGraphCopy::unsplit: indeg(u) != 2");

	edge e = first_in_edge(u);
	edge r = last_in_edge (u);

	edge er = reversal(e);
	edge rr = reversal(r);

	if (original(e) != original(rr) || original(r) != original(er))
		error_handler(1,"PlaneGraphCopy::unsplit: edge was not split!");

	edge x  = (rr != first_face_edge(face_of(rr))) ? e : rr;
	edge xr = (er != first_face_edge(face_of(er))) ? r : er;
	move_edge(x ,e,rr);
	move_edge(xr,r,er);
	set_reversal(x,xr);

	edge h = (x == e) ? rr : e;
	if (original(h) != nil) {
		list<edge> &L = _chain[original(h)];
		L.del_item(_in_chain[h]);
	}

	h = (xr == r) ? er : r;
	if (original(h) != nil) {
		list<edge> &L = _chain[original(h)];
		L.del_item(_in_chain[h]);
	}

	del_node(u);

	return x;
}

void PlaneGraphCopy::new_edge_chain (edge e, edge from, edge to,
	const list<edge>& crossings)
{
	edge e1, ed, ec;

	forall (ec, crossings) {
		e1 = split (ec);
		ed = split_face (from, e1);
		_in_chain[ed] = _chain[e_orig[ed] = e].append(ed);
		from = reversal(ec);
	}
	ed = split_face (from, to);
	_in_chain[ed] = _chain[e_orig[ed] = e].append(ed);
}


void PlaneGraphCopy::new_edge_chain (edge e, const leda::list<leda::edge>& crossings)
{
	edge ec, ed;
	node x, v = copy(source(e));
	forall(ec,crossings) {
		x = source(split(ec));
		ed = new_edge(v,x);
		_in_chain[ed] = _chain[e_orig[ed] = e].append(ed);
		set_reversal(ed,new_edge(x,v));
		v = x;
	}
	ed = new_edge(v,copy(target(e)));
	_in_chain[ed] = _chain[e_orig[ed] = e].append(ed);
	set_reversal(ed,new_edge(copy(target(e)),v));
}


void PlaneGraphCopy::del_edge_chain(edge e)
{
	if (_chain[e].empty()) return;

	node v = _copy[source(e)], w = _copy[target(e)];
	list<node> dummies;
	edge ec=0;
	forall(ec,_chain[e]) {
		if (source(ec) != v && source(ec) != w)
			dummies.append(source(ec));
		del_edge(reversal(ec));
		del_edge(ec);
	}

	node u;
	forall(u,dummies)
		unsplit(u);

	_chain[e].clear();
}


list<edge> PlaneGraphCopy::del_edge_chain_embedded(edge e)
{
	list<edge> updatedEdges;

	if (_chain[e].empty())
		return updatedEdges;

	node v = _copy[source(e)], w = _copy[target(e)];
	list<node> dummies;
	edge ec=0;
	forall(ec,_chain[e]) {
		if (source(ec) != v && source(ec) != w)
			dummies.append(source(ec));
		join_faces(ec);
	}

	node u;
	forall(u,dummies)
		updatedEdges.append(unsplit_embedded(u));

	_chain[e].clear();

	return updatedEdges;
}


face PlaneGraphCopy::expand (node v)
{
	edge e = 0;
	face f = 0;

	if (degree(v) < 6)
		error_handler(1,"PlaneGraphCopy::expand: deg(v) < 3");

	forall_out_edges(e,v)
		split(e);

	forall_out_edges(e,v)
		split_face (reversal(e), 
                            face_cycle_succ(cyclic_adj_pred(e)));


	forall_out_edges(e,v) {
		if (original(e) != nil) _chain[original(e)].pop_front();
		if (original(reversal(e)) != nil)
			_chain[original(reversal(e))].pop_back();

		f = join_faces(e);
	}

	del_node(v);

	return f;
}

node PlaneGraphCopy::collapse (face f, node orig)
{
	edge e, e1, e2;
	list<edge> edges = adj_edges(f);
	list_item it = edges.first();

	e1 = edges[it]; it = edges.succ(it);
	e2 = edges[it]; it = edges.succ(it);
	e1 = split_map_edge(split_face(e1,e2));

	node v = source(e1);
	v_orig[v] = orig;
	if (orig != nil) _copy[orig] = v;

	for (; it != nil; it = edges.succ(it))
		e1 = split_face(e1,edges[it]);

	while (!edges.empty()) {
		e = edges.pop();

		extend_chain(cyclic_adj_pred(e),reversal(cyclic_adj_succ(e)));
		join_faces(e);
	}

	return v;
}

// adds the dummy edge e2 to the chain to which e1 belongs
void PlaneGraphCopy::extend_chain(edge e1, edge e2)
{
	edge e1r = reversal(e1), e2r = reversal(e2);
	edge e1o = original(e1), e1ro = original(e1r);

	if (e1o != nil) {
		if (target(e1) == source(e2))
			_in_chain[e2] = _chain[e1o].push_back(e2);
		else
			_in_chain[e2] = _chain[e1o].push_front(e2);
		e_orig[e2] = e1o;
	}

	if (e1ro != nil) {
		if (target(e1r) == source(e2r))
			_in_chain[e2r] = _chain[e1ro].push_back(e2r);
		else
			_in_chain[e2r] = _chain[e1ro].push_front(e2r);
		e_orig[e2r] = e1ro;
	}
}

face PlaneGraphCopy::max_face() const
{
	int max = 0;
	face f, f_max = nil;

	forall_faces (f, *this) {
		if (size(f) >= max) {
			max = size(f);
			f_max = f;
		}
	}
	return f_max;
}

edge PlaneGraphCopy::add_edge(edge orig, edge rev_orig, edge e, node w,
	int dir)
{
	edge e1, e2;

	e_orig[e1 = new_edge(e,w,          dir)] = orig;
	e_orig[e2 = new_edge(w,reversal(e),dir)] = rev_orig;

	if (orig)     _chain[orig]     = list<edge>(e1);
	if (rev_orig) _chain[rev_orig] = list<edge>(e2);
	set_reversal(e1,e2);

	return e1;
}

edge PlaneGraphCopy::add_edge(edge orig, edge rev_orig, node v, node w)
{
	edge e1, e2;

	e_orig[e1 = new_edge(v,w)] = orig;
	e_orig[e2 = new_edge(w,v)] = rev_orig;

	if (orig)     _chain[orig]     = list<edge>(e1);
	if (rev_orig) _chain[rev_orig] = list<edge>(e2);
	set_reversal(e1,e2);

	return e1;
}

void PlaneGraphCopy::remove_deg1_nodes(const node_array<bool> &mark)
{
	node v;

	delete SS;
	SS = new StackDelNode;

	// delete marked nodes; store originals and reference edge in SS
	for(v = first_node(); v; v = succ_node(v)) {
		if (mark[v] || outdeg(v) == 0) continue;

		edge e, e_ref, e_ref_1;
		for (e_ref = first_adj_edge(v); e_ref && mark[target(e_ref)];
			e_ref = adj_succ(e_ref)) ;
		if (!e_ref) {
			forall_adj_edges(e,v) {
				SS->push(DelNode(original(e),original(reversal(e)),
					original(target(e)),nil));
				del_node(target(e));
			}
			continue;
		}
		for (e_ref_1 = e_ref, e = cyclic_adj_succ(e_ref); e != e_ref_1;
			e = cyclic_adj_succ(e)) {
			if (mark[target(e)]) {
				SS->push(DelNode(original(e),original(reversal(e)),
					original(target(e)),e_ref));
				del_node(target(e));
			} else
				e_ref = e;
		}
	}
	compute_faces();
}


void PlaneGraphCopy::restore_deg1_nodes(list<node> &deg1s)
{
	if (!SS) return;

	while (!SS->empty()) {
		node v;
		DelNode mmdn = SS->pop();
		if (mmdn.e_ref)
			add_edge(mmdn.eo, mmdn.eo_rev, mmdn.e_ref, v = add_node(mmdn.vo));
		else {
			node w = (mmdn.eo != nil) ? source(mmdn.eo) : target(mmdn.eo_rev);
			add_edge(mmdn.eo, mmdn.eo_rev, copy(w), v = add_node(mmdn.vo));
		}
		deg1s.append(v);
	}

	delete SS; SS = 0;
}


//-------------------------------------------------------------------
//                           LeveledGraph
//-------------------------------------------------------------------

LeveledGraph::LeveledGraph() : graph(), _set_level(true), _min_level(INT_MAX),
	_max_level(INT_MIN), _num_sources(0), _num_sinks(0), _nodes(list<node>())
{
	_level       .init(*this,0);
	_pos_in_nodes.init(*this,nil);
}


LeveledGraph::LeveledGraph(const LeveledGraph &LG) : graph(), _set_level(true),
	_nodes(list<node>())
{
	_init(LG);
}


void LeveledGraph::_init(const LeveledGraph &LG)
{
	_max_level   = LG._max_level;
	_min_level   = LG._min_level;
	_num_sources = LG._num_sources;
	_num_sinks   = LG._num_sinks;

	_level       .init(*this,0);
	_pos_in_nodes.init(*this,nil);

	node v, v_LG;
	for (v = first_node(), v_LG = LG.first_node(); v != nil;
		v = succ_node(v), v_LG = LG.succ_node(v_LG))
	{
		_pos_in_nodes[v] = _nodes[_level[v] = LG._level[v_LG]].append(v);
	}
}


LeveledGraph& LeveledGraph::operator=(const LeveledGraph &LG)
{
	graph::operator=(LG);
	_nodes = map<int,list<node> >(list<node>());
	_init(LG);

	return *this;
}


node LeveledGraph::new_node (int i)
{
	_set_level = false;
	node v = new_node();
	_set_level = true;
	_pos_in_nodes[v] = _nodes[_level[v]=i].append(v);

	if (i > _max_level)
		_max_level = i;

	if (i < _min_level)
		_min_level = i;

	return v;
}


void LeveledGraph::post_new_node_handler(node v)
{
	if (_set_level) {
		_pos_in_nodes[v] = _nodes[_level[v]=0].append(v);
		if (_max_level < 0)
			_max_level = 0;
		if (0 < _min_level)
			_min_level = 0;
	}
	_num_sources++;
	_num_sinks++;
}


void LeveledGraph::pre_del_node_handler(node v)
{
	_nodes[_level[v]].del_item(_pos_in_nodes[v]);

	_num_sources--; // if v gets deleted all incident edges get deleted
	_num_sinks--;   // before, so v is a sink and a source
}

void LeveledGraph::pre_new_edge_handler(node v, node w)
{
	if (outdeg(v) == 0) _num_sinks--;
	if (indeg (w) == 0) _num_sources--;
}

void LeveledGraph::post_del_edge_handler(node v, node w)
{
	if (outdeg(v) == 0) _num_sinks++;
	if (indeg (w) == 0) _num_sources++;
}

void LeveledGraph::pre_restore_edge_handler(edge e)
{
	if (outdeg(source(e)) == 0) _num_sinks--;
	if (indeg (target(e)) == 0) _num_sources--;
}

void LeveledGraph::post_hide_edge_handler(edge e)
{
	if (outdeg(source(e)) == 0) _num_sinks++;
	if (indeg (target(e)) == 0) _num_sources++;
}

void LeveledGraph::pre_move_edge_handler(edge e, node v, node w)
{
	node v1 = source(e), w1 = target(e);

	if (v1 != v) {
		if (outdeg(v1) == 1) _num_sinks++;
		if (outdeg(v)  == 0) _num_sinks--;
	}
	if (w1 != w) {
		if (indeg(w1) == 1) _num_sources++;
		if (indeg(w)  == 0) _num_sources--;
	}
}

void LeveledGraph::level(node v, int l_new)
{
	int l_old = _level[v];

	if (l_new != l_old) {
		_nodes[l_old].del_item(_pos_in_nodes[v]);
		_pos_in_nodes[v] = _nodes[_level[v]=l_new].append(v);

		if (l_new > _max_level)
			_max_level = l_new;

		if (l_new < _min_level)
			_min_level = l_new;
	}
}

void LeveledGraph::invert()
{
	const int sum = _min_level + _max_level;

	node v;
	forall_nodes(v,*this)
		_level[v] = sum - _level[v];

	for (int i = _min_level; i <= sum/2; i++)
		_nodes[i].swap(_nodes[sum-i]);

	rev_all_edges();
}
