/*******************************************************************************
+
+  AGD 1.2
+
+
+  layout.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                           layout.cpp
//
// implements concepts and layout algorithm base classes
//
//    author: Carsten Gutwenger
//    date:   january 2000
//-------------------------------------------------------------------


#include <AGD/GridLayoutModule.h>
#include <AGD/ConstLedaGraphConcept.h>
#include <AGD/ConstLedaGraphGridConcept.h>
#include <AGD/GraphCopyConcept.h>
#include <AGD/GraphCopyGridConcept.h>
#include <AGD/PlaneGraphCopyGridConcept.h>
#include <AGD/GraphGLI.h>
#include <AGD/GraphCopyGLI.h>
#include <AGD/PlaneGraphCopy.h>
#include <AGD/GraphCopyLI.h>
#include <AGD/MaxAllMapper.h>

using namespace leda;


//-------------------------------------------------------------------
//                        ConstLedaGraphConcept
//-------------------------------------------------------------------

void ConstLedaGraphConcept::update()
{
	_lip->update();
}


//-------------------------------------------------------------------
//                       ConstLedaGraphGridConcept
//-------------------------------------------------------------------

ConstLedaGraphGridConcept::ConstLedaGraphGridConcept (const graph &G,
	LayoutInterface &) : _cgp(&G), new_layout(true)
{
	_glp = new GridLayout(G);
}
	

ConstLedaGraphGridConcept::~ConstLedaGraphGridConcept()
{
	if (new_layout) delete _glp;
}

void ConstLedaGraphGridConcept::get_node_grid_size (GridCoordinateMapper &gcm,
	const LayoutInterface &A)
{
	GraphGLI gli(*_glp);
	gcm.get_node_grid_size (gli, A);
}

void ConstLedaGraphGridConcept::update (GridCoordinateMapper &gcm,
	LayoutInterface &A)
{
	GraphGLI gli(*_glp);
	gcm.call (gli, A);
}

void ConstLedaGraphGridConcept::update (GridLayout &)
{
	_glp->compact_all_bends();
}


//-------------------------------------------------------------------
//                          GraphCopyConcept
//-------------------------------------------------------------------

GraphCopyConcept::GraphCopyConcept (const graph &G, LayoutInterface &A)
{
	_gcp   = new GraphCopy(G);
	_gclip = new GraphCopyLI(*_gcp,A);
}
	

GraphCopyConcept::~GraphCopyConcept()
{
	delete _gcp;
	delete _gclip;
}


void GraphCopyConcept::update()
{
	_gclip->update();
}


//-------------------------------------------------------------------
//                         GridConceptBase
//-------------------------------------------------------------------

void GridConceptBase::set_zero_size()
{
	_glp->init_node_size(0,0);
}

void GridConceptBase::set_bbox (const IRect &bbox)
{
	_bbox = bbox;
	_bbox_set = true;
}

const IRect &GridConceptBase::get_bbox()
{
	if (!_bbox_set) {
		_bbox = _glp->bounding_box();
		_bbox_set = true;
	}
	return _bbox;
}


//-------------------------------------------------------------------
//                        GCGridConceptBase
//-------------------------------------------------------------------

IPolyline GCGridConceptBase::get_bends (edge e) const
{
	IPolyline ipl;
	IPoint ip;
	const list<edge>& ch = _gcp->chain(e);
	const node_map<int> &x = _glp->x();
	const node_map<int> &y = _glp->y();
	const edge_map<IPolyline> &bends = _glp->bends();
	edge e1;
	node v = _gcp->copy(source(e)), w = _gcp->copy(target(e));

	if (ch.empty()) {
		error_handler(0, "get_bends: chain(e) is empty!");
		return ipl;
	}

	forall (e1, ch) {
		ipl.append(IPoint(x[source(e1)],y[source(e1)]));

		if (!bends[e1].empty()) {
			forall (ip, bends[e1])
				ipl.append(ip);
		} else if (_gcp->is_map()) {
			forall_rev (ip, bends[_gcp->reversal(e1)])
				ipl.append(ip);
		}
	}
	ipl.append(IPoint(x[target(e1)],y[target(e1)]));

	if (_gcp->is_reversed(e))
		ipl.reverse_items();

	if (_glp->in_rect(ip = ipl[ipl.succ(ipl.first())], v))
		ipl.pop_front();

	if (ipl.length() < 2)
		error_handler(0, "get_bends: nodes are overlapping!");
	else {
		if (_glp->in_rect(ip = ipl[ipl.pred(ipl.last())], w))
			ipl.pop_back();
	}
	
	GridLayout::compact_bends(ipl);

	if (!ipl.empty() && ipl.head() == IPoint(x[v],y[v]))
		ipl.pop_front();
	if (!ipl.empty() && ipl.tail() == IPoint(x[w],y[w]))
		ipl.pop_back();

	return ipl;

}

void GCGridConceptBase::transform (GridLayout &dest)
{
	const graph& G = _gcp->original();
	dest = GridLayout(G);

	node v;
	forall_nodes (v,G) {
		node vc = _gcp->copy(v);
		dest.x(v) = _glp->x(vc);
		dest.y(v) = _glp->y(vc);
		dest.width (v, _glp->width (vc));
		dest.height(v, _glp->height(vc));
	}

	edge e;
	forall_edges (e,G) {
		dest.bends(e) = get_bends(e);
	}
}

void GCGridConceptBase::get_node_grid_size(const GridLayout &gl)
{
	if (_glp == &gl) return;

	const graph &G = _gcp->original();

	node v;

	if (_pgp == nil) {
		forall_nodes(v,G) {
			_glp->width (_gcp->copy(v),gl.width (v));
			_glp->height(_gcp->copy(v),gl.height(v));
		}
	} else {
		forall_nodes(v,G) {
			_glp->width (_gcp->copy(v),gl.width (_pgp->copy(v)));
			_glp->height(_gcp->copy(v),gl.height(_pgp->copy(v)));
		}
	}
}


//-------------------------------------------------------------------
//                       GraphCopyGridConcept
//-------------------------------------------------------------------

GraphCopyGridConcept::GraphCopyGridConcept (const graph &G,
	LayoutInterface &)
{
	_gcp = new GraphCopy(G);
	_glp = new GridLayout(*_gcp);
}
	
GraphCopyGridConcept::GraphCopyGridConcept (const graph &G,
	GridLayout &)
{
	_gcp = new GraphCopy(G);
	_glp = new GridLayout(*_gcp);
}


GraphCopyGridConcept::~GraphCopyGridConcept()
{
	delete _gcp;
	delete _glp;
}

void GraphCopyGridConcept::get_node_grid_size (GridCoordinateMapper &gcm,
	const LayoutInterface &A)
{
	GraphCopyGLI gli(*this);
	gcm.get_node_grid_size (gli, A);
}

void GraphCopyGridConcept::update (GridCoordinateMapper &gcm,
	LayoutInterface &A)
{
	GraphCopyGLI gli(*this);
	gcm.call (gli, A);
}

void GraphCopyGridConcept::update (GridLayout &gl)
{
	transform (gl);
}


//-------------------------------------------------------------------
//                       PlaneGraphCopyGridConcept
//-------------------------------------------------------------------

PlaneGraphCopyGridConcept::PlaneGraphCopyGridConcept (const graph &G,
	LayoutInterface &) : _ext_face(nil), new_graph(true), new_layout(true)
{
	_gcp = _pgcp = new PlaneGraphCopy(G);
	_glp = new GridLayout(*_pgcp);
}
	
PlaneGraphCopyGridConcept::PlaneGraphCopyGridConcept (const graph &G,
	face f, LayoutInterface &) : new_graph(true), new_layout(true)
{
	_gcp = _pgcp = new PlaneGraphCopy(G);
	_glp = new GridLayout(*_pgcp);
	_ext_face = (f != nil) ? _pgcp->copy_of_face(f) : nil;
}

PlaneGraphCopyGridConcept::PlaneGraphCopyGridConcept (PlaneGraphCopy &PG,
	face f, LayoutInterface &) : _pgcp(&PG), _ext_face(f),
	new_graph(false), new_layout(true)
{
	_gcp = _pgcp;
	_glp = new GridLayout(PG);
}

PlaneGraphCopyGridConcept::PlaneGraphCopyGridConcept (const graph &G,
	GridLayout &) : _ext_face(nil), new_graph(true), new_layout(true)
{
	_gcp = _pgcp = new PlaneGraphCopy(G);
	_glp = new GridLayout(*_pgcp);
}

PlaneGraphCopyGridConcept::PlaneGraphCopyGridConcept (const graph &G,
	face f, GridLayout &) : new_graph(true), new_layout(true)
{
	_gcp = _pgcp = new PlaneGraphCopy(G);
	_glp = new GridLayout(*_pgcp);
	_ext_face = (f != nil) ? _pgcp->copy_of_face(f) : nil;
}

PlaneGraphCopyGridConcept::PlaneGraphCopyGridConcept (PlaneGraphCopy &PG,
	face f, GridLayout &gl) : _pgcp(&PG), _ext_face(f),
	new_graph(false), new_layout(false)
{
	_gcp = _pgcp;
	_glp = &gl;
	gl = GridLayout(PG);
}


PlaneGraphCopyGridConcept::~PlaneGraphCopyGridConcept()
{
	if (new_graph)  delete _pgcp;
	if (new_layout) delete _glp;
}


void PlaneGraphCopyGridConcept::get_node_grid_size (GridCoordinateMapper &gcm,
	const LayoutInterface &A)
{
	if (new_graph) {
		GraphCopyGLI gli(*this);
		gcm.get_node_grid_size (gli, A);
	} else {
		GraphGLI gli(*_glp);
		gcm.get_node_grid_size (gli, A);
	}
}

void PlaneGraphCopyGridConcept::update (GridCoordinateMapper &gcm,
	LayoutInterface &A)
{
	if (new_graph) {
		GraphCopyGLI gli(*this);
		gcm.call (gli, A);
	} else {
		GraphGLI gli(*_glp);
		gcm.call (gli, A);
	}
}

void PlaneGraphCopyGridConcept::update (GridLayout &gl)
{
	if (!new_layout) return;

	transform (gl);
}


void PlaneGraphCopyGridConcept::set_postcondition(list<AgdKey> &post)
{
	post.append(key::planar_map_callable);
	post.append(key::PlaneGraphCopy_callable);
}

//-------------------------------------------------------------------
//                           LayoutModule
//-------------------------------------------------------------------

const int LayoutModule::in_graph = 1;
const int LayoutModule::out_layout = 1;

bool LayoutModule::check (const leda::graph& G, AgdKey &p) const
{
	return check_graph (in_graph, G, p);
}


//-------------------------------------------------------------------
//                         GridLayoutModule
//-------------------------------------------------------------------

GridLayoutModule::GridLayoutModule ()
{
	mapper.init(PreCond(),PostCond(),MaxAllMapper());
}

GridLayoutModule::GridLayoutModule (const PreCond &P, const PostCond &Q,
	const GridCoordinateMapper &M)
{
	mapper.init(P,Q,M);
}
