/*******************************************************************************
+
+  AGD 1.2
+
+
+  mappers.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                       MaxAllMapper.cpp
//
//    author: Carsten Gutwenger
//    date:   january 1999
//-------------------------------------------------------------------


#include <AGD/MaxAllMapper.h>
#include <AGD/MaxEachMapper.h>
#include <AGD/FixedGridMapper.h>
#include <AGD/FixedSizeMapper.h>
#include <AGD/LayoutInterface.h>

#include <LEDA/internal/std/math.h>

using namespace leda;


//-------------------------------------------------------------------
//                          MaxAllMapper
//-------------------------------------------------------------------

void MaxAllMapper::init (LayoutInterface &A)
{
	const graph& G = get_graph();
	double w_min = 0;

	node v;
	forall_nodes(v,G) {
		w_min = leda::max(w_min, (A.get_width (v)+_sep) / (grid_width (v)+1));
		w_min = leda::max(w_min, (A.get_height(v)+_sep) / (grid_height(v)+1));
	}

	increase_nodes(w_min, A);

	IRect bb = get_gbbox();

	ax = w_min;
	bx = _origin.xcoord() - w_min * bb.left();
	ay = w_min;
	by = _origin.ycoord() - w_min * bb.bottom();
}

void MaxAllMapper::increase_nodes(double w_min, LayoutInterface &A)
{
	const graph& G = get_graph();
	double beta = 2*_min_over;

	node v;
	forall_nodes(v,G) {
		A.set_width (v, leda::max(A.get_width (v),beta + grid_width (v)*w_min));
		A.set_height(v, leda::max(A.get_height(v),beta + grid_height(v)*w_min));
	}
}

void MaxAllMapper::do_get_node_grid_size(const LayoutInterface &A)
{
	if (_w <= 0) return; // illegal setting

	const graph& G = get_graph();

	node v;
	forall_nodes(v,G) {
		int gw = (int)(ceil((A.get_width(v)  + _sep)/_w - 1));
		if (gw > 0) grid_width(v,gw);

		int gh = (int)(ceil((A.get_height(v) + _sep)/_w - 1));
		if (gh > 0) grid_height(v,gh);
	}
}


//-------------------------------------------------------------------
//                        FixedGridMapper
//-------------------------------------------------------------------

void FixedGridMapper::init (LayoutInterface &A)
{
	const graph& G = get_graph();
	double beta_x = 2 * _min_over_x, beta_y = 2*_min_over_y;

	IRect bb = get_gbbox();
	xmin = bb.left();
	ymin = bb.bottom();

	node v;
	forall_nodes(v,G) {
		A.set_width (v, leda::max(A.get_width (v),beta_x + grid_width (v)*_wx));
		A.set_height(v, leda::max(A.get_height(v),beta_y + grid_height(v)*_wy));
	}
}

void FixedGridMapper::do_get_node_grid_size(const LayoutInterface &A)
{
	if (_wx <= 0 || _wy <= 0) return; // illegal setting

	const graph& G = get_graph();

	node v;
	forall_nodes(v,G) {
		int gw = (int)(ceil((A.get_width(v)  + _sep_x)/_wx - 1));
		if (gw > 0) grid_width(v,gw);

		int gh = (int)(ceil((A.get_height(v) + _sep_y)/_wy - 1));
		if (gh > 0) grid_height(v,gh);
	}
}


//-------------------------------------------------------------------
//                        FixedSizeMapper
//-------------------------------------------------------------------

void FixedSizeMapper::init (LayoutInterface &A)
{
	const graph& G = get_graph();
	double beta_x = 2 * _min_over_x, beta_y = 2*_min_over_y;

	IRect bb = get_gbbox();
	xmin = bb.left();
	ymin = bb.bottom();

	node v;
	forall_nodes(v,G) {
		double x, w = A.get_width (v);
		int gw = grid_width (v);
		if (w < (x = gw * _wx + beta_x)) A.set_width(v,x);
		else if (w > (x = (1+gw) * _wx - _sep_x)) A.set_width(v,x);

		double y, h = A.get_height (v);
		int gh = grid_height (v);
		if (h < (y = gh * _wy + beta_y)) A.set_height(v,y);
		else if (h > (y = (1+gh) * _wy - _sep_y)) A.set_height(v,y);
	}
}

void FixedSizeMapper::do_get_node_grid_size(const LayoutInterface &A)
{
	if (_wx <= 0 || _wy <= 0) return; // illegal setting

	const graph& G = get_graph();

	node v;
	forall_nodes(v,G) {
		int gw = (int)(ceil((A.get_width(v)  + _sep_x)/_wx - 1));
		if (gw > 0) grid_width(v,gw);

		int gh = (int)(ceil((A.get_height(v) + _sep_y)/_wy - 1));
		if (gh > 0) grid_height(v,gh);
	}
}


//-------------------------------------------------------------------
//                        MaxEachMapper
//-------------------------------------------------------------------

static MaxEachMapper *M;
static int f1(const node &v)
{
	return M->get_grid_width(v);
}
static int f2(const node &v)
{
	return M->get_grid_height(v);
}

void MaxEachMapper::init (LayoutInterface &A)
{
	const graph& G = get_graph();
	IRect bbox = get_gbbox();
	int i;
	node v;

	// compute width of columns
	const int xmin = bbox.left(), xmax = bbox.right();

	wc = array<double>(xmin,xmax);
	for (i = xmin; i <= xmax; i++)
		wc[i] = _sep_x + 2*_min_over_x;

	list<node> nodes = G.all_nodes();
	M = this;
	nodes.bucket_sort(0,xmax-xmin,f1);

	while(!nodes.empty()) {
		v = nodes.pop();
		const int xl = column(v), xr = column(v)+grid_width(v);

		double d = A.get_width(v) + _sep_x;
		for (i = xl; i <= xr; i++)
			d -= wc[i];

		if (d > 0)
			increase(wc,xl,xr,d);
	}

	// compute width of rows
	const int ymin = bbox.bottom(), ymax = bbox.top();
	wr = array<double>(ymin,ymax);
	for (i = ymin; i <= ymax; i++)
		wr[i] = _sep_y + 2*_min_over_y;

	nodes = G.all_nodes();
	nodes.bucket_sort(0,ymax-ymin,f2);

	while(!nodes.empty()) {
		v = nodes.pop();
		const int yb = row(v), yt = row(v)+grid_height(v);

		double d = A.get_height(v) + _sep_y;
		for (i = yb; i <= yt; i++)
			d -= wr[i];

		if (d > 0)
			increase(wr,yb,yt,d);
	}

	// enlarge nodes if necessary
	forall_nodes(v,G) {
		const int xl = column(v), xr = column(v)+grid_width(v);
		double w_min = 2*_min_over_x;
		for (i = xl+1; i < xr; i++)
			w_min += wc[i];
		if (xl < xr)
			w_min += 0.5*(wc[xl]+wc[xr]);

		A.set_width (v, leda::max(A.get_width (v), w_min));

		const int yb = row(v), yt = row(v)+grid_height(v);
		w_min = 2*_min_over_y;
		for (i = yb+1; i < yt; i++)
			w_min += wr[i];
		if (yb < yt)
			w_min += 0.5*(wr[yb]+wr[yt]);

		A.set_height (v, leda::max(A.get_height(v), w_min));
	}

	// compute coordinates for columns
	xpos = array<double>(xmin,xmax);
	xpos[xmin] = _origin.xcoord();

	for (i = xmin+1; i <= xmax; i++) {
		xpos[i] = xpos[i-1] + 0.5*(wc[i-1]+wc[i]);
	}

	// compute coordinates for rows
	ypos = array<double>(ymin,ymax);
	ypos[ymin] = _origin.ycoord();

	for (i = ymin+1; i <= ymax; i++) {
		ypos[i] = ypos[i-1] + 0.5*(wr[i-1]+wr[i]);
	}
}

DPoint MaxEachMapper::map_node (node v, LayoutInterface &A)
{
	double px, py;
	
	double w = A.get_width(v);
	int    xl = column(v), xr = xl+grid_width(v);
	double mx = 0.5 * (xpos[xl]+xpos[xr]);
	double left  = xpos[xl] + 0.5 * (_sep_x - wc[xl]);
	double right = xpos[xr] + 0.5 * (wc[xr] - _sep_x);

	if (w > (mx-left) * 2)
		px = left+w/2;
	else if (w > (right-mx)*2)
		px = right-w/2;
	else
		px = mx;

	double h = A.get_height(v);
	int    yb = row(v), yt = yb+grid_height(v);
	double my = 0.5 * (ypos[yb]+ypos[yt]);
	double low  = ypos[yb] + 0.5 * (_sep_y - wr[yb]);
	double high = ypos[yt] + 0.5 * (wr[yt] - _sep_y);

	if (h > (my-low) * 2)
		py = low+h/2;
	else if (h > (high-my)*2)
		py = high-h/2;
	else
		py = my;

	return DPoint(px,py);
}

void MaxEachMapper::cleanup(LayoutInterface &)
{
	wc = wr = xpos = ypos = array<double>();
}

int MaxEachMapper::cmp(const two_tuple<int,double> &a,const two_tuple<int,double> &b)
{
	return compare(a.second(), b.second());
}

void MaxEachMapper::increase(array<double> &w, int left, int right, double d)
{
	int i;
	list<two_tuple<int,double> > L;

	for (i = left; i <= right; i++)
		L.append(two_tuple<int,double>(i,w[i]));

	L.sort(cmp);

	list_item it;
	double delta, s = L.head().second();
	int k = 1;

	for (it = L.succ(L.first()); true; it = L.succ(it)) {
		if (it && (delta = (L[it].second() - s) * k) < d)
			d -= delta;
		else {
			s += d / k;
			break;
		}
		k++; s = L[it].second();
	}

	two_tuple<int,double> p;
	forall(p,L) {
		i = p.first();

		if (w[i] < s) w[i] = s;
		else break;
	}
}


void MaxEachMapper::do_get_node_grid_size(const LayoutInterface &A)
{
	if (_wx <= 0 || _wy <= 0) return; // illegal setting

	const graph& G = get_graph();

	node v;
	forall_nodes(v,G) {
		int gw = (int)(ceil((A.get_width(v)  + _sep_x)/_wx - 1));
		if (gw > 0) grid_width(v,gw);

		int gh = (int)(ceil((A.get_height(v) + _sep_y)/_wy - 1));
		if (gh > 0) grid_height(v,gh);
	}
}
