/*******************************************************************************
+
+  AGD 1.2
+
+
+  planarization.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        planarization.cpp
//
//
//    author: Carsten Gutwenger
//    date:   august 1997
//-------------------------------------------------------------------


#include <AGD/SubgraphPlanarizer.h>
#include <AGD/PlanarizationLayout.h>
#include <AGD/PlanarizationGridLayout.h>
#include <AGD/ShortestPathInserter.h>
#include <AGD/QuasiOrthogonalLayout.h>
#include <AGD/PlanarSubgraph.h>
#include <AGD/LEDAMakeAcyclic.h>
#include <AGD/PlaneGraphCopy.h>
#include <AGD/GraphCopyLI.h>
#include <AGD/GraphGLI.h>
#include <AGD/GraphCopyGLI.h>

#include <LEDA/graph/node_set.h>

using namespace leda;

//-------------------------------------------------------------------
//                        PlanarizerModule
//-------------------------------------------------------------------

const int PlanarizerModule::in_graph = 1;
const int PlanarizerModule::out_planarized_graph = 1;

bool PlanarizerModule::check (const GraphCopy& G, AgdKey& p) const
{
	return check_graph (in_graph, G, p);
}

void PlanarizerModule::call (const graph& G, PlaneGraphCopy& PG)
{
	do_call (G, PG);
	PG.compute_faces();
}


//-------------------------------------------------------------------
//                          LEDAMakeAcyclic
//-------------------------------------------------------------------

LEDAMakeAcyclic::LEDAMakeAcyclic()
{
	add_precondition (key::directed);
	add_post_rule (key::directed);
	add_post_rule (key::maximal_acyclic);
}

void LEDAMakeAcyclic::do_call (const graph& G, list<edge>& feedback)
{
	Is_Acyclic(G,feedback);
}

AGDModule *LEDAMakeAcyclic::clone() const
{
	return new LEDAMakeAcyclic;
}


//-------------------------------------------------------------------
//                           SubgraphModule
//-------------------------------------------------------------------

const int SubgraphModule::in_graph = 1;
const int SubgraphModule::out_subgraph = 1;

bool SubgraphModule::check (const graph& G, AgdKey &p) const
{
	return check_graph (in_graph, G, p);
}

void SubgraphModule::call (const graph& G, list<edge>& L)
{
	L.clear();

	do_call (G, L);

	_num_del_edges = L.length();
}

void SubgraphModule::call (graph& G)
{
	list<edge> L;
	edge e;

	do_call (G, L);

	_num_del_edges = L.length();
	forall (e, L) G.del_edge(e);
}

void SubgraphModule::call (GraphCopy& G, list<edge>& L)
{
	L.clear();

	do_call (G, L);

	_num_del_edges = L.length();

	list_item it;
	forall_items (it, L) {
		edge e = L[it];
		L[it] = G.original(e);
		G.del_edge(e);
	}
}


//-------------------------------------------------------------------
//                          SubgraphPlanarizer
//-------------------------------------------------------------------

SubgraphPlanarizer::SubgraphPlanarizer()
{
	add_precondition (key::no_self_loops);

	add_post_rule (key::planar);
	add_post_rule (key::simple, 1, key::simple);
	add_post_rule (key::no_self_loops);

	// initialize subgraph
	PreCond  PRE;
	PostCond POST;

	PRE [1] << key::no_self_loops
		<< key::no_reversals;
	POST[1] << key::planar << key::no_self_loops
		<< key::no_reversals;
	subgraph.init (PRE,POST,PlanarSubgraph());

	// initialize inserter
	PreCond  PRE_I(2);
	PostCond POST_I;
	inserter.init (PRE_I,POST_I,ShortestPathInserter());
}

AGDModule *SubgraphPlanarizer::clone () const
{
	SubgraphPlanarizer *M = new SubgraphPlanarizer;

	M->set_subgraph (subgraph.get());
	M->set_inserter (inserter.get());

	return M;
}

void SubgraphPlanarizer::do_call (const graph& G, PlaneGraphCopy& PG)
{
	list<edge> L;
	float T;

	used_time(T);
	subgraph.get().call (G, L);
	m_timeSubgraph = used_time(T);
	m_nEdges = L.length();

	PG = PlaneGraphCopy(G, L);
	used_time(T);
	inserter.get().call (PG, L);
	m_timeInsertion = used_time(T);
}


//-------------------------------------------------------------------
//                         PlanarizationLayout
//-------------------------------------------------------------------

PlanarizationLayout::PlanarizationLayout()
{
	add_precondition (key::no_self_loops);

	//
	// initialize used modules
	//

	PreCond  PRE;
	PostCond POST;

	PRE [1] << key::no_self_loops << key::directed;
	POST[1] << key::planar << key::no_self_loops;
	planarizer.init (PRE,POST,SubgraphPlanarizer());

	PRE[1].clear(); POST[1].clear();

	PRE [1] << key::simple << key::no_self_loops << key::planar << key::four_graph << key::connected;
	POST[0] << key::PlaneGraphCopy_callable;
	POST[1] << key::no_crossings;
	planar_layouter.init (PRE,POST,QuasiOrthogonalLayout());
}

AGDModule *PlanarizationLayout::clone () const
{
	PlanarizationLayout *M = new PlanarizationLayout;

	M->set_planarizer (planarizer.get());
	M->set_planar_layouter (planar_layouter.get());

	return M;
}

bool PlanarizationLayout::set_planar_layouter (const LayoutModule& M)
{
	if (planar_layouter.set(M)) {
		if (M.precondition(1).member(key::connected))
			add_precondition(key::connected);
		else
			rem_precondition(key::connected);
		if (M.precondition(1).member(key::four_graph))
			add_precondition(key::four_graph);
		else
			rem_precondition(key::four_graph);
		if (M.precondition(1).member(key::simple))
			add_precondition(key::simple);
		else
			rem_precondition(key::simple);
		return true;
	} else return false;
}

void PlanarizationLayout::do_call (ConstLedaGraphConcept &C)
{
	PlaneGraphCopy PG;

	planarizer.get().call (C.get_graph(), PG);
	_no_crossings = PG.number_of_nodes() - C.get_graph().number_of_nodes();

	GraphCopyLI B_PG(PG,C.get_layout());

	planar_layouter.get().call (PG, nil, B_PG);
}


//-------------------------------------------------------------------
//                         PlanGridConcept
//-------------------------------------------------------------------

PlanGridConcept::PlanGridConcept (const graph &G, LayoutInterface &) :
	_cgp(&G), _ext_face(nil), new_graph(true), new_layout(true), _is_planar(false)
{
	_gcp = new PlaneGraphCopy;
	_glp = new GridLayout(*_gcp);
}
	
PlanGridConcept::PlanGridConcept (const graph &G, face f,
	LayoutInterface &) : new_graph(true), new_layout(true), _is_planar(true)
{
	PlaneGraphCopy *pgcp = new PlaneGraphCopy(G);
	_gcp = pgcp;
	_glp = new GridLayout(*_gcp);
	_ext_face = (f != nil) ? pgcp->copy_of_face(f) : nil;
}

PlanGridConcept::PlanGridConcept (PlaneGraphCopy &PG, face f,
	LayoutInterface &) : _ext_face(f), new_graph(false), new_layout(true), _is_planar(true)
{
	_gcp = &PG;
	_glp = new GridLayout(*_gcp);
}

PlanGridConcept::PlanGridConcept (const graph &G, GridLayout &) :
	_cgp(&G), _ext_face(nil), new_graph(true), new_layout(true), _is_planar(false)
{
	_gcp = new PlaneGraphCopy;
	_glp = new GridLayout(*_gcp);
}

PlanGridConcept::PlanGridConcept (const graph &G, face f, GridLayout &) :
	new_graph(true), new_layout(true), _is_planar(true)
{
	PlaneGraphCopy *pgcp = new PlaneGraphCopy(G);
	_gcp = pgcp;
	_glp = new GridLayout(*_gcp);
	_ext_face = (f != nil) ? pgcp->copy_of_face(f) : nil;
}

PlanGridConcept::PlanGridConcept (PlaneGraphCopy &PG, face f,
	GridLayout &gl) : _ext_face(f), new_graph(false), new_layout(false), _is_planar(true)
{
	_gcp = &PG;
	_glp = &gl;
	gl = GridLayout(PG);
}

PlanGridConcept::~PlanGridConcept()
{
	if (new_graph)  delete _gcp;
	if (new_layout) delete _glp;
}

void PlanGridConcept::update (GridCoordinateMapper &gcm,
	LayoutInterface &A)
{
	if (new_graph) {
		GraphCopyGLI gli(*this);
		gcm.call (gli, A);
	} else {
		GraphGLI gli(*_glp);
		gcm.call (gli, A);
	}
}

void PlanGridConcept::update (GridLayout &gl)
{
	if (!new_layout) return;

	transform (gl);
}

void PlanGridConcept::set_postcondition(list<AgdKey> &post)
{
	post.append(key::planar_map_callable);
	post.append(key::PlaneGraphCopy_callable);
}


//-------------------------------------------------------------------
//                      PlanarizationGridLayout
//-------------------------------------------------------------------

PlanarizationGridLayout::PlanarizationGridLayout()
{
	add_precondition (key::no_self_loops);

	//
	// initialize used modules
	//

	PreCond  PRE;
	PostCond POST;

	PRE [1] << key::no_self_loops << key::directed;
	POST[1] << key::planar << key::no_self_loops;
	planarizer.init (PRE,POST,SubgraphPlanarizer());

	PRE[1].clear(); POST[1].clear();

	PRE [1] << key::simple << key::no_self_loops << key::planar << key::four_graph << key::connected;
	POST[0] << key::PlaneGraphCopy_callable;
	POST[1] << key::no_crossings;
	planar_layouter.init (PRE,POST,QuasiOrthogonalLayout());
}

AGDModule *PlanarizationGridLayout::clone () const
{
	PlanarizationGridLayout *M = new PlanarizationGridLayout;

	M->set_planarizer (planarizer.get());
	M->set_planar_layouter (planar_layouter.get());
	M->set_mapper(mapper.get());

	return M;
}

bool PlanarizationGridLayout::set_planar_layouter (const GridLayoutModule& M)
{
	if (planar_layouter.set(M)) {
		if (M.precondition(1).member(key::connected))
			add_precondition(key::connected);
		else
			rem_precondition(key::connected);
		if (M.precondition(1).member(key::four_graph))
			add_precondition(key::four_graph);
		else
			rem_precondition(key::four_graph);
		if (M.precondition(1).member(key::simple))
			add_precondition(key::simple);
		else
			rem_precondition(key::simple);
		return true;
	} else return false;
}

void PlanarizationGridLayout::do_call (PlanGridConcept &C)
{
	PlaneGraphCopy &PG = C.get_graph();

	if (!C.is_planar()) {
		planarizer.get().call (C.get_const_graph(), PG);
		_no_crossings = PG.number_of_nodes() - C.get_const_graph().number_of_nodes();
	} else
		_no_crossings = 0;

	planar_layouter.get().call_grid (C.get_graph(), C.get_exterior_face(), C.get_layout());
	C.set_bbox (planar_layouter.get().bounding_box());
}


//-------------------------------------------------------------------
//                         EdgeInsertionModule
//-------------------------------------------------------------------

const int EdgeInsertionModule::in_planar_map = 1;
const int EdgeInsertionModule::in_edge_list = 2;
const int EdgeInsertionModule::out_planarized_graph = 1;

/*
//-------------------------------------------------------------------
//                         ShortestPathInserter
//-------------------------------------------------------------------

void ShortestPathInserter::do_call (PlaneGraphCopy& PG, list<edge>& e_insert)
{
	edge e;

	//construct dual graph
	//--------------------

	GRAPH<int,edge> LG_dual; //edge-parameter is the corresponding edge in the primal
	edge_map<node> left_face(PG); //every edge in the primal graph knows its
                                      //left face
	face f;
	//construct nodes
	forall_faces(f,PG) {
		node dual_node=LG_dual.new_node(0); //for any face there is a node in the dual
    
		forall_face_edges(e,f) {
			left_face[e]=dual_node;  //edge in primal has pointer to that node
		}
	}

	//insert edges in the dual graph
	forall_edges(e,PG) 
		LG_dual.new_edge(left_face[e],left_face[PG.reversal(e)],e); 
  

	while (!e_insert.empty()) {

		edge current_edge = e_insert.pop();
		node src = PG.copy(source(current_edge)),
			 tgt = PG.copy(target(current_edge));


		//find possible source nodes for shortest path in LG_dual
		//-------------------------------------------------------
		//node source=source_nodes.pop();
		list<node> shortest_sources;
		forall_out_edges(e,src) {
			if (left_face[e]!=left_face[PG.reversal(e)]) 
				shortest_sources.append(left_face[e]);

			else if (shortest_sources.empty()) shortest_sources.append(left_face[e]);
		}
     
		//find possible target nodes for shortest path in LG_dual
		//-------------------------------------------------------
		//node target=target_nodes.pop();
		list<node> shortest_targets;
		forall_out_edges(e,tgt) {
			if (left_face[e]!=left_face[PG.reversal(e)])
				shortest_targets.append(left_face[e]);

			else if (shortest_targets.empty()) shortest_targets.append(left_face[e]);
		}
  
		//compute shortest path between shortest_sources and shortest_targets
		//-------------------------------------------------------------------
		list<node> unscanned;
		node_array<bool> visited(LG_dual,false);
    
		node_array<edge> pred(LG_dual,nil);
		face common_face=nil;
		node target_found;
		bool path_found=false;

		node i;
		node_set T(LG_dual);
		forall(i,shortest_targets) {
			T.insert(i);
		}

		forall(i,shortest_sources) {
			visited[i]=true;
			unscanned.append(i);
			if (T.member(i)) {
				edge e_help=LG_dual.first_adj_edge(i);
				e_help=LG_dual[e_help];
				common_face=PG.face_of(e_help);
				path_found=true;
			}
		}

		while (!unscanned.empty()&&!path_found) {
			i=unscanned.pop();

			edge e;
			forall_out_edges(e,i) {
				node j=LG_dual.opposite(i,e);
				if (!visited[j]) {
					pred[j]=e;
					visited[j]=true;
					unscanned.append(j);
				}

				if (T.member(j)) {
					target_found=j;
					path_found=true;
				}
			}
		}

		list<edge> shortest_path;
		while (common_face==nil && pred[target_found]!=nil) {
			shortest_path.push(pred[target_found]);
			target_found=LG_dual.opposite(target_found,pred[target_found]);
		}   
    

		//for any crossing put a dummy node
		//---------------------------------
		edge e_first,e_last;
		list<edge> crossed;

		if (shortest_path.empty()) {
			edge e_help;
			forall_face_edges(e_help,common_face) {
				if (PG.source(e_help)==src) e_first=e_help;
				if (PG.source(e_help)==tgt) e_last=e_help;
			}
		}

		else {
			edge e_help=LG_dual[shortest_path.head()];
			face f=PG.face_of(e_help);

			forall_face_edges(e_help,f) {
				if (PG.source(e_help)==src) e_first=e_help;
			}
      
			e_help=PG.reversal(LG_dual[shortest_path.tail()]);
			f=PG.face_of(e_help);

			forall_face_edges(e_help,f) {
				if (PG.source(e_help)==tgt) e_last=e_help;
			}
		}

		edge e;
		forall(e,shortest_path) {
			crossed.append(LG_dual[e]); 
		}

		PG.new_edge_chain (current_edge, e_first, e_last, crossed);

		while (!shortest_path.empty())
			LG_dual.del_node(source(shortest_path.pop()));
		LG_dual.del_node(left_face[e_last]);

		forall (e, PG.chain(current_edge)) {
			edge e1;
			face f_l = PG.face_of(e),
				 f_r = PG.face_of(PG.reversal(e));
			node v_l = LG_dual.new_node(0),
				 v_r = LG_dual.new_node(0);

			forall_face_edges (e1, f_l)
				left_face[e1] = v_l;
			forall_face_edges (e1, f_r)
				left_face[e1] = v_r;
		}

		forall (e, PG.chain(current_edge)) {
			edge e1;
			face f_l = PG.face_of(e),
				 f_r = PG.face_of(PG.reversal(e));
			node v_l = left_face[PG.first_face_edge(f_l)],
				 v_r = left_face[PG.first_face_edge(f_r)];

			forall_face_edges (e1, f_l) {
				node v1 = left_face[PG.reversal(e1)];
				LG_dual.new_edge(v_l,v1,e1);
				if (v1 != v_r) LG_dual.new_edge(v1,v_l,PG.reversal(e1));
			}
			forall_face_edges (e1, f_r) {
				node v1 = left_face[PG.reversal(e1)];
				LG_dual.new_edge(v_r,v1,e1);
				if (v1 != v_l) LG_dual.new_edge(v1,v_r,PG.reversal(e1));
			}
		}
	}
}
*/
