/*******************************************************************************
+
+  AGD 1.2
+
+
+  rules.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         rules.cpp
//
//
//    author: Carsten Gutwenger
//    date:   january 1999
//-------------------------------------------------------------------


#include <AGD/DefaultRuleAdmin.h>
#include <AGD/TaggedRule.h>
#include <AGD/KeySet.h>

using namespace leda;


//-------------------------------------------------------------------
//                          DefaultRuleAdmin
//-------------------------------------------------------------------

DefaultRuleAdmin::DefaultRuleAdmin(const DefaultKeymapper &M) : RuleAdmin(M)
{
	// undirected -> no_self_loops & no reversals
	add_rule(M, key::undirected, key::no_self_loops);
	add_rule(M, key::undirected, key::no_reversals);

	// acyclic & single source -> connected
	add_rule(M, key::acyclic,key::single_source, key::connected);

	// acyclic -> directed & no self loops & no reversals
	add_rule(M, key::acyclic, key::directed);
	add_rule(M, key::acyclic, key::no_self_loops);
	add_rule(M, key::acyclic, key::no_reversals);

	// st-graph <-> acyclic & single source & single sink
	add_rule(M, key::st_graph, key::acyclic);
	add_rule(M, key::st_graph, key::single_source);
	add_rule(M, key::st_graph, key::single_sink);
	add_rule(M, key::acyclic, key::single_source, key::single_sink, key::st_graph);

	// st_planar -> st_graph & planar
	add_rule(M, key::st_planar, key::st_graph);
	add_rule(M, key::st_planar, key::planar);

	// tree <-> connected & forest
	add_rule(M, key::connected, key::forest, key::tree);
	add_rule(M, key::tree, key::connected);
	add_rule(M, key::tree,  key::forest);

	// tree <-> forest & single source
	add_rule(M, key::forest,key::single_source, key::tree);
	add_rule(M, key::tree,  key::single_source);

	// forest -> acyclic & directed & planar
	add_rule(M, key::forest, key::acyclic);
	add_rule(M, key::forest, key::directed);
	add_rule(M, key::forest, key::planar);

	// single source -> directed
	add_rule(M, key::single_source, key::directed);

	// single sink -> directed
	add_rule(M, key::single_sink, key::directed);

	// triconnected -> biconnected
	add_rule(M, key::triconnected, key::biconnected);

	// biconnected -> connected
	add_rule(M, key::biconnected, key::connected);

	// maximum planar -> maximal planar
	add_rule(M, key::maximum_planar, key::maximal_planar);

	// maximal planar -> planar
	add_rule(M, key::maximal_planar, key::planar);

	// maximum acyclic -> maximal acyclic
	add_rule(M, key::maximum_acyclic, key::maximal_acyclic);

	// maximal acyclic -> acyclic
	add_rule(M, key::maximal_acyclic, key::acyclic);

	// reduced acyclic -> acyclic
	add_rule (M, key::reduced_acyclic, key::acyclic);
}


void DefaultRuleAdmin::add_rule(const DefaultKeymapper &M,
	key::Predefined p1, key::Predefined q)
{
	add(Rule(M.predefined(p1), M.predefined(q)));
}

void DefaultRuleAdmin::add_rule(const DefaultKeymapper &M,
	key::Predefined p1, key::Predefined p2, key::Predefined q)
{
	add(Rule(M.predefined(p1), M.predefined(p2), M.predefined(q)));
}

void DefaultRuleAdmin::add_rule(const DefaultKeymapper &M,
	key::Predefined p1, key::Predefined p2, key::Predefined p3, key::Predefined q)
{
	add(Rule(M.predefined(q)).And(M.predefined(p1)).
		And(M.predefined(p2)).And(M.predefined(p3)));
}


//-------------------------------------------------------------------
//                            RuleAdmin
//-------------------------------------------------------------------

RuleAdmin::~RuleAdmin()
{
	Rule *r;

	forall(r,_rules)
		delete r;
}


void RuleAdmin::add(const Rule &r)
{
	Rule *rp = new Rule(r);
	rp->index(_rules.length());
	_rules.append(rp);

	const list<AgdKey> &L = rp->left();
	AgdKey k;
	forall(k,L)
		on_left[k.index()].append(rp);
}


bool RuleAdmin::implies(const KeySet &P, const KeySet &Q) const
{
	return Q <= closure(P);
}


KeySet RuleAdmin::closure(const KeySet &P) const
{
	KeySet Pc = P;
	array<bool> mark(_mapper->number_of_keys());
	array<int>  count(number_of_rules());

	mark.init(false);

	const_iterator it;
	for(it = begin(); it.valid(); ++it) {
		if ((count[it->index()] = it->length()) == 0)
			Pc.insert(it->right());
	}

	KeySet::const_iterator itp;
	for(itp = Pc.begin(); itp != Pc.end(); ++itp) {
		visit(*itp,mark,count);
	}

	Keymapper::const_iterator k_it;
	for (k_it = _mapper->begin(); k_it != _mapper->end(); ++k_it) {
		if (mark[(*k_it).index()]) Pc.insert(*k_it);
	}

	return Pc;
}


void RuleAdmin::visit(AgdKey k, array<bool> &mark, array<int> &count) const
{
	const int id = k.index();

	if (mark[id] == false) {
		mark[id] = true;

		const list<Rule *> &L = on_left[id];
		Rule *rp;
		forall(rp,L) {
			if ((--count[rp->index()]) == 0) {
				visit(rp->right(),mark,count);
			}
		}
	}
}


//-------------------------------------------------------------------
//                            Rule
//-------------------------------------------------------------------

ostream& operator<< (ostream& os, const Rule& r)
{
	bool first = true;
	AgdKey p;

	forall (p,r.left()) {
		if (first) first = false; else os << " & ";
		os << p;
	}
	os << " -> " << r.right();

	return os;
}


//-------------------------------------------------------------------
//                            TaggedRule
//-------------------------------------------------------------------

ostream& operator<< (ostream& os, const TaggedRule& r)
{
	bool first = true;
	TRLiteral l;

	forall (l,r.left()) {
		if (first) first = false; else os << " & ";
		os << l;
	}
	os << " -> " << r.right();

	return os;
}

