/*******************************************************************************
+
+  AGD 1.2
+
+
+  clusterGraph2.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



/*********************************************************************

  FILENAME  : clusterGraph.cc
  VERSION   : 1.0.0
  AUTHOR    : Christoph Liebel
  LANGUAGE  : C++

*********************************************************************/

#include <LEDA/graph/graph.h>
#include <LEDA/graph/node_array.h>
#include <LEDA/internal/std/string.h>

#include <AGD/impl/cplanarity/clusterGraph.h>
#include <AGD/impl/PQTree/pqStack.h>

using namespace leda;

/*****************************************************************************/
/*****************************              **********************************/
/***************************** clusterGraph **********************************/
/*****************************              **********************************/
/*****************************************************************************/

clusterGraph::clusterGraph()
    : basicGraph()
{
    clusterNumber = 0;
    tree = 0;
    c_Nodes = 0;
}


/*****************************************************************************/
/*****************************                             *******************/
/***************************** clusterGraph(clusterGraph&) *******************/
/*****************************                             *******************/
/*****************************************************************************/

clusterGraph::clusterGraph(const clusterGraph& cGraph)
{
    // Initialising the data of the base-class 'basicGraph'

    n = cGraph.n;
    m = cGraph.m;
    mSingle = cGraph.mSingle;
    adjacencyList = NULL;
    direction = NULL;
    weight = NULL;

    if (cGraph.adjacencyList)
        {
            adjacencyList = new basicList<outEdge>[n];
            for (int j = 0; j < n; j++)
                adjacencyList[j].append(&cGraph.adjacencyList[j]);
        }

    if (cGraph.direction)
        {
            direction = new int[m];
            for (int i = 0; i < m; i++)
                direction[i] = cGraph.direction[i];
        }

    if (cGraph.weight)
        {
            weight = new double[m];
            for (int i = 0; i < m; i++)
                weight[i] = cGraph.weight[i];
        }

    strcpy(graphName,cGraph.graphName);

    // Initialising the data that belongs to the class 'clusterGraph' and
    // not to the base-class 'basicGraph'

    c_Nodes = NULL;
    tree = NULL;
    clusterNumber = cGraph.clusterNumber;
    
    if (cGraph.c_Nodes)
        {
            c_Nodes = new linkedList<int>[clusterNumber];
            for (int j = 0; j < clusterNumber; j++)
                c_Nodes[j].append(&cGraph.c_Nodes[j]);
        }

    if (cGraph.tree)
        {
            tree = new inclusionTree;
            constructTree(n,0);
        }
}


/*****************************************************************************/
/*****************************               *********************************/
/***************************** ~clusterGraph *********************************/
/*****************************               *********************************/
/*****************************************************************************/

clusterGraph::~clusterGraph()
{
    deleteGraph();
}


/*****************************************************************************/
/*****************************                  ******************************/
/***************************** constructC_Graph ******************************/
/*****************************                  ******************************/
/*****************************************************************************/

int clusterGraph::constructC_Graph(const ClusterGraph& C)
{
    deleteGraph();

    strcpy(graphName, "(none)");
    m=C.number_of_edges();
    n=C.number_of_nodes();

    if (n)
        adjacencyList = new basicList<outEdge>[n];

    clusterNumber=1;
    pqStack<Cluster*> clusterStack;
    pqStack<int> clusterN;
    Cluster *cl1,*cl2;
    clusterStack.push(C.get_rootcluster());
    while(!clusterStack.stackEmpty()) {
        cl1=clusterStack.pop();
        forall(cl2,cl1->get_subclusters()) {
            clusterStack.push(cl2);
            clusterNumber++;
        }
    }
    
    if (clusterNumber)
        c_Nodes = new linkedList<int>[clusterNumber];
    
    node_array<int> nodeNumber(C);
    node i;
    int count=0,count1;
    forall_nodes(i,C) nodeNumber[i]=count++;

    edge edg;
    int source,target,edgeNumber=0;

    // Collect graph edges

    forall_edges(edg,C) {
        source=nodeNumber[C.source(edg)];
        target=nodeNumber[C.target(edg)];
        adjacencyList[source].in(outEdge(target, edgeNumber));
        if (target != source)
            adjacencyList[target].in(outEdge(source, edgeNumber));
        edgeNumber++;
    }

    // Collect tree edges by DFS

    count=n;
    clusterStack.push(C.get_rootcluster());
    clusterN.push(count++);
    while(!clusterStack.stackEmpty()&&!clusterN.stackEmpty()) {
        cl1=clusterStack.pop();
        count1=clusterN.pop();
        forall(cl2,cl1->get_subclusters()) {
            c_Nodes[count1-n].in(count);
            clusterStack.push(cl2);
            clusterN.push(count++);
        }
        forall(i,cl1->get_nodes()) c_Nodes[count1-n].in(nodeNumber[i]);
    }

    tree = new inclusionTree;
    constructTree(n, 0);

    countMulti();
    return 1;
}

/*****************************************************************************/
/*****************************               *********************************/
/***************************** constructTree *********************************/
/*****************************               *********************************/
/*****************************************************************************/

treeElement* clusterGraph::constructTree(int cNumber, int position)
{
    int elementNumber;
    treeElement* storeActual;

    tree->insert(cNumber, position);
    storeActual = tree->actual;  // stores the pointer of the inserted element
    position = 0;  // next element is inserted as child
    c_Nodes[cNumber - n].reset();

    while (c_Nodes[cNumber - n].out(elementNumber))
        {
            if (elementNumber >= 0 && elementNumber < n)
                // insertion as a leaf
                tree->insert(elementNumber, position);
            else if (elementNumber >= n && elementNumber < clusterNumber + n)
                // insertion as internal pqNode
                tree->actual = constructTree(elementNumber, position);
            else
                cout << "ERROR (clusterGraph::constructTree):"
                    << "\n\twrong inclusion tree construction!" << endl;

            position = 1;
        }
    return storeActual;
}


/*****************************************************************************/
/*****************************             ***********************************/
/***************************** deleteGraph ***********************************/
/*****************************             ***********************************/
/*****************************************************************************/

void clusterGraph::deleteGraph()
{
    if (adjacencyList)
        {
            delete [] adjacencyList;
            adjacencyList = NULL;
        }

    if (direction)
        {
            delete [] direction;
            direction = NULL;
        }

    if (weight)
        {
            delete [] weight;
            weight = NULL;
        }

    if (tree)
        {
            delete tree;
            tree = NULL;
        }

    if (c_Nodes)
        {
            delete [] c_Nodes;
            c_Nodes = NULL;
        }

    n = m = mSingle = clusterNumber = 0;

    strcpy(graphName, "deleted");
}


/*****************************************************************************/
/*****************************            ************************************/
/***************************** getAdjList ************************************/
/*****************************            ************************************/
/*****************************************************************************/

basicList<outEdge>* clusterGraph::getAdjList() const
{
    return adjacencyList;
}


/*****************************************************************************/
/*****************************                 *******************************/
/***************************** getAdjList(int) *******************************/
/*****************************                 *******************************/
/*****************************************************************************/

basicList<outEdge>* clusterGraph::getAdjList(int number) const
{
    return &adjacencyList[number];
}


/*****************************************************************************/
/*****************************            ************************************/
/***************************** getC_Nodes ************************************/
/*****************************            ************************************/
/*****************************************************************************/

linkedList<int>* clusterGraph::getC_Nodes() const
{
    return c_Nodes;
}


/*****************************************************************************/
/*****************************              **********************************/
/***************************** getGraphName **********************************/
/*****************************              **********************************/
/*****************************************************************************/

char* clusterGraph::getGraphName()
{
    return graphName;
}


/*****************************************************************************/
/*****************************      ******************************************/
/***************************** getM ******************************************/
/*****************************      ******************************************/
/*****************************************************************************/

int clusterGraph::getM() const
{
    return m;
}


/*****************************************************************************/
/*****************************      ******************************************/
/***************************** getN ******************************************/
/*****************************      ******************************************/
/*****************************************************************************/

int clusterGraph::getN() const
{
    return n;
}


/*****************************************************************************/
/*****************************         ***************************************/
/***************************** getTree ***************************************/
/*****************************         ***************************************/
/*****************************************************************************/

inclusionTree* clusterGraph::getTree() const
{
    return tree;
}


/*****************************************************************************/
/*****************************           *************************************/
/***************************** operator= *************************************/
/*****************************           *************************************/
/*****************************************************************************/

const clusterGraph& clusterGraph::operator=(const clusterGraph& cGraph)
{
    // check, if 'cGraph' is identical with the actual object

    if (this == &cGraph)
        return *this;

    // change the data of the base-class 'basicGraph'

    deleteGraph();

    n = cGraph.n;
    m = cGraph.m;
    mSingle = cGraph.mSingle;

/*    if (adjacencyList)
        {
            delete [] adjacencyList;
            adjacencyList = NULL;
        }*/

    if (cGraph.adjacencyList)
        {
            adjacencyList = new basicList<outEdge>[n];
            for (int j = 0; j < n; j++)
                adjacencyList[j].append(&cGraph.adjacencyList[j]);
        }
    
/*    if (direction)
        {
            delete [] direction;
            direction = NULL;
        }*/

    if (cGraph.direction)
        {
            direction = new int[m];
            for (int i = 0; i < m; i++)
                direction[i] = cGraph.direction[i];
        }

/*    if (weight)
        {
            delete [] weight;
            weight = NULL;
        }*/

    if (cGraph.weight)
        {
            weight = new double[m];
            for (int i = 0; i < m; i++)
                weight[i] = cGraph.weight[i];
        }

    if (cGraph.graphName)
        strcpy(graphName,cGraph.graphName);

    // change the data that belongs to the class 'clusterGraph' and
    // not to the base-class 'basicGraph'

    clusterNumber = cGraph.clusterNumber;
    
/*    if (c_nodes)
        {
            delete [] c_nodes;
            c_nodes = NULL;
        }*/

    if (cGraph.c_Nodes)
        {
            c_Nodes = new linkedList<int>[clusterNumber];
            for (int j = 0; j < clusterNumber; j++)
                c_Nodes[j].append(&cGraph.c_Nodes[j]);
        }

/*    if (tree)
        {
            delete tree;
            tree = NULL;
        }*/

    if (cGraph.tree && cGraph.c_Nodes)
        {
            tree = new inclusionTree;
            constructTree(n,0);
        }

    return *this;
}




