/*******************************************************************************
+
+  AGD 1.2
+
+
+  sugiyama.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        SugiyamaLayout.cpp
//
//    author: Carsten Gutwenger
//    date:   decembre 1998
//-------------------------------------------------------------------


#include <AGD/SugiyamaLayout.h>
#include <AGD/LongestPathRanking.h>
#include <AGD/BarycenterHeuristic.h>
#include <AGD/GreedySwitchHeuristic.h>
#include <AGD/FastHierarchyLayout.h>
#include <AGD/GraphCopyLI.h>
#include <AGD/Hierarchy.h>
#include <LEDA/core/tuple.h>
#include <LEDA/internal/std/float.h>


using namespace leda;


//-------------------------------------------------------------------
//                        HierarchyLayoutModule
//-------------------------------------------------------------------

const int HierarchyLayoutModule::in_hierarchy = 1;
const int HierarchyLayoutModule::out_layout = 1;


//-------------------------------------------------------------------
//                        SugiyamaLayout
//-------------------------------------------------------------------

SugiyamaLayout::SugiyamaLayout() : _runs(3), _fails(4),
	_transpose(true), _init_order(true)
{
	// precondition
	add_precondition (key::no_self_loops);
	add_precondition (key::directed);

	// initialize used modules
	PreCond  PRE_1, PRE_2(2);
	PostCond POST_1;

	PRE_1 [1] << key::no_self_loops << key::directed;
	ranking.init (PRE_1,POST_1,LongestPathRanking());

	PRE_2 [1] = PRE_1 [1];
	cross_min.init (PRE_2,POST_1,BarycenterHeuristic());

	comp_coord.init (PreCond(),POST_1,FastHierarchyLayout());
}


AGDModule *SugiyamaLayout::clone() const
{
	SugiyamaLayout *M = new SugiyamaLayout;

	M->runs(runs());
	M->fails(fails());
	M->transpose(transpose());
	M->init_order(init_order());

	M->set_ranking(ranking.get());
	M->set_cross_min(cross_min.get());
	M->set_comp_coord(comp_coord.get());

	return M;
}


void SugiyamaLayout::do_call (GraphCopyConcept& C)
{
	GraphCopy &GC = C.get_graph();
	LayoutInterface &A = C.get_layout();

	if (GC.empty()) {
		best_nr_crossings = 0;
		return;
	}

	// we work always on a directed graph
	if (GC.is_undirected()) GC.make_directed();

	// compute the ranking
	node_array<int> rank(GC.original());
	ranking.get().call(GC.original(),rank);

	// build proper hierarchy and minimize crossings
	Hierarchy H(GC,rank);
	minimize_crossings(GC,H);

	// assign final coordinates
	comp_coord.get().call(GC,H,A);
	rem_dummy_sinks(GC);
}

void SugiyamaLayout::ext_call (const graph& G,
	node_array<int>& rank, LayoutInterface& A)
{
	GraphCopy GC(G);
	GraphCopyLI A_GC(GC,A);

	if (GC.empty()) {
		best_nr_crossings = 0;
		return;
	}

	if (GC.is_undirected()) GC.make_directed();

	// build proper hierarchy and minimize crossings
	Hierarchy H(GC,rank);
	minimize_crossings(GC,H);

	// assign final coordinates
	comp_coord.get().call(GC,H,A_GC);
	rem_dummy_sinks(GC);
	A_GC.update();
}

void SugiyamaLayout::reduce_crossings (GraphCopy& GC, Hierarchy& H)
{
	minimize_crossings(GC,H);
}


void SugiyamaLayout::rem_dummy_sinks(GraphCopy &GC)
{
	node v;
	forall_nodes(v,GC) {
		if (GC.is_dummy(v) && GC.outdeg(v) == 0)
			GC.rev_edge(GC.last_in_edge(v));
	}
}


void SugiyamaLayout::call_cross_min (const graph& G, Layer &L,
	TwoLayerCrossMin& minimizer)
{
	int i, j;
	list<two_tuple<node,int> > isolated;

	for (i = 0; i < L.length(); i++)
		if (L.adj_nodes(L[i]).empty())
			isolated.append(two_tuple<node,int>(L[i],i));

	minimizer.call(G,L);

	if (isolated.empty()) return;

	list<node> pos;

	j = 0;
	for (i = 0; i < L.length(); i++) {
		if (L.adj_nodes(L[i]).empty()) continue;
		while (!isolated.empty() && isolated.head().second() == j) {
			pos.append(isolated.pop().first());
			j++;
		}
		pos.append(L[i]); j++;
	}

	while (!isolated.empty())
		pos.append(isolated.pop().first());

	L.set_pos(pos);
}


int SugiyamaLayout::traverse_top_down(const graph& G,
	Hierarchy& H, TwoLayerCrossMin& minimizer)
{
	H.direction(Hierarchy::downward);
	for (int i = 1; i < H.number_of_layers(); i++) {
		H.set_changeable_layer(i);
		call_cross_min(G,H[i],minimizer);
	}
	if (transpose())
		do_transpose(H);
	return H.calc_crossings();
}


int SugiyamaLayout::traverse_bottom_up(const graph& G,
	Hierarchy& H, TwoLayerCrossMin& minimizer)
{
	H.direction(Hierarchy::upward);
	for (int i = H.number_of_layers()-2; i >= 0; i--) {
		H.set_changeable_layer(i);
		call_cross_min(G,H[i],minimizer);
	}
	if (transpose())
		do_transpose_rev(H);
	return H.calc_crossings();
}


int SugiyamaLayout::transpose_test (node v, node w, const Hierarchy &H)
{
	int d = 0;
	edge e1, e2;

	forall_out_edges(e1,v) {
		forall_out_edges(e2,w)
			update_d(H.pos(target(e1)),H.pos(target(e2)),d);
	}

	forall_in_edges(e1,v) {
		forall_in_edges(e2,w)
			update_d(H.pos(source(e1)),H.pos(source(e2)),d);
	}

	return d;
}


int SugiyamaLayout::transpose_test (node v, node x, node w, const Hierarchy &H)
{
	int d = transpose_test(v,w,H);
	edge e1, e2;

	forall_out_edges(e1,x) {
		forall_out_edges(e2,w)
			update_d( H.pos(target(e1)),H.pos(target(e2)),d);

		forall_out_edges(e2,v)
			update_d(H.pos(target(e2)),H.pos(target(e1)),d);
	}

	forall_in_edges(e1,x) {
		forall_in_edges(e2,w)
			update_d(H.pos(source(e1)),H.pos(source(e2)),d);
		
		forall_in_edges(e2,v)
			update_d(H.pos(source(e2)),H.pos(source(e1)),d);
	}

	return d;
}


bool SugiyamaLayout::transpose_layer(int i, Hierarchy &H)
{
	bool improved = false;

	if (changed_layer[i] || changed_layer[i-1] || changed_layer[i+1])
	{
		Layer &L = H[i];

		for (int j = 0; j < L.length()-1; j++) {
			if (transpose_test(L[j],L[j+1],H) > 0) {
				improved = true;
				L.swap(j,j+1);
			}

			if (j+2 < L.length()) {
				if (transpose_test(L[j],L[j+1],L[j+2],H) > 0) {
					improved = true;
					L.swap(j,j+2);
				}
			}
		}
	}

	return (changed_layer[i] = improved);
}


void SugiyamaLayout::reset_changed_layer()
{
	for (int i = 0; i <= changed_layer.high()-1; i++)
		changed_layer[i] = true;
}

void SugiyamaLayout::do_transpose(Hierarchy &H)
{
	bool improved = true;
	reset_changed_layer();

	while (improved) {
		improved = false;

		for (int i = 0; i < H.number_of_layers(); i++) {
			improved |= transpose_layer(i,H);
		}
	}
}


void SugiyamaLayout::do_transpose_rev(Hierarchy &H)
{
	bool improved = true;
	reset_changed_layer();

	while (improved) {
		improved = false;

		for (int i = H.number_of_layers()-1; i >= 0 ; i--) {
			improved |= transpose_layer(i,H);
		}
	}
}


void SugiyamaLayout::minimize_crossings(const graph& G, Hierarchy& H)
{
	best_nr_crossings = G.number_of_edges()*G.number_of_edges();

	if (G.number_of_edges() == 0) return;

	if (init_order()) initial_order(G,H);

	TwoLayerCrossMin &minimizer = cross_min.get();

	if (transpose()) {
		changed_layer = array<bool>(-1,H.number_of_layers());
		changed_layer[-1] = changed_layer[H.number_of_layers()] = false;
	}

	node_map<int> best_pos;

	minimizer.init(G,H);

	for (int i = 1; ;i++ ) {

		int fail_count = fails()+1;
		int new_nr_crossings, old_nr_crossings;

		old_nr_crossings = H.calc_crossings();
		if (old_nr_crossings < best_nr_crossings) {
			H.store_pos (best_pos);
			best_nr_crossings = old_nr_crossings;
		}

		do {
			new_nr_crossings = traverse_top_down(G,H,minimizer);

			if (new_nr_crossings < old_nr_crossings) {
				if (new_nr_crossings < best_nr_crossings) {
					H.store_pos(best_pos);
					best_nr_crossings = new_nr_crossings;
				}
				old_nr_crossings = new_nr_crossings;
				fail_count = fails()+1;
			} else
				--fail_count;

			new_nr_crossings = traverse_bottom_up(G,H,minimizer);

			if (new_nr_crossings < old_nr_crossings) {
				if (new_nr_crossings < best_nr_crossings) {
					H.store_pos(best_pos);
					best_nr_crossings = new_nr_crossings;
				}
				old_nr_crossings = new_nr_crossings;
				fail_count = fails()+1;
			} else
				--fail_count;

		} while (fail_count > 0);

		if (best_nr_crossings == 0 || i >= runs())
			break;

		H.shuffle();
	}

	H.restore_pos(best_pos);
	minimizer.cleanup();
	changed_layer = array<bool>();
}


void SugiyamaLayout::initial_order_dfs(node v, node_map<int> &pos,
	const Hierarchy &H, array<int> &counter)
{
	pos[v] = counter[H.rank(v)]++;

	node w;
	forall_adj_nodes(w,v)
		if (pos[w] == -1)
			initial_order_dfs(w,pos,H,counter);
}

void SugiyamaLayout::initial_order(const graph &G, Hierarchy &H)
{
	array<int> counter(H.number_of_layers());
	node_map<int> pos(G,-1);

	int i, j;
	for (i = 0; i < H.number_of_layers(); i++)
		counter[i] = 0;

	for (i = 0; i < H.number_of_layers(); i++)
		for (j = 0; j < H[i].length(); j++)
			if (pos[(H[i])[j]] == -1)
				initial_order_dfs((H[i])[j],pos,H,counter);

	H.restore_pos(pos);
}


//-------------------------------------------------------------------
//                            Layer
//-------------------------------------------------------------------

Layer::Layer (Hierarchy& H, int i, const list<node>& nodes) :
	_hierarchy (&H), _index(i)
{
	//_nodes = array<node>(nodes.length());
	_size = nodes.length();
	_nodes = LEDA_NEW_VECTOR(node,_size);

	node v;
	int j = 0;
	forall(v,nodes) {
		_nodes[j] = v;
		H._pos[v] = j++;
	}
}

const Layer& Layer::neighbored_layer() const
{
	return (_hierarchy->direction() == Hierarchy::downward) ?
		*_hierarchy->_layer[_index-1] :
		*_hierarchy->_layer[_index+1];
}


const list<node>& Layer::adj_nodes (node v) const
{
	Hierarchy::Direction d = _hierarchy->direction();
	int ch = _hierarchy->_changeable;

	return ((d == Hierarchy::upward && _index == ch) ||
		(d == Hierarchy::downward && _index != ch)) ?
		_hierarchy->_lower_adj_nodes[v] :
		_hierarchy->_upper_adj_nodes[v];
}

const list<edge>& Layer::adj_edges (node v) const
{
	Hierarchy::Direction d = _hierarchy->direction();
	int ch = _hierarchy->_changeable;

	return ((d == Hierarchy::upward && _index == ch) ||
		(d == Hierarchy::upward && _index != ch)) ?
		_hierarchy->_lower_adj_edges[v] :
		_hierarchy->_upper_adj_edges[v];
}

void Layer::swap (int i, int j)
{
	node v = _nodes[i];
	_hierarchy->_pos[_nodes[i] = _nodes[j]] = i;
	_hierarchy->_pos[_nodes[j] = v] = j;  
}

void Layer::quicksort (const node_array<double>& weight)
{
	if (length() == 0) return;

	_qs(0,length()-1,weight);

	int i;
	for (i = 0; i < length(); i++) {
		_hierarchy->_pos[_nodes[i]] = i;
	}
}

int Layer::cmp_qs (double w_v, int p_v, double w_w, int p_w)
{
	if ( w_w - w_v >= DBL_EPSILON ) return -1;
	if ( w_v - w_w >= DBL_EPSILON ) return 1;
	return p_v - p_w;
}
		 
void Layer::_qs (int l, int r, const node_array<double>& weight)
{
	random_source& rs = AGD::get_random_source();

	const node_map<int>& P = _hierarchy->_pos;
	int    i = l, j = r, m = rs(l,r), xp = P[_nodes[m]];
	double x = weight[_nodes[m]];

	do {
		while (cmp_qs(weight[_nodes[i]],P[_nodes[i]],x,xp) < 0) i++;
		while (cmp_qs(weight[_nodes[j]],P[_nodes[j]],x,xp) > 0) j--;
		if (i <= j) {
			leda::swap(_nodes[i++],_nodes[j--]);
		}
	} while (i <= j);

	if (l < j) _qs(l,j,weight);
	if (i < r) _qs(i,r,weight);
}

void Layer::set_pos (const node_map<int>& position)
{
	int i;
	node v;
	//array<node> N(length());
	node *N = LEDA_NEW_VECTOR(node,_size);

	for (i = 0; i < length(); i++) {
		v = _nodes[i];
		N[_hierarchy->_pos[v] = position[v]] = v;
	}

	//delete [] _nodes;
	LEDA_DEL_VECTOR(_nodes);
	_nodes = N;
}

void Layer::set_pos (const list<node> &pos)
{
	node v;
	int i = 0;

	forall(v,pos) {
		_nodes[_hierarchy->_pos[v] = i] = v;
		i++;
	}
}

void Layer::shuffle()
{
	//_nodes.permute();
	const int top = _size-1;
	if (top < 0) return; // prevents error if Layer is empty
	node *current = _nodes;

	random_source& rs = AGD::get_random_source();

	int i;
	for (i = 0; i < top; i++) {
		leda::swap(*current,_nodes[rs(i,top)]);
		_hierarchy->_pos[*current++] = i;
	}
	_hierarchy->_pos[*current] = i;
}

ostream& operator<< (ostream& os, const Layer& L)
{
	int i, l = L.length();

	os << "[ ";
	for (i = 0; i < l; i++) {
		os << index(L[i]);
		if (i < l-1) os << ",";
	}
	os << " ]";

	return os;
}

istream& operator>> (istream& is, Layer& )
{
	return is;
}


//-------------------------------------------------------------------
//                            Hierarchy
//-------------------------------------------------------------------

Hierarchy::Hierarchy (GraphCopy& GC, const node_array<int>& R)
{
	gc_p = &GC;
	_direction = downward;

	_rank.init(GC);
	_pos.init(GC);

	node v, w;
	edge e, e_orig;;

	forall_nodes(v,GC)
		_rank[v] = R[GC.original(v)];

	int i, rv, rw;
	list<edge> edges = GC.all_edges();

	forall(e,edges) {
		e_orig = GC.original(e);
		v = source(e); w = target(e);
		if (v == w) continue;
		rv = _rank[v]; rw =  _rank[w];

		if (rv > rw) {
			GC.rev_edge(e); leda::swap(rv,rw);
		}

		if (rv == rw) {
			e = GC.split(e);
			GC.rev_edge(e);
			_rank[target(e)] = rv+1;

		} else {
			for (rv++; rv < rw; rv++) {
				_rank[source(e = GC.split(e))] = rv;
			}
		}
	}

	int max_rank = 0;
	forall_nodes(v,GC)
		if (_rank[v] > max_rank) max_rank = _rank[v];

	array<list<node> > L(0,max_rank);
	forall_nodes(v,GC)
		L[_rank[v]].append(v);

	_layer = array<Layer *>(0,max_rank);
	for (i = 0; i <= max_rank; i++)
		_layer[i] = new Layer(*this,i,L[i]);

	_lower_adj_edges.init(GC);
	_lower_adj_nodes.init(GC);
	_upper_adj_edges.init(GC);
	_upper_adj_nodes.init(GC);

	forall_nodes(v,GC) {
		forall_out_edges(e,v) {
			if (target(e) == v) continue;
			_lower_adj_edges[v].append(e);
			_lower_adj_nodes[v].append(target(e));
		}
		forall_in_edges(e,v) {
			if (source(e) == v) continue;
			_upper_adj_edges[v].append(e);
			_upper_adj_nodes[v].append(source(e));
		}
	}

	// initialize node arrays used in crossing calculation only once
	high_occ.init(GC);
	sorted_adj_nodes.init(GC);
}

Hierarchy::~Hierarchy()
{
	int i;

	for (i = 0; i < number_of_layers(); i++)
		delete _layer[i];
}

void Hierarchy::store_pos (node_map<int>& old_pos)
{
	old_pos = _pos;
}

void Hierarchy::restore_pos (const node_map<int>& new_pos)
{
	node v;

	_pos = new_pos;

	forall_nodes(v, *gc_p)
		_layer[_rank[v]]->_nodes[_pos[v]] = v;
}

// caculate crossings between layer i and i+1
int Hierarchy::calc_crossings (int i)
{
	if (length(i) <= 1 || length(i+1) <= 1)
		return 0;

	int j, k;
	int deactivated, nodecross,
		allcross, crossing_number = 0;
	node uk, ul;
	list<node> upper_lower_list[2];
	list_item li_uj;
	int index[2] = { 0, 0 };

	for (j = i; j <= i+1; j++) {
		for (k = 0; k < length(j); k++) {
			uk = (*_layer[j])[k];
			high_occ[uk] = nil;
			sorted_adj_nodes[uk].clear();
		}
	}

	for (j = i; j <= i+1; j++) {
		for (k = 0; k < length(j); k++) {
			uk = (*_layer[j])[k];
			const list<node>& N = (j == i) ? _lower_adj_nodes[uk] : _upper_adj_nodes[uk];
			forall (ul, N)
				sorted_adj_nodes[ul].append(uk);
		}
	}
  
	j = 0;
	do {
		deactivated = nodecross = allcross = 0;
		ul = (*_layer[i+j])[index[j]];

		if (high_occ[ul] != nil) {
			forall_items (li_uj, upper_lower_list[j]) {
				if (upper_lower_list[j].inf(li_uj) == ul) {
					deactivated++;
					allcross += nodecross;
					upper_lower_list[j].erase(li_uj); // deletion of current element allowed since LEDA 3.5.1!
				} else
					nodecross++;
				if (high_occ[ul] == li_uj)
					break; // loop until high_occ[ ul ] (incl.)
			}
			crossing_number +=
				deactivated * upper_lower_list[(j == 0) ? 1 : 0].size() + allcross;
		}
    
		forall (uk, sorted_adj_nodes[ul])
			if (pos(uk) >  pos(ul) ||
				(pos(uk) == pos(ul) && j == 0))
				high_occ[uk] = upper_lower_list[(j == 0) ? 1 : 0].append(uk);

		index[j]++;
		if (index[1-j] < length(i+1-j))
			j = 1-j;
	} while (index[j] < length(i+j));

	return crossing_number;
}

int Hierarchy::calc_crossings()
{
	int i, sum = 0;

	for (i = 0; i < number_of_layers()-1; i++)
		sum += calc_crossings(i);

	return sum;
}

void Hierarchy::shuffle()
{
	for (int i = 0; i < number_of_layers(); i++)
		_layer[i]->shuffle();
}
