#include <LEDA/graphics/gl_camera.h>

#if defined(__BORLANDC__) || defined(_MSC_VER)
#include <windows.h>
#endif

#include <GL/gl.h>


namespace vgl {

gl_camera::gl_camera() 
{ eye  = 0.0, 0.0, 0.0; 
  look = 0.0, 0.0,-1.0;
  up   = 0.0, 1.0, 0.0; 
  initialize(); 
}

void gl_camera::operator()() 
{ glMatrixMode(GL_MODELVIEW);
  glLoadMatrixd(M);
}

void gl_camera::set(double ex, double ey, double ez,
         double lx, double ly, double lz,
         double ux, double uy, double uz)
{ eye  = ex, ey, ez;
  look = lx, ly, lz;
  up   = ux, uy, uz;
  initialize();  
}
        
void gl_camera::set(const vector3d& u0, const vector3d& v0, const vector3d& n0)   
{ u = u0; v = v0; n = n0; 
  set_model_matrix();
}    

void gl_camera::set_eye(double x, double y, double z)
{ eye = x, y, z;
  set_model_matrix();
}

template <class point>
void gl_camera::set_eye(const point& e)
{ eye[0] = e.xcoord();
  eye[1] = e.ycoord();
  eye[2] = e.zcoord();  
  set_model_matrix();
}

void gl_camera::set_eye(const vector3d& vec)
{ eye = vec;
  set_model_matrix();  
}

void gl_camera::get_eye(double& x, double& y, double& z) const 
{ x = eye[0]; y = eye[1]; z = eye[2]; }

template <class point>
void gl_camera::get_eye(point& p) const { p = point(eye[0],eye[1],eye[2]); }  

vector3d gl_camera::get_eye() const { return eye; }
vector3d gl_camera::get_n()   const { return n; }
vector3d gl_camera::get_u()   const { return u; }
vector3d gl_camera::get_v()   const { return v; } 

void gl_camera::rotate(double angle, double x, double y, double z)
{ d3_transform R; 
  R.rotate(angle,x,y,z);
  u = R(u); 
  v = R(v); 
  n = R(n);
  set_model_matrix();
}

void gl_camera::slide(double dx, double dy, double dz)
{ double x = eye[0] + dx * u[0] + dy * v[0] + dz * n[0];
  double y = eye[1] + dx * u[1] + dy * v[1] + dz * n[1];
  double z = eye[2] + dx * u[2] + dy * v[2] + dz * n[2];  
  eye = x, y, z;    
  set_model_matrix();
}

void gl_camera::roll(double angle) 
{ double c = cos(angle);
  double s = sin(angle);
  
  double ux = u[0], vx = v[0];
  double uy = u[1], vy = v[1];
  double uz = u[2], vz = v[2];
  
  u = c * ux - s * vx, c * uy - s * vy, c * uz - s * vz; 
  v = s * ux + c * vx, s * uy + c * vy, s * uz + c * vz;
  set_model_matrix();
}
  
void gl_camera::pitch(double angle) 
{ double c = cos(angle);
  double s = sin(angle);
  
  double nx = n[0], vx = v[0];
  double ny = n[1], vy = v[1];
  double nz = n[2], vz = v[2];
    
  n = c * nx - s * vx, c * ny - s * vy, c * nz - s * vz; 
  v = s * nx + c * vx, s * ny + c * vy, s * nz + c * vz;
  set_model_matrix();
}

void gl_camera::yaw(double angle) 
{ double c = cos(angle);
  double s = sin(angle);
  
  double ux = u[0], nx = n[0];
  double uy = u[1], ny = n[1];
  double uz = u[2], nz = n[2];
  
  u = c * ux - s * nx, c * uy - s * ny, c * uz - s * nz; 
  n = s * ux + c * nx, s * uy + c * ny, s * uz + c * nz;
  set_model_matrix();
}

void gl_camera::initialize()
{ n = norm(eye - look);
  u = norm(cross_product(up,n));
  v = cross_product(n,u);
  set_model_matrix();
}

void gl_camera::set_model_matrix()
{ vector3d ev(-eye[0],-eye[1],-eye[2]);  
  M = u[0], v[0], n[0], 0,
      u[1], v[1], n[1], 0,
      u[2], v[2], n[2], 0,
      ev*u, ev*v, ev*n, 1;
}

};
