// compression module includes
#include <LEDA/coding/BWT.h>
#include "../plain_c/fm_ds_ssort.h"

// if the following causes a problem, look at fm_ds_ssort.h!
typedef leda::coder_base::byte byte;
typedef leda::coder_base::sint32 sint32;

LEDA_BEGIN_NAMESPACE

/*
	The BWT code by Manzini and Ferragina makes use of static global variables.
	In order to protect this data in a multithreading environment, we use a
	mutex!
*/
static leda_lock BWT_mutex;

/// BWTCoder ////////////////////////////////////////////////////////////////////////////////

BWTCoder::BWTCoder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams, uint32 block_size)
: base(src_stream, tgt_stream, own_streams), 
  Buffer(0), SuffixArray(0), PosArray(0), BlockSizeEnc(0), BlockSizeDec(0)
{
	set_block_size(block_size);
}

BWTCoder::BWTCoder(const char* src_file_name, const char* tgt_file_name, uint32 block_size)
: base(src_file_name, tgt_file_name),
  Buffer(0), SuffixArray(0), PosArray(0), BlockSizeEnc(0), BlockSizeDec(0)
{
	set_block_size(block_size);
}

BWTCoder::~BWTCoder()
{
	free_buffers();
}

void 
BWTCoder::do_reset(reset_type t)
{
	base::do_reset(t);
	if (t == r_hard || t == r_soft) {
		free_buffers(); 
			// NOTE: we need different buffers and different sizes for encoding / decoding 
	}
	if (t == r_hard) {
		set_block_size(DefaultBlockSize);
	}
}

void
BWTCoder::set_block_size(uint32 block_size)
{
	if (can_change_parameters()) {
		BlockSizeEnc = block_size > 1024 ? block_size : 1024;
	}
	else error("BWTCoder::set_block_size: wrong state");
}

void 
BWTCoder::free_buffers()
{
	delete[] Buffer;      Buffer = 0;
	delete[] SuffixArray; SuffixArray = 0;
	delete[] PosArray;    PosArray = 0;
	BlockSizeDec = 0;
}

void 
BWTCoder::init_encoding()
{
	base::init_encoding();

	// allocate buffers for encoding
	if (Buffer != 0 || SuffixArray != 0 || PosArray != 0) free_buffers();
	int overshoot = init_ds_ssort(500, 2000);
	if (overshoot == 0) { error(); return; }
	Buffer      = new byte[BlockSizeEnc + overshoot];
	SuffixArray = new sint32[BlockSizeEnc];

	ByteCount = 0;

	BWT_mutex.lock();
}

void 
BWTCoder::encode_stream()
{
	int c;
	while ((c = read_byte()) >= 0 && c != EOF) {
		byte symbol = (byte) c;
		Buffer[ByteCount] = symbol;

		if (++ByteCount == BlockSizeEnc) {
			encode_block();
			ByteCount = 0;
		}
	}

	if (c == EOF) {
		if (ByteCount > 0) encode_block();
		BWT_mutex.unlock();
		encoding_finished();
		free_buffers();
	}
}

void 
BWTCoder::encode_block()
{
	ds_ssort(Buffer, SuffixArray, ByteCount);

	uint32 pos_eob = 0;

	write_uint32(ByteCount);

	write_byte(Buffer[ByteCount-1]); // Eob is smaller than any regular symbol
	for (uint32 i = 0; i < ByteCount; ++i) {
		uint32 pos = SuffixArray[i];
		if (pos == 0) pos_eob = i+1;
		else          write_byte(Buffer[pos-1]);
	}

	write_uint32(pos_eob);
}


void 
BWTCoder::init_decoding()
{
	base::init_decoding();

	if (SuffixArray != 0) free_buffers(); // should not occur!
		// NOTE: we do not free the buffers, if we have buffers that have been 
		//       allocated in decoding mode! Maybe, we can still use them ... (cf. seek)

	ByteCount = 0;

	BWT_mutex.lock();
}

void 
BWTCoder::decode_stream()
{
	if (ByteCount > 0) {
		if (!finish_writing_decoded_block()) return;
	}

	for (;;) {
		int c = read_byte();
		if (c == EOF) break;
		get_src_stream()->sputbackc((char) c);

		read_uint32(NextBlockSize);
		if (NextBlockSize > BlockSizeDec) {
			free_buffers();
			BlockSizeDec = NextBlockSize;
			Buffer   = new byte[BlockSizeDec];
			PosArray = new uint32[BlockSizeDec];
		}

		for (int i = 0; i < 256; ++i) 
			SymbolCount[i] = 0;

		while (ByteCount != NextBlockSize) {
			byte symbol = (byte) read_byte();
			Buffer[ByteCount++] = symbol;
			++SymbolCount[symbol];
		}

		if (!decode_block()) return;
	}

	BWT_mutex.unlock();
	decoding_finished();
	free_buffers();
}

bool
BWTCoder::decode_block()
{
	read_uint32(NextPosEob);
	if (NextPosEob > NextBlockSize) {
		error("BWTCoder::decode_block: input stream corrupted");
		return false;
	}

	// compute cumulative counts
	uint32 cur_total = 0;
	uint32 i;
	for (i = 0; i < 256; ++i) {
		uint32 count = SymbolCount[i];
		SymbolCount[i] = cur_total;
		cur_total += count;
	}

	for (i = 0; i < NextBlockSize; ++i)
		PosArray[SymbolCount[Buffer[i]]++] = i;

	NextPos = NextPosEob - 1;
	return finish_writing_decoded_block();
}

bool
BWTCoder::finish_writing_decoded_block()
{
	while (ByteCount > 0) {
		uint32 pos = PosArray[NextPos];
		byte symbol = Buffer[pos];
		NextPos = pos; if (pos < NextPosEob) --NextPos; // take care of Eob
		--ByteCount;
		if (write_byte(symbol) < 0) return false;
	}

	return true; // success, i.e. wrote all data to tgt stream
}

LEDA_END_NAMESPACE
