// compression module includes
#include <LEDA/coding/checksum.h>
#include <LEDA/coding/coder_strbuf_adapter.h>
#include <LEDA/coding/memory_streambuf.h>
#include <LEDA/coding/internal/strbuf_helpers.h>

// core library includes
#include <string.h> // -> we need the function memset

#if defined(LEDA_DEBUG)
#include <LEDA/system/assert.h>
#else
#undef assert
#define	assert(cond) ((void)0)
#endif

LEDA_BEGIN_NAMESPACE

typedef checksummer_base::uint16 uint16;
typedef checksummer_base::uint32 uint32;

#define ROL(x,n) (((x) << ((n) & 0x1F)) | ((x) >> (32-((n) & 0x1F))))
#define ROR(x,n) (((x) >> ((n) & 0x1F)) | ((x) << (32-((n) & 0x1F))))

#if defined(_MSC_VER) && defined(LEDA_CODER_EXACT_SIZE_INT32)
#include <stdlib.h> // get declarations of intrinsic rotation
#undef	ROL
#undef	ROR
#pragma intrinsic(_lrotl,_lrotr) // use intrinsic compiler rotations
#define	ROL(x,n)	_lrotl(x,n)			
#define	ROR(x,n)	_lrotr(x,n)
#endif


/// checksummer_base ////////////////////////////////////////////////////////////////////////

checksummer_base::checksummer_base(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams), 
  BlockSize(0), ReadAheadBuffer(MaxByteLengthOfCheckSum), 
  CheckSumsInStream(true), CheckSumState(cs_invalid)
{}

checksummer_base::checksummer_base(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name),
  BlockSize(0), ReadAheadBuffer(MaxByteLengthOfCheckSum), 
  CheckSumsInStream(true), CheckSumState(cs_invalid)
{}

void
checksummer_base::do_reset(reset_type t)
{
	base::do_reset(t);

	CheckSumState = cs_invalid;
	if (t == r_hard) set_checksums_in_stream_flag(true, 0);
}

string
checksummer_base::check()
{
	if (! get_tgt_stream()) {
		// we use a dummy streambuf!
		set_tgt_stream(new eat_all_streambuf(), true);
		decode();
		set_tgt_stream(0);
	}
	else decode();

	return get_checksum_as_hex_string();
}

string 
checksummer_base::get_checksum_as_hex_string() const
{
	string result;
	if (CheckSumState != cs_okay) return result;
	unsigned len = get_byte_length_of_checksum();
	for (unsigned i = 0; i < len; ++i) {
		byte b = get_ith_byte_of_checksum(i);
		result += string("%02x", b);
	}
	return result;
}

string 
checksummer_base::check(string str)
{
	memory_streambuf mbuf(str.cstring(), str.length());
	set_src_stream(&mbuf, false); set_tgt_stream(0);
	bool sav_flag = get_checksums_in_stream_flag(); uint32 sav_blk_sz = get_block_size();
	set_checksums_in_stream_flag(false);
	string result = check();
	set_checksums_in_stream_flag(sav_flag, sav_blk_sz);
	set_src_stream(0);
	return result;
}

void 
checksummer_base::set_checksums_in_stream_flag(bool sums_in_stream, uint32 blocksize)
{
	if (can_change_parameters()) {
		CheckSumsInStream = sums_in_stream; BlockSize = sums_in_stream ? blocksize : 0;
	}
	else error();
}

void 
checksummer_base::write_header_to_tgt_stream()
{
	write_uint32(BlockSize);
}

void 
checksummer_base::read_header_from_src_stream()
{
	read_uint32(BlockSize);
}

void 
checksummer_base::init_encoding()
{
	assert(MaxByteLengthOfCheckSum >= get_byte_length_of_checksum());

	if (! CheckSumsInStream) { process_no_id_in_streams(); process_no_header_in_streams(); }
	base::init_encoding();

	ByteCount = 0;
	reset_checksum();
	CheckSumState = cs_computing;
}

void 
checksummer_base::encode_stream()
{
	int c;
	while ((c = read_byte()) >= 0 && c != EOF) {
		byte symbol = (byte) c;
		
		update_checksum(symbol);

		write_byte(symbol);

		if (++ByteCount == BlockSize && CheckSumsInStream && BlockSize > 0) {
			finish_checksum();
			write_checksum();
			ByteCount = 0;
			reset_checksum();
		}
	}

	if (c == EOF) {
		finish_checksum();
		CheckSumState = (BlockSize == 0) ? cs_okay : cs_invalid;
		if (CheckSumsInStream && ByteCount > 0) {
			write_checksum();
		}
		encoding_finished();
	}
}

void
checksummer_base::write_checksum()
{
	unsigned bytes_to_write = get_byte_length_of_checksum();
	while (bytes_to_write-- > 0) {
		write_byte(get_ith_byte_of_checksum(bytes_to_write));
	}
}

void 
checksummer_base::init_decoding()
{
	assert(MaxByteLengthOfCheckSum >= get_byte_length_of_checksum());

	if (! CheckSumsInStream) { process_no_id_in_streams(); process_no_header_in_streams(); }
	base::init_decoding();

	ByteCount = 0;
	ReadAheadBuffer.clear();
	reset_checksum();
	CheckSumState = cs_computing;
}

void 
checksummer_base::decode_stream()
{
	if (!CheckSumsInStream) {
		int c;
		while ((c = read_byte()) != EOF) {
			byte symbol = (byte) c;

			update_checksum(symbol);

			if (write_byte(symbol) == coder_strbuf_adapter::StreamBlocked) return;
		}
		finish_checksum();
		CheckSumState = (CheckSumState == cs_computing) ? cs_okay : cs_invalid;
	}
	else {
		if (fill_read_ahead_buffer()) {
			int c;
			while ((c = read_byte()) != EOF) {
				++ByteCount;
				
				byte curr_symbol = ReadAheadBuffer.pop();
				ReadAheadBuffer.append((byte) c);

				update_checksum(curr_symbol);

				if (ByteCount == BlockSize && BlockSize > 0) {
					finish_checksum();
					compare_checksum_with_read_ahead();
					reset_checksum();
					ByteCount = 0;
					if (!fill_read_ahead_buffer()) break;
				}

				if (write_byte(curr_symbol) == coder_strbuf_adapter::StreamBlocked) return;
			}
		}

		finish_checksum();
		CheckSumState = (CheckSumState==cs_computing && BlockSize==0) ? cs_okay : cs_invalid;
		if (ByteCount > 0) {
			compare_checksum_with_read_ahead();
		}
	}

	decoding_finished();
}

bool
checksummer_base::fill_read_ahead_buffer()
{
	int read_ahead = (int) get_byte_length_of_checksum();
	while (ReadAheadBuffer.size() < read_ahead) {
		int c = read_byte();
		if (c == EOF) return false; // maybe file ends at block end
		ReadAheadBuffer.append((byte) c);
	}
	return true;
}

void
checksummer_base::compare_checksum_with_read_ahead()
{
	if (checksum_is_valid()) {
		unsigned bytes_to_compare = get_byte_length_of_checksum();
		while (bytes_to_compare-- > 0) {
			byte b = get_ith_byte_of_checksum(bytes_to_compare);
			if (b != ReadAheadBuffer.pop()) error("bad checksum in stream");
		}
	}
	else ReadAheadBuffer.clear(); // if not valid, then skip it
}


streampos 
checksummer_base::decode_seek(streampos current_pos, streampos new_pos)
{
	if (!decoding() || failed()) return base::decode_seek(current_pos, new_pos);

	CheckSumState = cs_invalid;

	if (new_pos < streamoff(0)) {
		// seek to the end
		set_state(s_decoding_finished);
		streampos size = seek_end_src_stream_relative_to_header();
		if (!CheckSumsInStream) return size;
		else if (BlockSize == 0) return size - streamoff(get_byte_length_of_checksum());
		else {
			streamoff blk_plus_sum_sz = streamoff(BlockSize + get_byte_length_of_checksum());
			streamoff num_blks = (streamoff(size) + blk_plus_sum_sz - 1) / blk_plus_sum_sz;
			return size - num_blks * streamoff(get_byte_length_of_checksum());
		}
	}

	// regular seek
	set_state(s_decoding); // make it running ...
	ReadAheadBuffer.clear();
	if (!CheckSumsInStream || BlockSize == 0) 
		return seek_pos_src_stream_relative_to_header(new_pos);
	else {
		ByteCount = uint32(new_pos) % BlockSize;
		streamoff num_full_blks = streamoff(new_pos) / streamoff(BlockSize);
		streamoff chksums_offset = num_full_blks * streamoff(get_byte_length_of_checksum());
		seek_pos_src_stream_relative_to_header(new_pos + chksums_offset);
		return new_pos;
	}
}


/// CRC32Coder //////////////////////////////////////////////////////////////////////////////

#include "../local_incl/crc32.tbl"

CRC32Coder::CRC32Coder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams)
{}

CRC32Coder::CRC32Coder(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name)
{}

void 
CRC32Coder::reset_checksum()
{
	CheckSum = 0xFFFFFFFFL;
}

void 
CRC32Coder::update_checksum(byte b)
{
	CheckSum = ((CheckSum >> 8) & 0x00FFFFFFL) ^ crc32_table[(CheckSum ^ b) & 0x000000FFL];
}


/// CCITTCoder //////////////////////////////////////////////////////////////////////////////

#define USE_CCITT_TABLE
#ifdef USE_CCITT_TABLE
#include "../local_incl/ccitt.tbl"
#endif

CCITTCoder::CCITTCoder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams)
{}

CCITTCoder::CCITTCoder(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name)
{}

void 
CCITTCoder::reset_checksum()
{
	CheckSum = 0;
}

void 
CCITTCoder::update_checksum(byte b)
{
#ifdef USE_CCITT_TABLE
	CheckSum = (CheckSum << 8) ^ ccitt_table[ (CheckSum >> 8) ^ b ];
	CheckSum &= 0xFFFF;
#else
	uint16 crc = (uint16) CheckSum;
	uint16 ch = b;
	ch <<= 8;
	for (int i = 8; i > 0; i--)
	{
		if ((ch ^ crc) & 0X8000)
			crc = (crc << 1) ^ 0x1021;
		else
			crc <<= 1;
		ch <<= 1;
	}
	CheckSum = crc;
#endif
}

/// Adler32Coder ////////////////////////////////////////////////////////////////////////////

/*
The specification for Adler32 may be found in RFC 1950. 
(ZLIB Compressed Data Format Specification version 3.3) 

From that document: 
"ADLER32 (Adler-32 checksum) This contains a checksum value of the uncompressed data 
(excluding any dictionary data) computed according to Adler-32 algorithm. This algorithm is 
a 32-bit extension and improvement of the Fletcher algorithm, used in the 
ITU-T X.224 / ISO 8073 standard. 

Adler-32 is composed of two sums accumulated per byte: s1 is the sum of all bytes, 
s2 is the sum of all s1 values. Both sums are done modulo 65521. 
s1 is initialized to 1, s2 to zero. The Adler-32 checksum is stored as s2*65536 s1 in 
most- significant-byte first (network) order." 

"8.2. The Adler-32 algorithm 

The Adler-32 algorithm is much faster than the CRC32 algorithm yet still provides an 
extremely low probability of undetected errors. 

The modulo on unsigned long accumulators can be delayed for 5552 bytes, so the modulo 
operation time is negligible. If the bytes are a, b, c, the second sum is 3a 2b c 3, 
and so is position and order sensitive, unlike the first sum, which is just a checksum. 
That 65521 is prime is important to avoid a possible large class of two-byte errors 
that leave the check unchanged. (The Fletcher checksum uses 255, which is not prime and 
which also makes the Fletcher check insensitive to single byte changes 0 <-> 255.) 

The sum s1 is initialized to 1 instead of zero to make the length of the sequence part of s2, 
so that the length does not have to be checked separately. (Any sequence of zeroes has a 
Fletcher checksum of zero.)" 
*/

Adler32Coder::Adler32Coder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams)
{}

Adler32Coder::Adler32Coder(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name)
{}

void 
Adler32Coder::reset_checksum()
{
	S1 = 1; S2 = 0; Cnt = 0;
}

void 
Adler32Coder::update_checksum(byte b)
{
	S1 += b; S2 += S1;
	if (++Cnt == 5552) {
		S1 %= 65521; S2 %= 65521; Cnt = 0;
	}
}

void 
Adler32Coder::finish_checksum()
{
	CheckSum = (S2 << 16) | S1;
}



/// MD5SumCoder /////////////////////////////////////////////////////////////////////////////

/* 
	Code is based on the public domain code by Colin Plumb (who claims no copyright)

	http://www.l2tpd.org/cxref/md5.c.src.html
	http://www.l2tpd.org/cxref/md5.h.src.html

	The MD5 Message-Digest Algorithm is described in rfc1321.txt.
*/
/* Original copyright notice:
 * This code implements the MD5 message-digest algorithm.
 * The algorithm is due to Ron Rivest.  This code was
 * written by Colin Plumb in 1993, no copyright is claimed.
 * This code is in the public domain; do with it what you wish.
 *
 * Equivalent code is available from RSA Data Security, Inc.
 * This code has been tested against that, and is equivalent,
 * except that you don't need to include two pages of legalese
 * with every copy.
 *
 * To compute the message digest of a chunk of bytes, declare an
 * MD5Context structure, pass it to MD5Init, call MD5Update as
 * needed on buffers full of bytes, and then call MD5Final, which
 * will fill a supplied 16-byte array with the digest.
 */

MD5SumCoder::MD5SumCoder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams)
{}

MD5SumCoder::MD5SumCoder(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name)
{}

MD5SumCoder::~MD5SumCoder()
{
	memset((void*)this, 0xFF, sizeof(MD5SumCoder));
	memset((void*)this, 0x55, sizeof(MD5SumCoder));
	memset((void*)this, 0x00, sizeof(MD5SumCoder));
}

void 
MD5SumCoder::reset_checksum()
{
/*
 * Start MD5 accumulation.  Set bit count to 0 and buffer to mysterious
 * initialization constants.
 */
	ByteCount = 0;

	Buf[0] = 0x67452301;
	Buf[1] = 0xefcdab89;
	Buf[2] = 0x98badcfe;
	Buf[3] = 0x10325476;
}

void 
MD5SumCoder::update_checksum(byte b)
{
	Chunk[ByteCount & 0x3F] = b;
	if ((++ByteCount & 0x3F) == 0) {
		md5_transform();
	}
}

void 
MD5SumCoder::finish_checksum()
{
/*
 * Final wrapup - pad to 64-byte boundary with the bit pattern 
 * 1 0* (64-bit count of bits processed, LSB-first)
 */

	unsigned cnt_in_chunk = unsigned(ByteCount & 0x3F);
	byte* ptr = &Chunk[0] + cnt_in_chunk;

	/* Set the first char of padding to 0x80.  This is safe since there is
	   always at least one byte free */
	*(ptr++) = 0x80; ++cnt_in_chunk;

	/* Bytes of padding needed to make 64 bytes */
	unsigned free_in_chunk = 64 - cnt_in_chunk;

	/* Pad out to 56 mod 64 */
	if (free_in_chunk < 8) {
		/* Two lots of padding:  Pad the first block to 64 bytes */
		memset(ptr, 0, free_in_chunk);
		md5_transform();
		/* Now fill the next block with 56 bytes */
		memset(Chunk, 0, 56);
	}
	else {
		/* Pad block to 56 bytes */
		memset(ptr, 0, free_in_chunk-8);
	}

#if !defined(LITTLE_ENDIAN_MACHINE) || !defined(LEDA_CODER_EXACT_SIZE_INT32)
	// setWord will simulate little endian ...
	setWord(14, (ByteCount << 3) & 0xFFFFFFFF); // lower  32 Bits of SumOfCounts number of BITS
	setWord(15, (ByteCount >> 29)& 0xFFFFFFFF); // higher 32 Bits of SumOfCounts number of BITS

	md5_transform();

	// setDigest will simulate little endian ...
	uint32 a = Buf[0];
	uint32 b = Buf[1];
	uint32 c = Buf[2];
	uint32 d = Buf[3];
	setDigest(0, a); setDigest(1, b); setDigest(2, c); setDigest(3, d);
#else
	Words[14] = (ByteCount << 3) & 0xFFFFFFFF; // lower  32 Bits of SumOfCounts number of BITS
	Words[15] = (ByteCount >> 29)& 0xFFFFFFFF; // higher 32 Bits of SumOfCounts number of BITS

	md5_transform();
#endif
}

/* The four core functions - F1 is optimized somewhat */

/* #define F1(x, y, z) (x & y | ~x & z) */
#define F1(x, y, z) (z ^ (x & (y ^ z)))
#define F2(x, y, z) F1(z, x, y)
#define F3(x, y, z) (x ^ y ^ z)
#define F4(x, y, z) (y ^ (x | ~z))

/* This is the central step in the MD5 algorithm. */
#ifdef LEDA_CODER_EXACT_SIZE_INT32
#define MD5STEP(f, w, x, y, z, data, s) \
	( w += f(x, y, z) + data,  w = ROL(w,s),  w += x )
#else
#define MD5STEP(f, w, x, y, z, data, s) \
	( w += f(x, y, z) + data, w &= 0xFFFFFFFF, w = ROL(w,s), w += x, w &= 0xFFFFFFFF )
#endif

/*
 * The core of the MD5 algorithm, this alters an existing MD5 hash to
 * reflect the addition of 16 longwords of new data.  MD5Update blocks
 * the data and converts bytes into longwords for this routine.
 */
void 
MD5SumCoder::md5_transform()
{
#ifdef LEDA_CODER_EXACT_SIZE_INT32
#ifndef LITTLE_ENDIAN_MACHINE
	for (int i = 0; i <= 15; ++i) swap_bytes_uint32(Words[i]);
#endif
#else
	for (int i = 15; i >= 0; --i) Words[i] = getWord(i);
#endif

	uint32 a, b, c, d;

	a = Buf[0];
	b = Buf[1];
	c = Buf[2];
	d = Buf[3];

	MD5STEP (F1, a, b, c, d, Words[0] + 0xd76aa478, 7);
	MD5STEP (F1, d, a, b, c, Words[1] + 0xe8c7b756, 12);
	MD5STEP (F1, c, d, a, b, Words[2] + 0x242070db, 17);
	MD5STEP (F1, b, c, d, a, Words[3] + 0xc1bdceee, 22);
	MD5STEP (F1, a, b, c, d, Words[4] + 0xf57c0faf, 7);
	MD5STEP (F1, d, a, b, c, Words[5] + 0x4787c62a, 12);
	MD5STEP (F1, c, d, a, b, Words[6] + 0xa8304613, 17);
	MD5STEP (F1, b, c, d, a, Words[7] + 0xfd469501, 22);
	MD5STEP (F1, a, b, c, d, Words[8] + 0x698098d8, 7);
	MD5STEP (F1, d, a, b, c, Words[9] + 0x8b44f7af, 12);
	MD5STEP (F1, c, d, a, b, Words[10] + 0xffff5bb1, 17);
	MD5STEP (F1, b, c, d, a, Words[11] + 0x895cd7be, 22);
	MD5STEP (F1, a, b, c, d, Words[12] + 0x6b901122, 7);
	MD5STEP (F1, d, a, b, c, Words[13] + 0xfd987193, 12);
	MD5STEP (F1, c, d, a, b, Words[14] + 0xa679438e, 17);
	MD5STEP (F1, b, c, d, a, Words[15] + 0x49b40821, 22);
		
	MD5STEP (F2, a, b, c, d, Words[1] + 0xf61e2562, 5);
	MD5STEP (F2, d, a, b, c, Words[6] + 0xc040b340, 9);
	MD5STEP (F2, c, d, a, b, Words[11] + 0x265e5a51, 14);
	MD5STEP (F2, b, c, d, a, Words[0] + 0xe9b6c7aa, 20);
	MD5STEP (F2, a, b, c, d, Words[5] + 0xd62f105d, 5);
	MD5STEP (F2, d, a, b, c, Words[10] + 0x02441453, 9);
	MD5STEP (F2, c, d, a, b, Words[15] + 0xd8a1e681, 14);
	MD5STEP (F2, b, c, d, a, Words[4] + 0xe7d3fbc8, 20);
	MD5STEP (F2, a, b, c, d, Words[9] + 0x21e1cde6, 5);
	MD5STEP (F2, d, a, b, c, Words[14] + 0xc33707d6, 9);
	MD5STEP (F2, c, d, a, b, Words[3] + 0xf4d50d87, 14);
	MD5STEP (F2, b, c, d, a, Words[8] + 0x455a14ed, 20);
	MD5STEP (F2, a, b, c, d, Words[13] + 0xa9e3e905, 5);
	MD5STEP (F2, d, a, b, c, Words[2] + 0xfcefa3f8, 9);
	MD5STEP (F2, c, d, a, b, Words[7] + 0x676f02d9, 14);
	MD5STEP (F2, b, c, d, a, Words[12] + 0x8d2a4c8a, 20);

	MD5STEP (F3, a, b, c, d, Words[5] + 0xfffa3942, 4);
	MD5STEP (F3, d, a, b, c, Words[8] + 0x8771f681, 11);
	MD5STEP (F3, c, d, a, b, Words[11] + 0x6d9d6122, 16);
	MD5STEP (F3, b, c, d, a, Words[14] + 0xfde5380c, 23);
	MD5STEP (F3, a, b, c, d, Words[1] + 0xa4beea44, 4);
	MD5STEP (F3, d, a, b, c, Words[4] + 0x4bdecfa9, 11);
	MD5STEP (F3, c, d, a, b, Words[7] + 0xf6bb4b60, 16);
	MD5STEP (F3, b, c, d, a, Words[10] + 0xbebfbc70, 23);
	MD5STEP (F3, a, b, c, d, Words[13] + 0x289b7ec6, 4);
	MD5STEP (F3, d, a, b, c, Words[0] + 0xeaa127fa, 11);
	MD5STEP (F3, c, d, a, b, Words[3] + 0xd4ef3085, 16);
	MD5STEP (F3, b, c, d, a, Words[6] + 0x04881d05, 23);
	MD5STEP (F3, a, b, c, d, Words[9] + 0xd9d4d039, 4);
	MD5STEP (F3, d, a, b, c, Words[12] + 0xe6db99e5, 11);
	MD5STEP (F3, c, d, a, b, Words[15] + 0x1fa27cf8, 16);
	MD5STEP (F3, b, c, d, a, Words[2] + 0xc4ac5665, 23);

	MD5STEP (F4, a, b, c, d, Words[0] + 0xf4292244, 6);
	MD5STEP (F4, d, a, b, c, Words[7] + 0x432aff97, 10);
	MD5STEP (F4, c, d, a, b, Words[14] + 0xab9423a7, 15);
	MD5STEP (F4, b, c, d, a, Words[5] + 0xfc93a039, 21);
	MD5STEP (F4, a, b, c, d, Words[12] + 0x655b59c3, 6);
	MD5STEP (F4, d, a, b, c, Words[3] + 0x8f0ccc92, 10);
	MD5STEP (F4, c, d, a, b, Words[10] + 0xffeff47d, 15);
	MD5STEP (F4, b, c, d, a, Words[1] + 0x85845dd1, 21);
	MD5STEP (F4, a, b, c, d, Words[8] + 0x6fa87e4f, 6);
	MD5STEP (F4, d, a, b, c, Words[15] + 0xfe2ce6e0, 10);
	MD5STEP (F4, c, d, a, b, Words[6] + 0xa3014314, 15);
	MD5STEP (F4, b, c, d, a, Words[13] + 0x4e0811a1, 21);
	MD5STEP (F4, a, b, c, d, Words[4] + 0xf7537e82, 6);
	MD5STEP (F4, d, a, b, c, Words[11] + 0xbd3af235, 10);
	MD5STEP (F4, c, d, a, b, Words[2] + 0x2ad7d2bb, 15);
	MD5STEP (F4, b, c, d, a, Words[9] + 0xeb86d391, 21);

	Buf[0] += a;
	Buf[1] += b;
	Buf[2] += c;
	Buf[3] += d;
}



/// SHACoder ////////////////////////////////////////////////////////////////////////////////

SHACoder::SHACoder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams)
{}

SHACoder::SHACoder(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name)
{}

SHACoder::~SHACoder()
{
	memset((void*)this, 0xFF, sizeof(SHACoder));
	memset((void*)this, 0x55, sizeof(SHACoder));
	memset((void*)this, 0x00, sizeof(SHACoder));
}

void 
SHACoder::reset_checksum()
{
	ByteCount = 0;

	Buf[0] = 0x67452301;
	Buf[1] = 0xefcdab89;
	Buf[2] = 0x98badcfe;
	Buf[3] = 0x10325476;
	Buf[4] = 0xc3d2e1f0;
}

void 
SHACoder::update_checksum(byte b)
{
	Chunk[ByteCount & 0x3F] = b;
	if ((++ByteCount & 0x3F) == 0) {
		sha_transform();
	}
}

void 
SHACoder::finish_checksum()
{
/*
 * Final wrapup - pad to 64-byte boundary with the bit pattern 
 * 1 0* (64-bit count of bits processed, LSB-first)
 */

	unsigned cnt_in_chunk = unsigned(ByteCount & 0x3F);
	byte* ptr = &Chunk[0] + cnt_in_chunk;

	/* Set the first char of padding to 0x80.  This is safe since there is
	   always at least one byte free */
	*(ptr++) = 0x80; ++cnt_in_chunk;

	/* Bytes of padding needed to make 64 bytes */
	unsigned free_in_chunk = 64 - cnt_in_chunk;

	/* Pad out to 56 mod 64 */
	if (free_in_chunk < 8) {
		/* Two lots of padding:  Pad the first block to 64 bytes */
		memset(ptr, 0, free_in_chunk);
		sha_transform();
		/* Now fill the next block with 56 bytes */
		memset(Chunk, 0, 56);
	}
	else {
		/* Pad block to 56 bytes */
		memset(ptr, 0, free_in_chunk-8);
	}

#if defined(LITTLE_ENDIAN_MACHINE) || !defined(LEDA_CODER_EXACT_SIZE_INT32)
	// setWord will simulate big endian ...
	setWord(14, (ByteCount >> 29)& 0xFFFFFFFF); // higher 32 Bits of SumOfCounts number of BITS
	setWord(15, (ByteCount << 3) & 0xFFFFFFFF); // lower  32 Bits of SumOfCounts number of BITS

	sha_transform();

	// setDigest will simulate big endian ...
	uint32 a = Buf[0];
	uint32 b = Buf[1];
	uint32 c = Buf[2];
	uint32 d = Buf[3];
	uint32 e = Buf[4];
	setDigest(0, a); setDigest(1, b); setDigest(2, c); setDigest(3, d); setDigest(4, e);
#else
	Words[14] = (ByteCount >> 29)& 0xFFFFFFFF; // higher 32 Bits of SumOfCounts number of BITS
	Words[15] = (ByteCount << 3) & 0xFFFFFFFF; // lower  32 Bits of SumOfCounts number of BITS

	sha_transform();
#endif
}

/* The four core functions - F1 is optimized somewhat */

#undef F1
#undef F2
#undef F3
#undef F4
/* #define F1(x, y, z) (x & y | ~x & z) */
#define F1(x, y, z) (z ^ (x & (y ^ z)))
#define F2(x, y, z) (x ^ y ^ z)
#define F3(x, y, z) ((x & y) | (x & z) | (y & z))
#define F4(x, y, z) (x ^ y ^ z)

/* This is the central step in the SHA-1 algorithm. */
#ifdef LEDA_CODER_EXACT_SIZE_INT32
#define SHASTEP(f, t, K) \
	{ uint32 tmp = ROL(a,5) + f(b,c,d) + e + Words[t] + K; \
	  e = d; d = c; c = ROL(b,30); b = a; a = tmp; }
#else
#define SHASTEP(f, t, K) \
	{ uint32 tmp = ROL(a,5) + f(b,c,d) + e + Words[t] + K; \
	  e = d; d = c; c = ROL(b,30); b = a; a = tmp & 0xFFFFFFFF; }
#endif

void 
SHACoder::sha_transform()
{
#ifdef LEDA_CODER_EXACT_SIZE_INT32
#ifdef LITTLE_ENDIAN_MACHINE
	for (int i = 0; i <= 15; ++i) swap_bytes_uint32(Words[i]);
#endif
#else
	for (int i = 15; i >= 0; --i) Words[i] = getWord(i);
#endif

	for (int t = 16; t <= 79; ++t)
		Words[t] = ROL((Words[t-3] ^ Words[t-8] ^ Words[t-14] ^ Words[t-16]), 1); 

	uint32 a, b, c, d, e;

	a = Buf[0];
	b = Buf[1];
	c = Buf[2];
	d = Buf[3];
	e = Buf[4];

	// rounds 0 .. 19
	SHASTEP (F1,  0, 0x5a827999);
	SHASTEP (F1,  1, 0x5a827999);
	SHASTEP (F1,  2, 0x5a827999);
	SHASTEP (F1,  3, 0x5a827999);
	SHASTEP (F1,  4, 0x5a827999);
	SHASTEP (F1,  5, 0x5a827999);
	SHASTEP (F1,  6, 0x5a827999);
	SHASTEP (F1,  7, 0x5a827999);
	SHASTEP (F1,  8, 0x5a827999);
	SHASTEP (F1,  9, 0x5a827999);
	SHASTEP (F1, 10, 0x5a827999);
	SHASTEP (F1, 11, 0x5a827999);
	SHASTEP (F1, 12, 0x5a827999);
	SHASTEP (F1, 13, 0x5a827999);
	SHASTEP (F1, 14, 0x5a827999);
	SHASTEP (F1, 15, 0x5a827999);
	SHASTEP (F1, 16, 0x5a827999);
	SHASTEP (F1, 17, 0x5a827999);
	SHASTEP (F1, 18, 0x5a827999);
	SHASTEP (F1, 19, 0x5a827999);
		
	// rounds 20 .. 39
	SHASTEP (F2, 20, 0x6ed9eba1);
	SHASTEP (F2, 21, 0x6ed9eba1);
	SHASTEP (F2, 22, 0x6ed9eba1);
	SHASTEP (F2, 23, 0x6ed9eba1);
	SHASTEP (F2, 24, 0x6ed9eba1);
	SHASTEP (F2, 25, 0x6ed9eba1);
	SHASTEP (F2, 26, 0x6ed9eba1);
	SHASTEP (F2, 27, 0x6ed9eba1);
	SHASTEP (F2, 28, 0x6ed9eba1);
	SHASTEP (F2, 29, 0x6ed9eba1);
	SHASTEP (F2, 30, 0x6ed9eba1);
	SHASTEP (F2, 31, 0x6ed9eba1);
	SHASTEP (F2, 32, 0x6ed9eba1);
	SHASTEP (F2, 33, 0x6ed9eba1);
	SHASTEP (F2, 34, 0x6ed9eba1);
	SHASTEP (F2, 35, 0x6ed9eba1);
	SHASTEP (F2, 36, 0x6ed9eba1);
	SHASTEP (F2, 37, 0x6ed9eba1);
	SHASTEP (F2, 38, 0x6ed9eba1);
	SHASTEP (F2, 39, 0x6ed9eba1);
		
	// rounds 40 .. 59
	SHASTEP (F3, 40, 0x8f1bbcdc);
	SHASTEP (F3, 41, 0x8f1bbcdc);
	SHASTEP (F3, 42, 0x8f1bbcdc);
	SHASTEP (F3, 43, 0x8f1bbcdc);
	SHASTEP (F3, 44, 0x8f1bbcdc);
	SHASTEP (F3, 45, 0x8f1bbcdc);
	SHASTEP (F3, 46, 0x8f1bbcdc);
	SHASTEP (F3, 47, 0x8f1bbcdc);
	SHASTEP (F3, 48, 0x8f1bbcdc);
	SHASTEP (F3, 49, 0x8f1bbcdc);
	SHASTEP (F3, 50, 0x8f1bbcdc);
	SHASTEP (F3, 51, 0x8f1bbcdc);
	SHASTEP (F3, 52, 0x8f1bbcdc);
	SHASTEP (F3, 53, 0x8f1bbcdc);
	SHASTEP (F3, 54, 0x8f1bbcdc);
	SHASTEP (F3, 55, 0x8f1bbcdc);
	SHASTEP (F3, 56, 0x8f1bbcdc);
	SHASTEP (F3, 57, 0x8f1bbcdc);
	SHASTEP (F3, 58, 0x8f1bbcdc);
	SHASTEP (F3, 59, 0x8f1bbcdc);
		
	// rounds 60 .. 79
	SHASTEP (F4, 60, 0xca62c1d6);
	SHASTEP (F4, 61, 0xca62c1d6);
	SHASTEP (F4, 62, 0xca62c1d6);
	SHASTEP (F4, 63, 0xca62c1d6);
	SHASTEP (F4, 64, 0xca62c1d6);
	SHASTEP (F4, 65, 0xca62c1d6);
	SHASTEP (F4, 66, 0xca62c1d6);
	SHASTEP (F4, 67, 0xca62c1d6);
	SHASTEP (F4, 68, 0xca62c1d6);
	SHASTEP (F4, 69, 0xca62c1d6);
	SHASTEP (F4, 70, 0xca62c1d6);
	SHASTEP (F4, 71, 0xca62c1d6);
	SHASTEP (F4, 72, 0xca62c1d6);
	SHASTEP (F4, 73, 0xca62c1d6);
	SHASTEP (F4, 74, 0xca62c1d6);
	SHASTEP (F4, 75, 0xca62c1d6);
	SHASTEP (F4, 76, 0xca62c1d6);
	SHASTEP (F4, 77, 0xca62c1d6);
	SHASTEP (F4, 78, 0xca62c1d6);
	SHASTEP (F4, 79, 0xca62c1d6);
		
	Buf[0] += a;
	Buf[1] += b;
	Buf[2] += c;
	Buf[3] += d;
	Buf[4] += e;
}

LEDA_END_NAMESPACE
