// compression module includes
#include <LEDA/coding/coder_strbuf_adapter.h>

// core LEDA includes
#include <LEDA/system/file.h>
#include <string.h> // for memcpy (see below)

#if defined(_MSC_VER) && _MSC_VER < 1300
#define LEDA_COMPRESSION_WA5
#endif

#ifndef LEDA_COMPRESSION_WA8
#define STREAM_BLOCKED StreamBlocked
#else
#define STREAM_BLOCKED EOF
#endif

#if defined(LEDA_DEBUG)
#include <LEDA/system/assert.h>
#else
#undef assert
#define	assert(cond) ((void)0)
#endif

LEDA_BEGIN_NAMESPACE

/// coder_strbuf_adapater ///////////////////////////////////////////////////////////////////

coder_strbuf_adapter::coder_strbuf_adapter
 (coder_base* coder, bool own_coder, streambuf* strbuf, bool own_strbuf, uint32 buf_size)
 : State(s_init), InternalBuffer(0),
   pCoder(coder), pEncodedStream(strbuf), 
   OwnCoder(own_coder), OwnEncodedStream(own_strbuf),
   pStreamForBuffering(0)
{
	set_buffer_size(buf_size);
}

coder_strbuf_adapter::~coder_strbuf_adapter()
{
	finalize();

	if (InternalBuffer != TinyBuffer) 
		delete[] InternalBuffer; 
	InternalBuffer = 0;
}

void
coder_strbuf_adapter::finalize()
// should (only) be called in destructor of coder_strbuf_adapter or derived classes
{
	if (State == s_finalized) return;

	// if we are still in the encoding of the first pass, we end it now
	if (State == s_encoding_first_pass && !SignalEof) 
		sputEOF(); 

	set_state(s_finalized); // also clears errors, if any
	set_coder(0); set_encoded_stream(0);
}

void
coder_strbuf_adapter::do_reset(coder_base::reset_type t)
{
	assert(t >= coder_base::r_hard && t <= coder_base::r_block);

	if (t == coder_base::r_hard || t == coder_base::r_soft) 
		set_state(s_init); // also clears errors
	else
		set_state( state( s_init | (State & (s_encoding | s_decoding))) );

	setg(0,0,0); setp(0,0);

	if (get_coder()) get_coder()->do_reset(t);
	if (t == coder_base::r_hard || t == coder_base::r_soft) set_encoded_stream(0);

	if (pStreamForBuffering) {
		destroy_internal_stream(pStreamForBuffering, TmpFileName);
		pStreamForBuffering = 0; TmpFileName = string();
	}
}

void
coder_strbuf_adapter::set_buffer_size(uint32 buf_size)
{
	if (! running()) {
		setg(0,0,0); setp(0,0);
		if (InternalBuffer != TinyBuffer) delete[] InternalBuffer;

		BufferSize = buf_size;
		if (BufferSize < 1) BufferSize = sizeof(TinyBuffer);

		if (BufferSize > sizeof(TinyBuffer))
			InternalBuffer = new char[BufferSize];
		else
			InternalBuffer = TinyBuffer;
	}
	else LEDA_EXCEPTION(1, "coder_strbuf_adapter::set_buffer_size: wrong state");
}

void 
coder_strbuf_adapter::set_coder(coder_base* coder, bool own_coder)
{
	if (! running()) {
		if (OwnCoder) delete pCoder;
		pCoder = coder; OwnCoder = own_coder;
	}
	else LEDA_EXCEPTION(1, "coder_strbuf_adapter::set_coder: wrong state");
}

coder_base* 
coder_strbuf_adapter::release_ownership_of_coder()
{
	OwnCoder = false; // okay, even if running
	return pCoder;
}

void 
coder_strbuf_adapter::set_encoded_stream(streambuf* stream, bool own_stream)
{
	if (! running()) {
		if (OwnEncodedStream) delete pEncodedStream;
		pEncodedStream = stream; OwnEncodedStream = own_stream;
	}
	else LEDA_EXCEPTION(1, "coder_strbuf_adapter::set_encoded_stream: wrong state");
}

void 
coder_strbuf_adapter::set_encoded_file(const char* file_name, ios::openmode mode)
{
	if (file_name == 0) { set_encoded_stream(0, false); return; }

	filebuf* stream = new filebuf;
	if (! stream->open(file_name, ios::openmode(mode | coder_base::binary))) {
		delete stream; stream = 0;
		LEDA_EXCEPTION(1, "coder_strbuf_adapter::set_encoded_file: open failed");
	}

	set_encoded_stream(stream, true); // own it and destroy it
}

void 
coder_strbuf_adapter::sputEOF()
{
	switch (State) {
	case s_init:		
	case s_init_encoding:		
		start_encoding();
		// fall through
	case s_encoding_first_pass:
		SignalEof = true;
		pubsync(); // will encode the internal buffer and change to s_encoding_later_pass
		if (! failed()) set_state(s_encoding_finished);
		break;
	default:
		error();
	}
}

void 
coder_strbuf_adapter::error(const char* err_msg)
{
	if (! failed()) set_state(s_error);
}

void 
coder_strbuf_adapter::set_state(state new_state)
{
	observe_state_change(new_state);
	State = new_state;

	if ( (State & (s_finished | s_finalized | s_error)) != 0 ) {
		SignalEof = true;
		if (pStreamForBuffering) {
			destroy_internal_stream(pStreamForBuffering, TmpFileName);
			pStreamForBuffering = 0; TmpFileName = string();
		}
	}
}

streambuf* 
coder_strbuf_adapter::create_internal_stream(string filename)
{
	filebuf* int_stream = new filebuf();
	if (! int_stream->open(filename, ios::openmode(ios::out | ios::in | ios::trunc | coder_base::binary))) {
		delete int_stream; int_stream = 0;
		error("coder_strbuf_adapter::create_internal_stream: could not open tmp file!");
	}
	return int_stream;
}

void 
coder_strbuf_adapter::destroy_internal_stream(streambuf* int_stream, string filename)
{
	delete int_stream;
	delete_file(filename);
}

int 
coder_strbuf_adapter::overflow(int c)
{
	switch (State) {
	case s_init:
	case s_init_encoding:
		start_encoding();
		return c != EOF ? sputc(c) : 0;
			// if the char is not EOF, put it into the buffer

	case s_encoding_first_pass:	
		encode_internal_buffer();
		return c != EOF ? sputc(c) : 0;
			// if the char is not EOF, put it into the buffer

	case s_decoding:
		if (c == EOF) return c;
		(*pptr()) = (char) c; NumOverFlowChars = 1;
		return STREAM_BLOCKED;

	default:
		error("coder_strbuf_adapter::overflow: wrong state"); 
		return EOF;
	}
}

int 
coder_strbuf_adapter::underflow()
{
	switch (State) {
	case s_init:
	case s_init_decoding:
		start_decoding(); // fall-through

	case s_decoding:	
#ifdef LEDA_COMPRESSION_WA5
		// the buggy streambuf version of MSVC 6.0 may have called although the 
		// buffer is NOT empty
		if (gptr() && gptr() < egptr()) return (int)(*gptr() & 0xff);
#endif
		assert(gptr() == 0 || gptr() == egptr());
		return decode_into_internal_buffer();

	case s_encoding_first_pass:
		if (!SignalEof) return STREAM_BLOCKED;
		set_state(s_encoding_later_pass);
		return EOF;

	case s_encoding_later_pass:
#ifdef LEDA_COMPRESSION_WA5
		if (gptr() != 0 && pStreamForBuffering) {
			// i.e. the buggy streambuf version of MSVC 6.0 should have called uflow()
			setg(0,0,0);
			return pStreamForBuffering->snextc();
		}
#endif
#ifdef LEDA_COMPRESSION_WA9
		assert(!gptr() || gptr() >= egptr());
		if (pStreamForBuffering) {
			int c = pStreamForBuffering->sbumpc();
			TinyBuffer[0] = (char) c;
			if (c != EOF) setg(TinyBuffer, TinyBuffer, TinyBuffer+1);
			return c;
		}
#else
		assert(gptr() == 0);
#endif
		return pStreamForBuffering ? pStreamForBuffering->sgetc() : EOF;

	case s_decoding_finished: // decoding has finished so return only EOFs
		return EOF;

	default:
		error("coder_strbuf_adapter::underflow: wrong state"); 
		return EOF;
	}
}

int 
coder_strbuf_adapter::uflow()
{
	switch (State) {
	case s_encoding_later_pass:
		return pStreamForBuffering ? pStreamForBuffering->sbumpc() : EOF;

	default:
		{
			int c = underflow();
			return c < 0 ? c : sbumpc();
		}
	}
}

int 
coder_strbuf_adapter::sync()
{
	if (State == s_encoding_first_pass) {	
		encode_internal_buffer();
	}

	if (encoding() && pEncodedStream && !failed()) {
		// NOTE: sync pEncodedStream ONLY if encoding!
		if (pEncodedStream->pubsync() != 0) {
			error("coder_strbuf_adapter::sync");
		}
	}

	return failed() ? -1 : 0; // failure or success
}

streampos 
coder_strbuf_adapter::seekoff(streamoff off, ios::seekdir way, ios::openmode which)
{
#ifdef LEDA_COMPRESSION_WA8
	/*
		If the an adapter function (namely overflow/underflow) returns EOF, 
		this function is used by coder_base::check_eof to distinguish between 
		a real EOF and StreamBlocked.
	*/
	if (way == ios::end && off == streamoff(1))
		return SignalEof ? streampos(EOF) : streampos(StreamBlocked);
#endif

	if (State == s_encoding_later_pass && pStreamForBuffering) {
		setg(0,0,0); setp(0,0);
		return pStreamForBuffering->pubseekoff(off, way, which);
	}

	// error(); -> NO error, so we can check safely whether seeks are supported
	return streampos(-1);
}

streampos 
coder_strbuf_adapter::seekpos(streampos sp, ios::openmode which)
{
	if (State == s_encoding_later_pass && pStreamForBuffering) {
		setg(0,0,0); setp(0,0);
		return pStreamForBuffering->pubseekpos(sp, which);
	}

	// error(); -> NO error, so we can check safely whether seeks are supported
	return streampos(-1);
}

streamsize 
coder_strbuf_adapter::xsgetn(char* buffer, streamsize count)
{
	streamsize total_read = 0;

	while (count > 0) {
		int pending;
		if (gptr() && (pending = (int)(egptr() - gptr())) > 0) {
			// copy from internal buffer
			if (count < pending) pending = (int)count;
			memcpy(buffer, gptr(), pending);
			buffer += pending; total_read += pending; count -= pending;
			gbump(pending);
			continue;
		}
		
		int c = uflow();
		if (c < 0) {
#ifdef LEDA_COMPRESSION_WA8
			assert(c == EOF);
			if (!SignalEOF) c = StreamBlocked;
#endif
			if (c == StreamBlocked) total_read = StreamBlocked - total_read;
			break;
		}

		*buffer++ = (char) c; ++total_read; --count;
	}

	return total_read;
}

streamsize 
coder_strbuf_adapter::xsputn(const char* buffer, streamsize count)
{
	streamsize total_written = 0;

	while (count > 0) {
		int space;
		if (pptr() && (space = (int)(epptr() - pptr())) > 0) {
			// copy from internal buffer
			if (count < space) space = (int)count;
			memcpy(pptr(), buffer, space);
			buffer += space; total_written += space; count -= space;
			pbump(space);
			continue;
		}
		
		int c = overflow((byte)*buffer);
		if (c < 0) {
			if (c == StreamBlocked) total_written = StreamBlocked - 1 - total_written;
				// NOTE: "-1" because one overflow character was written
			break;
		}

		++buffer; ++total_written; --count;
	}

	return total_written;
}

void 
coder_strbuf_adapter::start_encoding()
{
	if (!pCoder || pCoder->failed() || !pEncodedStream) {
		error("coder_strbuf_adapter::start_encoding: no coder or stream");
		return;
	}

	if (! pCoder->is_single_pass_encoder()) {
		TmpFileName = tmp_file_name();
		pStreamForBuffering = create_internal_stream(TmpFileName);
	}

	set_state(s_encoding_first_pass);
	SignalEof = false;
	pCoder->set_src_stream(this); pCoder->set_tgt_stream(pEncodedStream);
	setp(InternalBuffer, InternalBuffer + BufferSize); // internal puffer empty
}

void 
coder_strbuf_adapter::encode_internal_buffer()
{
	if (pStreamForBuffering) {
		int bytes_to_write = (int) (pptr() - pbase());
		bool ok = pStreamForBuffering->sputn(pbase(), bytes_to_write) == bytes_to_write;
		if (! ok) error("coder_strbuf_adapter::encode_internal_buffer: write to internal stream failed");
	}

	setg(pbase(), pbase(), pptr());

	pCoder->do_encode();
	if (pCoder->failed()) { setp(0,0); error(); return; }

	setp(InternalBuffer, InternalBuffer + BufferSize); // internal puffer empty again
}

void
coder_strbuf_adapter::start_decoding()
{
	if (!pCoder || pCoder->failed() || !pEncodedStream) {
		LEDA_EXCEPTION(1, "coder_strbuf_adapter::start_decoding: no coder or stream");
		return;
	}

	set_state(s_decoding);
	SignalEof = false;
	pCoder->set_src_stream(pEncodedStream); pCoder->set_tgt_stream(this);
}

int 
coder_strbuf_adapter::decode_into_internal_buffer()
{
	if (pCoder->finished()) { setg(0,0,0); set_state(s_decoding_finished); return EOF; }

	NumOverFlowChars = 0; 
	setp(InternalBuffer, InternalBuffer + BufferSize - 1); // internal puffer ready
		// "-1" -> save one character for overflow

	pCoder->do_decode();
	// NOTE: Even if pCoder may be finished, we may NOT change *this into the state
	//       s_decoding_finished, because the buffer may not be empty!

	if (pCoder->failed()) { setg(0,0,0); error(); return EOF; }	

	char* _gptr = pbase(); char* _egptr = pptr()+NumOverFlowChars;
	setg(_gptr, _gptr, _egptr);
	if (_gptr && (_gptr < _egptr)) return int((byte)*_gptr);
		// if the buffer is not empty, return the first char

	// But if the buffer is empty, then we can finish immediately ...
	assert(pCoder->finished());
	setg(0,0,0); set_state(s_decoding_finished); return EOF;
}

/// coder_strbuf_adapter_with_seek //////////////////////////////////////////////////////////

coder_strbuf_adapter_with_seek::coder_strbuf_adapter_with_seek
 (coder_base* coder, bool own_coder, streambuf* strbuf, bool own_strbuf, uint32 buf_size)
 : base(coder, own_coder, strbuf, own_strbuf, buf_size),
   PosCorrespToIntBuffer(0), PosEOF(-1)
{}

int
coder_strbuf_adapter_with_seek::underflow()
{
	if (get_state() == s_decoding) 
		PosCorrespToIntBuffer += streamoff(egptr() - eback());

	return base::underflow();
}

int
coder_strbuf_adapter_with_seek::overflow(int c)
{
	// NOTE: we do nothing if get_state() == s_init!
	if (get_state() == s_encoding_first_pass)
		PosCorrespToIntBuffer += streamoff(pptr() - pbase());
			// we do not have to count the overflow char!

	return base::overflow(c);
}

int
coder_strbuf_adapter_with_seek::sync()
{
	// since the buffer will be flushed, we have to increase PosCorrespToIntBuffer!
	if (get_state() == s_encoding_first_pass)
		PosCorrespToIntBuffer += streamoff(pptr() - pbase());

	return base::sync();
}

streampos 
coder_strbuf_adapter_with_seek::seekoff(streamoff off, ios::seekdir way, ios::openmode which)
{
#ifdef LEDA_COMPRESSION_WA8
	if (way == ios::end && off == streamoff(1)) return base::seekoff(off, way, which);
#endif

	if (later_pass()) return base::seekoff(off, way, which);

	switch (way) {
	case ios::beg:	return seekpos(off, which);
	case ios::cur:	return seekpos(get_current_pos() + off, which);
	case ios::end:	return seekpos(get_pos_EOF(which) + off, which);
	default:	error("unexpected case"); return EOF;
	}
	return streampos(-1);
}

streampos 
coder_strbuf_adapter_with_seek::seekpos(streampos new_pos, ios::openmode which)
{
	if (later_pass()) return base::seekpos(new_pos, which);

	if (failed()) return streampos(-1);

	if ( encoding() || (!decoding() && which == ios::out) )
		return encode_seekpos(new_pos, which);
	else
		return decode_seekpos(new_pos, which);
}

void 
coder_strbuf_adapter_with_seek::do_reset(coder_base::reset_type t)
{
	base::do_reset(t);
	PosCorrespToIntBuffer = 0;
	if (t == coder_base::r_hard || t == coder_base::r_soft) {
		PosEOF = -1;
	}
}

streampos 
coder_strbuf_adapter_with_seek::get_pos_EOF(ios::openmode which)
{
	if (!decoding() && (encoding() || which == ios::out)) return encode_get_current_pos();
		// we have to take care of the special case that state == s_init

	if (PosEOF == streampos(-1)) {
		// we do not know posEOF yet, so we compute it ...
		if (!get_coder()) { error(); return streampos(-1); }

		if (finished()) return PosEOF = decode_get_current_pos_of_coder();

		if (! running()) { start_decoding(); }
		PosEOF = get_coder()->decode_seek(decode_get_current_pos_of_coder(), -1);
		if (get_coder()->failed()) error();
		PosCorrespToIntBuffer = PosEOF; setg(0,0,0); // we changed the current pos!
	}
	return PosEOF;
}

streampos 
coder_strbuf_adapter_with_seek::encode_seekpos(streampos new_pos, ios::openmode which)
{
	assert( !decoding() && (which & ios::out)!=0 && !failed() );

	if (finished()) { error(); return streampos(-1); }
		// no restart in encoding mode (after sputEOF)

	streampos pos = encode_get_current_pos();
	if (pos == new_pos) return new_pos; // short-cut

	if (! running()) {
		start_encoding();
		pos = 0;
		if (failed()) return streampos(-1);
	}
	else encode_internal_buffer();

	// now the coder is running and its current position equals pos
	if (!get_coder()) { error(); return streampos(-1); }
	pos = get_coder()->encode_seek(pos, new_pos);
	if (get_coder()->failed()) error();

	// synchronize the buffered stream with the coder pos of the coder
	if (get_stream_for_buffering()) {
		bool ok = get_stream_for_buffering()->pubseekpos(new_pos, ios::out) == new_pos;
		if (!ok) error();
	}

	PosCorrespToIntBuffer = pos; setg(0,0,0);
	return pos;
}

streampos 
coder_strbuf_adapter_with_seek::decode_seekpos(streampos new_pos, ios::openmode which)
{
	assert( !encoding() && (which & ios::in)!=0 && !failed() );

	streampos pos = decode_get_current_pos();
	if (pos == new_pos) return new_pos; // short-cut

	if (! running()) {
		if (finished()) do_reset(coder_base::r_seek);
		start_decoding();
		pos = 0;
		if (failed()) return streampos(-1);
	}

	// now the coder is running and its current position equals pos
	if (!get_coder()) { error(); return streampos(-1); }
	streampos coder_pos = decode_get_current_pos_of_coder();

	// if new_pos falls into the internal buffer, we are lucky, 
	// and we not have to bother the coder
	if (PosCorrespToIntBuffer <= new_pos && new_pos <= coder_pos) {
		streamoff off = new_pos - pos;
		setg(eback(), gptr() + off, egptr()); 
		return new_pos;
	}

	// otherwise we leave up the seek to the coder
	pos = get_coder()->decode_seek(coder_pos, new_pos);
	if (get_coder()->failed()) error();

	PosCorrespToIntBuffer = pos; setg(0,0,0);
	return pos;
}

LEDA_END_NAMESPACE
