// compression module includes
#include <LEDA/coding/dict_coder.h>

#include <stdlib.h>

LEDA_BEGIN_NAMESPACE

/// DictCoder::dictionary ///////////////////////////////////////////////////////////////////

DictCoder::IndexType 
DictCoder::dictionary::find(CodeType curr_code, int in)
{
	IndexType ndx;
	int step = 11, pastzero = 0;

	ndx = ((IndexType) curr_code << 8) | in;

	ndx = ndx % Dictionary_Size;

	for (;;) {
		if (Entries[ndx].code == c_NIL)
			break;

		if (Entries[ndx].parent == curr_code && Entries[ndx].symbol == in)
		    break;

		ndx += step;
		if (ndx >= Dictionary_Size) {
			ndx -= Dictionary_Size;
			pastzero++;

			// safety check
			if (pastzero > 5) step = 1;
		}
	}

	return ndx;
}

/// DictCoder ///////////////////////////////////////////////////////////////////////////////

DictCoder::DictCoder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams), 
  DecodeBuffer(0), DecodeBufferSize(0)
{}

DictCoder::DictCoder(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name),
  DecodeBuffer(0), DecodeBufferSize(0)
{}

DictCoder::~DictCoder()
{
	if (DecodeBuffer) { free(DecodeBuffer); DecodeBuffer = 0; DecodeBufferSize = 0; }
}

void 
DictCoder::init_encoding()
{
	base::init_encoding();

	TotalBytesIn = 0;
	TotalBytesOut = 0;
	CurrBytesIn = 0;
	CurrBytesOut = 0;

	Dictionary.clear();

	// process first character
	int in = read_byte();
	if (in >= 0 && in != EOF) {
		CurrCode = (CodeType) in;
		CurrCode &= 0xFF;
		TotalBytesIn++;
		CurrBytesIn++;
	}
	else { // in < 0
		if (in == EOF) {
			write_code(dictionary::c_EndOfInput);
			encoding_finished();
		}
		else
			error("RLECoder::init_encoding: Unexpected STREAM_BLOCKED");
		return;
	}
}

void 
DictCoder::encode_stream()
{
	int in;
	while ((in = read_byte()) >= 0 && in != EOF) {
		TotalBytesIn++;
		CurrBytesIn++;

		IndexType idx = Dictionary.find(CurrCode, in);
		if (Dictionary[idx].code == dictionary::c_NIL) {
			// not a match
			write_code(CurrCode);

			// update the dictionary
			if (!Dictionary.full())
				Dictionary.insert_at_idx(idx, in, CurrCode);

			CurrCode = in;

			// had a miss; check compression efficiency
			if (CurrBytesIn >= 10000) {
				unsigned ratio;
				ratio = (TotalBytesOut * 100)/CurrBytesIn;

				CurrBytesIn = 0;
				CurrBytesOut = 0;
				if (ratio > BestRatio) {
					if (ratio > 50 && (ratio > 90 || ratio > LastRatio + 10)) {
						write_code(dictionary::c_NewDictionary);
						Dictionary.clear();
					}
				}
				else
					BestRatio = ratio;
				LastRatio = ratio;
			}
		}
		else // so far we match; go ahead
			CurrCode = Dictionary[idx].code;
	}

	if (in == EOF) {
		write_code(CurrCode);
		write_code(dictionary::c_EndOfInput);
		encoding_finished();
	}
}


void 
DictCoder::init_decoding()
{
	base::init_decoding();

	if (!DecodeBuffer) {
		DecodeBufferSize = DecodeBufferSizeStep;
		DecodeBuffer = (byte*) malloc(DecodeBufferSize);
	}

	NewDictionary = true;
	PendingDecodedBytesCount = 0;
}

void 
DictCoder::decode_stream()
{
	while (PendingDecodedBytesCount) {
		if (write_byte(DecodeBuffer[--PendingDecodedBytesCount]) < 0) return;
	}

	CodeType incode;
	for (;;) {
		read_code(incode);
		if (failed()) return;

		if (incode == dictionary::c_EndOfInput) {
			decoding_finished(); return;
		}

		if (incode == dictionary::c_NewDictionary) {
			NewDictionary = true; continue;
		}

		if (NewDictionary) {
			Dictionary.quick_clear();
			NewDictionary  = false;
			PreviousCode   = incode;
			PreviousSymbol = (byte) incode;
			if (write_byte((byte) incode) < 0) return;
			continue;
		}

		if (Dictionary.contains(incode)) {
			load_decode_buffer(0, incode);
		}
		else {
			load_decode_buffer(1, PreviousCode);
			DecodeBuffer[0] = PreviousSymbol;
		}
		
		// update dictionary
		PreviousSymbol = DecodeBuffer[PendingDecodedBytesCount-1];
		if (! Dictionary.full()) {
			Dictionary.append(PreviousSymbol, PreviousCode);
		}
		PreviousCode = incode;

		// output decoded bytes
		while (PendingDecodedBytesCount) {
			if (write_byte(DecodeBuffer[--PendingDecodedBytesCount]) < 0) return;
		}
	}

	error();
}

void 
DictCoder::load_decode_buffer(unsigned count, CodeType code)
{
	if (! Dictionary.contains(code)) {
		error("DictCoder::load_decode_buffer: invalid code");
		PendingDecodedBytesCount = 0;
		return;
	}

	while (code >= dictionary::c_PresetMax) {
		DecodeBuffer[count++] = Dictionary[code].symbol;
		code = Dictionary[code].parent;

		if (count == DecodeBufferSize) {
			DecodeBufferSize += DecodeBufferSizeStep;
			DecodeBuffer = (byte*) realloc(DecodeBuffer, DecodeBufferSize);
			if (!DecodeBuffer) {
				error("out of memory");
				DecodeBufferSize = PendingDecodedBytesCount = 0;
				return;
			}
		}
	}
	DecodeBuffer[count++] = (byte) code;

	PendingDecodedBytesCount = count;
}

LEDA_END_NAMESPACE
