// crypt module includes
#include <LEDA/coding/block_cipher_factory.h>
#include <LEDA/coding/block_ciphers.h>

// core LEDA includes
#include <LEDA/system/error.h>

LEDA_BEGIN_NAMESPACE

/// block_cipher_factory ////////////////////////////////////////////////////////////////////

#if defined(LEDA_MULTI_THREAD)
leda_lock block_cipher_factory::mutex;
#endif

block_cipher_factory BlockCipherFactory; // The one and only instance

block_cipher_factory::block_cipher_factory()
{
	if (this != &BlockCipherFactory)
		LEDA_EXCEPTION(1, "block_cipher_factory is a singleton!");
}

#define REGISTER_CIPHER(CIPHER) register_id(CIPHER::ID, construct_##CIPHER)

// define a constructor function for a cipher
#define DEF_CTOR_FUNC(CIPHER) \
	static block_cipher_base* construct_##CIPHER(block_cipher_factory::id_type id) \
	{ return new CIPHER; }

DEF_CTOR_FUNC(Blowfish);
DEF_CTOR_FUNC(Twofish);
DEF_CTOR_FUNC(Rijndael);
DEF_CTOR_FUNC(NoBlockCipher);

void 
block_cipher_factory::init()
{
	pCipherConstructorMap = new h_array<id_type, cipher_construction_func>(0);

	// register the LEDA block_ciphers ...
	REGISTER_CIPHER(Blowfish);
	REGISTER_CIPHER(Twofish);
	REGISTER_CIPHER(Rijndael);
	REGISTER_CIPHER(NoBlockCipher);
}

void 
block_cipher_factory::destroy()
{
	delete pCipherConstructorMap; pCipherConstructorMap = 0;
}

void 
block_cipher_factory::register_id(id_type id, cipher_construction_func ctor)
{
	lock();

	if (pCipherConstructorMap->defined(id)) { unlock(); return; }
	(*pCipherConstructorMap)[id] = ctor;

	unlock();
}

block_cipher_base* 
block_cipher_factory::construct_cipher_by_id(id_type id)
{
	lock();

	block_cipher_base* cipher = 0;
	if (pCipherConstructorMap->defined(id)) {
		cipher_construction_func ctor = (*pCipherConstructorMap)[id];
		cipher = ctor(id);
	}

	unlock();

	return cipher;
}

list<block_cipher_base*> 
block_cipher_factory::get_all_ciphers()
{
	list<block_cipher_base*> all_ciphers;

	lock();

	id_type id;
	forall_defined(id, *pCipherConstructorMap) {
		cipher_construction_func ctor = (*pCipherConstructorMap)[id];
		block_cipher_base* p_cipher = ctor(id);
		if (p_cipher) all_ciphers.append(p_cipher);
	}

	unlock();

	return all_ciphers;
}

void 
block_cipher_factory::print_registered_ciphers(ostream& out)
{
	list<block_cipher_base*> ciphers = get_all_ciphers();

	block_cipher_base* cipher;
	forall(cipher, ciphers) {
		out << string("%08x", cipher->get_id()) << " " << cipher->get_description();
		out << endl;
		delete cipher;
	}
}

/// block_cipher_factory_init ///////////////////////////////////////////////////////////////

unsigned block_cipher_factory_init::Count = 0;

block_cipher_factory_init::block_cipher_factory_init()
{
	if (Count++ == 0) {
		BlockCipherFactory.init();
	}
}

block_cipher_factory_init::~block_cipher_factory_init()
{
	if (--Count == 0) BlockCipherFactory.destroy();
}

LEDA_END_NAMESPACE
