// crypt module includes
#include <LEDA/coding/block_ciphers.h>

// core library includes
#include <string.h> // for memcpy

LEDA_BEGIN_NAMESPACE

#define ROL(x,n) (((x) << ((n) & 0x1F)) | ((x) >> (32-((n) & 0x1F))))
#define ROR(x,n) (((x) >> ((n) & 0x1F)) | ((x) << (32-((n) & 0x1F))))

#if defined(_MSC_VER) && defined(LEDA_CODER_EXACT_SIZE_INT32)
#include <stdlib.h> // get declarations of intrinsic rotation
#undef	ROL
#undef	ROR
#pragma intrinsic(_lrotl,_lrotr) // use intrinsic compiler rotations
#define	ROL(x,n)	_lrotl(x,n)			
#define	ROR(x,n)	_lrotr(x,n)
#endif

#define BSWAP(x) (((ROR(x,8) & 0xFF00FF00) | (ROL(x,8) & 0x00FF00FF)))
#define CONV_BYTES_TO_U32(b3, b2, b1, b0) ((b3 << 24) ^ (b2 << 16) ^ (b1 << 8) ^ b0)

#ifdef LITTLE_ENDIAN_MACHINE
#define BSWAP_LE(x) (x)
#define BSWAP_BE(x) BSWAP(x)
#else
#define BSWAP_LE(x) BSWAP(x)
#define BSWAP_BE(x) (x)
#endif

#if defined(LEDA_CODER_EXACT_SIZE_INT32)
#define LOAD_U32_LE(BYTES) BSWAP_LE( *((uint32*) (BYTES)) )
#define LOAD_U32_BE(BYTES) BSWAP_BE( *((uint32*) (BYTES)) )
#define STORE_U32_LE(x, BYTES) *((uint32*) (BYTES)) = BSWAP_LE(x)
#define STORE_U32_BE(x, BYTES) *((uint32*) (BYTES)) = BSWAP_BE(x)
#else
#define LOAD_U32_LE(BYTES) \
	(uint32((BYTES)[0])) | (uint32((BYTES)[1])<<8) | (uint32((BYTES)[2])<<16)  | (uint32((BYTES)[3])<<16)
#define LOAD_U32_BE(BYTES) \
	(uint32((BYTES)[0])<<24) | (uint32((BYTES)[1])<<16) | (uint32((BYTES)[2])<<8)  | (uint32((BYTES)[3]))
#define STORE_U32_LE(x, BYTES) \
	(BYTES)[0] = byte(x); \
	(BYTES)[1] = byte(x >> 8); \
	(BYTES)[2] = byte(x >> 16); \
	(BYTES)[3] = byte(x >> 24);
#define STORE_U32_BE(x, BYTES) \
	(BYTES)[0] = byte(x >> 24); \
	(BYTES)[1] = byte(x >> 16); \
	(BYTES)[2] = byte(x >> 8); \
	(BYTES)[3] = byte(x);
#endif

#define b0(x) ((byte)(x))
#define b1(x) ((byte)((x) >> 8))
#define b2(x) ((byte)((x) >> 16))
#define b3(x) ((byte)((x) >> 24))

#include "block_ciphers.tbl"

/// Blowfish ///////////////////////////////////////////////////////////////////////////////

// The following code is taken from Bruce Schneier ...
// "Blowfish is unpatented and license-free, and is available free for all uses."

Blowfish::Blowfish()
 : pState(0)
{
	pState = new state;
}

Blowfish::~Blowfish()
{
	if (pState) {
		wipe_memory(pState, sizeof(state));
		delete pState; pState = 0;
	}
}

void
Blowfish::set_key(const CryptKey& key, direction)
{
	if (key.get_size() < MinKeyBytes || key.get_size() > MaxKeyBytes) {
		LEDA_EXCEPTION(1, "Blowfish: wrong number of key bytes"); return;
	}

	memcpy(pState, &bf_initial_state, sizeof(state));

	int i, j = 0;
	for (i = 0; i < Rounds + 2; ++i) {
		uint32 data = 0x00000000;
		for (int k = 0; k < 4; ++k) {
			data = (data << 8) | key[j];
			if (++j >= key.get_size()) j = 0;
		}
		pState->P[i] = (pState->P[i]) ^ data;
	}

	uint32 datal = 0x00000000;
	uint32 datar = 0x00000000;

	for (i = 0; i < Rounds + 2; i += 2) {
		encipher_step(&datal, &datar);

		pState->P[i]     = datal;
		pState->P[i + 1] = datar;
	}

	for (i = 0; i < 4; ++i) {
		for (j = 0; j < 256; j += 2) {
			encipher_step(&datal, &datar);

			pState->S[i][j]     = datal;
			pState->S[i][j + 1] = datar;
		}
	}
}

void 
Blowfish::encipher(const byte* src, byte* tgt)
{
	uint32 xl = LOAD_U32_BE(src);
	uint32 xr = LOAD_U32_BE(src+4);
	encipher_step(&xl, &xr);
	STORE_U32_BE(xl, tgt);
	STORE_U32_BE(xr, tgt+4);
}

void 
Blowfish::encipher(byte* blk)
{
#if defined(LEDA_CODER_EXACT_SIZE_INT32)
	uint32* xl = (uint32*) blk;
	uint32* xr = (uint32*) (blk+4);
	*xl = BSWAP_BE(*xl); *xr = BSWAP_BE(*xr);
	encipher_step(xl, xr);
	*xl = BSWAP_BE(*xl); *xr = BSWAP_BE(*xr);
#else
	uint32 xl = LOAD_U32_BE(blk);
	uint32 xr = LOAD_U32_BE(blk+4);
	encipher_step(&xl, &xr);
	STORE_U32_BE(xl, blk);
	STORE_U32_BE(xr, blk+4);
#endif
}

void 
Blowfish::decipher(const byte* src, byte* tgt)
{
	uint32 xl = LOAD_U32_BE(src);
	uint32 xr = LOAD_U32_BE(src+4);
	decipher_step(&xl, &xr);
	STORE_U32_BE(xl, tgt);
	STORE_U32_BE(xr, tgt+4);
}

void 
Blowfish::decipher(byte* blk)
{
#if defined(LEDA_CODER_EXACT_SIZE_INT32)
	uint32* xl = (uint32*) blk;
	uint32* xr = (uint32*) (blk+4);
	*xl = BSWAP_BE(*xl); *xr = BSWAP_BE(*xr);
	decipher_step(xl, xr);
	*xl = BSWAP_BE(*xl); *xr = BSWAP_BE(*xr);
#else
	uint32 xl = LOAD_U32_BE(blk);
	uint32 xr = LOAD_U32_BE(blk+4);
	decipher_step(&xl, &xr);
	STORE_U32_BE(xl, blk);
	STORE_U32_BE(xr, blk+4);
#endif
}


Blowfish::uint32
Blowfish::F(uint32 x)
{
   uint16 d = x & 0x00FF;
   x >>= 8;
   uint16 c = x & 0x00FF;
   x >>= 8;
   uint16 b = x & 0x00FF;
   x >>= 8;
   uint16 a = x & 0x00FF;
   //y = ((S[0][a] + S[1][b]) ^ S[2][c]) + S[3][d];
   uint32 y = (pState->S[0][a]) + (pState->S[1][b]);
   y = y ^ (pState->S[2][c]);
   y = y + (pState->S[3][d]);

   return y;
}

void 
Blowfish::encipher_step(uint32* xl, uint32* xr)
{
	uint32 Xl = *xl;
	uint32 Xr = *xr;

	for (int i = 0; i < Rounds; ++i) {
		Xl = Xl ^ (pState->P[i]);
		Xr = F(Xl) ^ Xr;

		// swap Xl and Xr
		uint32 temp = Xl;
		Xl = Xr;
		Xr = temp;
	}

	// we do not undo the last swap explicitly ...
	Xl = Xl ^ (pState->P[Rounds]);
	Xr = Xr ^ (pState->P[Rounds + 1]);

	// ... but implicitly
	*xl = Xr; *xr = Xl;
}

void 
Blowfish::decipher_step(uint32* xl, uint32* xr)
{
	uint32 Xl = *xl;
	uint32 Xr = *xr;

	for (int i = Rounds + 1; i > 1; --i) {
		Xl = Xl ^ (pState->P[i]);
		Xr = F(Xl) ^ Xr;

		// swap Xl and Xr
		uint32 temp = Xl;
		Xl = Xr;
		Xr = temp;
	}

	// we do not undo the last swap explicitly ...
	Xl = Xl ^ (pState->P[1]);
	Xr = Xr ^ (pState->P[0]);

	// ... but implicitly
	*xl = Xr; *xr = Xl;
}

/// Twofish /////////////////////////////////////////////////////////////////////////////////

// The code follows the paper describing Twofish.

#undef TF_KEY
#define TF_KEY(k) (pState->ExpKey[k])

#undef TF_g
#define TF_g(x) \
	((pState->FKT[0][b0(x)]) ^ (pState->FKT[1][b1(x)]) ^ (pState->FKT[2][b2(x)]) ^ (pState->FKT[3][b3(x)]))

#undef TF_ENC_RND
#define TF_ENC_RND(R0, R1, R2, R3, rnd) \
    T0 = TF_g(R0); \
    T1 = TF_g(ROL(R1, 8)); \
    R2 = ROR(R2 ^ (T0 + T1 + TF_KEY(2*rnd+8)), 1); \
    R3 = ROL(R3,1) ^ (T0 + 2*T1 + TF_KEY(2*rnd+9)); 

#undef TF_DEC_RND
#define TF_DEC_RND(R0, R1, R2, R3, rnd) \
    T0 = TF_g(R0); \
    T1 = TF_g(ROL(R1, 8)); \
    R2 = ROL(R2, 1) ^ (T0 + T1 + TF_KEY(2*rnd+8)); \
    R3 = ROR(R3 ^ (T0 + 2*T1 + TF_KEY(2*rnd+9)), 1); 

Twofish::Twofish()
 : pState(0)
{
	pState = new state;
}

Twofish::~Twofish()
{
	if (pState) {
		wipe_memory(pState, sizeof(state));
		delete pState; pState = 0;
	}
}

void 
Twofish::set_key(const CryptKey& key, direction)
{
	if (key.get_size() < MinKeyBytes || key.get_size() > MaxKeyBytes) {
		LEDA_EXCEPTION(1, "Twofish: wrong number of key bytes"); return;
	}

	// determine k, N and m (pad m, if necessary)
	int k = (8*key.get_size() + 63) / 64; // = ceil(key.get_size() / 64)
	int N = k * 64;
	byte m[MaxKeyBytes];
	memset(m, 0, N/8);
	memcpy(m, key.get_bytes(), key.get_size());

	int i;

	// compute Mo and Me
    uint32 Mo[4], Me[4];
    for (i = 0; i < k; ++i) {
		Me[i] = LOAD_U32_LE(m + i*(2*4));
		Mo[i] = LOAD_U32_LE(m + i*(2*4)+4);
    }

	// compute ExpandedKey
	const uint32 RHO = 0x01010101L;
    for (i = 0; i < 20; ++i) {
		uint32 A = h(2*i*RHO, Me, k);
		uint32 B = ROL(h(2*i*RHO + RHO, Mo, k), 8);
		TF_KEY(2*i)   = A + B;
		TF_KEY(2*i+1) = ROL(A + 2*B, 9);
    }

	// compute S
	uint32 S[4];
	byte vector[8];
    for (i = 0; i < k; ++i) {
		memcpy(&vector, &m[8*i], 8);
		S[k-i-1] = RS_multiply(vector);
    }

	// precompute FKT
	key_fully(S, k);

	// wipe memory
	wipe_memory(m, sizeof(m));
	wipe_memory(Me, sizeof(Me));
	wipe_memory(Mo, sizeof(Mo));
	wipe_memory(S, sizeof(S));
}

Twofish::uint32 
Twofish::RS_multiply(byte vector[8])
{
    byte result[4];

	const uint32 RS_MOD = 0x14DL;

    for (int j = 0; j < 4; ++j) {
		byte t = 0;
		for (int k = 0; k < 8; k++) {
			t ^= poly_mult_mod(tf_RS[j][k], vector[k], RS_MOD);
		}
		result[j] = t;
    }

    return CONV_BYTES_TO_U32(result[3], result[2], result[1], result[0]);
}

Twofish::uint32 
Twofish::poly_mult_mod(uint32 a, uint32 b, uint32 mod)
{
    uint32 a_mult_b = 0;
    while (a) {
		if (a & 1) a_mult_b ^= b;
		b <<= 1;
		a >>= 1;
    }

    mod <<= 7;
    for (int i = 0; i < 8; ++i) {
		uint32 r = a_mult_b ^ mod;
		if (r < a_mult_b) a_mult_b = r;
		mod >>= 1;
    }

	return a_mult_b;
}

Twofish::uint32 
Twofish::h(uint32 X, uint32 L[4], int k)
{
    byte y0 = b0(X), y1 = b1(X), y2 = b2(X), y3 = b3(X);

    switch (k) {
	case 4:
	    y0 = tf_Q1[y0] ^ b0(L[3]);
	    y1 = tf_Q0[y1] ^ b1(L[3]);
	    y2 = tf_Q0[y2] ^ b2(L[3]);
	    y3 = tf_Q1[y3] ^ b3(L[3]);

	case 3:
	    y0 = tf_Q1[y0] ^ b0(L[2]);
	    y1 = tf_Q1[y1] ^ b1(L[2]);
	    y2 = tf_Q0[y2] ^ b2(L[2]);
	    y3 = tf_Q0[y3] ^ b3(L[2]);

	case 2:
	    y0 = tf_Q1[  tf_Q0 [ tf_Q0[y0] ^ b0(L[1]) ] ^ b0(L[0]) ];
	    y1 = tf_Q0[  tf_Q0 [ tf_Q1[y1] ^ b1(L[1]) ] ^ b1(L[0]) ];
	    y2 = tf_Q1[  tf_Q1 [ tf_Q0[y2] ^ b2(L[1]) ] ^ b2(L[0]) ];
	    y3 = tf_Q0[  tf_Q1 [ tf_Q1[y3] ^ b3(L[1]) ] ^ b3(L[0]) ];
    }

    // inline the MDS matrix multiplication
    byte z0 = y0            ^ tf_multEF[y1] ^ tf_mult5B[y2] ^ tf_mult5B[y3]; 
    byte z1 = tf_mult5B[y0] ^ tf_multEF[y1] ^ tf_multEF[y2] ^ y3; 
    byte z2 = tf_multEF[y0] ^ tf_mult5B[y1] ^ y2            ^ tf_multEF[y3]; 
    byte z3 = tf_multEF[y0] ^ y1            ^ tf_multEF[y2] ^ tf_mult5B[y3]; 

    return CONV_BYTES_TO_U32(z3, z2, z1, z0);
}

void
Twofish::key_fully(uint32 S[4], int k)
{
	// precompute g(y3|y2|y1|y0) = h(y3|y2|y1|y0, S, k) for all possible yj-values

    for (int i = 0; i < 256; ++i) {

		// run the Q permutations
		byte y0 = i, y1 = i, y2 = i, y3 = i;
		switch (k) {
		case 4:
			y0 = tf_Q1[y0] ^ b0(S[3]);
			y1 = tf_Q0[y1] ^ b1(S[3]);
			y2 = tf_Q0[y2] ^ b2(S[3]);
			y3 = tf_Q1[y3] ^ b3(S[3]);

		case 3:
			y0 = tf_Q1[y0] ^ b0(S[2]);
			y1 = tf_Q1[y1] ^ b1(S[2]);
			y2 = tf_Q0[y2] ^ b2(S[2]);
			y3 = tf_Q0[y3] ^ b3(S[2]);

		case 2:
			y0 = tf_Q1[  tf_Q0 [ tf_Q0[y0] ^ b0(S[1]) ] ^ b0(S[0]) ];
			y1 = tf_Q0[  tf_Q0 [ tf_Q1[y1] ^ b1(S[1]) ] ^ b1(S[0]) ];
			y2 = tf_Q1[  tf_Q1 [ tf_Q0[y2] ^ b2(S[1]) ] ^ b2(S[0]) ];
			y3 = tf_Q0[  tf_Q1 [ tf_Q1[y3] ^ b3(S[1]) ] ^ b3(S[0]) ];
		}

		// partial MDS matrix multiplication: multiply yj with column j of MDS
		pState->FKT[0][i] = ((tf_multEF[y0] << 24)
						  |  (tf_multEF[y0] << 16) 
						  |  (tf_mult5B[y0] << 8)
						  |  (y0));

		pState->FKT[1][i] = ((y1 << 24) 
						  |  (tf_mult5B[y1] << 16) 
						  |  (tf_multEF[y1] << 8)
						  |  (tf_multEF[y1]));

		pState->FKT[2][i] = ((tf_multEF[y2] << 24) 
						  |  (y2 << 16)
						  |  (tf_multEF[y2] << 8)
						  |  (tf_mult5B[y2]));

		pState->FKT[3][i] = ((tf_mult5B[y3] << 24) 
						  |  (tf_multEF[y3] << 16)
						  |  (y3 << 8) 
						  |  (tf_mult5B[y3]));
    }
}

void 
Twofish::encipher(byte* blk)
{
    uint32 T0, T1;
    uint32 R0, R1, R2, R3;

    R0 = TF_KEY(0) ^ LOAD_U32_LE(blk + 0);
    R1 = TF_KEY(1) ^ LOAD_U32_LE(blk + 4);
    R2 = TF_KEY(2) ^ LOAD_U32_LE(blk + 8);
    R3 = TF_KEY(3) ^ LOAD_U32_LE(blk + 12);

    TF_ENC_RND(R0, R1, R2, R3, 0);
    TF_ENC_RND(R2, R3, R0, R1, 1);
    TF_ENC_RND(R0, R1, R2, R3, 2);
    TF_ENC_RND(R2, R3, R0, R1, 3);
    TF_ENC_RND(R0, R1, R2, R3, 4);
    TF_ENC_RND(R2, R3, R0, R1, 5);
    TF_ENC_RND(R0, R1, R2, R3, 6);
    TF_ENC_RND(R2, R3, R0, R1, 7);
    TF_ENC_RND(R0, R1, R2, R3, 8);
    TF_ENC_RND(R2, R3, R0, R1, 9);
    TF_ENC_RND(R0, R1, R2, R3, 10);
    TF_ENC_RND(R2, R3, R0, R1, 11);
    TF_ENC_RND(R0, R1, R2, R3, 12);
    TF_ENC_RND(R2, R3, R0, R1, 13);
    TF_ENC_RND(R0, R1, R2, R3, 14);
    TF_ENC_RND(R2, R3, R0, R1, 15);

	STORE_U32_LE(R2 ^ TF_KEY(4), blk + 0); 
	STORE_U32_LE(R3 ^ TF_KEY(5), blk + 4); 
	STORE_U32_LE(R0 ^ TF_KEY(6), blk + 8); 
	STORE_U32_LE(R1 ^ TF_KEY(7), blk + 12); 
}

void 
Twofish::decipher(byte* blk)
{
    uint32 T0, T1;
    uint32 R0, R1, R2, R3;

    R0 = TF_KEY(4) ^ LOAD_U32_LE(blk + 0);
    R1 = TF_KEY(5) ^ LOAD_U32_LE(blk + 4);
    R2 = TF_KEY(6) ^ LOAD_U32_LE(blk + 8);
    R3 = TF_KEY(7) ^ LOAD_U32_LE(blk + 12);

    TF_DEC_RND(R0, R1, R2, R3, 15);
    TF_DEC_RND(R2, R3, R0, R1, 14);
    TF_DEC_RND(R0, R1, R2, R3, 13);
    TF_DEC_RND(R2, R3, R0, R1, 12);
    TF_DEC_RND(R0, R1, R2, R3, 11);
    TF_DEC_RND(R2, R3, R0, R1, 10);
    TF_DEC_RND(R0, R1, R2, R3, 9);
    TF_DEC_RND(R2, R3, R0, R1, 8);
    TF_DEC_RND(R0, R1, R2, R3, 7);
    TF_DEC_RND(R2, R3, R0, R1, 6);
    TF_DEC_RND(R0, R1, R2, R3, 5);
    TF_DEC_RND(R2, R3, R0, R1, 4);
    TF_DEC_RND(R0, R1, R2, R3, 3);
    TF_DEC_RND(R2, R3, R0, R1, 2);
    TF_DEC_RND(R0, R1, R2, R3, 1);
    TF_DEC_RND(R2, R3, R0, R1, 0);

	STORE_U32_LE(R2 ^ TF_KEY(0), blk + 0); 
	STORE_U32_LE(R3 ^ TF_KEY(1), blk + 4); 
	STORE_U32_LE(R0 ^ TF_KEY(2), blk + 8); 
	STORE_U32_LE(R1 ^ TF_KEY(3), blk + 12); 
}

void 
Twofish::encipher(const byte* src, byte* tgt)
{
	memcpy(tgt, src, BlkSz);
	encipher(tgt);
}

void 
Twofish::decipher(const byte* src, byte* tgt)
{
	memcpy(tgt, src, BlkSz);
	decipher(tgt);
}

/// Rijndael ////////////////////////////////////////////////////////////////////////////////

// The code is based on the following:
/**
 * rijndael-alg-fst.c
 *
 * @version 3.0 (December 2000)
 *
 * Optimised ANSI C code for the Rijndael cipher (now AES)
 *
 * @author Vincent Rijmen <vincent.rijmen@esat.kuleuven.ac.be>
 * @author Antoon Bosselaers <antoon.bosselaers@esat.kuleuven.ac.be>
 * @author Paulo Barreto <paulo.barreto@terra.com.br>
 *
 * This code is hereby placed in the public domain.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define RD_FULL_UNROLL

Rijndael::Rijndael()
 : Nr(0)
{}

Rijndael::~Rijndael()
{
	Nr = 0;
	wipe_memory(&Rk, sizeof(Rk));
}

void 
Rijndael::set_key(const CryptKey& key, direction dir)
{
	switch (dir) {
	case d_encipher:	key_setup_enc(key); break;
	case d_decipher:	key_setup_dec(key); break;
	case d_both:		LEDA_EXCEPTION(1, "Rijndael::set_key: d_both not allowed");
	}
}

void 
Rijndael::key_setup_enc(const CryptKey& key)
{
   	int i = 0;
	uint32 temp;

	uint32* rk = Rk;
	rk[0] = LOAD_U32_BE(key.get_bytes()     );
	rk[1] = LOAD_U32_BE(key.get_bytes() +  4);
	rk[2] = LOAD_U32_BE(key.get_bytes() +  8);
	rk[3] = LOAD_U32_BE(key.get_bytes() + 12);
	if (key.get_size() == 128 / 8) {
		for (;;) {
			temp  = rk[3];
			rk[4] = rk[0] ^
				(rd_Te4[(temp >> 16) & 0xff] & 0xff000000) ^
				(rd_Te4[(temp >>  8) & 0xff] & 0x00ff0000) ^
				(rd_Te4[(temp      ) & 0xff] & 0x0000ff00) ^
				(rd_Te4[(temp >> 24)       ] & 0x000000ff) ^
				rd_rcon[i];
			rk[5] = rk[1] ^ rk[4];
			rk[6] = rk[2] ^ rk[5];
			rk[7] = rk[3] ^ rk[6];
			if (++i == 10) {
				Nr = 10; return;
			}
			rk += 4;
		}
	}
	rk[4] = LOAD_U32_BE(key.get_bytes() + 16);
	rk[5] = LOAD_U32_BE(key.get_bytes() + 20);
	if (key.get_size() == 192 / 8) {
		for (;;) {
			temp = rk[ 5];
			rk[ 6] = rk[ 0] ^
				(rd_Te4[(temp >> 16) & 0xff] & 0xff000000) ^
				(rd_Te4[(temp >>  8) & 0xff] & 0x00ff0000) ^
				(rd_Te4[(temp      ) & 0xff] & 0x0000ff00) ^
				(rd_Te4[(temp >> 24)       ] & 0x000000ff) ^
				rd_rcon[i];
			rk[ 7] = rk[ 1] ^ rk[ 6];
			rk[ 8] = rk[ 2] ^ rk[ 7];
			rk[ 9] = rk[ 3] ^ rk[ 8];
			if (++i == 8) {
				Nr = 12; return;
			}
			rk[10] = rk[ 4] ^ rk[ 9];
			rk[11] = rk[ 5] ^ rk[10];
			rk += 6;
		}
	}
	rk[6] = LOAD_U32_BE(key.get_bytes() + 24);
	rk[7] = LOAD_U32_BE(key.get_bytes() + 28);
	if (key.get_size() == 256 / 8) {
        for (;;) {
        	temp = rk[ 7];
        	rk[ 8] = rk[ 0] ^
        		(rd_Te4[(temp >> 16) & 0xff] & 0xff000000) ^
        		(rd_Te4[(temp >>  8) & 0xff] & 0x00ff0000) ^
        		(rd_Te4[(temp      ) & 0xff] & 0x0000ff00) ^
        		(rd_Te4[(temp >> 24)       ] & 0x000000ff) ^
        		rd_rcon[i];
        	rk[ 9] = rk[ 1] ^ rk[ 8];
        	rk[10] = rk[ 2] ^ rk[ 9];
        	rk[11] = rk[ 3] ^ rk[10];
			if (++i == 7) {
				Nr = 14; return;
			}
        	temp = rk[11];
        	rk[12] = rk[ 4] ^
        		(rd_Te4[(temp >> 24)       ] & 0xff000000) ^
        		(rd_Te4[(temp >> 16) & 0xff] & 0x00ff0000) ^
        		(rd_Te4[(temp >>  8) & 0xff] & 0x0000ff00) ^
        		(rd_Te4[(temp      ) & 0xff] & 0x000000ff);
        	rk[13] = rk[ 5] ^ rk[12];
        	rk[14] = rk[ 6] ^ rk[13];
        	rk[15] = rk[ 7] ^ rk[14];

			rk += 8;
        }
	}

	LEDA_EXCEPTION(1, "Rijndael: wrong number of key bytes");
}

void
Rijndael::key_setup_dec(const CryptKey& key)
{
	int i, j;
	uint32 temp;
	uint32* rk = Rk;

	/* expand the cipher key: */
	key_setup_enc(key);
	/* invert the order of the round keys: */
	for (i = 0, j = 4*Nr; i < j; i += 4, j -= 4) {
		temp = rk[i    ]; rk[i    ] = rk[j    ]; rk[j    ] = temp;
		temp = rk[i + 1]; rk[i + 1] = rk[j + 1]; rk[j + 1] = temp;
		temp = rk[i + 2]; rk[i + 2] = rk[j + 2]; rk[j + 2] = temp;
		temp = rk[i + 3]; rk[i + 3] = rk[j + 3]; rk[j + 3] = temp;
	}
	/* apply the inverse MixColumn transform to all round keys but the first and the last: */
	for (i = 1; i < Nr; i++) {
		rk += 4;
		rk[0] =
			rd_Td0[rd_Te4[(rk[0] >> 24)       ] & 0xff] ^
			rd_Td1[rd_Te4[(rk[0] >> 16) & 0xff] & 0xff] ^
			rd_Td2[rd_Te4[(rk[0] >>  8) & 0xff] & 0xff] ^
			rd_Td3[rd_Te4[(rk[0]      ) & 0xff] & 0xff];
		rk[1] =
			rd_Td0[rd_Te4[(rk[1] >> 24)       ] & 0xff] ^
			rd_Td1[rd_Te4[(rk[1] >> 16) & 0xff] & 0xff] ^
			rd_Td2[rd_Te4[(rk[1] >>  8) & 0xff] & 0xff] ^
			rd_Td3[rd_Te4[(rk[1]      ) & 0xff] & 0xff];
		rk[2] =
			rd_Td0[rd_Te4[(rk[2] >> 24)       ] & 0xff] ^
			rd_Td1[rd_Te4[(rk[2] >> 16) & 0xff] & 0xff] ^
			rd_Td2[rd_Te4[(rk[2] >>  8) & 0xff] & 0xff] ^
			rd_Td3[rd_Te4[(rk[2]      ) & 0xff] & 0xff];
		rk[3] =
			rd_Td0[rd_Te4[(rk[3] >> 24)       ] & 0xff] ^
			rd_Td1[rd_Te4[(rk[3] >> 16) & 0xff] & 0xff] ^
			rd_Td2[rd_Te4[(rk[3] >>  8) & 0xff] & 0xff] ^
			rd_Td3[rd_Te4[(rk[3]      ) & 0xff] & 0xff];
	}
}

void 
Rijndael::encipher(const byte* pt, byte* ct)
{
	uint32* rk = Rk;

	uint32 s0, s1, s2, s3, t0, t1, t2, t3;
#ifndef RD_FULL_UNROLL
    int r;
#endif /* ?RD_FULL_UNROLL */

    /*
	 * map byte array block to cipher state
	 * and add initial round key:
	 */
	s0 = LOAD_U32_BE(pt     ) ^ rk[0];
	s1 = LOAD_U32_BE(pt +  4) ^ rk[1];
	s2 = LOAD_U32_BE(pt +  8) ^ rk[2];
	s3 = LOAD_U32_BE(pt + 12) ^ rk[3];
#ifdef RD_FULL_UNROLL
    /* round 1: */
   	t0 = rd_Te0[s0 >> 24] ^ rd_Te1[(s1 >> 16) & 0xff] ^ rd_Te2[(s2 >>  8) & 0xff] ^ rd_Te3[s3 & 0xff] ^ rk[ 4];
   	t1 = rd_Te0[s1 >> 24] ^ rd_Te1[(s2 >> 16) & 0xff] ^ rd_Te2[(s3 >>  8) & 0xff] ^ rd_Te3[s0 & 0xff] ^ rk[ 5];
   	t2 = rd_Te0[s2 >> 24] ^ rd_Te1[(s3 >> 16) & 0xff] ^ rd_Te2[(s0 >>  8) & 0xff] ^ rd_Te3[s1 & 0xff] ^ rk[ 6];
   	t3 = rd_Te0[s3 >> 24] ^ rd_Te1[(s0 >> 16) & 0xff] ^ rd_Te2[(s1 >>  8) & 0xff] ^ rd_Te3[s2 & 0xff] ^ rk[ 7];
   	/* round 2: */
   	s0 = rd_Te0[t0 >> 24] ^ rd_Te1[(t1 >> 16) & 0xff] ^ rd_Te2[(t2 >>  8) & 0xff] ^ rd_Te3[t3 & 0xff] ^ rk[ 8];
   	s1 = rd_Te0[t1 >> 24] ^ rd_Te1[(t2 >> 16) & 0xff] ^ rd_Te2[(t3 >>  8) & 0xff] ^ rd_Te3[t0 & 0xff] ^ rk[ 9];
   	s2 = rd_Te0[t2 >> 24] ^ rd_Te1[(t3 >> 16) & 0xff] ^ rd_Te2[(t0 >>  8) & 0xff] ^ rd_Te3[t1 & 0xff] ^ rk[10];
   	s3 = rd_Te0[t3 >> 24] ^ rd_Te1[(t0 >> 16) & 0xff] ^ rd_Te2[(t1 >>  8) & 0xff] ^ rd_Te3[t2 & 0xff] ^ rk[11];
    /* round 3: */
   	t0 = rd_Te0[s0 >> 24] ^ rd_Te1[(s1 >> 16) & 0xff] ^ rd_Te2[(s2 >>  8) & 0xff] ^ rd_Te3[s3 & 0xff] ^ rk[12];
   	t1 = rd_Te0[s1 >> 24] ^ rd_Te1[(s2 >> 16) & 0xff] ^ rd_Te2[(s3 >>  8) & 0xff] ^ rd_Te3[s0 & 0xff] ^ rk[13];
   	t2 = rd_Te0[s2 >> 24] ^ rd_Te1[(s3 >> 16) & 0xff] ^ rd_Te2[(s0 >>  8) & 0xff] ^ rd_Te3[s1 & 0xff] ^ rk[14];
   	t3 = rd_Te0[s3 >> 24] ^ rd_Te1[(s0 >> 16) & 0xff] ^ rd_Te2[(s1 >>  8) & 0xff] ^ rd_Te3[s2 & 0xff] ^ rk[15];
   	/* round 4: */
   	s0 = rd_Te0[t0 >> 24] ^ rd_Te1[(t1 >> 16) & 0xff] ^ rd_Te2[(t2 >>  8) & 0xff] ^ rd_Te3[t3 & 0xff] ^ rk[16];
   	s1 = rd_Te0[t1 >> 24] ^ rd_Te1[(t2 >> 16) & 0xff] ^ rd_Te2[(t3 >>  8) & 0xff] ^ rd_Te3[t0 & 0xff] ^ rk[17];
   	s2 = rd_Te0[t2 >> 24] ^ rd_Te1[(t3 >> 16) & 0xff] ^ rd_Te2[(t0 >>  8) & 0xff] ^ rd_Te3[t1 & 0xff] ^ rk[18];
   	s3 = rd_Te0[t3 >> 24] ^ rd_Te1[(t0 >> 16) & 0xff] ^ rd_Te2[(t1 >>  8) & 0xff] ^ rd_Te3[t2 & 0xff] ^ rk[19];
    /* round 5: */
   	t0 = rd_Te0[s0 >> 24] ^ rd_Te1[(s1 >> 16) & 0xff] ^ rd_Te2[(s2 >>  8) & 0xff] ^ rd_Te3[s3 & 0xff] ^ rk[20];
   	t1 = rd_Te0[s1 >> 24] ^ rd_Te1[(s2 >> 16) & 0xff] ^ rd_Te2[(s3 >>  8) & 0xff] ^ rd_Te3[s0 & 0xff] ^ rk[21];
   	t2 = rd_Te0[s2 >> 24] ^ rd_Te1[(s3 >> 16) & 0xff] ^ rd_Te2[(s0 >>  8) & 0xff] ^ rd_Te3[s1 & 0xff] ^ rk[22];
   	t3 = rd_Te0[s3 >> 24] ^ rd_Te1[(s0 >> 16) & 0xff] ^ rd_Te2[(s1 >>  8) & 0xff] ^ rd_Te3[s2 & 0xff] ^ rk[23];
   	/* round 6: */
   	s0 = rd_Te0[t0 >> 24] ^ rd_Te1[(t1 >> 16) & 0xff] ^ rd_Te2[(t2 >>  8) & 0xff] ^ rd_Te3[t3 & 0xff] ^ rk[24];
   	s1 = rd_Te0[t1 >> 24] ^ rd_Te1[(t2 >> 16) & 0xff] ^ rd_Te2[(t3 >>  8) & 0xff] ^ rd_Te3[t0 & 0xff] ^ rk[25];
   	s2 = rd_Te0[t2 >> 24] ^ rd_Te1[(t3 >> 16) & 0xff] ^ rd_Te2[(t0 >>  8) & 0xff] ^ rd_Te3[t1 & 0xff] ^ rk[26];
   	s3 = rd_Te0[t3 >> 24] ^ rd_Te1[(t0 >> 16) & 0xff] ^ rd_Te2[(t1 >>  8) & 0xff] ^ rd_Te3[t2 & 0xff] ^ rk[27];
    /* round 7: */
   	t0 = rd_Te0[s0 >> 24] ^ rd_Te1[(s1 >> 16) & 0xff] ^ rd_Te2[(s2 >>  8) & 0xff] ^ rd_Te3[s3 & 0xff] ^ rk[28];
   	t1 = rd_Te0[s1 >> 24] ^ rd_Te1[(s2 >> 16) & 0xff] ^ rd_Te2[(s3 >>  8) & 0xff] ^ rd_Te3[s0 & 0xff] ^ rk[29];
   	t2 = rd_Te0[s2 >> 24] ^ rd_Te1[(s3 >> 16) & 0xff] ^ rd_Te2[(s0 >>  8) & 0xff] ^ rd_Te3[s1 & 0xff] ^ rk[30];
   	t3 = rd_Te0[s3 >> 24] ^ rd_Te1[(s0 >> 16) & 0xff] ^ rd_Te2[(s1 >>  8) & 0xff] ^ rd_Te3[s2 & 0xff] ^ rk[31];
   	/* round 8: */
   	s0 = rd_Te0[t0 >> 24] ^ rd_Te1[(t1 >> 16) & 0xff] ^ rd_Te2[(t2 >>  8) & 0xff] ^ rd_Te3[t3 & 0xff] ^ rk[32];
   	s1 = rd_Te0[t1 >> 24] ^ rd_Te1[(t2 >> 16) & 0xff] ^ rd_Te2[(t3 >>  8) & 0xff] ^ rd_Te3[t0 & 0xff] ^ rk[33];
   	s2 = rd_Te0[t2 >> 24] ^ rd_Te1[(t3 >> 16) & 0xff] ^ rd_Te2[(t0 >>  8) & 0xff] ^ rd_Te3[t1 & 0xff] ^ rk[34];
   	s3 = rd_Te0[t3 >> 24] ^ rd_Te1[(t0 >> 16) & 0xff] ^ rd_Te2[(t1 >>  8) & 0xff] ^ rd_Te3[t2 & 0xff] ^ rk[35];
    /* round 9: */
   	t0 = rd_Te0[s0 >> 24] ^ rd_Te1[(s1 >> 16) & 0xff] ^ rd_Te2[(s2 >>  8) & 0xff] ^ rd_Te3[s3 & 0xff] ^ rk[36];
   	t1 = rd_Te0[s1 >> 24] ^ rd_Te1[(s2 >> 16) & 0xff] ^ rd_Te2[(s3 >>  8) & 0xff] ^ rd_Te3[s0 & 0xff] ^ rk[37];
   	t2 = rd_Te0[s2 >> 24] ^ rd_Te1[(s3 >> 16) & 0xff] ^ rd_Te2[(s0 >>  8) & 0xff] ^ rd_Te3[s1 & 0xff] ^ rk[38];
   	t3 = rd_Te0[s3 >> 24] ^ rd_Te1[(s0 >> 16) & 0xff] ^ rd_Te2[(s1 >>  8) & 0xff] ^ rd_Te3[s2 & 0xff] ^ rk[39];
    if (Nr > 10) {
        /* round 10: */
        s0 = rd_Te0[t0 >> 24] ^ rd_Te1[(t1 >> 16) & 0xff] ^ rd_Te2[(t2 >>  8) & 0xff] ^ rd_Te3[t3 & 0xff] ^ rk[40];
        s1 = rd_Te0[t1 >> 24] ^ rd_Te1[(t2 >> 16) & 0xff] ^ rd_Te2[(t3 >>  8) & 0xff] ^ rd_Te3[t0 & 0xff] ^ rk[41];
        s2 = rd_Te0[t2 >> 24] ^ rd_Te1[(t3 >> 16) & 0xff] ^ rd_Te2[(t0 >>  8) & 0xff] ^ rd_Te3[t1 & 0xff] ^ rk[42];
        s3 = rd_Te0[t3 >> 24] ^ rd_Te1[(t0 >> 16) & 0xff] ^ rd_Te2[(t1 >>  8) & 0xff] ^ rd_Te3[t2 & 0xff] ^ rk[43];
        /* round 11: */
        t0 = rd_Te0[s0 >> 24] ^ rd_Te1[(s1 >> 16) & 0xff] ^ rd_Te2[(s2 >>  8) & 0xff] ^ rd_Te3[s3 & 0xff] ^ rk[44];
        t1 = rd_Te0[s1 >> 24] ^ rd_Te1[(s2 >> 16) & 0xff] ^ rd_Te2[(s3 >>  8) & 0xff] ^ rd_Te3[s0 & 0xff] ^ rk[45];
        t2 = rd_Te0[s2 >> 24] ^ rd_Te1[(s3 >> 16) & 0xff] ^ rd_Te2[(s0 >>  8) & 0xff] ^ rd_Te3[s1 & 0xff] ^ rk[46];
        t3 = rd_Te0[s3 >> 24] ^ rd_Te1[(s0 >> 16) & 0xff] ^ rd_Te2[(s1 >>  8) & 0xff] ^ rd_Te3[s2 & 0xff] ^ rk[47];
        if (Nr > 12) {
            /* round 12: */
            s0 = rd_Te0[t0 >> 24] ^ rd_Te1[(t1 >> 16) & 0xff] ^ rd_Te2[(t2 >>  8) & 0xff] ^ rd_Te3[t3 & 0xff] ^ rk[48];
            s1 = rd_Te0[t1 >> 24] ^ rd_Te1[(t2 >> 16) & 0xff] ^ rd_Te2[(t3 >>  8) & 0xff] ^ rd_Te3[t0 & 0xff] ^ rk[49];
            s2 = rd_Te0[t2 >> 24] ^ rd_Te1[(t3 >> 16) & 0xff] ^ rd_Te2[(t0 >>  8) & 0xff] ^ rd_Te3[t1 & 0xff] ^ rk[50];
            s3 = rd_Te0[t3 >> 24] ^ rd_Te1[(t0 >> 16) & 0xff] ^ rd_Te2[(t1 >>  8) & 0xff] ^ rd_Te3[t2 & 0xff] ^ rk[51];
            /* round 13: */
            t0 = rd_Te0[s0 >> 24] ^ rd_Te1[(s1 >> 16) & 0xff] ^ rd_Te2[(s2 >>  8) & 0xff] ^ rd_Te3[s3 & 0xff] ^ rk[52];
            t1 = rd_Te0[s1 >> 24] ^ rd_Te1[(s2 >> 16) & 0xff] ^ rd_Te2[(s3 >>  8) & 0xff] ^ rd_Te3[s0 & 0xff] ^ rk[53];
            t2 = rd_Te0[s2 >> 24] ^ rd_Te1[(s3 >> 16) & 0xff] ^ rd_Te2[(s0 >>  8) & 0xff] ^ rd_Te3[s1 & 0xff] ^ rk[54];
            t3 = rd_Te0[s3 >> 24] ^ rd_Te1[(s0 >> 16) & 0xff] ^ rd_Te2[(s1 >>  8) & 0xff] ^ rd_Te3[s2 & 0xff] ^ rk[55];
        }
    }
    rk += Nr << 2;
#else  /* !RD_FULL_UNROLL */
    /*
	 * Nr - 1 full rounds:
	 */
    r = Nr >> 1;
    for (;;) {
        t0 =
            rd_Te0[(s0 >> 24)       ] ^
            rd_Te1[(s1 >> 16) & 0xff] ^
            rd_Te2[(s2 >>  8) & 0xff] ^
            rd_Te3[(s3      ) & 0xff] ^
            rk[4];
        t1 =
            rd_Te0[(s1 >> 24)       ] ^
            rd_Te1[(s2 >> 16) & 0xff] ^
            rd_Te2[(s3 >>  8) & 0xff] ^
            rd_Te3[(s0      ) & 0xff] ^
            rk[5];
        t2 =
            rd_Te0[(s2 >> 24)       ] ^
            rd_Te1[(s3 >> 16) & 0xff] ^
            rd_Te2[(s0 >>  8) & 0xff] ^
            rd_Te3[(s1      ) & 0xff] ^
            rk[6];
        t3 =
            rd_Te0[(s3 >> 24)       ] ^
            rd_Te1[(s0 >> 16) & 0xff] ^
            rd_Te2[(s1 >>  8) & 0xff] ^
            rd_Te3[(s2      ) & 0xff] ^
            rk[7];

        rk += 8;
        if (--r == 0) {
            break;
        }

        s0 =
            rd_Te0[(t0 >> 24)       ] ^
            rd_Te1[(t1 >> 16) & 0xff] ^
            rd_Te2[(t2 >>  8) & 0xff] ^
            rd_Te3[(t3      ) & 0xff] ^
            rk[0];
        s1 =
            rd_Te0[(t1 >> 24)       ] ^
            rd_Te1[(t2 >> 16) & 0xff] ^
            rd_Te2[(t3 >>  8) & 0xff] ^
            rd_Te3[(t0      ) & 0xff] ^
            rk[1];
        s2 =
            rd_Te0[(t2 >> 24)       ] ^
            rd_Te1[(t3 >> 16) & 0xff] ^
            rd_Te2[(t0 >>  8) & 0xff] ^
            rd_Te3[(t1      ) & 0xff] ^
            rk[2];
        s3 =
            rd_Te0[(t3 >> 24)       ] ^
            rd_Te1[(t0 >> 16) & 0xff] ^
            rd_Te2[(t1 >>  8) & 0xff] ^
            rd_Te3[(t2      ) & 0xff] ^
            rk[3];
    }
#endif /* ?RD_FULL_UNROLL */
    /*
	 * apply last round and
	 * map cipher state to byte array block:
	 */
	s0 =
		(rd_Te4[(t0 >> 24)       ] & 0xff000000) ^
		(rd_Te4[(t1 >> 16) & 0xff] & 0x00ff0000) ^
		(rd_Te4[(t2 >>  8) & 0xff] & 0x0000ff00) ^
		(rd_Te4[(t3      ) & 0xff] & 0x000000ff) ^
		rk[0];
	STORE_U32_BE(s0, ct     );
	s1 =
		(rd_Te4[(t1 >> 24)       ] & 0xff000000) ^
		(rd_Te4[(t2 >> 16) & 0xff] & 0x00ff0000) ^
		(rd_Te4[(t3 >>  8) & 0xff] & 0x0000ff00) ^
		(rd_Te4[(t0      ) & 0xff] & 0x000000ff) ^
		rk[1];
	STORE_U32_BE(s1, ct + 4 );
	s2 =
		(rd_Te4[(t2 >> 24)       ] & 0xff000000) ^
		(rd_Te4[(t3 >> 16) & 0xff] & 0x00ff0000) ^
		(rd_Te4[(t0 >>  8) & 0xff] & 0x0000ff00) ^
		(rd_Te4[(t1      ) & 0xff] & 0x000000ff) ^
		rk[2];
	STORE_U32_BE(s2, ct + 8 );
	s3 =
		(rd_Te4[(t3 >> 24)       ] & 0xff000000) ^
		(rd_Te4[(t0 >> 16) & 0xff] & 0x00ff0000) ^
		(rd_Te4[(t1 >>  8) & 0xff] & 0x0000ff00) ^
		(rd_Te4[(t2      ) & 0xff] & 0x000000ff) ^
		rk[3];
	STORE_U32_BE(s3, ct + 12);
}

void 
Rijndael::encipher(byte* blk)
{
	encipher(blk, blk);
}

void 
Rijndael::decipher(const byte* ct, byte* pt)
{
	uint32* rk = Rk;

	uint32 s0, s1, s2, s3, t0, t1, t2, t3;
#ifndef RD_FULL_UNROLL
    int r;
#endif /* ?RD_FULL_UNROLL */

    /*
	 * map byte array block to cipher state
	 * and add initial round key:
	 */
    s0 = LOAD_U32_BE(ct     ) ^ rk[0];
    s1 = LOAD_U32_BE(ct +  4) ^ rk[1];
    s2 = LOAD_U32_BE(ct +  8) ^ rk[2];
    s3 = LOAD_U32_BE(ct + 12) ^ rk[3];
#ifdef RD_FULL_UNROLL
    /* round 1: */
    t0 = rd_Td0[s0 >> 24] ^ rd_Td1[(s3 >> 16) & 0xff] ^ rd_Td2[(s2 >>  8) & 0xff] ^ rd_Td3[s1 & 0xff] ^ rk[ 4];
    t1 = rd_Td0[s1 >> 24] ^ rd_Td1[(s0 >> 16) & 0xff] ^ rd_Td2[(s3 >>  8) & 0xff] ^ rd_Td3[s2 & 0xff] ^ rk[ 5];
    t2 = rd_Td0[s2 >> 24] ^ rd_Td1[(s1 >> 16) & 0xff] ^ rd_Td2[(s0 >>  8) & 0xff] ^ rd_Td3[s3 & 0xff] ^ rk[ 6];
    t3 = rd_Td0[s3 >> 24] ^ rd_Td1[(s2 >> 16) & 0xff] ^ rd_Td2[(s1 >>  8) & 0xff] ^ rd_Td3[s0 & 0xff] ^ rk[ 7];
    /* round 2: */
    s0 = rd_Td0[t0 >> 24] ^ rd_Td1[(t3 >> 16) & 0xff] ^ rd_Td2[(t2 >>  8) & 0xff] ^ rd_Td3[t1 & 0xff] ^ rk[ 8];
    s1 = rd_Td0[t1 >> 24] ^ rd_Td1[(t0 >> 16) & 0xff] ^ rd_Td2[(t3 >>  8) & 0xff] ^ rd_Td3[t2 & 0xff] ^ rk[ 9];
    s2 = rd_Td0[t2 >> 24] ^ rd_Td1[(t1 >> 16) & 0xff] ^ rd_Td2[(t0 >>  8) & 0xff] ^ rd_Td3[t3 & 0xff] ^ rk[10];
    s3 = rd_Td0[t3 >> 24] ^ rd_Td1[(t2 >> 16) & 0xff] ^ rd_Td2[(t1 >>  8) & 0xff] ^ rd_Td3[t0 & 0xff] ^ rk[11];
    /* round 3: */
    t0 = rd_Td0[s0 >> 24] ^ rd_Td1[(s3 >> 16) & 0xff] ^ rd_Td2[(s2 >>  8) & 0xff] ^ rd_Td3[s1 & 0xff] ^ rk[12];
    t1 = rd_Td0[s1 >> 24] ^ rd_Td1[(s0 >> 16) & 0xff] ^ rd_Td2[(s3 >>  8) & 0xff] ^ rd_Td3[s2 & 0xff] ^ rk[13];
    t2 = rd_Td0[s2 >> 24] ^ rd_Td1[(s1 >> 16) & 0xff] ^ rd_Td2[(s0 >>  8) & 0xff] ^ rd_Td3[s3 & 0xff] ^ rk[14];
    t3 = rd_Td0[s3 >> 24] ^ rd_Td1[(s2 >> 16) & 0xff] ^ rd_Td2[(s1 >>  8) & 0xff] ^ rd_Td3[s0 & 0xff] ^ rk[15];
    /* round 4: */
    s0 = rd_Td0[t0 >> 24] ^ rd_Td1[(t3 >> 16) & 0xff] ^ rd_Td2[(t2 >>  8) & 0xff] ^ rd_Td3[t1 & 0xff] ^ rk[16];
    s1 = rd_Td0[t1 >> 24] ^ rd_Td1[(t0 >> 16) & 0xff] ^ rd_Td2[(t3 >>  8) & 0xff] ^ rd_Td3[t2 & 0xff] ^ rk[17];
    s2 = rd_Td0[t2 >> 24] ^ rd_Td1[(t1 >> 16) & 0xff] ^ rd_Td2[(t0 >>  8) & 0xff] ^ rd_Td3[t3 & 0xff] ^ rk[18];
    s3 = rd_Td0[t3 >> 24] ^ rd_Td1[(t2 >> 16) & 0xff] ^ rd_Td2[(t1 >>  8) & 0xff] ^ rd_Td3[t0 & 0xff] ^ rk[19];
    /* round 5: */
    t0 = rd_Td0[s0 >> 24] ^ rd_Td1[(s3 >> 16) & 0xff] ^ rd_Td2[(s2 >>  8) & 0xff] ^ rd_Td3[s1 & 0xff] ^ rk[20];
    t1 = rd_Td0[s1 >> 24] ^ rd_Td1[(s0 >> 16) & 0xff] ^ rd_Td2[(s3 >>  8) & 0xff] ^ rd_Td3[s2 & 0xff] ^ rk[21];
    t2 = rd_Td0[s2 >> 24] ^ rd_Td1[(s1 >> 16) & 0xff] ^ rd_Td2[(s0 >>  8) & 0xff] ^ rd_Td3[s3 & 0xff] ^ rk[22];
    t3 = rd_Td0[s3 >> 24] ^ rd_Td1[(s2 >> 16) & 0xff] ^ rd_Td2[(s1 >>  8) & 0xff] ^ rd_Td3[s0 & 0xff] ^ rk[23];
    /* round 6: */
    s0 = rd_Td0[t0 >> 24] ^ rd_Td1[(t3 >> 16) & 0xff] ^ rd_Td2[(t2 >>  8) & 0xff] ^ rd_Td3[t1 & 0xff] ^ rk[24];
    s1 = rd_Td0[t1 >> 24] ^ rd_Td1[(t0 >> 16) & 0xff] ^ rd_Td2[(t3 >>  8) & 0xff] ^ rd_Td3[t2 & 0xff] ^ rk[25];
    s2 = rd_Td0[t2 >> 24] ^ rd_Td1[(t1 >> 16) & 0xff] ^ rd_Td2[(t0 >>  8) & 0xff] ^ rd_Td3[t3 & 0xff] ^ rk[26];
    s3 = rd_Td0[t3 >> 24] ^ rd_Td1[(t2 >> 16) & 0xff] ^ rd_Td2[(t1 >>  8) & 0xff] ^ rd_Td3[t0 & 0xff] ^ rk[27];
    /* round 7: */
    t0 = rd_Td0[s0 >> 24] ^ rd_Td1[(s3 >> 16) & 0xff] ^ rd_Td2[(s2 >>  8) & 0xff] ^ rd_Td3[s1 & 0xff] ^ rk[28];
    t1 = rd_Td0[s1 >> 24] ^ rd_Td1[(s0 >> 16) & 0xff] ^ rd_Td2[(s3 >>  8) & 0xff] ^ rd_Td3[s2 & 0xff] ^ rk[29];
    t2 = rd_Td0[s2 >> 24] ^ rd_Td1[(s1 >> 16) & 0xff] ^ rd_Td2[(s0 >>  8) & 0xff] ^ rd_Td3[s3 & 0xff] ^ rk[30];
    t3 = rd_Td0[s3 >> 24] ^ rd_Td1[(s2 >> 16) & 0xff] ^ rd_Td2[(s1 >>  8) & 0xff] ^ rd_Td3[s0 & 0xff] ^ rk[31];
    /* round 8: */
    s0 = rd_Td0[t0 >> 24] ^ rd_Td1[(t3 >> 16) & 0xff] ^ rd_Td2[(t2 >>  8) & 0xff] ^ rd_Td3[t1 & 0xff] ^ rk[32];
    s1 = rd_Td0[t1 >> 24] ^ rd_Td1[(t0 >> 16) & 0xff] ^ rd_Td2[(t3 >>  8) & 0xff] ^ rd_Td3[t2 & 0xff] ^ rk[33];
    s2 = rd_Td0[t2 >> 24] ^ rd_Td1[(t1 >> 16) & 0xff] ^ rd_Td2[(t0 >>  8) & 0xff] ^ rd_Td3[t3 & 0xff] ^ rk[34];
    s3 = rd_Td0[t3 >> 24] ^ rd_Td1[(t2 >> 16) & 0xff] ^ rd_Td2[(t1 >>  8) & 0xff] ^ rd_Td3[t0 & 0xff] ^ rk[35];
    /* round 9: */
    t0 = rd_Td0[s0 >> 24] ^ rd_Td1[(s3 >> 16) & 0xff] ^ rd_Td2[(s2 >>  8) & 0xff] ^ rd_Td3[s1 & 0xff] ^ rk[36];
    t1 = rd_Td0[s1 >> 24] ^ rd_Td1[(s0 >> 16) & 0xff] ^ rd_Td2[(s3 >>  8) & 0xff] ^ rd_Td3[s2 & 0xff] ^ rk[37];
    t2 = rd_Td0[s2 >> 24] ^ rd_Td1[(s1 >> 16) & 0xff] ^ rd_Td2[(s0 >>  8) & 0xff] ^ rd_Td3[s3 & 0xff] ^ rk[38];
    t3 = rd_Td0[s3 >> 24] ^ rd_Td1[(s2 >> 16) & 0xff] ^ rd_Td2[(s1 >>  8) & 0xff] ^ rd_Td3[s0 & 0xff] ^ rk[39];
    if (Nr > 10) {
        /* round 10: */
        s0 = rd_Td0[t0 >> 24] ^ rd_Td1[(t3 >> 16) & 0xff] ^ rd_Td2[(t2 >>  8) & 0xff] ^ rd_Td3[t1 & 0xff] ^ rk[40];
        s1 = rd_Td0[t1 >> 24] ^ rd_Td1[(t0 >> 16) & 0xff] ^ rd_Td2[(t3 >>  8) & 0xff] ^ rd_Td3[t2 & 0xff] ^ rk[41];
        s2 = rd_Td0[t2 >> 24] ^ rd_Td1[(t1 >> 16) & 0xff] ^ rd_Td2[(t0 >>  8) & 0xff] ^ rd_Td3[t3 & 0xff] ^ rk[42];
        s3 = rd_Td0[t3 >> 24] ^ rd_Td1[(t2 >> 16) & 0xff] ^ rd_Td2[(t1 >>  8) & 0xff] ^ rd_Td3[t0 & 0xff] ^ rk[43];
        /* round 11: */
        t0 = rd_Td0[s0 >> 24] ^ rd_Td1[(s3 >> 16) & 0xff] ^ rd_Td2[(s2 >>  8) & 0xff] ^ rd_Td3[s1 & 0xff] ^ rk[44];
        t1 = rd_Td0[s1 >> 24] ^ rd_Td1[(s0 >> 16) & 0xff] ^ rd_Td2[(s3 >>  8) & 0xff] ^ rd_Td3[s2 & 0xff] ^ rk[45];
        t2 = rd_Td0[s2 >> 24] ^ rd_Td1[(s1 >> 16) & 0xff] ^ rd_Td2[(s0 >>  8) & 0xff] ^ rd_Td3[s3 & 0xff] ^ rk[46];
        t3 = rd_Td0[s3 >> 24] ^ rd_Td1[(s2 >> 16) & 0xff] ^ rd_Td2[(s1 >>  8) & 0xff] ^ rd_Td3[s0 & 0xff] ^ rk[47];
        if (Nr > 12) {
            /* round 12: */
            s0 = rd_Td0[t0 >> 24] ^ rd_Td1[(t3 >> 16) & 0xff] ^ rd_Td2[(t2 >>  8) & 0xff] ^ rd_Td3[t1 & 0xff] ^ rk[48];
            s1 = rd_Td0[t1 >> 24] ^ rd_Td1[(t0 >> 16) & 0xff] ^ rd_Td2[(t3 >>  8) & 0xff] ^ rd_Td3[t2 & 0xff] ^ rk[49];
            s2 = rd_Td0[t2 >> 24] ^ rd_Td1[(t1 >> 16) & 0xff] ^ rd_Td2[(t0 >>  8) & 0xff] ^ rd_Td3[t3 & 0xff] ^ rk[50];
            s3 = rd_Td0[t3 >> 24] ^ rd_Td1[(t2 >> 16) & 0xff] ^ rd_Td2[(t1 >>  8) & 0xff] ^ rd_Td3[t0 & 0xff] ^ rk[51];
            /* round 13: */
            t0 = rd_Td0[s0 >> 24] ^ rd_Td1[(s3 >> 16) & 0xff] ^ rd_Td2[(s2 >>  8) & 0xff] ^ rd_Td3[s1 & 0xff] ^ rk[52];
            t1 = rd_Td0[s1 >> 24] ^ rd_Td1[(s0 >> 16) & 0xff] ^ rd_Td2[(s3 >>  8) & 0xff] ^ rd_Td3[s2 & 0xff] ^ rk[53];
            t2 = rd_Td0[s2 >> 24] ^ rd_Td1[(s1 >> 16) & 0xff] ^ rd_Td2[(s0 >>  8) & 0xff] ^ rd_Td3[s3 & 0xff] ^ rk[54];
            t3 = rd_Td0[s3 >> 24] ^ rd_Td1[(s2 >> 16) & 0xff] ^ rd_Td2[(s1 >>  8) & 0xff] ^ rd_Td3[s0 & 0xff] ^ rk[55];
        }
    }
	rk += Nr << 2;
#else  /* !RD_FULL_UNROLL */
    /*
     * Nr - 1 full rounds:
     */
    r = Nr >> 1;
    for (;;) {
        t0 =
            rd_Td0[(s0 >> 24)       ] ^
            rd_Td1[(s3 >> 16) & 0xff] ^
            rd_Td2[(s2 >>  8) & 0xff] ^
            rd_Td3[(s1      ) & 0xff] ^
            rk[4];
        t1 =
            rd_Td0[(s1 >> 24)       ] ^
            rd_Td1[(s0 >> 16) & 0xff] ^
            rd_Td2[(s3 >>  8) & 0xff] ^
            rd_Td3[(s2      ) & 0xff] ^
            rk[5];
        t2 =
            rd_Td0[(s2 >> 24)       ] ^
            rd_Td1[(s1 >> 16) & 0xff] ^
            rd_Td2[(s0 >>  8) & 0xff] ^
            rd_Td3[(s3      ) & 0xff] ^
            rk[6];
        t3 =
            rd_Td0[(s3 >> 24)       ] ^
            rd_Td1[(s2 >> 16) & 0xff] ^
            rd_Td2[(s1 >>  8) & 0xff] ^
            rd_Td3[(s0      ) & 0xff] ^
            rk[7];

        rk += 8;
        if (--r == 0) {
            break;
        }

        s0 =
            rd_Td0[(t0 >> 24)       ] ^
            rd_Td1[(t3 >> 16) & 0xff] ^
            rd_Td2[(t2 >>  8) & 0xff] ^
            rd_Td3[(t1      ) & 0xff] ^
            rk[0];
        s1 =
            rd_Td0[(t1 >> 24)       ] ^
            rd_Td1[(t0 >> 16) & 0xff] ^
            rd_Td2[(t3 >>  8) & 0xff] ^
            rd_Td3[(t2      ) & 0xff] ^
            rk[1];
        s2 =
            rd_Td0[(t2 >> 24)       ] ^
            rd_Td1[(t1 >> 16) & 0xff] ^
            rd_Td2[(t0 >>  8) & 0xff] ^
            rd_Td3[(t3      ) & 0xff] ^
            rk[2];
        s3 =
            rd_Td0[(t3 >> 24)       ] ^
            rd_Td1[(t2 >> 16) & 0xff] ^
            rd_Td2[(t1 >>  8) & 0xff] ^
            rd_Td3[(t0      ) & 0xff] ^
            rk[3];
    }
#endif /* ?RD_FULL_UNROLL */
    /*
	 * apply last round and
	 * map cipher state to byte array block:
	 */
   	s0 =
   		(rd_Td4[(t0 >> 24)       ] & 0xff000000) ^
   		(rd_Td4[(t3 >> 16) & 0xff] & 0x00ff0000) ^
   		(rd_Td4[(t2 >>  8) & 0xff] & 0x0000ff00) ^
   		(rd_Td4[(t1      ) & 0xff] & 0x000000ff) ^
   		rk[0];
	STORE_U32_BE(s0, pt     );
   	s1 =
   		(rd_Td4[(t1 >> 24)       ] & 0xff000000) ^
   		(rd_Td4[(t0 >> 16) & 0xff] & 0x00ff0000) ^
   		(rd_Td4[(t3 >>  8) & 0xff] & 0x0000ff00) ^
   		(rd_Td4[(t2      ) & 0xff] & 0x000000ff) ^
   		rk[1];
	STORE_U32_BE(s1, pt + 4 );
   	s2 =
   		(rd_Td4[(t2 >> 24)       ] & 0xff000000) ^
   		(rd_Td4[(t1 >> 16) & 0xff] & 0x00ff0000) ^
   		(rd_Td4[(t0 >>  8) & 0xff] & 0x0000ff00) ^
   		(rd_Td4[(t3      ) & 0xff] & 0x000000ff) ^
   		rk[2];
	STORE_U32_BE(s2, pt + 8 );
   	s3 =
   		(rd_Td4[(t3 >> 24)       ] & 0xff000000) ^
   		(rd_Td4[(t2 >> 16) & 0xff] & 0x00ff0000) ^
   		(rd_Td4[(t1 >>  8) & 0xff] & 0x0000ff00) ^
   		(rd_Td4[(t0      ) & 0xff] & 0x000000ff) ^
   		rk[3];
	STORE_U32_BE(s3, pt + 12);
}

void 
Rijndael::decipher(byte* blk)
{
	decipher(blk, blk);
}

LEDA_END_NAMESPACE
