/*******************************************************************************
+
+  LEDA 7.0  
+
+
+  simple_coder.cpp
+
+
+  Copyright (c) 1995-2023
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/

#include "simple_coder.h"

#if defined(LEDA_STD_IO_HEADERS)
using std::streambuf;
using std::cout;
using std::endl;
#endif

simple_coder::simple_coder(streambuf* src_stream, streambuf* tgt_stream)
: base(src_stream, tgt_stream), ByteToAdd(DefaultByteToAdd)
{}

simple_coder::simple_coder(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name), ByteToAdd(DefaultByteToAdd)
{}

void 
simple_coder::do_reset(reset_type t)
{
	// call the base class
	base::do_reset(t);

	// and if t == r_hard, set all data to its initial values
	if (t == r_hard) {
		ByteToAdd = DefaultByteToAdd;
	}
}

void 
simple_coder::write_header_to_tgt_stream()
{
	base::write_header_to_tgt_stream();

	if (write_byte(ByteToAdd) < 0) error();
}

void 
simple_coder::read_header_from_src_stream()
{
	base::read_header_from_src_stream();

	int c = read_byte();
	if (c < 0) error();
	ByteToAdd = (byte) c;
}


/*
	Our coder is very simple:
	The encoder counts the number of bytes in the stream and 
	adds ByteToAdd to every byte.
	The decoder counts the number of bytes in the stream and 
	subtracts ByteToAdd from every byte.

	We want to make some remarks about our model of streams:
	Suppose that the encoder calls a function to read a character from a 
	stream, then the return value c can be one of the following:
	- a regular character, i.e. 0 <= c <= 0xFF
	- the value EOF (usually EOF == -1) to indicate the end of the stream
	- the value StreamBlocked (defined in coder_base.h) (explained below)

	We give some details on StreamBlocked, let us consider the following code:
	(1) encoding_ofstream<simple_coder> out("foo");
	(2) out << "Hel";
	(3) out << "lo!";
	(4) out.close();
	We analyse what happens from the point of view of simple_coder.
	When the output is sent to out in line (2), the coder gets initialized, 
	i.e. the function simple_coder::init_encoding() is called.
	Immediately after that the function simple_coder::encode_stream() is 
	called. This function reads the following data from the input stream:
		<H>, <e>, <l>, StreamBlocked
	StreamBlocked indicates that there is no more input data available now, 
	but it might be the case that the input stream will be able to deliver 
	more data in the future. Hence, the call of simple_coder::encode_stream() 
	returns.
	When line (3) is executed, more input data is sent to out and the function
	simple_coder::encode_stream() is called once more to process the new data.
	It reads the following input:
		<l>, <o>, <!>, StreamBlocked
	Upon receiving StreamBlocked it returns again.
	In line (4) the stream is closed, thus simple_coder::encode_stream() is 
	invoked one final time. It reads the EOF symbol from the input stream and
	calls the function coder_base::encoding_finished(), which does some clean-
	up work.
 */

void 
simple_coder::init_encoding()
{
	// call the function from the bass class first!
	base::init_encoding();

	// then perform your own initialization
	ByteCount = 0;
}

void 
simple_coder::encode_stream()
{
	/*
		in encoding mode "int c = read_byte()" may return the following values:
		- a regular character (i.e. 0 <= c <= 0xFF)
		- EOF (usually EOF == -1)
		- StreamBlocked, i.e. c == - 2
		  This means that currently there is no more input data available, but 
		  there might be some again later. So the coder should suspend its work 
		  and resume it later.
		and write_byte(symbol) may return
		- the written character "symbol" (to indicate success)
		- EOF to indicate failure

		we also explain the usage of read_bytes/write_bytes (not used below):
		"int res = read_bytes(buffer, buf_sz)" in encoding mode:
		- res == buf_sz -> read stopped because the buffer was full 
		  (there may be more input)
		- 0 <= res < buf_sz -> read stopped because EOF was detected
		  res equals the number of regular characters read and stored in buffer
		- res <= StreamBlocked < 0 -> read stopped as StreamBlocked occured
		  (StreamBlocked - res) equals the number of reg. chars read into buffer
		"int res = write_bytes(buffer, buf_sz)" in encoding mode:
		- res == buf_sz -> success, all bytes have been written to the stream
		- 0 <= res < buf_sz -> write stopped because an error occured
		  res equals the number of characters written
	 */
	int c;
	while ((c = read_byte()) >= 0 && c != EOF) {
		++ByteCount;
		byte symbol = (byte) c;
		if (write_byte(symbol+1) == EOF) error("write error");
	}

	if (c == EOF) {
//		std::cout << "encode EOF received after " << ByteCount << " bytes!\n";
		encoding_finished();
	}
	else {
//		std::cout << "encode STREAM_BLOCK received!\n";
	}
}


void 
simple_coder::init_decoding()
{
	// call the function from the bass class first!
	base::init_decoding();

	// then perform your own initialization
	ByteCount = 0;
}

void
simple_coder::decode_stream()
{
	/*
		in decoding mode read_byte() may return the following values:
		- a character, i.e. 0 <= c <= 0xFF
		- EOF (usually EOF == -1)
		and write_byte(symbol) may return
		- the written character "symbol" (to indicate success)
		- EOF to indicate failure
		- StreamBlocked
		  The call of write_byte(symbol) could succesfully write "symbol", but
		  the output stream cannot consume any additional data at the moment.
		  It might be able again later, so the coder should suspend its work 
		  now and prepare to resume later.

		read_bytes/write_bytes behave similar as in encoding mode; differences:
		- StreamBlocked cannot occur during reading, hence read_bytes() always
		  returns a non-negative number
		- StreamBlocked CAN occur during writing, so write_bytes() may return a
		  negative number res, when it receives a StreamBlocked before 
		  completion; the total number of successfully written characters can 
		  then be calculated as (StreamBlocked - res).
	 */
	int c;
	while ((c = read_byte()) != EOF) {
		++ByteCount;
		byte symbol = (byte) c;
		if ( (c=write_byte(symbol-1)) < 0) {
			if (c == EOF) { error("write error"); return; }
			else {
//				std::cout << "decode STREAM_BLOCK received!\n";
				return;
			}			
		}
	}

//	std::cout << "decode EOF received after " << ByteCount << " bytes!\n";
	decoding_finished();
}


streampos 
simple_coder::decode_seek(streampos current_pos, streampos new_pos)
/*
	Arguments:
	- current_pos: current position of coder 
	  (i.e. the pos after the last byte which has been read)
	  This is provided as an info to facilitate implementing decode_seek.
    - new_pos: (desired) new position of the coder
	  (if new_pos is negative, we are supposed to seek past the end of the 
	  stream)
	- return value: (actual) new position of the coder
	  (return streampos(-1) to indicate failue)
	All arguments to refer to positions in the decoded stream (which is equal 
	to the original data stream used for encoding)!!
	
	If new_pos is negative, the coder is supposed to determine the size of the
	decoded data stream and to call decoding_finished.
	Otherwise, the coder must go to the appropriate position in the encoded 
	data stream (i.e. the current src stream) and make sure the the next call 
	to decode_stream() outputs the data starting at position "new pos" in the 
	decoded stream.
*/
{
	// make sure that coder has been initialized and that it is okay
	if (!decoding() || failed()) return base::decode_seek(current_pos, new_pos);

	if (new_pos < streamoff(0)) {
		// seek past the end
		new_pos = seek_end_src_stream_relative_to_header();
		decoding_finished();
		return new_pos;
	}
	else {
		// regular seek
		set_state(s_decoding); // make it running ...
		return seek_pos_src_stream_relative_to_header(new_pos);

		// -> the next call to decode_stream() will return data at new_pos
	}
}

// register the coder so that it can be used with AutoDecoder
#include <LEDA/coding/coder_factory.h>
static leda::register_coder<simple_coder> _register_simple_coder_;
