/*******************************************************************************
+
+  LEDA 7.0
+
+
+  test_all_coders.cpp
+
+
+  Copyright (c) 1995-2023
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


#if defined(_MSC_VER) && _MSC_VER < 1300
#pragma optimize( "", off )
#endif

#include <LEDA/core/string.h>
#include <LEDA/core/random_source.h>
#include <LEDA/system/file.h>
#include <fstream>
#include <string.h> // just for memcmp
#include <cstdlib> // just for strtoul

#include <LEDA/coding/coder_factory.h>
#include <LEDA/coding/coder_strbuf_adapter.h>
#include <LEDA/coding/crypt.h>
#include <LEDA/coding/block_cipher_factory.h>

using namespace leda;

typedef coder_types::byte byte;
typedef coder_types::uint32 uint32;

#include "block_ciphers_test.tbl"

string crossplatform_org = "cross_platform.org";
string crossplatform_enc = "cross_platform.enc";
string crossplatform_key = "cross_platform.key";

const byte key_size = 128 / 8;
static const byte key_bytes[] = 
	{ 0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xB, 0xC, 0xD, 0xF };
CryptKey global_key(key_bytes, key_size);

typedef CoderPipe3<BWTCoder, MTFCoder, A0Coder> Coder;
typedef Coder CoderForCallbacks;

using std::cout;
using std::cerr;
using std::endl;
using std::flush;
using std::ostream;
using std::istream;
using std::ofstream;
using std::ifstream;
using std::streambuf;
using std::filebuf;
using std::ios;

using std::strtoul;

#ifndef __mipspro__
using std::streampos;
using std::streamoff;
using std::streamsize;
#endif

#ifdef LEDA_COMPRESSION_WA1
#define openmode open_mode
#define streamsize int
#define pubsync sync
#endif

#ifdef LEDA_COMPRESSION_WA2
#define filebuf leda_filebuf
#endif

#define open_in  coder_base::binary_in()
#define open_out coder_base::binary_out()

class coder_tester {
public:
	enum { BufferSize = 1024 };
	enum {
		// modules
		TestIOStreamLib = 0x0001,
		TestStreamIO = 0x0002, TestAdapterIO = 0x0004, TestMemoryChunks = 0x0008,
		TestEncDecStreams = 0x0010, TestSeekCoders = 0x0020, TestCrossPlatform = 0x0040, 
		TestCryptography = 0x0080, TestCallbacks = 0x0100,
		// file types
		TestEmptyFile = 0x0200, TestSmallFile = 0x0400, TestRandomFile = 0x0800,
		// coder types
		TestStandalone = 0x1000, TestSmallPipes = 0x2000, TestBigPipe = 0x4000,
		// everything
		TestAll = 0xFFFF
	};
	typedef coder_types::uint32 uint32;

public:
	coder_tester(int verbosity = 1)
	 : Verbosity(verbosity), Debug(false), DebugSize(1), TestMask(TestAll), 
	   CurrentCoder(0), ExitOnFailure(false), Okay(true), DeleteFiles(true) 
	{ set_coders_to_test(CoderFactory.get_standalone_coders(), true); }

	~coder_tester() { clear_coders_to_test(); }

	bool run(bool exit_on_failure = true);
	bool run(string test_file, string encoded_file, string decoded_file, bool exit_on_failure = true);
	bool generate_files_for_crossplatform_test(string input_file, string output_file, string key_file);

	void set_verbosity(int level) { Verbosity = level; }
	void inc_verbosity() { ++Verbosity; }

	void set_debug_mode(bool enable = true) { Debug = enable; if (enable) set_verbosity(3); }
	void set_debug_size(int sz) { set_debug_mode(); DebugSize = sz; }

	void set_test_mask(unsigned long mask = TestAll) { TestMask = mask; }
	void add_tests(unsigned long mask) { TestMask |= mask; }
	void remove_tests(unsigned long mask) { TestMask &= (TestAll ^ mask); }

	void set_coders_to_test(const list<coder_base*>& coders, bool own_coders = false);
	void remove_coder_to_test(const string& descr);
	void remove_all_coders_to_test_except(const string& descr);
	void clear_coders_to_test();

	bool failed() const { return !Okay; }

	void keep_files() { DeleteFiles = false; }

public:
	static bool encode_file(coder_base* coder, const char* src, const char* tgt, bool reset_coder = true);
	static bool decode_file(coder_base* coder, const char* src, const char* tgt, bool reset_coder = true);
	static bool compare_files(const char* file1, const char* file2);

	static bool copy_streambufs(streambuf* stream1, streambuf* stream2);
	static bool copy(const char* src, streambuf* tgt);
	static bool copy(streambuf* src, const char* tgt);
	static bool compare_streambufs(streambuf* stream1, streambuf* stream2);

private:
	// helpers
	void error(string msg);
	void encode_file(bool reset_coder = true);
	void decode_file(bool reset_coder = true);
	void compare_files();
	void generate_random_file(int size);

	void show_sizeof_coder_types();

	// test module: test iostream library
	void test_iostream_lib();

	// test module: test all coders
	void test_all_coders_with_files();
	void test_all_coders_with_file();
	void test_current_coder_with_file();
	void test_current_coder_with_streambufIO(bool reset_coder = true);
	void test_current_coder_with_adapterIO();
	void test_current_coder_with_memory_chunks();

	// test module: test encoding_ofstream/decoding_ifstream
	void test_one_coder_with_enc_dec_streams();
	void test_one_coder_with_enc_dec_streams_basic();
	void test_one_coder_with_enc_dec_streams_seek();
	void compare_streams_wrt_seek(istream& org, istream& cpy, int random_seeks = 0);

	// test module: test callbacks
	void test_one_coder_with_callbacks();

	// test module: test seek optimizing coders
	void test_seek_optimizing_coders();
	void test_coder_seek(bool is_cipher = false, int random_seeks = 0);
	void test_coder_pipe_seek(coder_pipe& pc, crypter_sym_key_base* c1 = 0, crypter_sym_key_base* c2 = 0, crypter_sym_key_base* c3 = 0, int random_seeks = 0);

	// test module: test cross platform
	void test_crossplatform();

	// test module: test cryptography
	void test_cryptography();
	void test_block_ciphers();
	void test_blowfish_one_block(byte* key, byte key_bytes, uint32 pl, uint32 pr, uint32 cl, uint32 cr);
	void test_twofish_run(int num_key_bytes, byte* final_ct);
	void test_rijndael_one_block(const byte* key, byte key_bytes, const byte* pt, const byte* ct);
	void test_omac_rijndael();
	void test_crypt_coders();
	list<block_cipher_mode_base*> construct_block_cipher_modes();
	list<crypter_sym_key_base*> construct_crypt_coders();
	void test_current_crypt_coder_with_streambufIO(bool reset_crypter = true);
	void test_current_block_cipher_mode_with_memory_chunks();
	void test_checksummers();

private:
	int						Verbosity;
	bool                    Debug;
	int						DebugSize;
	unsigned long			TestMask;
	list<coder_base*>		CodersToTest;
	bool					OwnCodersToTest;
	coder_base*				CurrentCoder;
	coder_strbuf_adapter	Adapter;
	string					TestFileName;
	string					TestFileDescription;
	string					EncodedFileName;
	string					DecodedFileName;
	bool					ExitOnFailure;
	bool					Okay;
	bool					DeleteFiles;
};


/// run >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

bool
coder_tester::run(bool exit_on_failure)
{
	return run(tmp_file_name(), tmp_file_name(), tmp_file_name(), exit_on_failure);
}

bool 
coder_tester::run(string test_file, string encoded_file, string decoded_file, bool exit_on_failure)
{
	if (Debug) {
		cout << "DEBUG mode enabled (size = " << DebugSize << ")" << endl << endl;
	}

	ExitOnFailure = exit_on_failure;
	TestFileName = test_file;
	EncodedFileName = encoded_file;
	DecodedFileName = decoded_file;

	Okay = true;

	show_sizeof_coder_types();
	test_iostream_lib();
	test_all_coders_with_files();
	test_one_coder_with_enc_dec_streams();
	test_one_coder_with_callbacks();
	test_seek_optimizing_coders();
	test_cryptography();
	test_crossplatform();

	return Okay;
}

/// run <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


/// manipulating CodersToTest >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

void 
coder_tester::set_coders_to_test(const list<coder_base*>& coders, bool own_coders) 
{
	clear_coders_to_test(); 
	CodersToTest = coders; OwnCodersToTest = own_coders;
}

void
coder_tester::remove_coder_to_test(const string& descr)
{
	list_item it;
	forall_items(it, CodersToTest) {
		coder_base* coder = CodersToTest[it];
		if (coder->get_description().index(descr) >= 0) {
			if (OwnCodersToTest) delete coder;
			CodersToTest.del_item(it);
		}
	}
}

void
coder_tester::remove_all_coders_to_test_except(const string& descr)
{
	list_item it;
	forall_items(it, CodersToTest) {
		coder_base* coder = CodersToTest[it];
		if (coder->get_description().index(descr) < 0) {
			if (OwnCodersToTest) delete coder;
			CodersToTest.del_item(it);
		}
	}
}

void 
coder_tester::clear_coders_to_test()
{
	if (OwnCodersToTest) {
          forall(CurrentCoder, CodersToTest) delete CurrentCoder; 
        }

	CurrentCoder = 0;
	CodersToTest.clear(); OwnCodersToTest = false;
}

/// manipulating CodersToTest <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


/// public static helpers >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

bool 
coder_tester::encode_file(coder_base* coder, const char* src, const char* tgt, bool reset_coder)
{
	delete_file(tgt); 

	bool ok;
	{
		if (reset_coder) coder->reset();
		coder->set_src_file(src); 
                coder->set_tgt_file(tgt);
		coder->encode();
		ok = !coder->failed();
	}
	return ok;
}

bool 
coder_tester::decode_file(coder_base* coder, const char* src, const char* tgt, bool reset_coder)
{
//	delete_file(tgt); 

	bool ok;
	{
		if (reset_coder) coder->reset();
		coder->set_src_file(src); coder->set_tgt_file(tgt);
		coder->decode();
		ok = !coder->failed();
	}
	return ok;
}

bool 
coder_tester::compare_files(const char* file1, const char* file2)
{
	filebuf file1_fb;
	file1_fb.open(file1, open_in);
	filebuf file2_fb;
	file2_fb.open(file2, open_in);

	return compare_streambufs(&file1_fb, &file2_fb);
}

bool 
coder_tester::copy_streambufs(streambuf* stream1, streambuf* stream2)
{
	char buffer[BufferSize];
	int bytes_read, bytes_written;
	do {
		bytes_read    = stream1->sgetn(buffer, BufferSize);
		bytes_written = stream2->sputn(buffer, bytes_read);
		if (bytes_read != bytes_written) { 
			return false;
		}
	} while (bytes_read == BufferSize);

	return true;
}

bool 
coder_tester::copy(const char* src, streambuf* tgt)
{
	filebuf src_fb;
	src_fb.open(src, open_in);

	return copy_streambufs(&src_fb, tgt);
}

bool 
coder_tester::copy(streambuf* src, const char* tgt)
{
	filebuf tgt_fb;
	tgt_fb.open(tgt, open_out);

	return copy_streambufs(src, &tgt_fb);
}

bool 
coder_tester::compare_streambufs(streambuf* stream1, streambuf* stream2)
{
	char buffer1[BufferSize], buffer2[BufferSize];
	int bytes_read1, bytes_read2;
	bool equal = true;
	do {
		bytes_read1 = stream1->sgetn(buffer1, BufferSize);
		bytes_read2 = stream2->sgetn(buffer2, BufferSize);
		if ((bytes_read1 != bytes_read2) || memcmp(buffer1, buffer2, bytes_read1)) { 
			equal = false; break;
		}
	} while (bytes_read1 == BufferSize);

	return equal;
}

/// public static helpers <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


/// private helpers >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

void 
coder_tester::error(string msg)
{
	Okay = false;

	if (CurrentCoder) {
		cerr << "FAILURE: " << CurrentCoder->get_description() << " " << msg << endl;
	}
	else {
		cerr << "FAILURE: " << msg << endl;
	}

	if (ExitOnFailure) {
		cerr << "TestFile:    " << TestFileName << endl;
		cerr << "EncodedFile: " << EncodedFileName << endl;
		cerr << "DecodedFile: " << DecodedFileName << endl;
		exit(1);
	}
}


void
coder_tester::encode_file(bool reset_coder)
{
	bool ok = encode_file(CurrentCoder, TestFileName, EncodedFileName, reset_coder);
	if (!ok) error("file encode");
}

void
coder_tester::decode_file(bool reset_coder)
{
	bool ok = decode_file(CurrentCoder, EncodedFileName, DecodedFileName, reset_coder);
	if (!ok) error("file decode");
}

void 
coder_tester::compare_files()
{
	bool ok = compare_files(TestFileName, DecodedFileName);
	if (!ok) error("files compare");
}

void 
coder_tester::generate_random_file(int size)
{
	if (Debug) {
		size = 1;
		TestFileDescription = string("_DEB_ %d", size);

		ofstream out(TestFileName);
		while (--size >= 0) {
			char ch;
			ch = char('A' + size);
			out.put(ch);
		}
		out.close();
	}
	else {
		random_source S;
		int seed = int(used_time()*1000);
		S.set_seed(seed);

		TestFileDescription = string("random %d", size);

		ofstream out(TestFileName);
		while (--size >= 0) {
			char ch;
			S >> ch;
			out.put(ch);
		}
		out << "seed:" << seed;
		out.close();
	}
}

void 
coder_tester::show_sizeof_coder_types()
{
	if (Verbosity < 2) return;

	cout << "sizes of coder types:" << endl;
	cout << "  coder_types::byte:    " << sizeof(coder_types::byte) << endl;
	cout << "  coder_types::uint16:  " << sizeof(coder_types::uint16) << endl;
	cout << "  coder_types::sint16:  " << sizeof(coder_types::sint16) << endl;
	cout << "  coder_types::uint32:  " << sizeof(coder_types::uint32) << endl;
	cout << "  coder_types::sint32:  " << sizeof(coder_types::sint32) << endl;
	cout << "  coder_types::id_type: " << sizeof(coder_types::id_type) << endl;
	cout << endl;
}

/// private helpers <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


/// module: test iostream library >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

void
coder_tester::test_iostream_lib()
{
	if ((TestMask & TestIOStreamLib) == 0) return;

	if (Verbosity) cout << "testing iostream library on this platform ..." << endl;

	// test ostream::write and istream::read
	{
		const char* test_str = "ABCD";
		char buffer[sizeof(test_str)+1];
		memset(buffer, 0, sizeof(buffer));

		ofstream out(TestFileName, open_out);
		out.write(test_str, sizeof(test_str));
		out.close();

		ifstream in(TestFileName, open_in);
		in.read(buffer, sizeof(test_str));
		in.close();

		if (strncmp(test_str, buffer, strlen(test_str)) != 0) {
			error("ostream::write or istream::read is buggy");
		}
	}

	if (Verbosity) cout << endl;

	if (DeleteFiles) delete_file(TestFileName);
}

/// module: test iostream library <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


/// module: test all coders >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

void
coder_tester::test_all_coders_with_files()
{
	if (TestMask & TestEmptyFile) {
		// generate empty test file
		ofstream out(TestFileName);
		out.close();

		TestFileDescription = "empty file";
		test_all_coders_with_file();
	}

	if (TestMask & TestSmallFile) {
		// generate small test file
		ofstream out(TestFileName);
		int i;
		for (i = 0; i < 256; ++i) {
			for (int j = 0; j < 20; ++j)
				out.put(char(i));
		}
//		out << "Hallo!" << endl;
		out.put((char)0);
		out.close();

		TestFileDescription = "small";
		test_all_coders_with_file();
	}

	if (TestMask & TestRandomFile) {
		generate_random_file(4*1024);
		test_all_coders_with_file();
	}

	if (DeleteFiles) delete_file(TestFileName);
}

void
coder_tester::test_all_coders_with_file()
{
	if (Verbosity) cout << "testing all coders on file (" << TestFileDescription << ") ..." << endl;

	if (TestMask & TestStandalone) {
		// test standalone coders
		forall(CurrentCoder, CodersToTest) {
			if (Verbosity > 1) cout << "  testing coder " << CurrentCoder->get_description() << endl;
			test_current_coder_with_file();
		}
	}

	if (TestMask & TestSmallPipes) {
		// test coder_pipe (size 3)
		if (CodersToTest.size() >= 3) {
			list_item it;
			forall_items(it, CodersToTest) {
				coder_base* coder1 = CodersToTest[it];
				it = CodersToTest.cyclic_succ(it);
				coder_base* coder2 = CodersToTest[it];
				it = CodersToTest.cyclic_succ(it);
				coder_base* coder3 = CodersToTest[it];

				GenericCoderPipe pipe(3);
				pipe.set_coder(1, coder1);
				pipe.set_coder(2, coder2);
				pipe.set_coder(3, coder3);

				CurrentCoder = &pipe;
				if (Verbosity >= 2) cout << "  testing coder " << CurrentCoder->get_description() << endl;
				test_current_coder_with_file();
			}
		}
	}

	if (TestMask & TestBigPipe) {
		// test coder_pipe (big)
		// (we create one big pipe that contains all standalone coders ...)
		{
			GenericCoderPipe* p_small_pipe = new GenericCoderPipe(2);
			p_small_pipe->set_coder(1, new Coder(), true);
			p_small_pipe->set_coder(2, new Coder(), true);

			GenericCoderPipe big_pipe(CodersToTest.size()+1);
			coder_base* coder; int k = 0;
			forall(coder, CodersToTest) {
				big_pipe.set_coder(++k, coder);
			}
			big_pipe.set_coder(++k, p_small_pipe, true);

			CurrentCoder = &big_pipe;
			if (Verbosity >= 2) cout << "  testing coder " << CurrentCoder->get_description() << endl;
			test_current_coder_with_file();
		}
	}

	CurrentCoder = 0;

	if (Verbosity) cout << endl;

	if (DeleteFiles) { delete_file(EncodedFileName); delete_file(DecodedFileName); }
}

void 
coder_tester::test_current_coder_with_file()
{
	test_current_coder_with_streambufIO();
	test_current_coder_with_adapterIO();
	test_current_coder_with_memory_chunks();
}

void 
coder_tester::test_current_coder_with_streambufIO(bool reset_coder)
{
	if ((TestMask & TestStreamIO) == 0) return;
	if (Verbosity >= 3) cout << "    testing coder with streambuf IO ..." << endl;

	// test encoding and decoding with coder
	if (Verbosity >= 4) cout << "      encoding ..." << endl;
	encode_file(reset_coder); if (failed()) return;

	if (Verbosity >= 4) cout << "      decoding ..." << endl;
	decode_file(); if (failed()) return;

	if (Verbosity >= 4) cout << "      comparing ..." << endl;
	compare_files(); if (failed()) return;

	// test decoding with AutoDecoder
	if (Verbosity >= 4) cout << "      auto decoding ..." << endl;
	AutoDecoder auto_decoder;
	coder_base* sav_coder = CurrentCoder;
	CurrentCoder = &auto_decoder;
	decode_file();
	if (!failed()) {
		if (Verbosity >= 4) cout << "      comparing (auto) ..." << endl;
		compare_files();
	}
	CurrentCoder = sav_coder;
}

void
coder_tester::test_current_coder_with_adapterIO()
{
	if ((TestMask & TestAdapterIO) == 0) return;
	if (Verbosity >= 3) cout << "    testing coder with adapter IO ..." << endl;

	bool ok;

	// test adapter with CurrentCoder
	Adapter.set_coder(CurrentCoder, false);

	// encode with adapter
	if (Verbosity >= 4) cout << "      encoding ..." << endl;
	Adapter.reset();
	Adapter.set_encoded_file(EncodedFileName, ios::out);
	ok = copy(TestFileName, &Adapter); Adapter.sputEOF();
	if (!ok || Adapter.failed()) {
		error("adapter encode");
		return;
	}

	// decode with adapter
	if (Verbosity >= 4) cout << "      decoding ..." << endl;
	Adapter.reset();
	Adapter.set_encoded_file(EncodedFileName, ios::in);
	ok = copy(&Adapter, DecodedFileName);
	if (!ok || Adapter.failed()) {
		error("adapter decode");
		return;
	}

	// compare files
	if (Verbosity >= 4) cout << "      comparing ..." << endl;
	compare_files();
	if (failed()) return;

	// decode with adapter and AutoDecoder
	if (Verbosity >= 4) cout << "      auto decoding ..." << endl;
	AutoDecoder auto_decoder;
	coder_base* sav_coder = CurrentCoder;
	CurrentCoder = &auto_decoder;
	Adapter.reset();
	Adapter.set_encoded_file(EncodedFileName, ios::in);
	ok = copy(&Adapter, DecodedFileName);
	Adapter.reset(); Adapter.set_coder(0);
	if (!ok || Adapter.failed()) {
		error("adapter AutoDecoder decode");
		return;
	}
	if (!failed()) {
		if (Verbosity >= 4) cout << "      comparing (auto) ..." << endl;
		compare_files();
	}
	CurrentCoder = sav_coder;
}

void
coder_tester::test_current_coder_with_memory_chunks()
{
	if ((TestMask & TestMemoryChunks) == 0) return;
	if (Verbosity >= 3) cout << "    testing coder with memory chunks ..." << endl;

	// prepare memory buffers and read TestFile into memory
	uint32 orig_sz    = uint32( size_of_file(TestFileName) );
	uint32 max_enc_sz = 4 * (orig_sz + 1024);
	char* buf_orig = new char[orig_sz];
	char* buf_enc  = new char[max_enc_sz];
	char* buf_dec  = new char[orig_sz];
	{
		ifstream in(TestFileName, open_in);
		in.read(buf_orig, orig_sz);
	}

	// encode memory chunk
	if (Verbosity >= 4) cout << "      encoding ..." << endl;
	CurrentCoder->reset();
	uint32 enc_sz = CurrentCoder->encode_memory_chunk(buf_orig, orig_sz, buf_enc, max_enc_sz);
	if (CurrentCoder->failed()) {
		error("memory encode");
		return;
	}

	// decode memory chunk
	if (Verbosity >= 4) cout << "      decoding ..." << endl;
	CurrentCoder->reset();
	uint32 dec_sz = CurrentCoder->decode_memory_chunk(buf_enc, enc_sz, buf_dec, orig_sz);
	if (CurrentCoder->failed()) {
		error("memory decode");
		return;
	}

	// compare memory chunks
	if (Verbosity >= 4) cout << "      comparing ..." << endl;
	if (orig_sz != dec_sz || memcmp(buf_orig, buf_dec, orig_sz) != 0) {
		error("memory compare");
		return;
	}

	// free memory
	delete[] buf_orig;
	delete[] buf_enc;
	delete[] buf_dec;
}

/// module: test all coders <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


/// module: test encoding_ofstream/decoding_ifstream >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

void
coder_tester::test_one_coder_with_enc_dec_streams()
{
	if ((TestMask & TestEncDecStreams) == 0) return;
	if (Verbosity) {
		Coder coder;
		cout << "testing encoding_ofstream / decoding_ifstream";
		if (Verbosity >= 2)
			cout << endl << "        with coder " << coder.get_description();
		cout << " ..." << endl;
	}

	test_one_coder_with_enc_dec_streams_basic();
	test_one_coder_with_enc_dec_streams_seek();

	if (Verbosity) cout << endl;

	if (DeleteFiles) {
		delete_file(TestFileName); delete_file(EncodedFileName);
	}
}

void
coder_tester::test_one_coder_with_enc_dec_streams_basic()
{
	if (Verbosity >= 3) cout << "   basic I/O ..." << endl;

	int num_lines = 20;
	char buffer1[256], buffer2[256];
	int i;

	// encoding
	{
		if (Verbosity >= 4) cout << "      encoding ..." << endl;

		ofstream                 out1(TestFileName);
		encoding_ofstream<Coder> out2(EncodedFileName);

		for (i = 0; i < num_lines; ++i) {
			string line("Hallo Test %d", i);
			out1.seekp(2, ios::cur); out1 << line;
			out2.seekp(2, ios::cur); out2 << line;
		}
		out1.close(); out2.close();

		if (out1.fail()) {
			error("ofstream (out1)");
			return;
		}
		if (out2.fail()) {
			error("encoding_ofstream (out2)");
			return;
		}
	}

	// decoding
	{
		if (Verbosity >= 4) cout << "      decoding ..." << endl;

		ifstream                 in1(TestFileName);
		decoding_ifstream<Coder> in2(EncodedFileName);

		for (i = 0; i < num_lines; ++i) {
			if (in1.get() != 0 || in1.get() != 0) { error("zeros (in1)"); return; }
			in1.get(buffer1, 256, (char) 0);
			if (in2.get() != 0 || in2.get() != 0) { error("zeros (in2)"); return; }
			in2.get(buffer2, 256, (char) 0);
			if (string(buffer1) != string(buffer2)) { error("strings differ"); return; }
		}
		in1.close(); in2.close();

		if (in1.fail()) {
			error("decoding_ifstream (in1)");
			return;
		}
		if (in2.fail()) {
			error("decoding_ifstream (in2)");
			return;
		}
	}

	// auto decoding
	{
		if (Verbosity >= 4) cout << "      auto decoding ..." << endl;
	
		ifstream              in1(TestFileName);
		autodecoding_ifstream in2(EncodedFileName);

		for (i = 0; i < num_lines; ++i) {
			if (in1.get() != 0 || in1.get() != 0) { error("auto zeros (in1)"); return; }
			in1.get(buffer1, 256, (char) 0);
			if (in2.get() != 0 || in2.get() != 0) { error("auto zeros (in2)"); return; }
			in2.get(buffer2, 256, (char) 0);
			if (string(buffer1) != string(buffer2)) { error("auto strings differ"); return; }
		}
		in1.close(); in2.close();

		if (in1.fail()) {
			error("ifstream (in1)");
			return;
		}
		if (in2.fail()) {
			error("autodecoding_ifstream (in2)");
			return;
		}
	}
}

void
coder_tester::test_one_coder_with_enc_dec_streams_seek()
{
	if (Verbosity >= 3) cout << "   seek ..." << endl;

	// decode seek
	{
		if (Verbosity >= 4) cout << "      decode seek ..." << endl;

		ifstream                 in1(TestFileName, open_in);
		decoding_ifstream<Coder> in2(EncodedFileName);
		compare_streams_wrt_seek(in1, in2);
	}

	// auto decode seek
	{
		if (Verbosity >= 4) cout << "      auto decode seek ..." << endl;

		ifstream              in1(TestFileName, open_in);
		autodecoding_ifstream in2(EncodedFileName);
		compare_streams_wrt_seek(in1, in2);
	}
}

void 
coder_tester::compare_streams_wrt_seek(istream& org, istream& cpy, int random_seeks)
{
	const int buf_size = 256;
	char buf_org[buf_size], buf_cpy[buf_size];
	int n,i;

	// determine size of original stream
	org.seekg(streamoff(0), ios::end);
	streamoff size = streamoff(org.tellg());
	org.seekg(streampos(0));

	// perform successive reads, (skipping some chars before each read)
	const int num_seeks = 12; const int to_skip = 10;
	n = size / num_seeks - to_skip; if (n > buf_size) n = buf_size;
	if (n > 0) {
		for (i = 0; i < num_seeks; ++i) {
			org.seekg(streamoff(to_skip), ios::cur); cpy.seekg(streamoff(to_skip), ios::cur);
			if (org.tellg() != cpy.tellg()) error("cmp stream seek - tellg (1)");
			org.read(buf_org, n); cpy.read(buf_cpy, n);
			if (org.gcount() != cpy.gcount()) error("cmp stream seek - gcounts differ (1)");
			if (memcmp(buf_org, buf_cpy, n) != 0) error("cmp stream seek - buffers differ (1)");
		}
	}

	// compare the sizes
	org.seekg(streamoff(0), ios::end); cpy.seekg(streamoff(0), ios::end);
	if (org.tellg() != cpy.tellg()) error("cmp stream seek - tellg (2)");
	if (size == 0) return;

	// check seek to end and seek to beginning
	org.seekg(streamoff(-1), ios::end); cpy.seekg(streamoff(-1), ios::end);
	if (org.get() != cpy.get()) error("cmp stream seek - get (1)");
	org.seekg(streampos(0)); cpy.seekg(streampos(0));
	if (org.get() != cpy.get()) error("cmp stream seek - get (2)");

	// two larger reads in the middle and in the last quarter
	n = size / 4 - 1; if (n > buf_size) n = buf_size;
	if (n > 0) {
		org.seekg(size/2, ios::beg); cpy.seekg(size/2, ios::beg);
		org.read(buf_org, n); cpy.read(buf_cpy, n);
		if (org.gcount() != cpy.gcount()) error("cmp stream seek - gcounts differ (2)");
		if (memcmp(buf_org, buf_cpy, n) != 0) error("cmp stream seek - buffers differ (2)");

		org.seekg(-size/4, ios::end); cpy.seekg(-size/4, ios::end);
		org.read(buf_org, n); cpy.read(buf_cpy, n);
		if (org.gcount() != cpy.gcount()) error("cmp stream seek - gcounts differ (3)");
		if (memcmp(buf_org, buf_cpy, n) != 0) error("cmp stream seek - buffers differ (3)");
		if (org.tellg() != cpy.tellg()) error("cmp stream seek - tellg (3)");
	}

	// try random seeks (if appropriate)
	if (random_seeks > 0 && size > buf_size) {
		random_source rs;
		rs.set_seed(0);
		rs.set_range(0, int(size - buf_size - 1));
		{	
			for (int i = 1; i <= random_seeks; ++i) {
				streampos pos = streamoff(rs());
				org.seekg(pos); cpy.seekg(pos);
				org.read(buf_org, buf_size); cpy.read(buf_cpy, buf_size);
				if (memcmp(buf_org, buf_cpy, buf_size) != 0) error("cmp stream seek - buffers differ (4)");
			}
		}
	}

	// compare the streams from beginning till end
	org.seekg(streampos(0)); cpy.seekg(streampos(0));
	do {
		org.read(buf_org, buf_size); cpy.read(buf_cpy, buf_size);
		n = org.gcount();
		if (org.gcount() != cpy.gcount()) error("cmp stream seek - gcounts differ (5)");
		if (memcmp(buf_org, buf_cpy, n) != 0) error("cmp stream seek - buffers differ (5)");
	} while (n == buf_size);
}

/// module: test encoding_ofstream/decoding_ifstream <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


/// module: test callbacks >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

streamsize cb_underflow_read_from_stream(char* buf, streamsize buf_sz, void* info)
{
	if (info == 0) {
		cout << "underflow: no stream" << endl;
		return false;
	}

	istream& in = *(istream*)info;
	if (!in.good()) return 0;
/*
	streamsize cnt = 0;
	while (cnt < buf_sz) {
		in.get( *(buf++) );
		if (in.eof()) return cnt;
		++cnt;
	}
	return cnt;
*/
	in.read(buf, buf_sz);
	return in.gcount();
}

bool cb_overflow_write_to_stream(const char* buf, streamsize sz, void* info)
{
	if (info == 0) {
		cout << "overflow: no stream" << endl;
		return false;
	}

	ostream& out = *(ostream*)info;
	out.write(buf, sz);
	return out.good();
}

void
coder_tester::test_one_coder_with_callbacks()
{
	if ((TestMask & TestCallbacks) == 0) return;
	if (Verbosity) {
		CoderForCallbacks coder;
		cout << "testing callbacks";
		if (Verbosity >= 2)
			cout << endl << "        with coder " << coder.get_description();
		cout << " ..." << endl;
	}

	generate_random_file(4*1024);

	// encoding
	{
		if (Verbosity >= 4) cout << "      encoding ..." << endl;

		ifstream in(TestFileName, coder_base::binary_in());
		mem_strbuf_with_callbacks in_buf(256);
		in_buf.set_underflow_callback(cb_underflow_read_from_stream, &in);

		ofstream out(EncodedFileName, coder_base::binary_out() | ios::trunc);
		mem_strbuf_with_callbacks out_buf(256);
		out_buf.set_overflow_callback(cb_overflow_write_to_stream, &out);

		CoderForCallbacks coder(&in_buf, &out_buf);
		CurrentCoder = &coder;
		coder.encode();
		if (coder.failed() || !coder.finished()) error("encode");
		CurrentCoder = 0;

		out_buf.pubsync(); // flush the buffer!
	}

	// decoding
	{
		if (Verbosity >= 4) cout << "      decoding ..." << endl;

		ifstream in(EncodedFileName, coder_base::binary_in());
		mem_strbuf_with_callbacks in_buf(256);
		in_buf.set_underflow_callback(cb_underflow_read_from_stream, &in);

		ofstream out(DecodedFileName, coder_base::binary_out() | ios::trunc);
		mem_strbuf_with_callbacks out_buf(256);
		out_buf.set_overflow_callback(cb_overflow_write_to_stream, &out);

		CoderForCallbacks coder(&in_buf, &out_buf);
		CurrentCoder = &coder;
		coder.decode();
		if (coder.failed() || !coder.finished()) error("decode");
		CurrentCoder = 0;

		out_buf.pubsync(); // flush the buffer!
	}

	// compare files
	if (Verbosity >= 4) cout << "      comparing ..." << endl;
	compare_files();
	if (failed()) return;

	if (Verbosity) cout << endl;

	if (DeleteFiles) {
		delete_file(TestFileName); delete_file(EncodedFileName); delete_file(DecodedFileName);
	}
}

/// module: test callbacks <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


/// module: test seek optimizing coders >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

void
coder_tester::test_seek_optimizing_coders()
{
	if ((TestMask & TestSeekCoders) == 0) return;

	if (Verbosity) {
		cout << "testing seek optimizing coders ..." << endl;
	}

	generate_random_file(16 * 1024);

// Adler32Coder (checksummer)
	Adler32Coder chk1;
//	chk1.set_checksums_in_stream_flag(true, 1024);
	CurrentCoder = &chk1;
	test_coder_seek(false, 2000); // is not a crypter and try random seeks

// SHACoder (checksummer)
	SHACoder chk2;
	chk2.set_checksums_in_stream_flag(true, 1024);
	CurrentCoder = &chk2;
	test_coder_seek(false, 2000); // is not a crypter and try random seeks

// ecb_coder
	ECBCoder<> ecb;
	CurrentCoder = &ecb;
	test_coder_seek(true, 2000); // is crypter and try random seeks

// cbc_coder
	CBCCoder<> cbc;
	CurrentCoder = &cbc;
	test_coder_seek(true, 2000); // is crypter and try random seeks

// omac_coder
	OMACCoder<> oc;
	CurrentCoder = &oc;
	test_coder_seek(true, 5000); // is crypter and try random seeks

// block_coder
	BlockCoder<Coder> bc;
	bc.set_block_size(0); // minimum possible
	CurrentCoder = &bc;
	test_coder_seek(false, 100);

// block_coder< block_coder<...> >
	BlockCoder< BlockCoder<Coder> > bbc;
	bbc.get_coder()->set_block_size(bbc.get_block_size()/10);
	CurrentCoder = &bbc;
	test_coder_seek(false, 100);

// coder_pipes
	// all optimize seek
	CoderPipe3< OMACCoder<>, BlockCoder<PPMIICoder>, CBCCoder<> > pc1;
	test_coder_pipe_seek(pc1, pc1.get_coder1(), 0, pc1.get_coder3(), 100);

	// first does not optimize seek
	CoderPipe3< DeflateCoder, BlockCoder<MTFCoder>, CFBCoder<> > pc2;
	test_coder_pipe_seek(pc2, 0, 0, pc2.get_coder3(), 0);

	// second does not optimize seek
	CoderPipe3< ECBCoder<>, PPMIICoder, BlockCoder<CCITTCoder> > pc3;
	test_coder_pipe_seek(pc3, pc3.get_coder1(), 0, 0, 0);

	// third does not optimize seek
	CoderPipe3< MD5SumCoder, BlockCoder<HuffmanCoder>, DeflateCoder > pc4;
	test_coder_pipe_seek(pc4, 0, 0, 0, 0);

	// no coder optimizes seek
	CoderPipe3< DictCoder, DeflateCoder, A0sCoder > pc5;
	test_coder_pipe_seek(pc5, 0, 0, 0, 0);

#ifdef _SIMPLE_CODER_H
// simple_coder
	simple_coder _simple_coder_;
	CurrentCoder = &_simple_coder_;
	test_coder_seek(false, 1000); // is not a crypter and try random seeks
#endif

	CurrentCoder = 0;

	if (Verbosity) cout << endl;

	if (DeleteFiles) {
		delete_file(TestFileName); delete_file(EncodedFileName); delete_file(DecodedFileName);
	}
}

void
coder_tester::test_coder_seek(bool is_cipher, int random_seeks)
{
	if (Verbosity >= 2) {
		cout << "  testing " << CurrentCoder->get_description() << " ..." << endl;
	}

	unsigned long save_test_mask = TestMask; TestMask |= TestStreamIO;
	if (is_cipher)
		test_current_crypt_coder_with_streambufIO(false); // do not reset crypter before encoding!
	else
		test_current_coder_with_streambufIO(false); // do not reset coder before encoding!
	TestMask = save_test_mask;

	if (Verbosity >= 4) cout << "      auto decode seek ..." << endl;
	ifstream                            in1(TestFileName, open_in);
	decoding_ifstream<CryptAutoDecoder> in2(EncodedFileName);
	if (is_cipher) in2.get_coder()->add_key(global_key);
	compare_streams_wrt_seek(in1, in2, random_seeks);

	if (CurrentCoder) CurrentCoder->reset(); // close & release files
}

void
coder_tester::test_coder_pipe_seek(coder_pipe& pc, 
								   crypter_sym_key_base* c1, crypter_sym_key_base* c2, crypter_sym_key_base* c3, 
								   int random_seeks)
{
	CurrentCoder = &pc;

	if (Verbosity >= 2) {
		cout << "  testing " << CurrentCoder->get_description() << " ..." << endl;
	}

	if (Verbosity >= 4) cout << "      encoding test file ..." << endl;
	if (c1) c1->set_key(global_key);
	if (c2) c2->set_key(global_key);
	if (c3) c3->set_key(global_key);
	encode_file(false);

	if (Verbosity >= 4) cout << "      decoding test file ..." << endl;
	pc.reset();
	if (c1) c1->set_key(global_key);
	if (c2) c2->set_key(global_key);
	if (c3) c3->set_key(global_key);
	decode_file(false);

	if (Verbosity >= 4) cout << "      comparing files ..." << endl;
	compare_files();

	if (Verbosity >= 4) cout << "      auto decode seek ..." << endl;
	ifstream                             in1(TestFileName, open_in);
	decoding_ifstream<CryptAutoDecoder> in2(EncodedFileName);
	in2.get_coder()->add_key(global_key); // can't harm
	in2.get_coder()->add_key(global_key); // ...
	in2.get_coder()->add_key(global_key); // ...
	compare_streams_wrt_seek(in1, in2, random_seeks);

	if (CurrentCoder) CurrentCoder->reset(); // close & release files
}

/// module: test seek optimizing coders <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


/// module: test cross platform >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

void
coder_tester::test_crossplatform()
{
	if ((TestMask & TestCrossPlatform) == 0) return;
	
	if (!is_file(crossplatform_org) || !is_file(crossplatform_enc) || !is_file(crossplatform_key)) {
		if (Verbosity)
			cout << "cannot perform cross platform test because files are missing:" << endl
			     << "  " << crossplatform_org 
				 << " / " << crossplatform_enc
				 << " / " << crossplatform_key << endl << endl;
		return;
	}

	if (Verbosity) {
		cout << "performing cross platform test";
		if (Verbosity >= 2)
			cout << endl << "   files: " 
			     << crossplatform_org << " / " << crossplatform_enc << " / " << crossplatform_key;
		cout << " ..." << endl;
	}

	CryptAutoDecoder coder;

	coder_base* coder_sav = CurrentCoder; CurrentCoder = &coder;
	string tf_sav = TestFileName, ef_sav = EncodedFileName, df_sav = DecodedFileName;
	TestFileName = crossplatform_org;

	string tmp1 = tmp_file_name(), tmp2 = tmp_file_name();

	if (Verbosity >= 3) cout << "     step 1: decoding with ppmii ... " << endl;
	EncodedFileName = crossplatform_enc;
	DecodedFileName = tmp1;
	decode_file();
	if (Verbosity >= 4)	cout << "     coder: " << coder.get_description() << endl;
	coder.reset();

	if (Verbosity >= 3) cout << "     step 2: deciphering ... " << endl;
	EncodedFileName = DecodedFileName; // use output from previous step
	DecodedFileName = tmp2;
	coder.add_keys_in_file(crossplatform_key);
	decode_file(false); // no reset -> keep the key stream
	if (Verbosity >= 4)	cout << "     coder: " << coder.get_description() << endl;
	coder.reset();

	if (Verbosity >= 3) cout << "     step 3: decoding ... " << endl;
	EncodedFileName = DecodedFileName; // use output from previous step
	DecodedFileName = df_sav;
	decode_file();
	if (Verbosity >= 4)	cout << "     coder: " << coder.get_description() << endl;
	coder.reset();
	compare_files();

	CurrentCoder = coder_sav;
	TestFileName = tf_sav; EncodedFileName = ef_sav;

	if (DeleteFiles) {
		delete_file(tmp1); delete_file(tmp2); delete_file(DecodedFileName);
	}

	if (Verbosity) cout << endl;
}

bool 
coder_tester::generate_files_for_crossplatform_test(string input_file, string output_file, string key_file)
{
	// step 1: encode input_file with BigPipe
	set_coders_to_test(CoderFactory.get_standalone_coders(), true);
	set_test_mask(TestBigPipe | TestSmallFile | TestStreamIO);
	keep_files();
	string tmp1 = tmp_file_name(), tmp2 = tmp_file_name();
	run(input_file, tmp1, tmp2, false);
	delete_file(tmp2);

	list<crypter_sym_key_base*> crypters = construct_crypt_coders();

	// prepare key_stream
	ofstream key_stream; key_stream.open(key_file, open_out);
	CryptKey key = global_key;

	// set up the crypter pipe
	GenericCoderPipe cipher_pipe(crypters.size() + 1);
	crypter_sym_key_base* crypter;
	int i = 0;
	forall(crypter, crypters) {
		// set key and write it to the key file
		crypter->set_key(key);
		key_stream << key;

		// insert the crypter into the pipe
		cipher_pipe.set_coder(++i, crypter, true);
	}
	key_stream.close();

	// add a block_coder as final coder to the pipe
	BlockCoder<Coder>* p_blk_coder = new BlockCoder<Coder>;
	cipher_pipe.set_coder(++i, p_blk_coder, true);

	// step 2: encipher the output of BigPipe with the crypter pipe
	CurrentCoder = &cipher_pipe;
	cipher_pipe.set_src_file(tmp1); cipher_pipe.set_tgt_file(tmp2);
	cipher_pipe.encode();
	Okay &= !cipher_pipe.failed();
	if (cipher_pipe.failed()) error("enciphering failed!");
	CurrentCoder = 0;

	// step 3: encode with PPMIICoder and low memory (debug mode)
	PPMIICoder ppmii;
	CurrentCoder = &ppmii;
	ppmii.set_src_file(tmp2); ppmii.set_tgt_file(output_file);
	ppmii.set_model_memory_bound(0-1);
	ppmii.encode();
	Okay &= !ppmii.failed();
	if (ppmii.failed()) error("encoding with PPMII failed!");
	CurrentCoder = 0;

	delete_file(tmp1); delete_file(tmp2);
	return Okay;
}

/// module: test cross platform <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


/// module: test cryptography >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

void
coder_tester::test_cryptography()
{
	if ((TestMask & TestCryptography) == 0) return;
	
	if (Verbosity) {
		cout << "testing cryptography ..." << endl;
	}

	test_block_ciphers();
	test_crypt_coders();
	test_checksummers();

	if (Verbosity) cout << endl;
}

void
coder_tester::test_block_ciphers()
{
	if (Verbosity >= 2) {
		cout << "  testing block crypters ..." << endl;
	}

	coder_base* coder_sav = CurrentCoder; CurrentCoder = 0;

	int t = 0;
	byte i;

// Blowfish
	if (Verbosity >= 3)
		cout << "     Blowfish ..." << endl;

	if (Verbosity >= 4)
		cout << "       variable key test ..." << endl;
	for (i = 0; i < BF_NUM_VARIABLE_KEY_TESTS; ++i,++t) {
		test_blowfish_one_block(bf_variable_key[i], BF_VARIABE_KEY_LEN, 
								bf_plaintext_l[t],  bf_plaintext_r[t],
								bf_ciphertext_l[t], bf_ciphertext_r[t]);
	}
	
	if (Verbosity >= 4)
		cout << "       key set test ..." << endl;
	for (i = 1; i <= BF_NUM_SET_KEY_TESTS; ++i,++t) {
		if (i < Blowfish::MinKeyBytes) continue;
		test_blowfish_one_block(bf_set_key, i, 
								bf_plaintext_set_l, bf_plaintext_set_r,
								bf_ciphertext_l[t], bf_ciphertext_r[t]);
	}
	
// Twofish
	if (Verbosity >= 3)
		cout << "     Twofish ..." << endl;

	for (t = 0; t < 3; ++t) {
		int key_bits = 128 + 64 * t;

		if (Verbosity >= 4)
			cout << "       " << key_bits << "-bit key test ..." << endl;
		test_twofish_run(key_bits / 8, tf_final_ciphertext[t]);
	}
	
// Rijndael
	if (Verbosity >= 3)
		cout << "     Rijndael ..." << endl;

	if (Verbosity >= 4)	cout << "       128-bit key test ..." << endl;
	test_rijndael_one_block(rd_128_key, 128 / 8, rd_pt, rd_128_ct);	

	if (Verbosity >= 4)	cout << "       196-bit key test ..." << endl;
	test_rijndael_one_block(rd_196_key, 196 / 8, rd_pt, rd_196_ct);	

	if (Verbosity >= 4)	cout << "       256-bit key test ..." << endl;
	test_rijndael_one_block(rd_256_key, 256 / 8, rd_pt, rd_256_ct);	

	if (Verbosity >= 4)	cout << "       omac test ..." << endl;
	test_omac_rijndael();

	CurrentCoder = coder_sav;
}

void 
coder_tester::test_blowfish_one_block(byte* key, byte num_key_bytes,
									  uint32 pl, uint32 pr, 
									  uint32 cl, uint32 cr)
{
	Blowfish bf;
	bf.set_key(CryptKey(key, num_key_bytes));
	byte src[8], tgt[8];
	src[0] = byte((pl & 0xFF000000) >> 24);
	src[1] = byte((pl & 0x00FF0000) >> 16);
	src[2] = byte((pl & 0x0000FF00) >> 8);
	src[3] = byte((pl & 0x000000FF));
	src[4] = byte((pr & 0xFF000000) >> 24);
	src[5] = byte((pr & 0x00FF0000) >> 16);
	src[6] = byte((pr & 0x0000FF00) >> 8);
	src[7] = byte((pr & 0x000000FF));
	bf.encipher(src, tgt);
	uint32 el = (uint32(tgt[0]) << 24)
					 | (uint32(tgt[1]) << 16)
					 | (uint32(tgt[2]) <<  8)
					 | (uint32(tgt[3]));
	uint32 er = (uint32(tgt[4]) << 24)
					 | (uint32(tgt[5]) << 16)
					 | (uint32(tgt[6]) <<  8)
					 | (uint32(tgt[7]));
	bf.decipher(tgt);
	if (el != cl || er != cr) error("Blowfish - wrong ciphertext");
	if (memcmp(src, tgt, 8)) error("Blowfish - decipher != plaintext");
}

void 
coder_tester::test_twofish_run(int num_key_bytes, byte* final_ct)
{
	Twofish tf;
	Twofish::byte key[Twofish::MaxKeyBytes];
	Twofish::byte pt[Twofish::BlkSz], enc[Twofish::BlkSz], dec[Twofish::BlkSz];

	memset(&key, 0, Twofish::MaxKeyBytes);
	memset(&pt,  0, Twofish::BlkSz);

	for (int i = 1; i <= 49; ++i) {
		tf.set_key(CryptKey(key, num_key_bytes));
		tf.encipher(pt, enc);
		tf.decipher(enc, dec);
		if (memcmp(dec, pt, Twofish::BlkSz)) error("Twofish - decipher != plaintext");
		memmove(key + Twofish::BlkSz, key, Twofish::MaxKeyBytes - Twofish::BlkSz);
		memcpy(key, pt,  Twofish::BlkSz);
		memcpy(pt,  enc, Twofish::BlkSz);
	}

	if (memcmp(enc, final_ct, Twofish::BlkSz)) error("Twofish - wrong final crypter");
}

void 
coder_tester::test_rijndael_one_block(const byte* key_bytes, byte num_key_bytes,
									  const byte* pt, const byte* ct)
{
	Rijndael rd;
	Rijndael::byte enc[Rijndael::BlkSz], dec[Rijndael::BlkSz];
	CryptKey key(key_bytes, num_key_bytes);

	rd.set_enc_key(key);
	rd.encipher(pt, enc);
	rd.set_dec_key(key);
	rd.decipher(enc, dec);

	if (memcmp(dec, pt, Rijndael::BlkSz)) error("Rijndael - decipher != plaintext");
	if (memcmp(enc, ct, Rijndael::BlkSz)) error("Rijndael - wrong final crypter");
}

void 
coder_tester::test_omac_rijndael()
{
	omac_mode om;
	om.set_block_cipher(new Rijndael, true);

	for (int i = 0; i < NUM_OMAC_TV; ++i) {
		omac_tv& tv = omac_tvs[i];

		memory_streambuf mbuf((char*) tv.msg, tv.msg_len, false, false);
		om.reset();
		om.set_MAC_in_stream_flag(false);
		om.set_key(CryptKey(tv.key, tv.key_len));
		om.set_src_stream(&mbuf);
		const byte* mac = om.check();
		if (memcmp(mac, tv.mac, Rijndael::BlkSz)) error("omac_mode<Rijndael>");
	}
}

void
coder_tester::test_crypt_coders()
{
	if (Verbosity >= 2) {
		cout << "  testing crypter coders ..." << endl;
	}

	generate_random_file(4 * 1024);

	if (Debug) {
		ECBCoder<NoBlockCipher> coder;
		CurrentCoder = &coder;
		test_current_crypt_coder_with_streambufIO();
	}
	else {
		list<crypter_sym_key_base*> crypters = construct_crypt_coders();
		crypter_sym_key_base* crypter;
		forall(crypter, crypters) {
			CurrentCoder = crypter;
			test_current_crypt_coder_with_streambufIO();
			delete CurrentCoder; CurrentCoder = 0;
		}

		list<block_cipher_mode_base*> bc_modes = construct_block_cipher_modes();
		block_cipher_mode_base* bc_mode;
		forall(bc_mode, bc_modes) {
			CurrentCoder = bc_mode;
			test_current_block_cipher_mode_with_memory_chunks();
			delete CurrentCoder; CurrentCoder = 0;
		}
	}

	if (DeleteFiles) { delete_file(EncodedFileName); delete_file(DecodedFileName); }
}

list<block_cipher_mode_base*>
coder_tester::construct_block_cipher_modes()
{
	list<block_cipher_mode_base*> result;

	list<block_cipher_base*> bcs;
	list<block_cipher_mode_base*> bc_modes = CoderFactory.get_block_cipher_modes();
	block_cipher_mode_base* bc_mode;
	forall(bc_mode, bc_modes) {
		if (bcs.empty()) {
			bcs = BlockCipherFactory.get_all_ciphers();
			bcs.permute();
		}
		bc_mode->set_block_cipher(bcs.pop(), true);
		result.append(bc_mode);
	}
	while (! bcs.empty()) delete bcs.pop(); // free remaining crypters

	return result;
}

list<crypter_sym_key_base*>
coder_tester::construct_crypt_coders()
{
	list<crypter_sym_key_base*> result;

	// so far we have only block cipher modes ...
	list<block_cipher_mode_base*> bc_modes = construct_block_cipher_modes();
	block_cipher_mode_base* bc_mode;
	forall(bc_mode, bc_modes) {
		result.append(bc_mode);
	}

	return result;
}

void 
coder_tester::test_current_crypt_coder_with_streambufIO(bool reset_crypter)
// NOTE: crypters must be treated differently from other coders because they need a key
{
	crypter_sym_key_base* crypter = static_cast<crypter_sym_key_base*>(CurrentCoder);
	CryptKey key = global_key;

	if (Verbosity >= 3) cout << "    testing " << crypter->get_description() << " with streambuf IO ..." << endl;

	// test encoding and decoding with coder
	if (Verbosity >= 4) cout << "      encoding ..." << endl;
	if (reset_crypter) crypter->reset();
	crypter->set_key(key);
	encode_file(false); if (failed()) return;

	if (Verbosity >= 4) cout << "      decoding ..." << endl;
	crypter->reset();
	crypter->set_key(key);
	decode_file(false); if (failed()) return;

	if (Verbosity >= 4) cout << "      comparing ..." << endl;
	compare_files(); if (failed()) return;

	// test decoding with CryptAutoDecoder
	if (Verbosity >= 4) cout << "      auto decoding ..." << endl;
	CryptAutoDecoder auto_decoder;
	auto_decoder.add_key(key);
	coder_base* sav_coder = CurrentCoder;
	CurrentCoder = &auto_decoder;
	decode_file(false);
	if (!failed()) {
		if (Verbosity >= 4) cout << "      comparing (auto) ..." << endl;
		compare_files();
	}
	CurrentCoder = sav_coder;
}

void 
coder_tester::test_current_block_cipher_mode_with_memory_chunks()
// NOTE: crypters must be treated differently from other coders because they need a key
{
	block_cipher_mode_base* crypter = static_cast<block_cipher_mode_base*>(CurrentCoder);
	CryptKey key = global_key;

	if (Verbosity >= 3) cout << "    testing " << crypter->get_description() << " with memory chunks ..." << endl;

	// prepare memory buffers and read TestFile into memory
	crypter->reset(); crypter->set_key(key);
	uint32 orig_sz    = uint32( size_of_file(TestFileName) );
	uint32 max_enc_sz = crypter->calculate_length_of_encoded_data(orig_sz);
	char* buf_orig = new char[orig_sz];
	char* buf_enc  = new char[max_enc_sz];
	char* buf_dec  = new char[orig_sz];
	{
		ifstream in(TestFileName, open_in);
		in.read(buf_orig, orig_sz);
	}

	uint32 start_sz = orig_sz - 20; if (start_sz < 0) start_sz = 0;
	for (uint32 sz = start_sz; sz <= orig_sz; ++ sz) {
		if (Verbosity >= 4) cout << "      input size = " << sz << flush;

		// encode memory chunk
		CurrentCoder->reset();
		uint32 calc_enc_sz = crypter->calculate_length_of_encoded_data(sz);
		if (Verbosity >= 4) cout << " / calc. output size = " << calc_enc_sz << flush;
		uint32 enc_sz = CurrentCoder->encode_memory_chunk(buf_orig, sz, buf_enc, max_enc_sz);
		if (Verbosity >= 4) cout << " / output size = " << enc_sz << endl;
		if (CurrentCoder->failed()) {
			error("memory encode");
			return;
		}
		if (enc_sz != calc_enc_sz) {
			error("memory encode (wrong size)");
			return;
		}

		// decode memory chunk
		CurrentCoder->reset();
		uint32 dec_sz = CurrentCoder->decode_memory_chunk(buf_enc, enc_sz, buf_dec, orig_sz);
		if (CurrentCoder->failed()) {
			error("memory decode");
			return;
		}

		// compare memory chunks
		if (sz != dec_sz || memcmp(buf_orig, buf_dec, sz) != 0) {
			error("memory compare");
			return;
		}
	}

	// free memory
	delete[] buf_orig;
	delete[] buf_enc;
	delete[] buf_dec;
}

void
coder_tester::test_checksummers()
{
	if (Verbosity >= 2) {
		cout << "  testing checksummers ..." << endl;
	}

	coder_base* coder_sav = CurrentCoder; CurrentCoder = 0;

	const string str1 = "abc";
	const string str2 = "abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq";

// MD5Sum
	if (Verbosity >= 3)
		cout << "    MD5Sum ..." << endl;
	MD5SumCoder md5sum;
	if (md5sum.check(str1) != "900150983cd24fb0d6963f7d28e17f72") error("md5sum - string 1");
	md5sum.reset();
	if (md5sum.check(str2) != "8215ef0796a20bcaaae116d3876c664a") error("md5sum - string 2");

// SHA
	if (Verbosity >= 3)
		cout << "    SHA ..." << endl;
	SHACoder sha;
	if (sha.check(str1) != "a9993e364706816aba3e25717850c26c9cd0d89d") error("SHA - string 1");
	sha.reset();
	if (sha.check(str2) != "84983e441c3bd26ebaae4aa1f95129e5e54670f1") error("SHA - string 2");

	CurrentCoder = coder_sav;
}

/// module: test cryptography <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


/// global functions ************************************************************************

static void usage()
{
	cout << "usage: test_all_coders [options] [test_file encoded_file decoded_file]" << endl;
	cout << " options:" << endl;
	cout << "   -v: include verbosity level" << endl;
	cout << "   -d: enable debugging" << endl;
	cout << "   -D sz: set file size for debugging" << endl;
	cout << "   -k: keep temporary files" << endl;
	cout << "   -g: generate files for cross-platform test" << endl;
	cout << "   -t mask: set test mask, which selects the test to perform" << endl;
	cout << "   -T mask: perform all tests except for those matching mask" << endl;
	cout << "   -e coder: exclude the specified coder from tests" << endl;
	cout << "   -o coder: use only the specified coder for tests" << endl;
	cout << " files (if not specified temporary files are used):" << endl;
	cout << "   test_file:    name of input file for the tests" << endl;
	cout << "   encoded_file: name of encoded file for the tests" << endl;
	cout << "   decoded_file: name of decoded file for the tests" << endl;
	cout << " test mask:" << endl;
	cout << "   TestIOStreamLib: \t" << string("0x%04x", coder_tester::TestIOStreamLib) << endl;
	cout << "   TestStreamIO: \t" << string("0x%04x", coder_tester::TestStreamIO) << endl;
	cout << "   TestAdapterIO: \t" << string("0x%04x", coder_tester::TestAdapterIO) << endl;
	cout << "   TestMemoryChunks: \t" << string("0x%04x", coder_tester::TestMemoryChunks) << endl;
	cout << "   TestEncDecStreams: \t" << string("0x%04x", coder_tester::TestEncDecStreams) << endl;
	cout << "   TestSeekCoders: \t" << string("0x%04x", coder_tester::TestSeekCoders) << endl;
	cout << "   TestCrossPlatform: \t" << string("0x%04x", coder_tester::TestCrossPlatform) << endl;
	cout << "   TestCryptography: \t" << string("0x%04x", coder_tester::TestCryptography) << endl;
	cout << "   TestCallbacks: \t" << string("0x%04x", coder_tester::TestCallbacks) << endl;
	cout << "   TestEmptyFile: \t" << string("0x%04x", coder_tester::TestEmptyFile) << endl;
	cout << "   TestSmallFile: \t" << string("0x%04x", coder_tester::TestSmallFile) << endl;
	cout << "   TestRandomFile: \t" << string("0x%04x", coder_tester::TestRandomFile) << endl;
	cout << "   TestStandalone: \t" << string("0x%04x", coder_tester::TestStandalone) << endl;
	cout << "   TestSmallPipes: \t" << string("0x%04x", coder_tester::TestSmallPipes) << endl;
	cout << "   TestBigPipe: \t" << string("0x%04x", coder_tester::TestBigPipe) << endl;
}

static int generate_files_for_crossplatform_test()
{
	coder_tester test(4);
	test.generate_files_for_crossplatform_test(crossplatform_org, crossplatform_enc, crossplatform_key);

	if (test.failed()) {
		cerr << "Generation of cross-platform test files failed!";
		return 1;
	}
	return 0;
}

int main(int argc, char* argv[]) 
{
	coder_tester test;

	while (argc > 1 && argv[1][0] == '-') {
		switch (argv[1][1]) {
		case 'v': test.inc_verbosity(); break;
		case 'd': test.set_debug_mode(); break;
		case 'D': {
					--argc; ++argv; if (argc <= 1) break;
					char* stop; int sz = (int) strtol(argv[1], &stop, 0);
					if (*stop == 0) test.set_debug_size(sz);
				  }
				  break;
		case 'k': test.keep_files(); break;
		case 'g': return generate_files_for_crossplatform_test();
		case 't': {
					--argc; ++argv; if (argc <= 1) break;
					char* stop; unsigned long mask = strtoul(argv[1], &stop, 0);
					if (*stop == 0) test.set_test_mask(mask);
				  }
				  break;
		case 'T': {
					--argc; ++argv; if (argc <= 1) break;
					char* stop; unsigned long mask = strtoul(argv[1], &stop, 0);
					if (*stop == 0) test.remove_tests(mask);
				  }
				  break;
		case 'e': {
					--argc; ++argv; if (argc <= 1) break;
					test.remove_coder_to_test( string(argv[1]) );
				  }
				  break;
		case 'o': {
					--argc; ++argv; if (argc <= 1) break;
					test.remove_all_coders_to_test_except( string(argv[1]) );
				  }
				  break;
		default:  usage(); return 1;
		}
		--argc; ++argv;
	}

	bool ok;
	if (argc < 4) ok = test.run();
	else          ok = test.run(argv[1], argv[2], argv[3]);

	if (! ok) {
		cout << "There have been FAILURES!!!" << endl;
		return 1;
	}

	cout << "all tests completed successfully!" << endl;

	return 0;
}
