/*******************************************************************************
+
+  LEDA 7.0
+
+
+  string_matching.cpp
+
+
+  Copyright (c) 1995-2023
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/

#include<LEDA/string/string_matching.h>
#include<string.h>
#include<cstdlib>

LEDA_BEGIN_NAMESPACE

int minShift[XSIZE];
int freq[ASIZE];

int getTransition(char *x, int m, int p, array<List> &L,
                  char c) {
   List cell;
 
   if (p < m - 1 && x[p + 1] == c)
      return(p + 1);
   else if (p > -1) {
      cell = L[p];
      while (cell != NULL)
         if (x[cell->element] == c)
            return(cell->element);
         else
            cell = cell->next;
      return(-1);
   }
   else
      return(-1);
}
void setTransition(int p, int q, array<List> &L) {
   List cell;
 
   cell = (List)malloc(sizeof(struct _cell));
   if (cell == NULL)
      LEDA_EXCEPTION(1,"SIMON/setTransition");
   cell->element = q;
   cell->next = L[p];
   L[p] = cell;
}




  
/* returns a new data structure for
   a graph with v vertices and e edges */
Graph newGraph(int v, int e) {
   Graph g;

   g = (Graph)calloc(1, sizeof(struct _graph));
   if (g == NULL)
      LEDA_EXCEPTION(1,"newGraph");
   g->vertexNumber  = v;
   g->edgeNumber    = e;
   g->initial       = 0;
   g->vertexCounter = 1;
   return(g);
}


/* returns a new data structure for
   a automaton with v vertices and e edges */
Graph newAutomaton(int v, int e) {
   Graph aut;

   aut = newGraph(v, e);
   aut->target = (int *)calloc(e, sizeof(int));
   if (aut->target == NULL)
      LEDA_EXCEPTION(1,"newAutomaton");
   aut->terminal = (int *)calloc(v, sizeof(int));
   if (aut->terminal == NULL)
      LEDA_EXCEPTION(1,"newAutomaton");
   return(aut);
}


/* returns a new data structure for
   a suffix automaton with v vertices and e edges */
Graph newSuffixAutomaton(int v, int e) {
   Graph aut;

   aut = newAutomaton(v, e);
   memset(aut->target, UNDEFINED, e*sizeof(int));
   aut->suffixLink = (int *)calloc(v, sizeof(int));
   if (aut->suffixLink == NULL)
      LEDA_EXCEPTION(1,"newSuffixAutomaton");
   aut->length = (int *)calloc(v, sizeof(int));
   if (aut->length == NULL)
      LEDA_EXCEPTION(1,"newSuffixAutomaton");
   aut->position = (int *)calloc(v, sizeof(int));
   if (aut->position == NULL)
      LEDA_EXCEPTION(1,"newSuffixAutomaton");
   aut->shift = (int *)calloc(e, sizeof(int));
   if (aut->shift == NULL)
      LEDA_EXCEPTION(1,"newSuffixAutomaton");
   return(aut);
}
/* returns a new data structure for
   a trie with v vertices and e edges */
Graph newTrie(int v, int e) {
   Graph aut;
 
   aut = newAutomaton(v, e);
   memset(aut->target, UNDEFINED, e*sizeof(int));
   aut->suffixLink = (int *)calloc(v, sizeof(int));
   if (aut->suffixLink == NULL)
      LEDA_EXCEPTION(1,"newTrie");
   aut->length = (int *)calloc(v, sizeof(int));
   if (aut->length == NULL)
      LEDA_EXCEPTION(1,"newTrie");
   aut->position = (int *)calloc(v, sizeof(int));
   if (aut->position == NULL)
      LEDA_EXCEPTION(1,"newTrie");
   aut->shift = (int *)calloc(e, sizeof(int));
   if (aut->shift == NULL)
      LEDA_EXCEPTION(1,"newTrie");
   return(aut);
}


/* returns a new vertex for graph g */
int newVertex(Graph g) {
   if (g != NULL && g->vertexCounter <= g->vertexNumber)
      return(g->vertexCounter++);
   LEDA_EXCEPTION(1,"newVertex");
   return 0;
}


/* returns the initial vertex of graph g */
int getInitial(Graph g) {
   if (g != NULL)
      return(g->initial);
   LEDA_EXCEPTION(1,"getInitial");
   return 0;
}


/* returns true if vertex v is terminal in graph g */
boolean isTerminal(Graph g, int v) {
   if (g != NULL && g->terminal != NULL &&
       v < g->vertexNumber)
      return(g->terminal[v]);
   LEDA_EXCEPTION(1,"isTerminal");
   return 0;
}


/* set vertex v to be terminal in graph g */
void setTerminal(Graph g, int v) {
   if (g != NULL && g->terminal != NULL &&
       v < g->vertexNumber)
      g->terminal[v] = 1;
   else
      LEDA_EXCEPTION(1,"isTerminal");
}


/* returns the target of edge from vertex v
   labelled by character c in graph g */
int getTarget(Graph g, int v, unsigned char c) {
   if (g != NULL && g->target != NULL &&
       v < g->vertexNumber && v*c < g->edgeNumber)
      return(g->target[v*(g->edgeNumber/g->vertexNumber) +
                       c]);
   LEDA_EXCEPTION(1,"getTarget");
   return 0;
}


/* add the edge from vertex v to vertex t
   labelled by character c in graph g */
void setTarget(Graph g, int v, unsigned char c, int t) {
   if (g != NULL && g->target != NULL &&
       v < g->vertexNumber &&
       v*c <= g->edgeNumber && t < g->vertexNumber)
      g->target[v*(g->edgeNumber/g->vertexNumber) + c] = t;
   else
      LEDA_EXCEPTION(1,"setTarget");
}


/* returns the suffix link of vertex v in graph g */
int getSuffixLink(Graph g, int v) {
   if (g != NULL && g->suffixLink != NULL &&
       v < g->vertexNumber)
      return(g->suffixLink[v]);
   LEDA_EXCEPTION(1,"getSuffixLink");
   return 0;
}


/* set the suffix link of vertex v
   to vertex s in graph g */
void setSuffixLink(Graph g, int v, int s) {
   if (g != NULL && g->suffixLink != NULL &&
       v < g->vertexNumber && s < g->vertexNumber)
      g->suffixLink[v] = s;
   else
      LEDA_EXCEPTION(1,"setSuffixLink");
}


/* returns the length of vertex v in graph g */
int getLength(Graph g, int v) {
   if (g != NULL && g->length != NULL &&
       v < g->vertexNumber)
      return(g->length[v]);
   LEDA_EXCEPTION(1,"getLength");
   return 0;
}


/* set the length of vertex v to integer ell in graph g */
void setLength(Graph g, int v, int ell) {
   if (g != NULL && g->length != NULL &&
       v < g->vertexNumber)
      g->length[v] = ell;
   else
      LEDA_EXCEPTION(1,"setLength");
}


/* returns the position of vertex v in graph g */
int getPosition(Graph g, int v) {
   if (g != NULL && g->position != NULL &&
       v < g->vertexNumber)
      return(g->position[v]);
   LEDA_EXCEPTION(1,"getPosition");
   return 0;
}


/* set the length of vertex v to integer ell in graph g */
void setPosition(Graph g, int v, int p) {
   if (g != NULL && g->position != NULL &&
       v < g->vertexNumber)
      g->position[v] = p;
   else
      LEDA_EXCEPTION(1,"setPosition");
}


/* returns the shift of the edge from vertex v
   labelled by character c in graph g */
int getShift(Graph g, int v, unsigned char c) {
   if (g != NULL && g->shift != NULL &&
       v < g->vertexNumber && v*c < g->edgeNumber)
      return(g->shift[v*(g->edgeNumber/g->vertexNumber) +
             c]);
   LEDA_EXCEPTION(1,"getShift");
   return 0;
}


/* set the shift of the edge from vertex v
   labelled by character c to integer s in graph g */
void setShift(Graph g, int v, unsigned char c, int s) {
   if (g != NULL && g->shift != NULL &&
       v < g->vertexNumber && v*c <= g->edgeNumber)
      g->shift[v*(g->edgeNumber/g->vertexNumber) + c] = s;
   else
      LEDA_EXCEPTION(1,"setShift");
}


/* copies all the characteristics of vertex source
   to vertex target in graph g */
void copyVertex(Graph g, int target, int source) {
   if (g != NULL && target < g->vertexNumber &&
       source < g->vertexNumber) {
      if (g->target != NULL)
         memcpy(g->target +
                target*(g->edgeNumber/g->vertexNumber),
                g->target +
                source*(g->edgeNumber/g->vertexNumber),
                (g->edgeNumber/g->vertexNumber)*
                sizeof(int));
      if (g->shift != NULL)
         memcpy(g->shift +
                target*(g->edgeNumber/g->vertexNumber),
                g->shift +
                source*(g->edgeNumber/g->vertexNumber),
                (g->edgeNumber/g->vertexNumber)* sizeof(int));
      if (g->terminal != NULL)
         g->terminal[target] = g->terminal[source];
      if (g->suffixLink != NULL)
         g->suffixLink[target] = g->suffixLink[source];
      if (g->length != NULL)
         g->length[target] = g->length[source];
      if (g->position != NULL)
         g->position[target] = g->position[source];
   }
   else
      LEDA_EXCEPTION(1,"copyVertex");
}


void buildSuffixAutomaton(char *x, int m, Graph aut) {
   int i, art, init, last, p, q, r;
   char c;
  
   init = getInitial(aut);
   art = newVertex(aut);
   setSuffixLink(aut, init, art);
   last = init;
   for (i = 0; i < m; ++i) {
      c = x[i];
      p = last;
      q = newVertex(aut);
      setLength(aut, q, getLength(aut, p) + 1);
      setPosition(aut, q, getPosition(aut, p) + 1);
      while (p != init &&
             getTarget(aut, p, c) == UNDEFINED) {
         setTarget(aut, p, c, q);
         setShift(aut, p, c, getPosition(aut, q) -
                             getPosition(aut, p) - 1);
         p = getSuffixLink(aut, p);
      }
      if (getTarget(aut, p, c) == UNDEFINED) {
         setTarget(aut, init, c, q);
         setShift(aut, init, c,
                  getPosition(aut, q) -
                  getPosition(aut, init) - 1);
         setSuffixLink(aut, q, init);
      }
      else
         if (getLength(aut, p) + 1 ==
             getLength(aut, getTarget(aut, p, c)))
            setSuffixLink(aut, q, getTarget(aut, p, c));
         else {
            r = newVertex(aut);
            copyVertex(aut, r, getTarget(aut, p, c));
            setLength(aut, r, getLength(aut, p) + 1);
            setSuffixLink(aut, getTarget(aut, p, c), r);
            setSuffixLink(aut, q, r);
            while (p != art &&
                   getLength(aut, getTarget(aut, p, c)) >=
                   getLength(aut, r)) {
               setShift(aut, p, c,
                        getPosition(aut,
                                    getTarget(aut, p, c)) -
                        getPosition(aut, p) - 1);
               setTarget(aut, p, c, r);
               p = getSuffixLink(aut, p);
            }
         }
      last = q;
   }
   setTerminal(aut, last);
   while (last != init) {
      last = getSuffixLink(aut, last);
      setTerminal(aut, last);
   }
}


int getTransition(char *x, int p, array<List> &L, char c) {
   List cell;

   if (p > 0 && x[p - 1] == c)
      return(p - 1);
   else {
      cell = L[p];
      while (cell != NULL)
         if (x[cell->element] == c)
            return(cell->element);
         else
            cell = cell->next;
      return(UNDEFINED);
   }
}
/* Maximal Shift pattern comparison function. */
int maxShiftPcmp(const void *pat1, const void *pat2) {
   int dsh;
   pattern* patp1 = (pattern*)pat1;
   pattern* patp2 = (pattern*)pat2;
   
   dsh = minShift[patp2->loc] - minShift[patp1->loc];
   return(dsh ? dsh : (patp2->loc - patp1->loc));
}



void orderPattern(char *x, int m, int (*pcmp)(const void*, const void*), pattern *pat) {
   int i;

   for (i = 0; i <= m; ++i) {
      pat[i].loc = i;
      pat[i].c = x[i];
   }
   pattern* patptr = &pat[0];
   std::qsort(patptr, m, sizeof(pattern), pcmp);
}


/* Optimal Mismatch pattern comparison function. */
int optimalPcmp(const void *pat1, const void *pat2) {
   //float fx;
   pattern* patp1 = (pattern*)pat1;
   pattern* patp2 = (pattern*)pat2;
   int fx = freq[int(patp1->c)] - freq[int(patp2->c)];
   return(fx ? (fx > 0 ? 1 : -1) :
               (patp2->loc - patp1->loc));
}

void string_matching::readPattern() {
	cout << "Please enter the Pattern for string matching: " << endl;
	cin >> x;
	m = x.length();
}

void string_matching::readString() {
	cout << "Please enter the String for string matching: " << endl;
	cin >> y;
	n = y.length();
}

void string_matching::reportResult() {
	int i;
	forall(i, result)
		cout << i << " ";
	cout << endl;
}

list<int> string_matching::getResult() const {
	return result;
}

void string_matching::setPattern(const string& str)
{
	x = str;
	m = x.length();
}
void string_matching::setString(const string& str)
{
	y = str;
	n = y.length();
}
string string_matching::getPattern()
{
	return x;
}
string string_matching::getString()
{
	return y;
}
void string_matching::setStrings(const string& str1, const string& str2)
{
	x = str1;
	m = x.length();
	y = str2;
	n = y.length();
}

void string_matching::setAlgorithm(alg A) {
	current = A;
}

void string_matching::setAlgorithm(int A) {
	if (A < 1 || A > 19)
		LEDA_EXCEPTION(1, "distance::setAlgorithm: index out of range!");
	current = alg(A);
}

list<int> string_matching::getAllOccurences() {
	switch(current) {
		case 1: karp_rabin();
			break;
		case 2: morris_pratt();
			break;
		case 3: knuth_morris_pratt();
			break;
		case 4: galil_seiferas();
			break;
		case 5: boyer_moore();
			break;
		case 6: turbo_boyer_moore();
			break;
		case 7: tuned_boyer_moore();
			break;
		case 8: zhu_takaoka();
			break;
		case 9: berry_ravindran();
			break;
		case 10: smith();
			break;
		case 11: raita();
			break;
		case 12: horspool();
			break;
		case 13: brute_force();
			break;
		case 14: shift_or();
			break;
		case 15: reverse_colussi();
			break;
		case 16: quick_search();
			break;
		case 17: reverse_factor();
			break;
		case 18: turbo_reverse_factor();
			break;
		case 19: optimal_mismatch();
			break;
		break;
	}	
	return result;
}

int string_matching::getFirstOccurence() {
	mode = 0;
	next_occurence = -1;
	switch(current) {
		case 1: karp_rabin();
			break;
		case 2: morris_pratt();
			break;
		case 3: knuth_morris_pratt();
			break;
		case 4: galil_seiferas();
			break;
		case 5: boyer_moore();
			break;
		case 6: turbo_boyer_moore();
			break;
		case 7: tuned_boyer_moore();
			break;
		case 8: zhu_takaoka();
			break;
		case 9: berry_ravindran();
			break;
		case 10: smith();
			break;
		case 11: raita();
			break;
		case 12: horspool();
			break;
		case 13: brute_force();
			break;
		case 14: shift_or();
			break;
		case 15: reverse_colussi();
			break;
		case 16: quick_search();
			break;
		case 17: reverse_factor();
			break;
		case 18: turbo_reverse_factor();
			break;
		case 19: optimal_mismatch();
			break;
		break;
	}	
	mode = -1;
	return next_occurence;
}

int string_matching::getNextOccurence(int i) {
	mode = i;
	next_occurence = -1;
	switch(current) {
		case 1: karp_rabin();
			break;
		case 2: morris_pratt();
			break;
		case 3: knuth_morris_pratt();
			break;
		case 4: galil_seiferas();
			break;
		case 5: boyer_moore();
			break;
		case 6: turbo_boyer_moore();
			break;
		case 7: tuned_boyer_moore();
			break;
		case 8: zhu_takaoka();
			break;
		case 9: berry_ravindran();
			break;
		case 10: smith();
			break;
		case 11: raita();
			break;
		case 12: horspool();
			break;
		case 13: brute_force();
			break;
		case 14: shift_or();
			break;
		case 15: reverse_colussi();
			break;
		case 16: quick_search();
			break;
		case 17: reverse_factor();
			break;
		case 18: turbo_reverse_factor();
			break;
		case 19: optimal_mismatch();
			break;
		break;
	}	
	mode = -1;
	return next_occurence;
}


// the knuth morris pratt algorithm

void string_matching::preKmp(char* xad, array<int> &kmpNext) {
   int i, j;

   i = 0;
   j = kmpNext[0] = -1;
   while (i < m) {
      while (j > -1 && xad[i] != xad[j])
         j = kmpNext[j];
      i++;
      j++;
      if (xad[i] == xad[j])
         kmpNext[i] = kmpNext[j];
      else
         kmpNext[i] = j;
   }
}


void string_matching::knuth_morris_pratt() {
	char* xad = &(x[0]);
	char* yad = &(y[0]);
	array<int> kmpNext(m);
	int i, j;
	result.clear();
   /* Preprocessing */
   preKmp(xad, kmpNext);

   /* Searching */
   i = j = 0;
   if (mode == -1)
	   while (j < n) {
		   while (i > -1 && xad[i] != yad[j])
			   i = kmpNext[i];
		   i++;
		   j++;
		   if (i >= m) {
			   result.append(j - i);
			   i = kmpNext[i];
		   }   
	   }
   else {
	   j = mode;
	   while (j < n) {
		   while (i > -1 && xad[i] != yad[j])
			   i = kmpNext[i];
		   i++;
		   j++;
		   if (i >= m) {
			   next_occurence = (j - i);
			   return;
		   }   
	   }
   }

}


// the brute force method

void string_matching::brute_force() { 
	char* xad = &(x[0]);
	char* yad = &(y[0]);
	

	result.clear();
	char *yb;
	/* Searching */ 
	if (mode == -1)
	{
		for (yb = yad; *yad != EOS; ++yad) 
			if (memcmp(xad, yad, m) == 0) 
				result.append(int(yad - yb));
	}
	else
	{
		yb = yad;
		for (yad+=mode; *yad != EOS; ++yad) 
			if (memcmp(xad, yad, m) == 0)
			{
				next_occurence = int(yad - yb);
				return;
			}
	}


}



void string_matching::preAut(char* xad, Graph aut) {
   int i, state, target, oldTarget;
 
   for (state = getInitial(aut), i = 0; i < m; ++i) {
      oldTarget = getTarget(aut, state, xad[i]);
      target = newVertex(aut);
      setTarget(aut, state, xad[i], target);
      copyVertex(aut, target, oldTarget);
      state = target;
   }
   setTerminal(aut, state);
}
void string_matching::automaton() {
	char* xad = &(x[0]);
	result.clear();

   int j, state;
   Graph aut;
 
   /* Preprocessing */
   aut = newAutomaton(m + 1, (m + 1)*ASIZE);
   preAut(xad, aut);
 
   /* Searching */
   for (state = getInitial(aut), j = 0; j < n; ++j) {
      state = getTarget(aut, state, y[j]);
      if (isTerminal(aut, state))
         result.append(j - m + 1);
   }
}

// the Karp-Rabin algorithm
void string_matching::karp_rabin() {
	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();
   int d, hx, hy, i, j;

   /* Preprocessing */
   /* computes d = 2^(m-1) with
      the left-shift operator */
   for (d = i = 1; i < m; ++i)
      d = (d<<1);

   for (hy = hx = i = 0; i < m; ++i) {
      hx = ((hx<<1) + xad[i]);
      hy = ((hy<<1) + yad[i]);
   }

   /* Searching */
   if(mode == -1) {
	   j = 0;
	   while (j <= n-m) {
		   if (hx == hy && memcmp(xad, yad + j, m) == 0)
			   result.append(j);
		   hy = REHASH(yad[j], yad[j + m], hy);
		   ++j;
	   }
   }
   else {
	   for (j = 0; j < mode; j++)
		   hy = REHASH(yad[j], yad[j + m], hy);
	   j = mode;
	   while (j <= n-m) {
		   if (hx == hy && memcmp(xad, yad + j, m) == 0) {
			   next_occurence = j;
			   return;
		   }
		   hy = REHASH(yad[j], yad[j + m], hy);
		   ++j;
	   }
   }
}

// the shift or algorithm



int string_matching::preSo(char* xx, unsigned int S[]) { 
    unsigned int j, lim; 
    int i; 
    for (i = 0; i < ASIZE; ++i) 
      S[i] = ~0; 
    for (lim = i = 0, j = 1; i < m; ++i, j <<= 1) { 
      S[int(xx[i])] &= ~j; 
      lim |= j; 
    } 
    lim = ~(lim>>1); 
    return(lim); 
  } 

void string_matching::shift_or() { 
	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

    unsigned int lim, state; 
    unsigned int S[ASIZE]; 
    int j; 
    if (m > WORD) 
      LEDA_EXCEPTION(1,"SO: Use pattern size <= word size"); 

    /* Preprocessing */ 
    lim = preSo(xad, S); 

    /* Searching */ 
	if (mode == -1)
	{
    for (state = ~0, j = 0; j < n; ++j) { 
      state = (state<<1) | S[int(yad[j])]; 
      if (state < lim) 
        result.append(j - m + 1); 
    } 
	}
	else
	{
    for (state = ~0, j = mode; j < n; ++j) { 
      state = (state<<1) | S[int(yad[j])]; 
      if (state < lim) 
	  {
		  next_occurence = j-m+1;
		  return;
	  } 
    } 
	}
  } 

// the morris pratt algorithm

void string_matching::preMp(char *xx, array<int> &mpNext) {
   int i, j;

   i = 0;
   j = mpNext[0] = -1;
   while (i < m) {
      while (j > -1 && xx[i] != xx[j])
         j = mpNext[j];
      mpNext[++i] = ++j;
   }
}


void string_matching::morris_pratt() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

	int i, j;
	array<int> mpNext(m);

   /* Preprocessing */
	preMp(xad, mpNext);

   /* Searching */
   if (mode == -1) {
	   i = j = 0;
	   while (j < n) {
		   while (i > -1 && xad[i] != yad[j])
			   i = mpNext[i];
		   i++;
		   j++;
		   if (i >= m) {
			   result.append(j - i);
			   i = mpNext[i];
		   }
	   }
   }
   else {
	   i = 0;
	   j = mode;
	   while (j < n) {
		   while (i > -1 && xad[i] != yad[j])
			   i = mpNext[i];
		   i++;
		   j++;
		   if (i >= m) {
			   next_occurence = (j - i);
			   return;
		   }
	   }
   }
}





int preColussi(char *x, int m, int h[], int next[], int shift[]) {
   int i, k, nd, q, r(0), s;
   array<int> hmax(m);
   array<int> kmin(m);
   array<int> nhd0(m);
   array<int> rmin(m);
//   int hmax[m], kmin[m], nhd0[m], rmin[m];

   // Computation of hmax
   i = k = 1;
   do {
      while (x[i] == x[i - k])
         i++;
      hmax[k] = i;
      q = k + 1;
      while (hmax[q - k] + k < i) {
         hmax[q] = hmax[q - k] + k;
         q++;
      }
      k = q;
      if (k == i + 1)
         i = k;
   } while (k <= m);

   // Computation of kmin
   int* kptr = &kmin[0];
   memset(kptr, 0, m*sizeof(int));
   for (i = m; i >= 1; --i)
      if (hmax[i] < m)
         kmin[hmax[i]] = i;

   // Computation of rmin
   for (i = m - 1; i >= 0; --i) {
      if (hmax[i + 1] == m)
         r = i + 1;
      if (kmin[i] == 0)
         rmin[i] = r;
      else
         rmin[i] = 0;
   }

   // Computation of h
   s = -1;
   r = m;
   for (i = 0; i < m; ++i)
      if (kmin[i] == 0)
         h[--r] = i;
      else
         h[++s] = i;
   nd = s;
 
   // Computation of shift 
   for (i = 0; i <= nd; ++i)
      shift[i] = kmin[h[i]];
   for (i = nd + 1; i < m; ++i)
      shift[i] = rmin[h[i]];
   shift[m] = rmin[0];

   // Computation of nhd0
   s = 0;
   for (i = 0; i < m; ++i) {
      nhd0[i] = s;
      if (kmin[i] > 0)
         ++s;
   }


   // Computation of next
   for (i = 0; i <= nd; ++i)
      next[i] = nhd0[h[i] - kmin[h[i]]];
   for (i = nd + 1; i < m; ++i)
      next[i] = nhd0[m - rmin[h[i]]];
   next[m] = nhd0[m - rmin[h[m - 1]]];

   return(nd);
}

// the reverse colussi algorithm


void string_matching::preRc(char *x, array<int> &h, int rcBc[ASIZE][XSIZE], array<int> &rcGs) {
   int a, i, j, k, q, r(0), s;
   array<int> hmin(m);
   array<int> kmin(m);
   array<int> link(m);
   array<int> rmin(m);
   int   locc[ASIZE];
 
   /* Computation of link and locc */
   for (a = 0; a < ASIZE; ++a)
      locc[a] = -1;
   link[0] = -1;
   for (i = 0; i < m - 1; ++i) {
      link[i + 1] = locc[int(x[i])];
      locc[int(x[i])] = i;
   }
 
   /* Computation of rcBc */
   for (a = 0; a < ASIZE; ++a)
      for (s = 1; s <= m; ++s) {
         i = locc[a];
         j = link[m - s];
         while (i - j != s && j >= 0)
            if (i - j > s)
               i = link[i + 1];
            else
               j = link[j + 1];
         while (i - j > s)
            i = link[i + 1];
         rcBc[a][s] = m - i - 1;
      }
 
   /* Computation of hmin */
   k = 1;
   i = m - 1;
   while (k <= m) {
      while (i - k >= 0 && x[i - k] == x[i])
         --i;
      hmin[k] = i;
      q = k + 1;
      while (hmin[q - k] - (q - k) > i) {
         hmin[q] = hmin[q - k];
         ++q;
      }
      i += (q - k);
      k = q;
      if (i == m)
         i = m - 1;
   }
 
   /* Computation of kmin */
   int* kptr = &kmin[0];
   memset(kptr, 0, m * sizeof(int));
   for (k = m; k > 0; --k)
      kmin[hmin[k]] = k;
 
   /* Computation of rmin */
   for (i = m - 1; i >= 0; --i) {
      if (hmin[i + 1] == i)
         r = i + 1;
      rmin[i] = r;
   }
 
   /* Computation of rcGs */
   i = 1;
   for (k = 1; k <= m; ++k)
      if (hmin[k] != m - 1 && kmin[hmin[k]] == k) {
         h[i] = hmin[k];
         rcGs[i++] = k;
      }
   i = m-1;
   for (j = m - 2; j >= 0; --j)
      if (kmin[j] == 0) {
         h[i] = j;
         rcGs[i--] = rmin[j];
      }
   rcGs[m] = rmin[0];
}
void string_matching::reverse_colussi() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

	int i, j, s, rcBc[ASIZE][XSIZE];
	array<int> rcGs(m);
	array<int> h(m);
 
   /* Preprocessing */
   preRc(xad, h, rcBc, rcGs);
 
   /* Searching */
   if (mode == -1)
   {
   j = 0;
   s = m;
   while (j <= n - m) {
      while (j <= n - m && xad[m - 1] != yad[j + m - 1]) {
         s = rcBc[int(yad[j + m - 1])][s];
         j += s;
      }
      for (i = 1; i < m && xad[h[i]] == yad[j + h[i]]; ++i);
      if (i >= m && j < n-m+1)
         result.append(j);
      s = rcGs[i];
      j += s;
   }
   }
   else
   {
   j = mode;
   s = m;
   while (j <= n - m) {
      while (j <= n - m && xad[m - 1] != yad[j + m - 1]) {
         s = rcBc[int(yad[j + m - 1])][s];
         j += s;
      }
      for (i = 1; i < m && xad[h[i]] == yad[j + h[i]]; ++i);
      if (i >= m && j < n-m+1)
	  {
		  next_occurence = j;
		  return;
	  }
      s = rcGs[i];
      j += s;
   }
   }
}

// the horspool algorithm

void string_matching::preBmBc(char *x, int bmBc[]) {
   int i;
 
   for (i = 0; i < ASIZE; ++i)
      bmBc[i] = m;
   for (i = 0; i < m - 1; ++i)
      bmBc[int(x[i])] = m - i - 1;
}

void string_matching::horspool() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

   int j, bmBc[ASIZE];
   char c;

   /* Preprocessing */
   preBmBc(xad, bmBc);

   /* Searching */
   if (mode == -1)
   {
   j = 0;
   while (j <= n - m) {
      c = yad[j + m - 1];
      if (xad[m - 1] == c && memcmp(xad, yad + j, m - 1) == 0)
         result.append(j);
      j += bmBc[int(c)];
   }
   }
   else
   {
   j = mode;
   while (j <= n - m) {
      c = yad[j + m - 1];
      if (xad[m - 1] == c && memcmp(xad, yad + j, m - 1) == 0)
	  {
		  next_occurence = j;
		  return;
	  }
      j += bmBc[int(c)];
   }
   }
}

// the not so naive algorithm


void string_matching::not_so_naive() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();
   int j, k, ell;
   if (m == 1)
	   LEDA_EXCEPTION(1, "length has to be > 2");

   /* Preprocessing */
   if (xad[0] == xad[1]) {
      k = 2;
      ell = 1;
   }
   else {
      k = 1;
      ell = 2;
   }
  
   /* Searching */
   if (mode == -1)
   {
   j = 0;
   while (j <= n - m)
      if (xad[1] != yad[j + 1])
         j += k;
      else {
         if (memcmp(xad + 2, yad + j + 2, m - 2) == 0 && xad[0] == yad[j])
            result.append(j);
         j += ell;
      }
   }
   else
   {
   j = mode;
   while (j <= n - m)
      if (xad[1] != yad[j + 1])
         j += k;
      else {
		  if (memcmp(xad + 2, yad + j + 2, m - 2) == 0 && xad[0] == yad[j]){
			  next_occurence = j;
			  return;
		  }
         j += ell;
      }
   }
}

// the boyer moore algorithm

void string_matching::suffixes(char *x, array<int> &suff) {
   int f(0), g, i;
 
   suff[m - 1] = m;
   g = m - 1;
   for (i = m - 2; i >= 0; --i) {
      if (i > g && suff[i + m - 1 - f] < i - g)
         suff[i] = suff[i + m - 1 - f];
      else {
         if (i < g)
            g = i;
         f = i;
         while (g >= 0 && x[g] == x[g + m - 1 - f])
            --g;
         suff[i] = f - g;
      }
   }
}
void string_matching::preBmGs(char *x, array<int> &bmGs) {
   int i, j;
   array<int> suff(m);
 
   suffixes(x, suff);
 
   for (i = 0; i < m; ++i)
      bmGs[i] = m;
   j = 0;
   for (i = m - 1; i >= -1; --i)
      if (i == -1 || suff[i] == i + 1)
         for (; j < m - 1 - i; ++j)
            if (bmGs[j] == m)
               bmGs[j] = m - 1 - i;
   for (i = 0; i <= m - 2; ++i)
      bmGs[m - 1 - suff[i]] = m - 1 - i;
}
void string_matching::boyer_moore() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();
	int i, j;
	array<int> bmGs(m);
	int bmBc[ASIZE];
 
   /* Preprocessing */
   preBmGs(xad, bmGs);
   preBmBc(xad, bmBc);
 
   /* Searching */
   if (mode == -1) {
	   j = 0;
	   while (j <= n - m) {
		   for (i = m - 1; i >= 0 && xad[i] == yad[i + j]; --i);
		   if (i < 0) {
			   result.append(j);
			   j += bmGs[0];
		   }
		   else
			   j += max(bmGs[i], bmBc[int(yad[i + j])] - m + 1 + i);
	   }
   }
   else {
	   j = mode;
	   while (j <= n - m) {
		   for (i = m - 1; i >= 0 && xad[i] == yad[i + j]; --i);
		   if (i < 0) {
			   next_occurence = j;
			   return;
		   }
		   else
			   j += max(bmGs[i], bmBc[int(yad[i + j])] - m + 1 + i);
	   }
   }
}

// the turbo boyer moore algorithm

void string_matching::turbo_boyer_moore() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();
	int bcShift, i, j, shift, u, v, turboShift, bmBcx[ASIZE];
	array<int> bmGs(m);

   /* Preprocessing */
   preBmGs(xad, bmGs);
   preBmBc(xad, bmBcx);

   /* Searching */
   if (mode == -1)
   {
   j = u = 0;
   shift = m;
   while (j <= n - m) {
      i = m - 1;
      while (i >= 0 && xad[i] == yad[i + j]) {
         --i;
         if (u != 0 && i == m - 1 - shift)
            i -= u;
      }
      if (i < 0) {
         result.append(j);
         shift = bmGs[0];
         u = m - shift;
      }
      else {
         v = m - 1 - i;
         turboShift = u - v;
         bcShift = bmBcx[int(yad[i + j])] - m + 1 + i;
         shift = max(turboShift, bcShift);
         shift = max(shift, bmGs[i]);
         if (shift == bmGs[i])
            u = min(m - shift, v);
         else {
           if (turboShift < bcShift)
              shift = max(shift, u + 1);
           u = 0;
         }
      }
      j += shift;
   }
   }
   else {
	   u = 0;
   j = mode;
   shift = m;
   while (j <= n - m) {
      i = m - 1;
      while (i >= 0 && xad[i] == yad[i + j]) {
         --i;
         if (u != 0 && i == m - 1 - shift)
            i -= u;
      }
      if (i < 0) {
         next_occurence = j;
		 return;
      }
      else {
         v = m - 1 - i;
         turboShift = u - v;
         bcShift = bmBcx[int(yad[i + j])] - m + 1 + i;
         shift = max(turboShift, bcShift);
         shift = max(shift, bmGs[i]);
         if (shift == bmGs[i])
            u = min(m - shift, v);
         else {
           if (turboShift < bcShift)
              shift = max(shift, u + 1);
           u = 0;
         }
      }
      j += shift;
   }
   }
}



// the quick search algorithm

void string_matching::preQsBc(char *x, int qsBc[]) {
   int i;

   for (i = 0; i < ASIZE; ++i)
      qsBc[i] = m + 1;
   for (i = 0; i < m; ++i)
      qsBc[int(x[i])] = m - i;
}


void string_matching::quick_search() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

   int j, qsBc[ASIZE];

   /* Preprocessing */
   preQsBc(xad, qsBc);
 
   /* Searching */
   if (mode == -1)
   {
   j = 0;
   while (j <= n - m) {
      if (memcmp(xad, yad + j, m) == 0)
         result.append(j);
      j += qsBc[int(yad[j + m])];               /* shift */
   }
   }
   else
   {
   j = mode;
   while (j <= n - m) {
      if (memcmp(xad, yad + j, m) == 0)
	  {
		  next_occurence = j;
		  return;
	  }
      j += qsBc[int(yad[j + m])];               /* shift */
   }
   }
}

// the tuned boyer moore algorithm

void string_matching::tuned_boyer_moore() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

   int j, k, shift, bmBc[ASIZE];
 
   /* Preprocessing */
   preBmBc(xad, bmBc);
   shift = bmBc[int(xad[m - 1])];
   bmBc[int(xad[m - 1])] = 0;
   memset(yad + n, xad[m - 1], m);
 
   /* Searching */
   if (mode == -1)
   {
   j = 0;
   while (j < n) {
      k = bmBc[int(yad[j + m -1])];
      while (k !=  0) {
         j += k; k = bmBc[int(yad[j + m -1])];
         j += k; k = bmBc[int(yad[j + m -1])];
         j += k; k = bmBc[int(yad[j + m -1])];
      }
      if (memcmp(xad, yad + j, m - 1) == 0 && j < n-m+1)
         result.append(j);
      j += shift;                          /* shift */
   }
   }
   else
   {
   j = mode;
   while (j < n) {
      k = bmBc[int(yad[j + m -1])];
      while (k !=  0) {
         j += k; k = bmBc[int(yad[j + m -1])];
         j += k; k = bmBc[int(yad[j + m -1])];
         j += k; k = bmBc[int(yad[j + m -1])];
      }
      if (memcmp(xad, yad + j, m - 1) == 0 && j < n-m+1)
	  {
         next_occurence = j;
		 return;
	  }
      j += shift;                          /* shift */
   }
   }
}

// the zhu takaoka algorithm

void string_matching::preZtBc(char *x, int ztBc[ASIZE][ASIZE]) {
   int i, j;

   for (i = 0; i < ASIZE; ++i)
      for (j = 0; j < ASIZE; ++j)
         ztBc[i][j] = m;
   for (i = 0; i < ASIZE; ++i)
      ztBc[i][int(x[0])] = m - 1;
   for (i = 1; i < m - 1; ++i)
      ztBc[int(x[i - 1])][int(x[i])] = m - 1 - i;
}

void string_matching::zhu_takaoka() {
	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

	int i, j, ztBc[ASIZE][ASIZE];
	array<int> bmGs(m);

   // Preprocessing
   preZtBc(xad, ztBc);
   preBmGs(xad, bmGs);

   // Searching
   if (mode == -1)
   {
   j = 0;
   while (j <= n - m) {
      i = m - 1;
      while (i < m && xad[i] == yad[i + j])
         --i;
      if (i < 0) {
         result.append(j);
         j += bmGs[0];
      }
      else
         j += max(bmGs[i], ztBc[int(yad[j + m - 2])] [int(yad[j + m - 1])] );
   }
   }
   else
   {
   j = mode;
   while (j <= n - m) {
      i = m - 1;
      while (i < m && xad[i] == yad[i + j])
         --i;
      if (i < 0) {
         next_occurence = j;
         return;
      }
      else
         j += max(bmGs[i], ztBc[int(yad[j + m - 2])] [int(yad[j + m - 1])] );
   }
   }
}

// the berry ravindran algorithm

void string_matching::preBrBc(char *x, int brBc[ASIZE][ASIZE]) {

   int a, b, i;

   for (a = 0; a < ASIZE; ++a)
      for (b = 0; b < ASIZE; ++b)
         brBc[a][b] = m + 2;
   for (a = 0; a < ASIZE; ++a)
      brBc[a][int(x[0])] = m + 1;
   for (i = 0; i < m - 1; ++i)
      brBc[int(x[i])][int(x[i + 1])] = m - i;
   for (a = 0; a < ASIZE; ++a)
      brBc[int(x[m - 1])][a] = 1;
}


void string_matching::berry_ravindran() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

   int j, brBc[ASIZE][ASIZE];

   /* Preprocessing */
   preBrBc(xad, brBc);

   /* Searching */
   yad[n + 1] = '\0';

   if (mode == -1)
   {
   j = 0;
   while (j <= n - m) {
      if (memcmp(xad, yad + j, m) == 0)
         result.append(j);
      j += brBc[int(yad[j + m])] [int(yad[j + m + 1])];
   }
   }
   else
   {
   j = mode;
   while (j <= n - m) {
      if (memcmp(xad, yad + j, m) == 0)
	  {
         next_occurence = j;
		 return;
	  }
      j += brBc[int(yad[j + m])] [int(yad[j + m + 1])];
   }
   }
}

// the smith algorithm

void string_matching::smith() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

   int j, bmBc[ASIZE], qsBc[ASIZE];

   /* Preprocessing */
   preBmBc(xad, bmBc);
   preQsBc(xad, qsBc);

   /* Searching */
   if (mode == -1)
   {
   j = 0;
   while (j<= n - m) {
      if (memcmp(xad, yad + j, m) == 0)
         result.append(j);
      j += max(bmBc[int(yad[j + m - 1])], qsBc[int(yad[j + m])]);
   }
   }
   else
   {
   j = mode;
   while (j<= n - m) {
      if (memcmp(xad, yad + j, m) == 0)
	  {
		  next_occurence = j;
		  return;
	  }
      j += max(bmBc[int(yad[j + m - 1])], qsBc[int(yad[j + m])]);
   }
   }
}

// the raita algorithm

// the backward nondeterministic dawg matching algorithm

void string_matching::backward_nondeterministic_dawg_matching() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

  int B[ASIZE];
  int i, j, s, d, last;
  if (m > int(sizeof(WORD)))
    LEDA_EXCEPTION(1,"BNDM");

  // Pre processing
  memset(B,0,ASIZE*sizeof(int));
  s=1;
  for (i=m-1; i>=0; i--){
    B[int(xad[i])] |= s;
    s <<= 1;
  }

  // Searching phase 
  j=0;
  while (j <= n-m){
    i=m-1; last=m;
    d = ~0;
    while (i>=0 && d!=0) {
      d &= B[int(yad[j+i])];
      i--;
      if (d != 0){
    if (i >= 0)
      last = i+1;
    else
      result.append(j);
       }
       d <<= 1;
     }
     j += last;
   }
  }


void string_matching::raita() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

   int j, bmBc[ASIZE];
   char c, firstCh, *secondCh, middleCh, lastCh;

   /* Preprocessing */
   preBmBc(xad, bmBc);
   firstCh = xad[0];
   secondCh = xad + 1;
   middleCh = xad[m/2];
   lastCh = xad[m - 1];

   /* Searching */
   if (mode == -1)
   {
   j = 0;
   while (j <= n - m) {
      c = yad[j + m - 1];
      if (lastCh == c && middleCh == yad[j + m/2] &&
          firstCh == yad[j])
		  if ((memcmp(secondCh, yad + j + 1, m - 2) == 0) || (m == 1))
          result.append(j);
      j += bmBc[int(c)];
   }
   }
   else
   {
   j = mode;
   while (j <= n - m) {
      c = yad[j + m - 1];
      if (lastCh == c && middleCh == yad[j + m/2] &&
          firstCh == yad[j])
		  if ((memcmp(secondCh, yad + j + 1, m - 2) == 0) || (m == 1))
		  {
			  next_occurence = j;
			  return;
		  }
      j += bmBc[int(c)];
   }
   }
	
}

// the reverse factor algorithm

char *string_matching::reverse(char *x) {
   char *xR;
   int i;
 
   xR = (char *)malloc((m + 1)*sizeof(char));
   for (i = 0; i < m; ++i)
      xR[i] = x[m - 1 - i];
   xR[m] = '\0';
   return(xR);
}
void string_matching::reverse_factor() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();


   int i, j, shift, period, init, state;
   Graph aut;
   char *xR;
 
   /* Preprocessing */
   aut = newSuffixAutomaton(2*(m + 2), 2*(m + 2)*ASIZE);
   xR = reverse(xad);
   buildSuffixAutomaton(xR, m, aut);
   init = getInitial(aut);
   period = m;
 
   /* Searching */
   if (mode == -1)
   {
   j = 0;
   while (j <= n - m) {
      i = m - 1;
      state = init;
      shift = m;
      while (i + j >= 0 &&
             getTarget(aut, state, yad[i + j]) !=
             UNDEFINED) {
         state = getTarget(aut, state, yad[i + j]);
         if (isTerminal(aut, state)) {
            period = shift;
            shift = i;
         }
         --i;
      }
      if (i < 0) {
         result.append(j);
         shift = period;
      }
      j += shift;
   }
   }
   else
   {
   j = mode;
   while (j <= n - m) {
      i = m - 1;
      state = init;
      shift = m;
      while (i + j >= 0 &&
             getTarget(aut, state, yad[i + j]) !=
             UNDEFINED) {
         state = getTarget(aut, state, yad[i + j]);
         if (isTerminal(aut, state)) {
            period = shift;
            shift = i;
         }
         --i;
      }
      if (i < 0) {
		  next_occurence = j;
		  return;
      }
      j += shift;
   }
   }
   //return 0;
}

// the turbo reverse factor algorithm

void string_matching::turbo_reverse_factor() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

   int period, i, j, shift, u, periodOfU, disp, init, state, mu;
   array<int> mpNext(m + 1);
   char *xR;
   Graph aut;
  
   /* Preprocessing */
   aut = newSuffixAutomaton(2*(m + 2), 2*(m + 2)*ASIZE);
   xR = reverse(xad);
   buildSuffixAutomaton(xR, m, aut);
   init = getInitial(aut);
   preMp(xad, mpNext);
   period = m - mpNext[m];
   i = 0;
   shift = m;
  
   /* Searching */
   if (mode == -1)
   {
   j = 0;
   while (j <= n - m) {
      i = m - 1;
      state = init;
      u = m - 1 - shift;
      periodOfU = (shift != m ?
                   m - shift - mpNext[m - shift] : 0);
      shift = m;
      disp = 0;
      while (i > u &&
             getTarget(aut, state, yad[i + j]) !=
             UNDEFINED) {
         disp += getShift(aut, state, yad[i + j]);
         state = getTarget(aut, state, yad[i + j]);
         if (isTerminal(aut, state))
            shift = i;
         --i;
      }
      if (i <= u)
      {  if (disp == 0) {
            result.append(j);
            shift = period;
         }
         else {
            mu = (u + 1)/2;
            if (periodOfU <= mu) {
               u -= periodOfU;
               while (i > u &&
                      getTarget(aut, state, yad[i + j]) !=
                      UNDEFINED) {
                  disp += getShift(aut, state, yad[i + j]);
                  state = getTarget(aut, state, yad[i + j]);
                  if (isTerminal(aut, state))
                     shift = i;
                  --i;
               }
               if (i <= u)
                  shift = disp;
            }
            else {
               u = u - mu - 1;
               while (i > u &&
                      getTarget(aut, state, yad[i + j]) !=
                      UNDEFINED) {
                  disp += getShift(aut, state, yad[i + j]);
                  state = getTarget(aut, state, yad[i + j]);
                  if (isTerminal(aut, state))
                     shift = i;
                  --i;
               }
            }
         }
      }
      j += shift;
   }
   }
   else
   {
   j = mode;
   while (j <= n - m) {
      i = m - 1;
      state = init;
      u = m - 1 - shift;
      periodOfU = (shift != m ?
                   m - shift - mpNext[m - shift] : 0);
      shift = m;
      disp = 0;
      while (i > u &&
             getTarget(aut, state, yad[i + j]) !=
             UNDEFINED) {
         disp += getShift(aut, state, yad[i + j]);
         state = getTarget(aut, state, yad[i + j]);
         if (isTerminal(aut, state))
            shift = i;
         --i;
      }
      if (i <= u)
      {  if (disp == 0) {
			 next_occurence = j;
			 return;
         }
         else {
            mu = (u + 1)/2;
            if (periodOfU <= mu) {
               u -= periodOfU;
               while (i > u &&
                      getTarget(aut, state, yad[i + j]) !=
                      UNDEFINED) {
                  disp += getShift(aut, state, yad[i + j]);
                  state = getTarget(aut, state, yad[i + j]);
                  if (isTerminal(aut, state))
                     shift = i;
                  --i;
               }
               if (i <= u)
                  shift = disp;
            }
            else {
               u = u - mu - 1;
               while (i > u &&
                      getTarget(aut, state, yad[i + j]) !=
                      UNDEFINED) {
                  disp += getShift(aut, state, yad[i + j]);
                  state = getTarget(aut, state, yad[i + j]);
                  if (isTerminal(aut, state))
                     shift = i;
                  --i;
               }
            }
         }
      }
      j += shift;
   }
   }
}

// the forward dawg matching algorithm

void string_matching::forward_dawg_matching() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

   int j, init, ell, state;
   Graph aut;
 
   /* Preprocessing */
   aut = newSuffixAutomaton(2*(m + 2), 2*(m + 2)*ASIZE);
   buildSuffixAutomaton(xad, m, aut);
   init = getInitial(aut);
 
   /* Searching */
   ell = 0;
   state = init;
   for (j = 0; j < n; ++j) {
      if (getTarget(aut, state, yad[j]) != UNDEFINED) {
         ++ell;
         state = getTarget(aut, state, yad[j]);
      }
      else {
         while (state != init &&
                getTarget(aut, state, yad[j]) == UNDEFINED)
            state = getSuffixLink(aut, state);
         if (getTarget(aut, state, yad[j]) != UNDEFINED) {
            ell = getLength(aut, state) + 1;
            state = getTarget(aut, state, yad[j]);
         }
         else {
            ell = 0;
            state = init;
         }
      }
      if (ell == m)
         result.append(j - m + 1);
   }
   //return 0;
}



// the backward oracle matching algorithm

void string_matching::oracle(char *x, array<char> &T, array<List> &L) {
   int i, p, q(0);
   array<int> S(m + 1);
   char c;

   S[m] = m + 1;
   for (i = m; i > 0; --i) {
      c = x[i - 1];
      p = S[i];
      while (p <= m &&
             (q = getTransition(x, p, L, c)) ==
             UNDEFINED) {
         setTransition(p, i - 1, L);
         p = S[p];
      }
      S[i - 1] = (p == m + 1 ? m : q);
   }
   p = 0;
   while (p <= m) {
      T[p] = 1;
      p = S[p];
   }
}


void string_matching::backward_oracle_matching() {


	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

   array<char> T(m + 1);
   array<List> L(m + 1);
   int i, j, p, period(0), q, shift;

   /* Preprocessing */
   char* tptr = &T[0];
   List* lptr = &L[0];

   memset(lptr, 0, (m + 1)*sizeof(List));
   memset(tptr, 0, (m + 1)*sizeof(char));
   oracle(xad, T, L);

   /* Searching */
   j = 0;
   while (j <= n - m) {
      i = m - 1;
      p = m;
      shift = m;
      while (i + j >= 0 &&
             (q = getTransition(xad, p, L, yad[i + j])) !=
             UNDEFINED) {
         p = q;
         if (T[p] == 1) {
            period = shift;
            shift = i;
         }
         --i;
      }
      if (i < 0) {
         result.append(j);
         shift = period;
      }
      j += shift;
   }
}

// the galil seiferas algorithm
void string_matching::search(char* xad, char* yad) {
	if (mode == -1)
	{
	while (pgs <= n - m) {
      while (pgs + sgs + qgs < n && xad[sgs + qgs] == yad[pgs + sgs + qgs])
         ++qgs;
      if (qgs == m - sgs && memcmp(xad, yad + pgs, sgs + 1) == 0)
         result.append(pgs);
      if (qgs == pgs1 + qgs1) {
         pgs += pgs1;
         qgs -= pgs1;
      }
      else {
         pgs += (qgs/kgs + 1);
         qgs = 0;
      }
   }
	}
	else
	{
		pgs += mode;
   while (pgs <= n - m) {
      while (pgs + sgs + qgs < n && xad[sgs + qgs] == yad[pgs + sgs + qgs])
         ++qgs;
      if (qgs == m - sgs && memcmp(xad, yad + pgs, sgs + 1) == 0)
	  {
		  next_occurence = pgs;
		  return;
	  }
      if (qgs == pgs1 + qgs1) {
         pgs += pgs1;
         qgs -= pgs1;
      }
      else {
         pgs += (qgs/kgs + 1);
         qgs = 0;
      }
   }
	}
}


//void newP1(char* xad, char* yad);

void string_matching::parse(char* xad, char* yad) {
   while (1) {
      while (xad[sgs + qgs1] == xad[sgs + pgs1 + qgs1])
         ++qgs1;
      while (pgs1 + qgs1 >= kgs*pgs1) {
         sgs += pgs1;
         qgs1 -= pgs1;
      }
      pgs1 += (qgs1/kgs + 1);
      qgs1 = 0;
      if (pgs1 >= pgs2)
         break;
   }
   newP1(xad, yad);
}
void string_matching::newP2(char* xad, char* yad) {
   while (xad[sgs + qgs2] == xad[sgs + pgs2 + qgs2] && pgs2 + qgs2 < kgs*pgs2)
      ++qgs2;
   if (pgs2 + qgs2 == kgs*pgs2)
      parse(xad, yad);
   else
      if (sgs + pgs2 + qgs2 == m)
         search(xad, yad);
      else {
         if (qgs2 == pgs1 + qgs1) {
            pgs2 += pgs1;
            qgs2 -= pgs1;
         }
         else {
            pgs2 += (qgs2/kgs + 1);
            qgs2 = 0;
         }
         newP2(xad, yad);
      }
}
void string_matching::newP1(char* xad, char* yad) {
   while (xad[sgs + qgs1] == xad[sgs + pgs1 + qgs1])
      ++qgs1;
   if (pgs1 + qgs1 >= kgs*pgs1) {
      pgs2 = qgs1;
      qgs2 = 0;
      newP2(xad, yad);
   }
   else {
      if (sgs + pgs1 + qgs1 == m)
         search(xad, yad);
      else {
         pgs1 += (qgs1/kgs + 1);
         qgs1 = 0;
         newP1(xad, yad);
      }
   }
}


void string_matching::galil_seiferas() {
   char* xad = &(x[0]);
   char* yad = &(y[0]);
   result.clear();
   kgs = 4;
   pgs = qgs = sgs = qgs1 = pgs2 = qgs2 = 0;
   pgs1 = 1;
   newP1(xad, yad);
}

// the two way algorithm

/* Computing of the maximal suffix for <= */
int string_matching::maxSuf(char *x, int *p) {
   int ms, j, k;
   char a, b;

   ms = -1;
   j = 0;
   k = *p = 1;
   while (j + k < m) {
      a = x[j + k];
      b = x[ms + k];
      if (a < b) {
         j += k;
         k = 1;
         *p = j - ms;
      }
      else
         if (a == b)
            if (k != *p)
               ++k;
            else {
               j += *p;
               k = 1;
            }
         else { /* a > b */
            ms = j;
            j = ms + 1;
            k = *p = 1;
         }
   }
   return(ms);
} 
/* Computing of the maximal suffix for >= */
int string_matching::maxSufTilde(char *x, int *p) {
   int ms, j, k;
   char a, b;

   ms = -1;
   j = 0;
   k = *p = 1;
   while (j + k < m) {
      a = x[j + k];
      b = x[ms + k];
      if (a > b) {
         j += k;
         k = 1;
         *p = j - ms;
      }
      else
         if (a == b)
            if (k != *p)
               ++k;
            else {
               j += *p;
               k = 1;
            }
         else { /* a < b */
            ms = j;
            j = ms + 1;
            k = *p = 1;
         }
   }
   return(ms);
}


/* Two Way string matching algorithm. */
void string_matching::two_way() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

   int i, j, ell, memory, p, per, q;

   /* Preprocessing */
   i = maxSuf(xad, &p);
   j = maxSufTilde(xad, &q);
   if (i > j) {
      ell = i;
      per = p;
   }
   else {
      ell = j;
      per = q;
   }

   /* Searching */
   if (memcmp(xad, xad + per, ell + 1) == 0) {
      j = 0;
      memory = -1;
      while (j <= n - m) {
         i = max(ell, memory) + 1;
         while (i < m && xad[i] == yad[i + j])
            ++i;
         if (i >= m) {
            i = ell;
            while (i > memory && xad[i] == yad[i + j])
               --i;
            if (i <= memory)
               result.append(j);
            j += per;
            memory = m - per - 1;
         }
         else {
            j += (i - ell);
            memory = -1;
         }
      }
   }
   else {
      per = max(ell + 1, m - ell - 1) + 1;
      j = 0;
      while (j <= n - m) {
         i = ell + 1;
         while (i < m && xad[i] == yad[i + j])
            ++i;
         if (i >= m) {
            i = ell;
            while (i >= 0 && xad[i] == yad[i + j])
               --i;
            if (i < 0)
               result.append(j);
            j += per;
         }
         else
            j += (i - ell);
      }
   }
}

// the on ordered alphabet algorithm

/* Compute the next maximal suffix. */
void nextMaximalSuffix(char *x, int m,
                       int *i, int *j, int *k, int *p) {
   char a, b;
 
   while (*j + *k < m) {
      a = x[*i + *k];
      b = x[*j + *k];
      if (a == b)
         if (*k == *p) {
            (*j) += *p;
            *k = 1;
         }
         else
            ++(*k);
      else
         if (a > b) {
            (*j) += *k;
            *k = 1;
            *p = *j - *i;
         }
         else {
            *i = *j;
            ++(*j);
            *k = *p = 1;
         }
   }
}
/* String matching on ordered alphabets algorithm. */

void string_matching::on_ordered_alphabet() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

   int i, ip, j, jp, k, p;
 
   /* Searching */
   ip = -1;
   i = j = jp = 0;
   k = p = 1;
   while (j <= n - m) {
      while (i + j < n && i < m && xad[i] == yad[i + j])
         ++i;
      if (i == 0) {
         ++j;
         ip = -1;
         jp = 0;
         k = p = 1;
      }
      else {
         if (i >= m)
            result.append(j);
         nextMaximalSuffix(yad + j, i+1, &ip, &jp, &k, &p);
         if (ip < 0 ||
             (ip < p &&
              memcmp(yad + j, yad + j + p, ip + 1) == 0)) {
            j += p;
            i -= p;
            if (i < 0)
               i = 0;
            if (jp - ip > p)
               jp -= p;
            else {
               ip = -1;
               jp = 0;
               k = p = 1;
            }
         }
         else {
            j += (max(ip + 1,
                      min(i - ip - 1, jp + 1)) + 1);
            i = jp = 0;
            ip = -1;
            k = p = 1;
         }
      }
   }
}


// the optimal mismatch algorithm



/* Construct an ordered pattern from a string. */
/* Find the next leftward matching shift for
   the first ploc pattern elements after a
   current shift or lshift. */
int string_matching::matchShift(char *x, int ploc, int lshift, pattern *pat) {
   int i, j;

   for (; lshift < m; ++lshift) {
      i = ploc;
      while (--i >= 0) {
         if ((j = (pat[i].loc - lshift)) < 0)
            continue;
         if (pat[i].c != x[j])
            break;
      }
      if (i < 0)
         break;
   }
   return(lshift);
}


/* Constructs the good-suffix shift table
   from an ordered string. */
void string_matching::preAdaptedGs(char *x, array<int> &adaptedGs, pattern *pat) {
   int lshift, i, ploc;

   adaptedGs[0] = lshift = 1;
   for (ploc = 1; ploc <= m; ++ploc) {
      lshift = matchShift(x, ploc, lshift, pat);
      adaptedGs[ploc] = lshift;
   }
   for (ploc = 0; ploc <= m; ++ploc) {
      lshift = adaptedGs[ploc];
      while (lshift < m) {
         i = pat[ploc].loc - lshift;
         if (i < 0 || pat[ploc].c != x[i])
            break;
         ++lshift;
         lshift = matchShift(x, ploc, lshift, pat);
      }
      adaptedGs[ploc] = lshift;
   }
}


/* Optimal Mismatch string matching algorithm. */
void string_matching::optimal_mismatch() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

	int i, j;
	array<int> adaptedGs(m);
	pattern* pat = new pattern[m];
	int qsBc[ASIZE];
   

   /* Preprocessing */
   orderPattern(xad, m, optimalPcmp, pat);
   preQsBc(xad, qsBc);
   preAdaptedGs(xad, adaptedGs, pat);

   /* Searching */
   if (mode == -1)
   {
   j = 0;
   while (j <= n - m) {
      i = 0;
      while (i < m && pat[i].c == yad[j + pat[i].loc])
         ++i;
      if (i >= m)
         result.append(j);
      j += max(adaptedGs[i],qsBc[int(y[j + m])]);
   }
   }
   else
   {
   j = mode;
   while (j <= n - m) {
      i = 0;
      while (i < m && pat[i].c == yad[j + pat[i].loc])
         ++i;
      if (i >= m)
	  {
		  next_occurence = j;
		  return;
	  }
      j += max(adaptedGs[i],qsBc[int(y[j + m])]);
   }
   }
}

// the maximal shift algorithm



/* Computation of the MinShift table values. */
void string_matching::computeMinShift(char *x) {
   int i, j;

   for (i = 0; i < m; ++i) {
      for (j = i - 1; j >= 0; --j)
          if (x[i] == x[j])
             break;
      minShift[i] = i - j;
   }
}




/* Maximal Shift string matching algorithm. */
void string_matching::maximal_shift() {

	char* xad = &(x[0]);
	result.clear();

   int i, j, qsBc[ASIZE];
   array<int> adaptedGs(m);
   pattern* pat = new pattern[m];

   /* Preprocessing */
   computeMinShift(xad);
   orderPattern(xad, m, maxShiftPcmp, pat);
   preQsBc(xad, qsBc);
   preAdaptedGs(xad, adaptedGs, pat);
   /* Searching */
   j = 0;
   while (j <= n - m) {
      i = 0;
      while (i < m && pat[i].c == y[j + pat[i].loc])
         ++i;
      if (i >= m)
         result.append(j);
      j += max(adaptedGs[i], qsBc[int(y[j + m])]);
   }
}

// the skip search algorithm


void string_matching::skip_search() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();

   int i, j;
   List ptr, z[ASIZE];

   /* Preprocessing */
   memset(z, 0, ASIZE*sizeof(List));
   for (i = 0; i < m; ++i) {
      ptr = (List)malloc(sizeof(struct _cell));
      if (ptr == NULL)
         LEDA_EXCEPTION(1,"SKIP");
      ptr->element = i;
      ptr->next = z[int(xad[i])];
      z[int(xad[i])] = ptr;
   }
  
   /* Searching */
   for (j = m - 1; j < n; j += m)
      for (ptr = z[int(y[j])]; ptr != NULL; ptr = ptr->next)
         if (memcmp(xad, yad + j - ptr->element, m) == 0)
            result.append(j - ptr->element);
}

// the kmp skip search algorithm

int string_matching::attempt(char *y, char *x, int start, int wall) {
   int k;

   k = wall - start;
   while (k < m && x[k] == y[k + start])
      ++k;
   return(k);
}


void string_matching::kmp_skip_search() {

	char* xad = &(x[0]);
	char* yad = &(y[0]);
	result.clear();



    int i, j, k, kmpStart, per, start, wall;
	array<int> kmpNext(m);
    array<int> list(m);
    array<int> mpNext(m);
    int z[ASIZE];

   /* Preprocessing */
   preMp(xad, mpNext);
   preKmp(xad, kmpNext);
   memset(z, -1, ASIZE*sizeof(int));
   int* listptr = &list[0];
   memset(listptr, -1, m*sizeof(int));
   z[int(xad[0])] = 0;
   for (i = 1; i < m; ++i) {
      list[i] = z[int(xad[i])];
      z[int(xad[i])] = i;
   }

   /* Searching */
   wall = 0;
   per = m - kmpNext[m];
   i = j = -1;
   do {
      j += m;
   } while (j < n && z[int(yad[j])] < 0);
   if (j >= n)
     return;
   i = z[int(yad[j])];
   start = j - i;
   while (start <= n - m) {
      if (start > wall)
         wall = start;
      k = attempt(yad, xad, start, wall);
      wall = start + k;
      if (k == m) {
         result.append(start);
         i -= per;
      }
      else
         i = list[i];
      if (i < 0) {
         do {
            j += m;
         } while (j < n && z[int(yad[j])] < 0);
         if (j >= n)
            return;
         i = z[int(yad[j])];
      }
      kmpStart = start + k - kmpNext[k];
      k = kmpNext[k];
      start = j - i;
      while (start < kmpStart ||
             (kmpStart < start && start < wall)) {
         if (start < kmpStart) {
            i = list[i];
            if (i < 0) {
               do {
                  j += m;
               } while (j < n && z[int(yad[j])] < 0);
               if (j >= n)
                  return;
               i = z[int(yad[j])];
            }
            start = j - i;
         }
         else {
            kmpStart += (k - mpNext[k]);
            k = mpNext[k];
         }
      }
   }
}

LEDA_END_NAMESPACE


