#ifdef _MSC_IDE

#include <LEDA/system/file.h>

#if defined(LEDA_STD_IO_HEADERS)
using std::cout;
using std::cerr;
using std::endl;
using std::flush;
using std::ostream;
using std::istream;
using std::ofstream;
using std::ifstream;
using std::streambuf;
using std::filebuf;
using std::ios;
#endif

using namespace leda;

#endif // ifdef _MSC_IDE

#include "DirectoryTree.h"

typedef DirectoryTreeNode Node;

/// DirectoryTreeNode //////////////////////////////////////////////////////////////////////

DirectoryTreeNode::DirectoryTreeNode(const leda::string& name)
 : Name(name), Father(0), ItemInFather(0), Mark(false)
{}

/// access ///

string
DirectoryTreeNode::get_path() const
{
	string path;
	if (Father) {
		path = Father->get_path();
		if (path != string()) path += get_directory_delimiter();
	}
	path += Name;
	return path;
}

Node* 
DirectoryTreeNode::get_child(string name) const
{
	Node* child;
	forall(child, Children) if (child->get_name() == name) return child;
	return 0;
}

DirectoryTreeNode* 
DirectoryTreeNode::get_left_sibling() const
{
	if (!Father || !ItemInFather) return 0; // no father
	list_item it = Father->Children.pred(ItemInFather);
	if (!it) return 0; // this is first child
	return Father->Children[it];
}

DirectoryTreeNode* 
DirectoryTreeNode::get_right_sibling() const
{
	if (!Father || !ItemInFather) return 0; // no father
	list_item it = Father->Children.succ(ItemInFather);
	if (!it) return 0; // this is last child
	return Father->Children[it];
}

DirectoryTreeNode* 
DirectoryTreeNode::get_predecessor() const
{
	// DFS order
	Node* current = get_left_sibling();
	if (! current) return Father;

	while (!current->is_leaf()) {
		current = current->get_last_child();
	}
	return current;
}

DirectoryTreeNode* 
DirectoryTreeNode::get_successor() const
{
	// DFS order
	if (! is_leaf()) return get_first_child();

	Node* current = const_cast<Node*>(this);
	while (current->Father && current->Father->get_last_child() == current  ) {
		current = current->Father;
	}
	return current->get_right_sibling();
}

Node* 
DirectoryTreeNode::follow_path(string path) const
{
	string delim = get_directory_delimiter();
	// remove delimiter at the beginning, if one is there
	if (path.head(delim.length()) == delim) path = path(delim.length(), path.length()-1);
	// if path is empty, return this
	if (path == "") return const_cast<Node*>(this);
	// extract the first name on path
	int pos = path.pos(delim);
	if (pos < 0) pos = path.length();
	string child_name = path(0, pos-1);
	// ... and lookup the child
	Node* child = get_child(child_name); if (! child) return 0;
	// remove child name from path and perform a recursive call on the child
	path = path(pos+delim.length(), path.length()-1);
	return child->follow_path(path);
}

/// update ///

void 
DirectoryTreeNode::set_mark_of_ancestors(bool mark)
{
	Node* current = this;
	do {
		current->Mark = mark;
		current = current->get_father();
	} while (current);
}

void 
DirectoryTreeNode::set_mark_of_descendants(bool mark)
{
	Mark = mark;
	Node* child;
	forall(child, Children) {
		child->set_mark_of_descendants(mark);
	}
}

void 
DirectoryTreeNode::clear()
{
	if (Father) Father->remove_child(this);
	clear_children();
}

void 
DirectoryTreeNode::clear_children()
{
	Node* child;
	forall(child, Children) {
		child->Father = 0; child->ItemInFather = 0; // avoid side effects
		delete child;
	}
	Children.clear();
}

void 
DirectoryTreeNode::add_child(Node* child)
{
	child->ItemInFather = Children.append(child);
	child->Father = this;
}

void 
DirectoryTreeNode::remove_child(Node* child)
{
	if (!child || child->Father != this || !child->ItemInFather) return;

	Children.del_item(child->ItemInFather);
	child->Father = 0; child->ItemInFather = 0;
}

Node* 
DirectoryTreeNode::remove_child(string name)
{
	Node* child = get_child(name);
	remove_child(child);
	return child;
}

/// DirectoryTree //////////////////////////////////////////////////////////////////////////

DirectoryTree::DirectoryTree(const leda::string& path_to_root)
 : PathToRoot(path_to_root), Root(0)
{
	rebuild();
}

string 
DirectoryTree::get_absolute_path(Node* n) const
{
	if (n == Root) return PathToRoot;
	else           return PathToRoot + get_directory_delimiter() + n->get_path();
}

void
DirectoryTree::clear()
{
	delete Root; Root = 0;
}

void
DirectoryTree::rebuild()
{
	clear();
	Root = new Node();
	rebuild_children(Root);
}

void
DirectoryTree::rebuild_children(Node* n)
{
	if (n == 0) return;

	n->clear_children();

	list<string> sub_dirs = get_directories(get_absolute_path(n));
	string sub_dir;
	forall(sub_dir, sub_dirs) {
		if (sub_dir == "." || sub_dir == "..") continue;
		Node* child = new Node(sub_dir);
		n->add_child(child);
	}

	list<Node*> children = n->get_children();
	Node* child;
	forall(child, children) {
		rebuild_children(child);
	}
}

void 
DirectoryTree::print(ostream& out) const
{
	cout << PathToRoot;
	for (Iter it = begin(); it != end(); ++it) {
		int d = it->get_depth();
		for (int i = 0; i < d; ++i) out << "  ";
		out << it->get_name() << endl;
	}
}

void 
DirectoryTree::remove_nodes_with_name(string name)
{
	Iter it = begin();
	while (it.valid()) {
		if (it->get_name() == name) {
			Node* current = it;
			current->clear_children(); // NOTE: this affects "++it" below!
			++it;
			clear_subtree(current);
		}
		else ++it;
	}
}

list<string> 
DirectoryTree::find_files(string pattern) const
{
	list<string> result;
	for (Iter it = begin(); it != end(); ++it) {
		string path = get_absolute_path(it);
		if (! is_directory(path)) continue;

		list<string> files = get_files(path, pattern);
		string file;
		forall(file, files) {
			string file_path = path + get_directory_delimiter() + file;
			result.append(file_path);
		}
	}
	return result;
}

void 
DirectoryTree::clear_subtree(Node* subtree_root)
{
	delete subtree_root;
	if (subtree_root == Root) Root = 0;
}

void 
DirectoryTree::clear_subtree(string relative_path)
{
	if (!Root) return;
	clear_subtree(Root->follow_path(relative_path));
}
