#define LEDA_CHECKING_OFF

#include <LEDA/graph/graph.h>
#include <LEDA/graph/static_fgraph.h>
#include <LEDA/graph/node_slot1.h>
#include <LEDA/graph/edge_slot1.h>
#include <LEDA/graph/min_cut.h>
#include <LEDA/core/array2.h>
#include <LEDA/core/partition.h>
#include <LEDA/system/assert.h>


using namespace leda;
using std::cout;
using std::cin;
using std::endl;
using std::flush;
using std::ostream;

#include "max_flow8.h"


typedef max_flow<int>::node_data nd_type;
typedef max_flow<int>::edge_data ed_type;
  
const int ns = sizeof(nd_type)/3;
const int es = sizeof(ed_type)/4;
  
typedef static_fgraph<opposite_graph,data_slots<ns+1>,data_slots<es+2> > 
                                                                     st_graph;

typedef st_graph::node st_node;
typedef st_graph::edge st_edge;
  
typedef node_slot1<nd_type,st_graph,0>  node_data_array;
typedef node_slot1<node,st_graph,ns>    orig_array;

typedef edge_slot1<ed_type,st_graph,0> edge_data_array;
typedef edge_slot<int,st_graph,es>     cap_array;
typedef edge_slot<int,st_graph,es+1>   flow_array;

typedef  max_flow<int, st_graph, node_data_array, edge_data_array> MaxFlow;


st_node min_weight_node(const st_graph& G, const cap_array& cap)
{
  node_array<int,st_graph> W(G,0);

  st_node v;
  forall_nodes(v,G)
  { int sum = 0;
    st_edge e;
    forall_out_edges(e,v) sum += cap[e];
    W[v] = sum;
   }

  st_node u = G.first_node();
  forall_nodes(v,G) 
      if (W[v] < W[u]) u = v;

  return u;
}


node min_weight_node(const graph& G, const edge_array<int>& weight, int& val)
{
  node_array<int> W(G,0);

  node v;
  forall_nodes(v,G)
  { int sum = 0;
    edge e;
    forall_inout_edges(e,v) sum += weight[e];
    W[v] = sum;
   }

  node u = G.first_node();
  forall_nodes(v,G) 
      if (W[v] < W[u]) u = v;

  val = W[u];

  return u;
}




int min_cut_mf0(graph& G0, const edge_array<int>& weight ) 
{
  
  node_array<st_node> V(G0);

  int n = G0.number_of_nodes();
  int m = G0.number_of_edges();

  st_graph G;
  G.start_construction(n,2*m);

  cap_array  cap(G);
  flow_array flow(G);

  node v;
  forall_nodes(v,G0) V[v] =  G.new_node();

  st_node s = G.first_node(); 

  forall_nodes(v,G0)
  { edge e;
    forall_inout_edges(e,v)
    { node w = opposite(v,e);
      st_edge x = G.new_edge(V[v],V[w]);
      cap[x] = weight[e];
     }
   }

  G.finish_construction();

  MaxFlow mf;

  int best_value = MAXINT;

  st_node t;
  forall_nodes(t,G)
  { if (s == t) continue;
    int f = mf.run(G,s,t,cap,flow);
    if (f < best_value) best_value = f; 
  }

  return best_value; 
}



int min_cut_mf1(graph& G0, const edge_array<int>& weight ) 
{
  node_array<st_node> V(G0);

  int n = G0.number_of_nodes();
  int m = G0.number_of_edges();

  st_graph G;
  G.start_construction(n,2*m);

  cap_array  cap(G);
  flow_array flow(G);

  node v;
  forall_nodes(v,G0) V[v] =  G.new_node();



  forall_nodes(v,G0)
  { edge e;
    forall_inout_edges(e,v)
    { node w = opposite(v,e);
      st_edge x = G.new_edge(V[v],V[w]);
      cap[x] = weight[e];
     }
   }

  G.finish_construction();

//st_node s = G.first_node(); 
  st_node s = min_weight_node(G,cap);

  MaxFlow mf;

  int best_value = MAXINT;

  st_node t;
  forall_nodes(t,G)
  { if (s == t) continue;
    int f = mf.min_cut(G,s,t,cap,flow,best_value);
    if (f < best_value) best_value = f; 
  }

  return best_value; 
}



//------------------------------------------------------------------------------

  
void contract(graph& G, node a, node b,
                        edge_array<int>& weight, 
                        node_array<edge>& s_edge,
                        list<edge>& L, int best_val)
{ 
  if (G.degree(b) == 0) return;

  edge e;
  forall_inout_edges(e,a) s_edge[G.opposite(a,e)] =  e; 

  list<edge> del;
  list<edge> mov;

  forall_inout_edges(e,b)
  { node v = G.opposite(b,e);
    if ( v == a) { del.append(e); continue; }
    edge x = s_edge[v];
    if (x == nil)
      mov.append(e);
    else 
      { weight[x] += weight[e]; 
        if (weight[x] >= best_val) L.append(x);
        del.append(e);
       }
  }

  forall_inout_edges(e,a)  s_edge[G.opposite(a,e)] =  nil; 
 

  forall(e,mov)
  { node v = G.opposite(b,e);
    G.move_edge(e,a,v);
   }

  //G.del_edges(del);
  forall(e,del) G.hide_edge(e);

  assert(G.degree(b) == 0);
}




void construct_contract_graph(GRAPH<node,int>& G, const graph& H, const edge_array<int>& weight)
{
  node_array<node> v_in_G(H); 
  node v;
  forall_nodes(v,H) v_in_G[v] = G.new_node(v);
  edge e;
  forall_edges(e,H)
  { node x = v_in_G[source(e)];
    node y = v_in_G[target(e)];
    G.new_edge(x,y,weight[e]);
/*
    edge ee = G.new_edge(x,y,0);
    G[ee] = weight[e];
*/
  }
}



st_node construct_flow_graph(st_graph& G, orig_array& orig, 
                                          cap_array& cap, 
                                          const graph& H, 
                                          node s,
                                          const edge_array<int>& weight,
                                          node_array<st_node>& V)
{
  V.init(H,0);

  int n = H.number_of_nodes();
  int m = H.number_of_edges();

//cout << "construct flow graph: n = " << n << "  m = " << m << endl;

  G.start_construction(n+1,2*m+1);

  st_node s0 = G.new_node();
  orig[s0] = 0;

  node v;
  forall_nodes(v,H) 
  { st_node u =  G.new_node();
    V[v] = u;
    orig[u] = v;
   }

  st_edge x = G.new_edge(s0,V[s]);
  cap[x] = 0;
  
  forall_nodes(v,H)
  { edge e;
    forall_inout_edges(e,v)
    { node w = opposite(v,e);
      st_edge x = G.new_edge(V[v],V[w]);
      cap[x] = weight[e];
/*
      int we = weight[e];
      cap[x] = we;
*/
      assert(cap[x] == weight[e]);
     }
   }

  G.finish_construction();

//cout << "construct flow graph finished"  << endl;

  return s0;
}




int min_cut_mf2(const graph& G_orig, edge_array<int>& weight_orig,
                                     list<node>& C, bool use_heuristic, double frac) 
{
  GRAPH<node,int> G;
  construct_contract_graph(G,G_orig,weight_orig);
  edge_array<int>& weight = G.edge_data();

  st_graph F;
  orig_array V_orig(F);
  cap_array  cap(F);
  flow_array flow(F);

  node_array<st_node> V(G,0);
  node_array<edge> s_edge(G,0);

  int best_value = MAXINT;
  node s = min_weight_node(G,weight,best_value);
  node s_orig = G[s];

  best_value++;

  st_node s0 = construct_flow_graph(F,V_orig,cap,G,s,weight,V);
  st_edge e0 = F.first_out_edge(s0);
  int m0 = G.number_of_edges();

  MaxFlow mf;
  node t_best = 0;
  int mf_count = 0;

  node t;
  forall_nodes(t,G)
  { 
    if (t == s || G.degree(t) == 0) continue;

    cap[e0] = best_value;
    int f = mf.min_cut(F,s0,V[t],cap,flow,best_value);
    mf_count++;

    bool new_best_value = false;
    if (f < best_value) 
    { best_value = f;
      new_best_value = true;
      t_best = G[t];
     }

    if (use_heuristic)
    { list<edge> L;
      contract(G,s,t,weight,s_edge,L,best_value);

      while (!L.empty())
      { edge e = L.pop();
        node x = source(e);
        node y = target(e);
        if (y == s) swap(x,y);
        contract(G,x,y,weight,s_edge,L,best_value);
       }
  
      if (new_best_value) 
      { list<edge> L;
        edge e;
        forall_edges(e,G)
          if (weight[e] >= best_value) L.append(e);
  
        while (!L.empty())
        { edge e = L.pop();
          node x = source(e);
          node y = target(e);
          if (y == s) swap(x,y);
          contract(G,x,y,weight,s_edge,L,best_value);
         }
       }
     }

     if (G.number_of_edges() == 0) break;

     if (G.number_of_edges() < frac*m0)
     { //cout << string("rebuild: mf = %d   f = %d   |E| = %d   |F| = %d", 
       //                mf_count, best_value, G.number_of_edges(), 
       //                F.number_of_edges()) << endl;
       s0 = construct_flow_graph(F,V_orig,cap,G,s,weight,V);
       e0 = F.first_out_edge(s0);
       m0 = G.number_of_edges();
      }

  }

  assert(t_best != 0);

  construct_flow_graph(F,V_orig,cap,G_orig,s_orig,weight_orig,V);

  int f = mf.run(F,V[s_orig],V[t_best],cap,flow);
//cout << "f = " << f << "  best_value = " << best_value << endl;
  assert(f == best_value);
  
  list<st_node> cut;
  mf.compute_min_st_cut(F,V[s_orig],V[t_best],cap,flow,cut);

  C.clear();
  st_node v;
  forall(v,cut) C.append(V_orig[v]);

  return best_value; 
}
  
  

//------------------------------------------------------------------------------


void graph_gen_random(graph& G, edge_array<int>& weight, int n, int m, 
                                int c_min, int c_max)
{ 
   if (m > n*(n-1)/2) m = n*(n-1)/2;

   random_simple_undirected_graph(G,n,m);
   weight.init(G);
   edge e;
   forall_edges(e,G) weight[e] = rand_int(c_min,c_max);
}

void graph_gen_path(graph& G, edge_array<int>& weight, int n)
{
  node v = 0;

  for(int i=1;i<n;i++)
  { node u = G.new_node();
    if (v) G.new_edge(v,u);
    v = u;
   }

  weight.init(G);
  int count = 0;
  edge e;
  forall_edges(e,G) weight[e] = ++count;
}



void construct_cluster(graph& G, array<node>& V, edge_array<int>& weight, 
                                                 int n, int w)
{
  for(int i=0; i< n; i++) V[i] = G.new_node();

  for(int i=0; i<n; i++) 
    for(int j=i+1; j<n; j++) 
    { edge e = G.new_edge(V[i],V[j]);
      weight[e] = w;
     }
}



void graph_gen_torus(graph& G, edge_array<int>& weight, 
                           int cols, int rows, int c_sz, int c_min, int c_max)
{
  int c_num = cols*rows;

  int n = c_sz*c_num;
  int m = c_sz*(c_sz-1)*c_num/2 + 2*c_num;

  weight.init(G,m,0);

  array2<node> north(rows,cols);
  array2<node> south(rows,cols);
  array2<node> west(rows,cols);
  array2<node> east(rows,cols);

  array<node> V(c_sz);

  int w = 0;
  for(int i=0; i<cols; i++)
  { for(int j=0; j<cols; j++)
    { construct_cluster(G,V,weight,c_sz,w++);
      north(i,j) = V[0];
      south(i,j) = V[1];
      west(i,j)  = V[2];
      east(i,j)  = V[3];
     }
   }


  for(int i=0; i<cols; i++)
  { for(int j=0; j<cols; j++)
    { node a = north(i,j);
      node b = south((i+1)%rows,j);
      node c = east(i,j);
      node d = west(i,(j+1)%cols);
      edge x = G.new_edge(a,b);
      edge y = G.new_edge(c,d);
      weight[x] = rand_int(1,w);
      weight[y] = rand_int(1,w);
     }
   }

  assert(G.number_of_nodes() == n);
  assert(G.number_of_edges() == m);

  edge e;
  forall_edges(e,G) weight[e] = rand_int(c_min,c_max);
}



int main()
{
  cout << endl;
  int nn = read_int("n = ");

  char gen = 'r';
  cout << "generator: r(random)/t(orus)/p(ath)  ";
  cin >> gen;

  int c_min = read_int("c_min = ");
  int c_max = read_int("c_max = ");

  int j_min = 4;
  int j_max = 10;
  int j_step = 1;

  if (gen == 't')
     j_max = read_int("max cluster size : ");

  cout << endl;
  for (int j = j_min; j <= j_max; j+=j_step) cout << string("%19d",j);
  cout << endl; 
  cout << endl; 

  for (int i = 1; i <= 100; i++)
  { 
    int n = i*nn;

    if (gen == 't')
      cout << string("%3dx%3d  ",n,n);
    else
       cout << string("%5d  ",n);
    
    for (int j = j_min; j <= j_max; j+=j_step)
    { 
      int m = 10*j*n;
      graph G;
      edge_array<int> weight(G);

      switch (gen) {
      case 'p': graph_gen_path(G,weight,n);
                break;
      case 'r': graph_gen_random(G,weight,n,m,c_min,c_max);
                break;
      case 't': graph_gen_torus(G,weight,n,n,j,c_min,c_max);
                break;
      }

  
      float T = used_time(); 
  
  
      list<node> C0;
      int ct0 = MIN_CUT(G,weight,C0,true);
//cout << " |C0| = " << C0.length() << "  ct0 = " << ct0 << endl;
      cout << string("%5.2f/",used_time(T)) << flush;

      list<node> C2;
      int ct2 = min_cut_mf2(G,weight,C2,true,0.05);
      cout << string("%5.2f/",used_time(T)) << flush;
//cout << " |C2| = " << C2.length() << "  ct2 = " << ct2 << endl;
      assert(ct0 == ct2);

      list<node> C3;
      int ct3 = min_cut_mf2(G,weight,C3,true,0.10);
      cout << string("%5.2f  ",used_time(T)) << flush;
//cout << " |C3| = " << C3.length() << "  ct3 = " << ct3 << endl;
      assert(ct0 == ct3);

    }
    cout << endl;

  }

return 0;
}

