// #define LEDA_CHECKING_OFF

#include <LEDA/core/array.h>
#include <LEDA/graph/graph.h>
#include <LEDA/graph/static_fgraph.h>
#include <LEDA/graph/node_slot1.h>
#include <LEDA/graph/edge_slot1.h>
#include <LEDA/graph/node_pq.h>
#include <LEDA/graph/min_cut.h>
#include <LEDA/graph/dimacs.h>
#include <LEDA/system/assert.h>

using namespace leda;
using std::cout;
using std::cin;
using std::endl;
using std::flush;
using std::ostream;



#define MF_UNDIRECTED
#include <LEDA/graph/templates/max_flow.h>

typedef max_flow<int>::node_data nd_type;
typedef max_flow<int>::edge_data ed_type;
  
const int ns = sizeof(nd_type)/3;
const int es = sizeof(ed_type)/4;
  
typedef static_fgraph<opposite_graph,data_slots<ns+1>,data_slots<es+2> > 
                                                                     st_graph;
typedef st_graph::node st_node;
typedef st_graph::edge st_edge;
  
typedef node_slot1<nd_type,st_graph,0>  node_data_array;
typedef node_slot1<node,st_graph,ns>    orig_array;

typedef edge_slot1<ed_type,st_graph,0> edge_data_array;
typedef edge_slot<int,st_graph,es>     cap_array;
typedef edge_slot<int,st_graph,es+1>   flow_array;

typedef  max_flow<int, st_graph, node_data_array, edge_data_array> MaxFlow;

void construct_flow_graph(st_graph& G, orig_array& orig, 
                                          cap_array& cap, 
                                          const graph& H, 
                                          node s,
                                          const edge_array<int>& weight,
                                          node_array<st_node>& V)
{
  V.init(H,0);

  int n = H.number_of_nodes();
  int m = H.number_of_edges();

  G.start_construction(n,2*m);

  node v;
  forall_nodes(v,H) 
  { st_node u =  G.new_node();
    V[v] = u;
    orig[u] = v;
   }
  
  forall_nodes(v,H)
  { edge e;
    forall_out_edges(e,v)
    { node w = opposite(v,e);
      st_edge x = G.new_edge(V[v],V[w]);
      cap[x] = weight[e];
     }
   }

  G.finish_construction();
}


int compute_st_cut(const graph& G, node s, node t, 
                   const edge_array<int>& weight, int best_val)
{
  st_graph F;
  orig_array V_orig(F);
  cap_array  cap(F);
  flow_array flow(F);
  node_array<st_node> V(G,0);
  construct_flow_graph(F,V_orig,cap,G,s,weight,V);
  MaxFlow mf;
  return mf.mincut_value(F,V[s],V[t],cap,flow,best_val);
}







int sum_incident_weights(const edge_array<int>& weight, node v)
{ int sum = 0;
  edge e;
  forall_inout_edges(e,v) sum += weight[e];
  return sum;
}

void sort_nodes_by_weight(graph& G, const edge_array<int>& weight)
{ node_array<int> W(G);
  node v;
  forall_nodes(v,G) W[v] = sum_incident_weights(weight,v);
  G.sort_nodes(W);
}


void contract(graph& G, node a, node b, edge_array<int>& weight, 
                                        node_array<int>& cut,
                                        node_array<edge>& s_edge,
                                        list<edge>& L,
                                        edge_array<bool>& in_L,
                                        int& best_val)
{
  assert(a != b);
  assert(G.degree(a) > 0);
  assert(G.degree(b) > 0);

  //cout << string("contract(%d,%d)",G.index(a),G.index(b)) << endl;

  double cut_a = cut[a];

  cut_a += cut[b];

  edge e;
  forall_inout_edges(e,a) s_edge[G.opposite(a,e)] =  e; 

  list<edge> del;
  list<edge> mov;

  forall_inout_edges(e,b)
  { node v = G.opposite(b,e);

    if (v == a) 
    { del.append(e); 
      cut_a -= weight[e];
      continue; 
     }

    edge x = s_edge[v];

    if (x == nil)
      mov.append(e);
    else 
      { weight[x] += weight[e]; 
        //if (weight[x] >= best_val) L.append(x);
        del.append(e);
       }
  }


  forall(e,mov)
  { node v = G.opposite(b,e);
    G.move_edge(e,a,v);
   }

  forall(e,del) G.hide_edge(e);
  //G.del_edges(del);

  assert(cut_a <= MAXINT);
  cut[a] = int(cut_a);

  if (cut[a] < best_val) best_val = cut[a];

  forall_inout_edges(e,a)  
  { 
    node u = G.opposite(a,e);

    if (!in_L[e] && s_edge[u] == e) 
    { L.append(e);
      in_L[e] = true;
     }

    if (s_edge[u] == nil) continue;

/*
    edge x;
    forall_inout_edges(x,u)  
    if (!in_L[x]) 
    { L.append(x);
      in_L[x] = true;
     }
*/

    s_edge[u] = nil;
   }

  assert(G.degree(b) == 0);
  //assert(G.degree(a) > 0);
  //G.del_node(b);
}



int is_contractible(graph& G, edge x, edge_array<int>& weight, 
                               node_array<int>& cut,
                               node_array<edge>& s_edge,
                               int  best_val)
{
  int w = weight[x];
  if (w >= best_val) return 1;
return 0;

  node a = G.source(x);
  node b = G.target(x);

  assert(a != b);

  if (2*w >= cut[a] || 2*w >= cut[b]) return 2;

return 0;

  int result = 0;

  edge e;
  forall_inout_edges(e,a) s_edge[G.opposite(a,e)] =  e; 

  forall_inout_edges(e,b)
  { node v = G.opposite(b,e);
    edge y = s_edge[v];
    if (y == nil) continue;
    int w1 = 2*(weight[y]+weight[x]);
    int w2 = 2*(weight[e]+weight[x]);
/*
cout << w1 << " " <<cut[a] << endl;
cout << w2 << " " <<cut[b] << endl;
cout <<endl;
*/

    if (w1 >= cut[a] && w2 >= cut[b]) 
    {  result = 3;
       break;
     }
  }

  forall_inout_edges(e,a) s_edge[G.opposite(a,e)] =  0; 

  return result;
}



void read_dimacs(graph& G, edge_array<int>& cap)
{
  node s=0;
  node t=0;

  cout << endl;
  cout << "reading DIMACS maxflow problem from std input ... ";
  cout << flush;

  dimacs_mf D(cin);
  D.translate(G,s,t,cap);
  cout << endl;

  int n = G.number_of_nodes();
  int m = G.number_of_edges();

  cout << endl;
  cout << string("%6d nodes",n) << endl;
  cout << string("%6d edges",m) << endl;
  cout << endl;
}


void read_koeln(graph& G, edge_array<int>& cap)
{
  cout << endl;
  cout << "reading min cut problem from std input ... ";
  cout << flush;

  int n,m;

  cin >> n >> m;

  array<node> V(n);
  array<double> W(m);

  for(int i=0; i<n; i++) V[i] = G.new_node();

  for(int i=0; i<m; i++) 
  { int a,b;
    cin >> a >> b >> W[i];
    G.new_edge(V[a],V[b]);
   }

  cap.init(G);
  int j=0;
  edge e;
  //forall_edges(e,G) cap[e] = int(1000*W[j++]);
  forall_edges(e,G) cap[e] = int(100*W[j++]);

  cout << endl;
  cout << string("%6d nodes",n) << endl;
  cout << string("%6d edges",m) << endl;
  cout << endl;
}



void compute_shrink_list(const graph& G, const node_array<int>& cut,
                                         const edge_array<int>& weight)
{

  node y;
  forall_nodes(y,G)
  { list<node> L;
    edge e;
    forall_inout_edges(e,y)
    { int w = weight[e];
      node u = G.opposite(y,e);
      if (2*w > cut[u]) L.append(u); 
     }
   }
}



int main()
{
 graph G;
 edge_array<int> weight(G);

/*
 int n = 1000;
 int m = 100000;

 random_simple_undirected_graph(G,n,m);

 weight.init(G);
 edge e;
 forall_edges(e,G) weight[e] = rand_int(1,1000);
*/

 //read_dimacs(G,weight);
 read_koeln(G,weight);

 //int n = G.number_of_nodes();
 //int m = G.number_of_edges();

 float T = used_time();

 float time_mf = 0;


 node_array<int> cut(G,0);

 node_array<edge> s_edge;
 s_edge.use_node_data(G,0);

 node best_node = 0;
 int best_val = MAXINT;

 node v;
 forall_nodes(v,G)
 { cut[v] = sum_incident_weights(weight,v);
   if (cut[v] < best_val)
   { best_node = v;
     best_val = cut[v];
    }
  }

 cout << "best_val = " << best_val << endl;

 int st_count = 0;
 //node s = best_node;
 node s = G.first_node();

 st_graph F;
 orig_array V_orig(F);
 cap_array  cap(F);
 flow_array flow(F);
 node_array<st_node> V(G,0);
 construct_flow_graph(F,V_orig,cap,G,s,weight,V);
 MaxFlow mf;

/*
edge e;
forall_edges(e,G) is_contractible(G,e,weight,cut,s_edge,best_val);
*/


 array<int> case_counter(4);
 case_counter.init(0);

 int contract_count = 0;

 node t;
 forall_nodes(t,G)
 { 
   if (t == s) continue;
   if (G.degree(t) == 0) 
   { //cout << "degree(t) = 0" << endl;
     continue;
    }

   //assert(G.degree(s) > 0);
   if (G.degree(s) == 0)
   { s = G.first_node();
     while (s != nil && (s == t || G.degree(s) == 0)) s = G.succ_node(s);
     if (s == nil) break;
    }

   cout << string("maxflow(%d,%d)",G.index(s),G.index(t)) << endl;
  
   float T1 = used_time();
   int f = mf.mincut_value(F,V[s],V[t],cap,flow,best_val);
   time_mf += used_time(T1);

   st_count++;


   list<edge> L;
   edge_array<bool> in_L(G,false);

   if (f >= best_val) 
     { contract(G,s,t,weight,cut,s_edge,L,in_L,best_val);
       contract_count++;

       //if (st_count < n/2) continue;

       //cout << "L = " << L.length() << endl;
      }
   else
   {  cout << "new best_val = " << f << endl;
      best_val = f;
      edge e;
      forall_edges(e,G) 
         if (weight[e] > best_val) L.append(e);
    }

    cout << "|L| = " << L.length() << " ";
    cout << case_counter[1] << " ";
    cout << case_counter[2] << " ";
    cout << case_counter[3] << endl;


    while (!L.empty())
    { edge e = L.pop();
      if (G.is_hidden(e)) continue;
      int c = is_contractible(G,e,weight,cut,s_edge,best_val);
      if (c > 0)
      { case_counter[c]++;
        node u = G.source(e);
        node v = G.target(e);
        if (v == s) swap(u,v);
        contract(G,u,v,weight,cut,s_edge,L,in_L,best_val);
        contract_count++;
      }
    }
 }


 cout << endl;
 cout << "nodes    = " << G.number_of_nodes() << endl;
 cout << "edges    = " << G.number_of_edges() << endl;
 cout << "st_count = " << st_count <<endl;
 cout << "result   = " << best_val <<endl;
 cout << "contract1 = " << case_counter[1] <<endl;
 cout << "contract2 = " << case_counter[2] <<endl;
 cout << "contract3 = " << case_counter[3] <<endl;
 cout << endl;
 cout << string("total time:    %5.2f sec", used_time(T)) << endl;
 cout << string("maxflow time:  %5.2f sec", time_mf) << endl;
 cout << endl;
 cout << "contractions: " << contract_count;
 cout << endl;

 return 0;

}
