#include <LEDA/graph/graph.h>
#include <LEDA/graph/dimacs_graph.h> 
#include <LEDA/graph/graph_alg.h> 

#include <LEDA/core/queue.h>
#include <LEDA/core/b_queue.h>
#include <LEDA/core/array.h>
#include <LEDA/core/list.h>

#include <LEDA/system/assert.h>
#include <LEDA/system/multi_thread.h>
#include <LEDA/system/thread.h>

#include <iostream>
#include <fstream>

#define push_flow(e,x)   { flow[e] += (e.out() ? x : -(x)); }


using namespace leda;
using std::cout;
using std::cin;
using std::endl;
using std::cerr;


//*************************************************************
// **************************** fifo_set  *********************
//*************************************************************

class fifo_set{

  list<node> L;

public:

  fifo_set(){}

  node del() { if (!L.empty()) return L.pop(); else return nil; }
 
  void insert(node v, int)  { L.append(v); }
  void insert0(node v, int) { L.append(v); }                            

  void split(int& thr_num, array<fifo_set>& F, node_array<int>& dist)
  {
          int g(0);
          node v;
          forall(v,L)
          {
                  int number = g % thr_num;
                  F[number].insert(v, dist[v]);
                  g++;
          }
  } 

  int size() { return L.size(); }

  bool empty() { return L.empty(); }

  void clear() { L.clear(); }

  ~fifo_set(){}
};



struct thread_data
{
  graph* G;
  node s;
  node t;
  edge_array<int>* cap;
  edge_array<int>* flow;
  node_array<int>* dist;
  node_array<int>* excess;
  fifo_set* liste;
  int* num_pushes; 
  int* num_edge_inspections; 
  int* num_relabels; 
  int limit_heur;
  int id; 
  bool local_relabel;
  node_array<string>* name;
};


//******************************************************************
// **************************** Graphgenerator *********************
//******************************************************************

void maxflow_gen(graph& G, node& s, node& t, int& n, int& m)
{ G.clear();
  random_simple_graph(G,n,m);
  s = G.first_node(); t = G.last_node();
  node v;
  forall_nodes(v,G) { G.new_edge(s,v); G.new_edge(v,t); }
}


//************************************************************************

void compute_dist_t(const graph& G, 
                    node t, 
                    const edge_array<int>& flow, 
                    const edge_array<int>& cap, 
                    const node_array<int>& excess, 
                    node_array<int>& dist, 
                    fifo_set& U, 
                    b_queue<node>& Q, 
                    int& BFS_inspections)
{ 
  int n = G.number_of_nodes();
  dist.init(G,n);

  Q.append(t);
  dist[t] = 0;  

  while ( !Q.empty() )
  { node v = Q.pop();
    int  d = dist[v] + 1;
    if ( d >= n) break;
    edge e;

    for(e = G.first_adj_edge(v); e; e = G.adj_succ(e))
    { BFS_inspections++;
                        if ( flow[e] == 0 ) continue;
      node u = target(e); 
      int& du = dist[u];
      if ( du >= n ) 
      { du = d;
        Q.append(u); 
        if ( excess[u] > 0 ) U.insert(u,d); 
       }
     }

    for(e = G.first_in_edge(v); e; e = G.in_succ(e))
    { BFS_inspections++;
                        if ( cap[e] == flow[e] ) continue;
      node u = source(e); 
      int& du = dist[u];
      if ( du >= n )  
      { du = d;  
        Q.append(u); 
        if (excess[u] > 0) U.insert(u,d); 
      }
    }
  }
}


//**************************************************************
// ***************************   maxflow    ********************
//**************************************************************


void maxflow(const graph& G, 
             node s, node t, 
             const edge_array<int>& cap, 
             edge_array<int>& flow, 
             node_array<int>& dist, 
             node_array<int>& excess,
             fifo_set& U, 
             int& num_pushes, 
             int& num_edge_inspections,
             int& num_relabels, 
             int limit_heur,
             int id, 
             bool local_relabel)
{   
  int n = G.number_of_nodes();  
  int local_num_pushes = 0; 
  int local_num_edge_inspections = 0; 
  int local_num_relabels = 0;

  for(;;) 
  {
    node v = U.del();

    if (v == nil) break;

    if(excess[v] == 0 || v == t || dist[v] == n) continue;

    assert(dist[v] < n);
    assert(excess[v] >= 0);
  
    int dmin = MAXINT;

    if ( excess[v] > 0 )
    { 
      graph::incident_edge e; 
      forall_incident_edges(e,v)
      { 
        if (excess[v] == 0) break;
        local_num_edge_inspections++;
        int x = (e.out() ? (cap[e] - flow[e]) : flow[e]);            

        if (x == 0) continue;
        node w = G.opposite(e,v);

        if ( dist[w] == dist[v] - 1 ) 
        { 
          local_num_pushes++;

          while(1) // garantiert, dass push ausgeführt wird
          {
            if (excess[v] < x) x = excess[v]; 

            if(Fetch_And_Add(&excess[v],-x) >= x) 
            { // sonst negativer excess!!
              if (Fetch_And_Add(&excess[w],x) == 0) U.insert(w, dist[w]);
               // sonst kann w in 2 Listen eingefügt werden
              push_flow(e,x);
              break;
             }
            else Fetch_And_Add(&excess[v],x); // push findet nicht statt!!
           }
           assert(excess[v] >= 0);
        }
        else 
          if ( dist[w] < dmin ) dmin = dist[w]; 
      }

    }
                
    if (excess[v] > 0 && local_relabel) 
    { 
      if (local_num_edge_inspections > limit_heur) break;

      dmin++; 
      local_num_relabels++; 
      if ( dmin < n)
      { dist[v] = dmin;
        U.insert(v,dmin);
        }
    }

  }

  Fetch_And_Add(&num_pushes, local_num_pushes);
  Fetch_And_Add(&num_edge_inspections, local_num_edge_inspections);
  Fetch_And_Add(&num_relabels, local_num_relabels);       
}



//************************************************************************

void* maxflow_thread(void *threadarg)
{ 
  thread_data* my_data = (thread_data *) threadarg;       
  graph& G = *(my_data->G);       
  node s = my_data->s;    
  node t = my_data->t;            
  edge_array<int>& cap = *(my_data->cap);
  edge_array<int>& flow = *(my_data->flow);
  node_array<int>& dist = *(my_data->dist);
  node_array<int>& excess = *(my_data->excess);
  fifo_set U = *(my_data->liste);         
  int& num_pushes = *(my_data->num_pushes);
  int& num_edge_inspections = *(my_data->num_edge_inspections);
  int& num_relabels = *(my_data->num_relabels);
  int limit_heur = my_data->limit_heur;
  int id = my_data->id;
  bool local_relabel = my_data->local_relabel;

  maxflow(G, s, t, cap, flow, dist, excess, U, 
          num_pushes, num_edge_inspections, num_relabels, 
          limit_heur, id, local_relabel);

  return 0;
}



//***************************************************************************


int main(int argc, char** argv)
{       
  if ( argc != 2)
  { cerr << "usage: mf3-4 (0 or 1)" << endl;
    cerr << "0 for loading dimacs network" << endl;
    cerr << "1 for infinite loop with random graphs" << endl;
    return 1;
  }

  int choice = atoi(argv[1]);

  cout << "n m h Graph Versuchnr #threads real_time used_time #relabels #global_relabels #BFS_inspections #pushes #edge_inspections Berechnungen Berech/sec Fluss" << endl;

  int nr = 1;
  bool local_relabel(1);
  float h = 2;


  while (1) 
  {
    // Variablen
    graph G(2,2);

    node s;
    node t;

    edge_array<int> cap;
    assert(cap.use_edge_data(G));

    edge_array<int> flow;
    assert(flow.use_edge_data(G));

    node_array<int>  excess;
    assert(excess.use_node_data(G));

    node_array<int> dist;
    assert(dist.use_node_data(G));

    int n, m;

    if(choice == 0) 
    { // Graph einlesen
      read_dimacs_mf(cin,G,s,t,cap);
      if (s == t) LEDA_EXCEPTION(1,"MAXFLOW: source == sink");
      n = G.number_of_nodes(); 
      m = G.number_of_edges();
     }
     else
     { n = rand_int(500,5000); 
       m = rand_int(n*3,(n*(n-1)));
       maxflow_gen(G, s, t, n, m);
       edge e;
       forall_edges(e,G) cap[e] = rand_int(100,500);
     }

     float tt = used_time();
     int fluss = MAX_FLOW(G, s, t, cap, flow);
     float time_complete = used_time(tt);

     string sg("%d %d ? ", n, m);
     string tg("random %d 1 %f / / / / / / / / %d", nr, time_complete, fluss);

     cout << sg + tg << endl;


     int num_pushes, num_edge_inspections, num_relabels, num_global_relabels, BFS_inspections;
     int thr_num;


     for(thr_num=1; thr_num <= 4; thr_num *= 2)
     { flow.init(G,0);
       excess.init(G,0);
       num_relabels = 0;
       num_pushes = 0;
       num_edge_inspections = 0;
       BFS_inspections = 0;
       num_global_relabels  = 0;


       // saturate all edges leaving s
       edge e;
       forall_out_edges(e,s) 
       { int c = cap[e];
         if (c == 0) continue;
         node v = target(e);
         flow[e] = c;
         excess[s] -= c;
         excess[v] += c;
       } 
         
       fifo_set U;
       b_queue<node> Q(n);
       list<node> S;

       int limit_heur = (int) (h*m / (thr_num));
       array<fifo_set> F(thr_num);

       float t3 = used_time(); 
       float t4 = real_time();


       // Threads
       thread_data thread_data_array[thr_num];

       // Threadargumente
       for(int i=0; i<thr_num; i++) 
       { thread_data_array[i].G = &G;
         thread_data_array[i].s = s;
         thread_data_array[i].t = t;
         thread_data_array[i].cap = &cap;
         thread_data_array[i].flow = &flow;
         thread_data_array[i].dist = &dist;
         thread_data_array[i].excess = &excess;
         thread_data_array[i].liste = &F[i];                     
         thread_data_array[i].num_pushes = &num_pushes;
         thread_data_array[i].num_edge_inspections = &num_edge_inspections;
         thread_data_array[i].num_relabels = &num_relabels;
         thread_data_array[i].limit_heur = limit_heur;
         thread_data_array[i].id = i;
         thread_data_array[i].local_relabel = local_relabel;
       }


       while(1)
       {
         num_global_relabels++;
         compute_dist_t(G,t,flow,cap,excess,dist,U,Q,BFS_inspections);

         if ( U.empty() ) break;

         if(U.size() <= 10*thr_num)
         {
           int id = 0;                     
           maxflow(G, s, t, cap, flow, dist, excess, U, 
                   num_pushes, num_edge_inspections, num_relabels, 
                   limit_heur, id, local_relabel);
         }
         else
         { thread_t* thr = new thread_t[thr_num];
           U.split(thr_num, F, dist);
           U.clear();
           for(int i=0; i<thr_num; i++) 
            thr[i] = thread_create(maxflow_thread,(void*)&thread_data_array[i]);
           for(int i=0; i<thr_num; i++) thread_join(thr[i]);
         }
       }


       float time_complete_real = real_time(t4);
       float time_complete_used = used_time(t3);

       int total_ops = num_relabels + num_pushes + num_edge_inspections + BFS_inspections;

       string s1("%d %d %d ", n, m, int(h));
       string t1("random %d %d %f %f %d %d %d %d %d %d %d %d", 
                  nr, thr_num, time_complete_real, time_complete_used, 
                  num_relabels,num_global_relabels, BFS_inspections, 
                  num_pushes, num_edge_inspections, total_ops,
                  int(total_ops/time_complete_real), excess[t]);

       cout << s1 + t1 << endl;

       if(excess[t] != fluss)
       { cout << "Fehler! Falscher Fluss!" << endl;
         return 0;
        }
      }
      cout << endl;
      nr++;

      if (choice == 0) break;
   }
}

