#define LEDA_CHECKING_OFF

#include <LEDA/graph/graph.h>
#include <LEDA/graph/static_graph.h>
#include <LEDA/graph/node_slot.h>
#include <LEDA/core/b_stack.h>
#include <LEDA/core/array.h>
#include <LEDA/core/tuple.h>
#include <assert.h>

#if !defined(forall_rev_out_edges)
#define forall_rev_out_edges(e,v) \
for(void* loop_stop = ((e = (v+1)->first_out_edge()-1), v->first_out_edge()-1);\
    e != loop_stop; e = e-1)
#endif



using namespace leda;

#include "cm_rec.h"
#include "cm_nonrec.h"
#include "tarjan_rec.h"
#include "tarjan_nonrec.h"
#include "kosaraju.h"
#include "dfsnum.h"



using std::cout;
using std::flush;
using std::endl;


template <class st_graph>
void construct_static_graph(st_graph& Gs, const graph& G) 
{ 
  // constructs a static copy of G

  typedef typename st_graph::node st_node;
  typedef typename st_graph::edge st_edge;

  int n = G.number_of_nodes();
  int m = G.number_of_edges();

  Gs.start_construction(n,m);
  node_array<st_node> N(G);

  node v;
  forall_nodes(v,G) N[v] = Gs.new_node();

  forall_nodes(v,G)
  { st_node x = N[v];
    for (edge e = G.first_out_edge(v); e; e = G.next_out_edge(e))
    { st_node y = N[G.target(e)]; 
      Gs.new_edge(x,y);
     }
   }
  Gs.finish_construction();
}


void acyclic_dense_graph(graph& G, int n)
{ array<node> V(n);
  for(int i=0; i<n; i++) V[i] = G.new_node();

  for(int i=0; i<n; i++) 
     for(int j=i+1; j<n; j++) 
        G.new_edge(V[i],V[j]);
}
  

void SCC_generator1(graph& G, int n, int m)
{ 
  array<node> V(n);

  for (int i = 0; i < n; i++) V[i] = G.new_node();

  int k = m/n;

  for (int i = 1; i < n; i++) G.new_edge(V[0],V[i]);

  for (int i = 1; i <= k+1; i++) G.new_edge(V[i],V[0]);

  for (int i = k+2; i < n; i++) 
    for (int j = 0; j < k; j++) 
       G.new_edge(V[i],V[j + 1]);
}

void SCC_generator2(graph& G, int n, int m)
{ 
  array<node> V(n);

  for (int i = 0; i < n; i++) V[i] = G.new_node();

  int k = m/n;

  for (int i = 1; i < n; i++) G.new_edge(V[0],V[i]);

  for (int i = 1; i <= k+1; i++) G.new_edge(V[i],V[0]);

  for (int i = k+2; i < n; i++) 
    for (int j = 0; j < k; j++) 
       G.new_edge(V[i],V[k - j]);
}



int main(int argc, char** argv) 
{
  int n,m;

  if (argc < 3) 
  { n = read_int("n = ");
    m = read_int("m = ");
   }
  else
  { n = atoi(argv[1]);
    m = atoi(argv[2]);
   }

  graph GG;

  //acyclic_dense_graph(GG,n);

  //SCC_generator1(GG,n,m);
  random_graph(GG,n,m);


  typedef static_graph<directed_graph,data_slots<1>,data_slots<0> > st_graph;
  typedef st_graph::node node;
  typedef st_graph::edge edge;

  st_graph G;
  construct_static_graph(G,GG);


  node_array<int,st_graph> compnum0(G);   // for comparing results
  STRONG_COMPONENTS_CM0<st_graph> scc0;
  int scc_num = scc0(G,compnum0);

  cout << "scc_num = " << scc_num << endl;
  cout << endl;

  node_slot<int,st_graph,0> compnum(G);


  cout << endl;
  cout << "      Cheriyan/Mehlhorn                 Tarjan                  DFS";
  cout << endl;
  cout << " orig  rec1  rec2  non1  non2    orig  rec1  rec2  non1  non2";
  cout << endl;
  cout << endl;


{
  STRONG_COMPONENTS_CM0<st_graph> scc;
  int s = scc(G,compnum);
  cout << string("%5.2f ", scc.cpu_time()) << flush;
  node v;
  forall_nodes(v,G) assert(compnum0[v] == compnum[v]);
  assert(s == scc_num);
}


{
  STRONG_COMPONENTS_CM1<st_graph>     scc;
  int s = scc(G,compnum);
  cout << string("%5.2f ", scc.cpu_time()) << flush;
  node v;
  forall_nodes(v,G) assert(compnum0[v] == compnum[v]);
  assert(s == scc_num);
}

{
  STRONG_COMPONENTS_CM2<st_graph>     scc;
  int s = scc(G,compnum);
  cout << string("%5.2f ", scc.cpu_time()) << flush;
  node v;
  forall_nodes(v,G) assert(compnum0[v] == compnum[v]);
  assert(s == scc_num);
}

{
  STRONG_COMPONENTS_CM_NR1<st_graph>  scc;
  int s = scc(G,compnum);
  cout << string("%5.2f ", scc.cpu_time()) << flush;
  node v;
  forall_nodes(v,G) assert(compnum0[v] == compnum[v]);
  assert(s == scc_num);
}


{
  STRONG_COMPONENTS_CM_NR2<st_graph>  scc;
  int s = scc(G,compnum);
  cout << string("%5.2f ", scc.cpu_time()) << flush;
  node v;
  forall_nodes(v,G) assert(compnum0[v] == compnum[v]);
  assert(s == scc_num);
}



  cout << "  " << flush;


{
  STRONG_COMPONENTS_TARJAN0<st_graph>    scc;
  int s = scc(G,compnum);
  cout << string("%5.2f ", scc.cpu_time()) << flush;
  node v;
  forall_nodes(v,G) assert(compnum0[v] == compnum[v]);
  assert(s == scc_num);
}

{
  STRONG_COMPONENTS_TARJAN1<st_graph>    scc;
  int s = scc(G,compnum);
  cout << string("%5.2f ", scc.cpu_time()) << flush;
  node v;
  forall_nodes(v,G) assert(compnum0[v] == compnum[v]);
  assert(s == scc_num);
}


{
  STRONG_COMPONENTS_TARJAN2<st_graph>    scc;
  int s = scc(G,compnum);
  cout << string("%5.2f ", scc.cpu_time()) << flush;
  node v;
  forall_nodes(v,G) assert(compnum0[v] == compnum[v]);
  assert(s == scc_num);
}



{
  STRONG_COMPONENTS_TARJAN_NR1<st_graph> scc;
  int s = scc(G,compnum);
  cout << string("%5.2f ", scc.cpu_time()) << flush;
  node v;
  forall_nodes(v,G) assert(compnum0[v] == compnum[v]);
  assert(s == scc_num);
}


{
  STRONG_COMPONENTS_TARJAN_NR2<st_graph> scc;
  int s = scc(G,compnum);
  cout << string("%5.2f ", scc.cpu_time()) << flush;
  node v;
  forall_nodes(v,G) assert(compnum0[v] == compnum[v]);
  assert(s == scc_num);
}

  cout << " " << flush;

{
  DFS_NUM<st_graph> dfs;
  int s = dfs(G,compnum);
  cout << string("%5.2f ", dfs.cpu_time()) << flush;
}

/*
{
  STRONG_COMPONENTS_KOSARAJU<st_graph> scc;
  int s = scc(G,compnum);
  cout << string("%5.2f ", scc.cpu_time()) << flush;
  assert(s == scc_num);
}
*/

  cout << endl;

  return 0;
}


