/*******************************************************************************
+
+  LEDA 7.0  
+
+
+  _x11.c
+
+
+  Copyright (c) 1995-2023
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



#include <LEDA/system/basic.h>
#include <LEDA/graphics/x_window.h>
#include <LEDA/core/string.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>


LEDA_BEGIN_NAMESPACE

extern int  INIT_GRAPHICS(int mode,int& width,int& height,int& depth,int& dpi);

extern void UPDATE_SCREEN_AREA(unsigned int* pixels, int w, int h, 
                             int x0, int y0, int x1, int y1,
                             int x, int y);


extern int  CHECK_NEXT_EVENT(int& val,int& x,int& y,unsigned long& t,int msec);

extern void SET_MOUSE_CURSOR(int id);


static Display*      display = NULL;
static Visual*       visual = NULL;

static int x_depth = 24;
static int x_width= 2000;
static int x_height = 1500;
static int x_dpi = 96;

static int screen = 0;

static Window        root_window;
static Window        win;
static XGCValues     gc_val;
static GC            gc;
static Cursor        cursor = 0;

/*
static Colormap      color_map;
*/

int INIT_GRAPHICS(int mode, int& width, int& height, int& depth, int& dpi)
{
  if (mode == 0)
  { XDestroyWindow(display,win);
    XCloseDisplay(display); 
    display = NULL;
    return 0;
   }

  if (display != NULL) return ConnectionNumber(display);

  if ((display = XOpenDisplay(0)) == NULL)	
  { fprintf(stderr, "Can\'t open display: \n");
    abort();
   }

  screen = DefaultScreen(display);
  root_window = RootWindow(display,screen);

  x_width = int(0.9 * DisplayWidth(display,screen));
  x_height= int(0.9 * DisplayHeight(display,screen));

  int pixels = DisplayWidth(display,screen);
  int millis = DisplayWidthMM(display,screen);
  x_dpi = int(0.5 + (25.4 * pixels)/millis);

  XVisualInfo vinfo;
  XMatchVisualInfo(display,screen,24,TrueColor, &vinfo);
  visual = vinfo.visual;
  x_depth = vinfo.depth;

  depth = x_depth;
  width = x_width;
  height = x_height;
  dpi = x_dpi;

/*
  color_map = XCreateColormap(display,root_window,visual,AllocNone);
*/

  gc_val.background = WhitePixel(display,screen);
  gc_val.foreground = BlackPixel(display,screen);

  gc = XCreateGC(display,RootWindow(display,screen),
                      GCBackground | GCForeground, &(gc_val));
  
  XSetLineAttributes(display,gc,1,LineSolid,CapButt,JoinMiter);

  // open window
  
  XSetWindowAttributes attrib;
  attrib.backing_store = Always;

  win = XCreateWindow(display, RootWindow(display,screen),
                      0, 0, width,height, 2,
                      DefaultDepth(display,screen), InputOutput, 
                      DefaultVisual(display,screen), 
                      CWBackingStore, &attrib);
  
  XSelectInput(display,win, EnterWindowMask | LeaveWindowMask    |
                            KeyPressMask    | KeyReleaseMask     |
                            ButtonPressMask | ButtonReleaseMask  |
                            PointerMotionMask  | 
                            ExposureMask    | StructureNotifyMask); 
    
  XSetWindowBackground(display,win,WhitePixel(display, screen)); 
   
  XStoreName(display,win,"LEDA");
  XSetIconName(display,win,"LEDA");

  int xpos = 50;
  int ypos = 75;
  XMoveResizeWindow(display,win,xpos,ypos,width,height);

  XSizeHints size_hints;
  size_hints.flags = PPosition;
  size_hints.x = xpos;
  size_hints.y = ypos;
  XSetWMProperties(display,win,0,0,0,0,&size_hints,0,0);

  XMapWindow(display,win);
  
  XEvent e;
  do XMaskEvent(display,StructureNotifyMask,&e);
  while (e.type != MapNotify);

  while (XCheckWindowEvent(display, win, ExposureMask, &e));

  return ConnectionNumber(display);
}



/*
void UPDATE_SCREEN_AREA1(unsigned int* pixels, int w, int h, 
                       int x0, int y0, int x1, int y1,
                       int x, int y)
{ 
  // copy rectangle [x0,y0,x1,y1] from (w x h) pixel buffer
  // to canvas to position (x,y)

  if (x1 > w-1) x1 = w-1;
  if (y1 > h-1) y1 = h-1;
  if (x0 < 0) x0 = 0;
  if (y0 < 0) y0 = 0;

  if (x0 > x1 || y0 > y1) return;

  int width = x1-x0+1;
  int height = y1-y0+1;

  int sz = width*height;
  unsigned int* buf = new unsigned int[sz];
  unsigned int* p = buf;

  assert(sz >= 0);

  for (int j = y0; j <= y1; j++)
    for(int i = x0; i <= x1; i++)
        *p++ = pixels[j*w + i];


  XImage* im = XCreateImage(display, visual,
                                     x_depth,
                                     ZPixmap,      //format
                                     0,            //offset
                                     (char*)buf,
                                     width,
                                     height,
                                     32,     // bitmap_pad
                                     4*width // bytes_per_line
                                     );
  XPutImage(display,win,gc,im,0,0,x,y,width,height);
  XDestroyImage(im);
}
*/


void UPDATE_SCREEN_AREA(unsigned int* pixels, int w, int h, 
                     int x0, int y0, int x1, int y1,
                     int x, int y)
{ 
  // copy rectangle [x0,y0,x1,y1] from (w x h) pixel buffer
  // to canvas to position (x,y)

  if (x0 < 0) { x -= x0; x0 = 0; }
  if (y0 < 0) { y -= y0; y0 = 0; }

  if (x1 > w-1) x1 = w-1;
  if (y1 > h-1) y1 = h-1;

  if (x0 > x1 || y0 > y1) return;

  XImage* img = XCreateImage(display, visual,
                             x_depth,
                             ZPixmap,      //format
                             0,            //offset
                             (char*)pixels,
                             w,h,
                             32, // bitmap_pad
                             4*w // bytes_per_line
                             );

  int width = x1-x0+1;
  int height = y1-y0+1;

  XPutImage(display,win,gc,img,x0,y0,x,y,width,height);

  img->data = 0;
  XDestroyImage(img);
}



//------------------------------------------------------------------------------
// event handling
//------------------------------------------------------------------------------

XEvent event;

int CHECK_NEXT_EVENT(int& val, int& x, int& y, unsigned long& t, int msec)
{
  // msec > 0: blocking / timeout
  // msec = 0: blocking / no timeout
  // msec < 0: non-blocking (check)

  int e = no_event;

  bool blocking = (msec >= 0);

  if (msec > 0 && XPending(display) == 0)
  { int fd_disp = ConnectionNumber(display);

    timeval polltime;
    polltime.tv_sec  = msec / 1000;
    polltime.tv_usec = 1000 * (msec % 1000);

    fd_set rdset, wrset, xset;
    FD_ZERO(&rdset);
    FD_SET(fd_disp,&rdset);
    FD_ZERO(&wrset);
    FD_ZERO(&xset);
    FD_SET(fd_disp,&xset);

    int sel = select(fd_disp+1,&rdset,&wrset,&xset,&polltime);

    if (sel <= 0)
    { //w = 0;
      return no_event;
     }
  }


  if (blocking)
      XNextEvent(display, &event);
  else
  { if (XCheckMaskEvent(display,
                      EnterWindowMask | LeaveWindowMask    |
                      KeyPressMask    | PointerMotionMask  |
                      ButtonPressMask | ButtonReleaseMask  |
                      ExposureMask    | StructureNotifyMask, &event) == 0)
    return no_event;
   }


  switch (event.type) {

 case Expose:  e = exposure_event;
               x = event.xexpose.x;
               y = event.xexpose.y;
               val = event.xexpose.width;
               t = event.xexpose.height;
               break;


  case ConfigureNotify: e = configure_event;
                        x = event.xconfigure.x;
                        y = event.xconfigure.y;
                        //if (x == 0 && y == 0) x_window_to_screen(*w,x,y);
                        break;


  case DestroyNotify: //e = destroy_event;
                      break;



  case ButtonPress: val = event.xbutton.button;
                    x = event.xbutton.x;
                    y = event.xbutton.y;
                    t = event.xbutton.time;
                    e = button_press_event;
                    if (event.xbutton.state & Mod1Mask)    val = 2; 
                    if (event.xbutton.state & ShiftMask)   val |= 256; 
                    if (event.xbutton.state & ControlMask) val |= 512;
                    XUngrabPointer(display,CurrentTime);
                    break;

  case ButtonRelease: val = event.xbutton.button;
                      x = event.xbutton.x;
                      y = event.xbutton.y;
                      t = event.xbutton.time;
                      if (event.xbutton.state & Mod1Mask)    val = 2; 
                      if (event.xbutton.state & ShiftMask)   val |= 256; 
                      if (event.xbutton.state & ControlMask) val |= 512;
                      e = button_release_event;
                      break;

  case LeaveNotify:
  case EnterNotify:
  case MotionNotify: x = event.xmotion.x;
                     y = event.xmotion.y;
                     t = event.xbutton.time;
                     e = motion_event;
                     break;

  case KeyRelease:
  case KeyPress: { char c = 0;
                   KeySym keysym;
                   XComposeStatus status;
                   XLookupString((XKeyEvent*)&event,&c,1, &keysym, &status);

                   //val = (c) ? c : keysym;

                   switch (keysym) {
  
                     case XK_Print:     c = KEY_PRINT;
                                        break;
                     case XK_BackSpace: c = KEY_BACKSPACE;
                                        break;
                     case XK_Return:    c = KEY_RETURN;
                                        break;
                     case XK_Escape:    c = KEY_ESCAPE;
                                        break;
                     case XK_Left:      c = KEY_LEFT;
                                        break;
                     case XK_Right:     c = KEY_RIGHT;
                                        break;
                     case XK_Up:        c = KEY_UP;
                                        break;
                     case XK_Down:      c = KEY_DOWN;
                                        break;
                     case XK_Home:      c = KEY_HOME;
                                        break;
                     case XK_End:       c = KEY_END;
                                        break;
                     case XK_Tab:       c = KEY_TAB;
                                        break;
                     case XK_Insert:    c = KEY_INSERT;
                                        break;
                     case XK_Delete:    c = KEY_DELETE;
                                        break;
                     case XK_Page_Up:   c = KEY_PAGE_UP;
                                        break;
                     case XK_Page_Down: c = KEY_PAGE_DOWN;
                                        break;

                     case XK_F1:        c = KEY_F1;
                                        break;
                     case XK_F2:        c = KEY_F2;
                                        break;
                     case XK_F3:        c = KEY_F3;
                                        break;
                     case XK_F4:        c = KEY_F4;
                                        break;
                     case XK_F5:        c = KEY_F5;
                                        break;
                     case XK_F6:        c = KEY_F6;
                                        break;
                     case XK_F7:        c = KEY_F7;
                                        break;
                     case XK_F8:        c = KEY_F8;
                                        break;
                     case XK_F9:        c = KEY_F9;
                                        break;
                     case XK_F10:       c = KEY_F10;
                                        break;
  
                    }


                   val = c;

                   if (event.xkey.state & ShiftMask)   val |= 256;
                   if (event.xkey.state & ControlMask) val |= 512;
                   if (event.xkey.state & Mod1Mask)    val |= 1024; // alt

                   x = event.xmotion.x;
                   y = event.xmotion.y;
                   t = event.xkey.time;

                   if (event.type == KeyPress)
                     e = key_press_event;
                   else
                     e = key_release_event;

                   break;
                  }
   }


  //printf("x11:  e = %d  val = %d  x = %d   y = %d\n",e,val,x,y);

  return e;
}

//------------------------------------------------------------------------------
// mouse cursor
//------------------------------------------------------------------------------


void SET_MOUSE_CURSOR(int id)
{
  if (cursor) {
    XFreeCursor(display,cursor);
    cursor = 0;
  }

  if (id >= 0)
  { cursor = XCreateFontCursor(display,id);
    XDefineCursor(display,win,cursor);
   }
  else
    XUndefineCursor(display,win);

  XFlush(display);
}


LEDA_END_NAMESPACE
