/*******************************************************************************
+
+  LEDA 7.0  
+
+
+  _x_basic.c
+
+
+  Copyright (c) 1995-2023
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//------------------------------------------------------------------------------
//  basic graphics functions 
//------------------------------------------------------------------------------

#include <LEDA/system/basic.h>
/*
#include <LEDA/system/assert.h>
*/
#include <LEDA/core/string.h>
#include <LEDA/core/b_queue.h>

#include <LEDA/graphics/x_window.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <assert.h>

#include <LEDA/graphics/pixmaps/leda_icon.xpm>

#include "ttf/roman.64"
#include "ttf/fixed.36"
#include "ttf/bold.36"
#include "ttf/italic.36"

LEDA_BEGIN_NAMESPACE

#include <LEDA/graphics/x_basic.h>

#include "x_image.h"

typedef int Window;

typedef unsigned int pixel;

//--------------------------------------------------------------------------
// device dependent functions
//--------------------------------------------------------------------------

extern int INIT_GRAPHICS(int mode,int& width,int& height,int& depth,int& dpi);

extern void UPDATE_SCREEN_AREA(unsigned int* pixels, int w, int h, 
                             int x0, int y0, int x1, int y1, 
                             int x, int y); 
/*
   copy rectangle [x0,y0,x1,y1] from (w x h) pixel buffer (pixels)
   to screen at position (x,y)
*/



extern int CHECK_NEXT_EVENT(int& val, int& x, int& y,unsigned long& t,int msec);

/* 
  // get next event:
  // msec > 0: blocking / timeout
  // msec = 0: blocking / no timeout
  // msec < 0: non-blocking (check)
*/


extern void SET_MOUSE_CURSOR(int id);




//--------------------------------------------------------------------------

inline 
double AREA(double px, double py, double qx, double qy, double rx, double ry)
{ return ((px-qx) * (py-ry) - (py-qy) * (px-rx))/2; }

inline int SIGN(double x) {
   if (x > +0.5) return +1;
   if (x < -0.5) return -1;
   return 0;
}

inline bool TEST_SIGN(double x, double y) {
//const double eps = 0.1;
const double eps = 1.0;
   //if (fabs(x) < eps || fabs(y) < eps) return true;
   if (fabs(x) < eps || fabs(y) < eps) return false;
   return (x > 0 && y > 0) || (x < 0 && y < 0);
}


//------------------------------------------------------------------------------
// window
//------------------------------------------------------------------------------

struct gg_window
{
  int   width;
  int   height;
  int   xpos;
  int   ypos;
  int   x0;
  int   y0;
  int   x1;
  int   y1;

  int   save_width;
  int   save_height;
  int   save_xpos;
  int   save_ypos;
  int   save_x0;
  int   save_y0;
  int   save_x1;
  int   save_y1;


  int          COLOR;
  int          LINEWIDTH;
  line_style   LINESTYLE;
  drawing_mode MODE;
  text_mode    TEXTMODE;

  char         FONT_NAME[64];
  char         FONT_NAME_TMP[64];

  int            FONT_HEIGHT = 0;
  int*           FONT_WIDTH = 0;
  int*           FONT_OFFSET = 0;
  unsigned char* FONT_ALPHA = 0;
  float          FONT_SCALE = 0.8f;

  drawing_mode save_mode;
  line_style   save_ls;
  int          save_lw;
  int          save_clr;

  int   border_clr;
  int   border_w;
  int   cursor;

  int   bg_clr;
  int   save_bg_clr;
  int   label_clr;
  char  header[512];
  char  label[512];

  char* bg_pixmap;
  char* icon_pixmap;
  char  minimized;


  Window  id;

  void* inf;
  int   mapped;
  bool  flush;

  gg_window* pwin; // parent

  x_image* canvas;
  x_image* canvas_save;
  x_image* buffer;

  void  (*redraw)(void*,int,int,int,int,int);

/*
  ~gg_window() { cout << "gg_window destructor" << endl; }
*/
};


//------------------------------------------------------------------------------
// events
//------------------------------------------------------------------------------

struct gg_event {
  int ev;
  int val;
  int x;
  int y;
  gg_event() { ev = -1; val = x = y = 0; }
};

static gg_event event_buf;
static gg_event last_event;



//------------------------------------------------------------------------------
// display
//------------------------------------------------------------------------------

static int display_width = 0;
static int display_height = 0;
static int display_depth = 0;
static int display_dpi = 0;
static int display_fd = 0;

static bool do_not_open_display = false;

//------------------------------------------------------------------------------

// window manager

static int HEADER_W = 47;

static int BORDER_W = HEADER_W/4;
static int RESIZE_W = HEADER_W/3;
static int BUTTON_W = HEADER_W/3;
static int BUTTON_Y = int(0.36 * HEADER_W);
  
static int BUTTON_X[] = { int(3.7*HEADER_W),
                          int(2.5*HEADER_W), 
                          int(1.3*HEADER_W) };

void redraw_root();

//----------------------------------------------------------------------------


const int win_max = 256;

static gg_window* win_list[win_max];
static gg_window* win_stack[win_max];

static gg_window* root_win = 0;
static gg_window* grab_win = 0;

static int app_count = 0;
static int win_count = 0;
static int win_top = 0;

static int active_win = 0;

static int active_win_button = 0;


inline gg_window* set_draw_window(Window w) { return win_list[w]; }



/*
#if !defined(M_PI)
#define	M_PI   3.14159265358979323846
#endif
*/

template<class T>
inline void SWAP(T& a, T& b)
{ T t = a;
  a = b;
  b = t;
 }



void CopyArea(x_image* src_img, int src_x, int src_y, 
              x_image* dst_img, int dst_x, int dst_y, 
              int width, int height, bool use_alpha=false) 
{ 
  // copy/blend pixel rectangle from src_img  to dst_img

  x_image* root_canv = root_win->canvas;

  assert(src_img != root_canv);

/*
  // root canvas possible (if we are drawing window decorations)
  assert(dst_img != root_canv);
*/

  pixel* src = src_img->buf; 
  int src_w = src_img->w;
  int src_h = src_img->h;

  pixel* dst = dst_img->buf; 
  int dst_w = dst_img->w;
  int dst_h = dst_img->h;

  assert(src != dst);

  int dx = dst_x - src_x;
  int dy = dst_y - src_y;

  // src --> dst with offset (dx,dy)

  int x0 = src_x;
  int y0 = src_y;
  int x1 = src_x + width;
  int y1 = src_y + height;

  // clip to src

  if (x0 < 0) x0 = 0;
  if (y0 < 0) y0 = 0;
  if (x1 >= src_w) x1 = src_w-1;
  if (y1 >= src_h) y1 = src_h-1;

  int clip_x0 = dst_img->clip_x0;
  int clip_y0 = dst_img->clip_y0;
  int clip_x1 = dst_img->clip_x1;
  int clip_y1 = dst_img->clip_y1;

  for(int src_y = y0; src_y <= y1; src_y++)
  { for(int src_x = x0; src_x <= x1; src_x++) 
    { int dst_x = src_x + dx;
      int dst_y = src_y + dy;

      if (dst_x < clip_x0 || dst_x > clip_x1 || 
          dst_y < clip_y0 || dst_y > clip_y1) continue;

      pixel* p = src + src_x + src_y*src_w;
      pixel* q = dst + dst_x + dst_y*dst_w;

      *q  = x_image::blend_pixels(*q,*p,use_alpha);
    }
  }

}


void CopyWindowAreaToRoot(gg_window* win, int x0, int y0, int x1, int y1,
                          int& vis_x0, int& vis_y0, int& vis_x1, int& vis_y1)
{ 
  // copy pixel rectangle from src_img rectangle to root canvas
  // pixels obscured by windows above win in win_stack are not copied
  // bounding box of visible pixels is returned [vis_x0,vis_y0,vis_x1,vis_y0]
  // (src coordinates)

  // src --> dst: add dx = win->xpos and  dy = win->ypos to coords

  int stack_p = win->mapped;

  x_image* win_canvas = win->canvas;
  pixel* src = win_canvas->buf; 
  int src_w = win_canvas->w;
  int src_h = win_canvas->h;

  x_image* root_canvas = root_win->canvas;
  pixel* dst = root_canvas->buf; 
  int dst_w = root_canvas->w;
  int dst_h = root_canvas->h;

  int dx = win->xpos;
  int dy = win->ypos;

  assert(src != dst);

  // clip to src

  if (x0 < 0) x0 = 0;
  if (y0 < 0) y0 = 0;
  if (x1 >= src_w) x1 = src_w-1;
  if (y1 >= src_h) y1 = src_h-1;

  vis_x0 = x1; 
  vis_y0 = y1;
  vis_x1 = x0;
  vis_y1 = y0;


  for(int src_y = y0; src_y <= y1; src_y++)
  { for(int src_x = x0; src_x <= x1; src_x++) 
    { int dst_x = src_x + dx;
      int dst_y = src_y + dy;

      if (dst_x < 0 || dst_x >= dst_w || dst_y < 0 || dst_y >= dst_h) continue;

      bool visible = true ;
      if (stack_p > 0) 
      { int k = stack_p + 1;
        while (visible && k <= win_top) 
        { gg_window* wp = win_stack[k++];
          if (dst_x >= wp->x0 && dst_x <= wp->x1 && 
              dst_y >= wp->y0 && dst_y <= wp->y1) visible = false;
        }
      }

      if (visible) 
      { if (src_x < vis_x0) vis_x0 = src_x;
        if (src_x > vis_x1) vis_x1 = src_x;
        if (src_y < vis_y0) vis_y0 = src_y;
        if (src_y > vis_y1) vis_y1 = src_y;
        dst[dst_x + dst_y*dst_w] = src[src_x + src_y*src_w];
      }

    }
  }

}

void CopyWindowAreaToRoot(gg_window* win, int x0, int y0, int x1, int y1)
{ int vis_x0 = 0;
  int vis_y0 = 0;
  int vis_x1 = 0;
  int vis_y1 = 0;
  CopyWindowAreaToRoot(win,x0,y0,x1,y1,vis_x0,vis_y0,vis_x1,vis_y1);
}




void SetAlpha(x_image* img, int x0, int y0, int x1, int y1, int alpha, int op) 
{ 
  // op = src_mode
  // op = or_mode
  // op = and_mode

  int w = img->w;
  int h = img->h;

  if (x0 < 0) x0 = 0;
  if (y0 < 0) y0 = 0;
  if (x1 >= w) x1 = w-1;
  if (y1 >= h) y1 = h-1;

  int aa = (alpha << 24);

  for(int y = y0; y <= y1; y++)
    for(int x = x0; x <= x1; x++) 
    {  pixel* q = img->buf + y*w + x; 
       int rgb = *q & 0xffffff;
       switch (op) {
         case src_mode: *q = (aa | rgb);
                        break;
         case or_mode:  *q |= aa;
                        break;
         case and_mode: *q &= (aa | 0xffffff);
                        break;
       }
     }
 }


void FlushDisplay(gg_window* win, int x0, int y0,int x1, int y1)
{ 
  if (win->flush == false) return;

  x_image* canv = win->canvas;
  x_image* root_canv = root_win->canvas;

/*
// clipping
  if (x0 < canv->clip_x0) x0 = canv->clip_x0;
  if (y0 < canv->clip_y0) y0 = canv->clip_y0;
  if (x1 > canv->clip_x1) x1 = canv->clip_x1;
  if (y1 > canv->clip_y1) y1 = canv->clip_y1;
*/

  if (win != root_win) {
     CopyWindowAreaToRoot(win,x0,y0,x1,y1, x0,y0,x1,y1);
  }

  x0 += win->xpos;
  y0 += win->ypos;
  x1 += win->xpos;
  y1 += win->ypos;

  //cout << string("%4d %4d %4d %4d",x0,y0,x1,y1) << endl;

  UPDATE_SCREEN_AREA(root_canv->buf,root_canv->w,root_canv->h,
                   x0,y0,x1,y1,x0,y0);
}


void FillArea(x_image* dst_image, int dst_x, int dst_y, 
                                   int width,  int height, pixel clr) 
{ 
  pixel* dst = dst_image->buf; 
  int dst_w = dst_image->w;
  //int dst_h = dst_image->h;

  int x0 = dst_x;
  int y0 = dst_y;
  int x1 = dst_x + width-1;
  int y1 = dst_y + height-1;

  // clip to dst

  if (x0 < dst_image->clip_x0) x0 = dst_image->clip_x0;
  if (y0 < dst_image->clip_y0) y0 = dst_image->clip_y0;
  if (x1 > dst_image->clip_x1) x1 = dst_image->clip_x1;
  if (y1 > dst_image->clip_y1) y1 = dst_image->clip_y1;

  for(int y = y0; y < y1; y++)
  { pixel* q = dst + x0 + y*dst_w;
    for(int x = x0; x < x1; x++) *q++ = clr;
   }
}


const char* x_display_info(int& width, int& height, int& dpi)  
{ 
  x_open_display();

  width = display_width;
  height = display_height;
  dpi = display_dpi;
  return "xx"; 
} 



int x_window_bits_saved(int) { return 1; }


// fonts

int x_set_font(Window w, const char* font_name) 
{ gg_window* win = win_list[w];

  strcpy(win->FONT_NAME,font_name);

  float sz = atoi(font_name+1);

  // adjust sz to dpi resolution

  //sz *= (display_dpi/140.0);

  sz *= (display_dpi/180.0);


  switch(font_name[0]) {

     case 'T': win->FONT_WIDTH  = ROMAN_FONT_WIDTH;
               win->FONT_HEIGHT = ROMAN_FONT_HEIGHT;
               win->FONT_OFFSET = ROMAN_FONT_OFFSET;
               win->FONT_ALPHA  = ROMAN_FONT_ALPHA;
               win->FONT_SCALE  = sz/ROMAN_FONT_HEIGHT;
               break;
    
     case 'F': sz *= 0.8;
               win->FONT_WIDTH  = FIXED_FONT_WIDTH;
               win->FONT_HEIGHT = FIXED_FONT_HEIGHT;
               win->FONT_OFFSET = FIXED_FONT_OFFSET;
               win->FONT_ALPHA  = FIXED_FONT_ALPHA;
               win->FONT_SCALE  = sz/FIXED_FONT_HEIGHT;
               break;
    
     case 'B': win->FONT_WIDTH  = BOLD_FONT_WIDTH;
               win->FONT_HEIGHT = BOLD_FONT_HEIGHT;
               win->FONT_OFFSET = BOLD_FONT_OFFSET;
               win->FONT_ALPHA  = BOLD_FONT_ALPHA;
               win->FONT_SCALE  = sz/BOLD_FONT_HEIGHT;
               break;
    
     case 'I': win->FONT_WIDTH  = ITALIC_FONT_WIDTH;
               win->FONT_HEIGHT = ITALIC_FONT_HEIGHT;
               win->FONT_OFFSET = ITALIC_FONT_OFFSET;
               win->FONT_ALPHA  = ITALIC_FONT_ALPHA;
               win->FONT_SCALE  = sz/ITALIC_FONT_HEIGHT;
               break;
    }

  return 1;
}


void x_save_font(Window w) 
{ gg_window* win = win_list[w];
  strcpy(win->FONT_NAME_TMP,win->FONT_NAME);
}

void x_restore_font(Window w) 
{ gg_window* win = win_list[w];
  char tmp[64];
  strcpy(tmp,win->FONT_NAME_TMP);
  x_set_font(w,tmp);
 }


// predefined fonts

/*
void x_set_text_font(Window w)   { x_set_font(w,"T27"); }
void x_set_italic_font(Window w) { x_set_font(w,"I27"); }
void x_set_bold_font(Window w)   { x_set_font(w,"B28"); }
void x_set_fixed_font(Window w)  { x_set_font(w,"F22"); }
void x_set_button_font(Window w) { x_set_font(w,"T28"); }
*/

void x_set_text_font(Window w)   { x_set_font(w,"T35"); }
void x_set_italic_font(Window w) { x_set_font(w,"I35"); }
void x_set_bold_font(Window w)   { x_set_font(w,"B34"); }
void x_set_fixed_font(Window w)  { x_set_font(w,"F35"); }
void x_set_button_font(Window w) { x_set_font(w,"T35"); }



int x_set_cursor(Window w, int c) 
{ int c_old = win_list[w]->cursor;
  win_list[w]->cursor  = c;
  if (w == active_win) SET_MOUSE_CURSOR(c);
  return c_old;
}

int x_get_cursor(Window w) { return win_list[w]->cursor; }


drawing_mode x_get_mode(Window w)       { return win_list[w]->MODE; }
text_mode    x_get_text_mode(Window w)  { return win_list[w]->TEXTMODE; }
line_style   x_get_line_style(Window w) { return win_list[w]->LINESTYLE; }

int x_get_line_width(Window w)   { return win_list[w]->LINEWIDTH; }
int x_get_color(Window w)        { return win_list[w]->COLOR; }
int x_get_border_color(Window w) { return win_list[w]->border_clr; }
int x_get_border_width(Window w) { return win_list[w]->border_w; }


int x_set_line_width(Window w, int width)
{ gg_window* win = win_list[w];
  int save = win->LINEWIDTH;
  win->LINEWIDTH = width;
  return save;
}

text_mode x_set_text_mode(Window w, text_mode mode)
{ gg_window* win = win_list[w];
  text_mode save = win->TEXTMODE;
  win->TEXTMODE = mode;
  return save;
}

line_style x_set_line_style(Window w, line_style style)
{ gg_window* win = win_list[w];
  line_style save = win->LINESTYLE;
  win->LINESTYLE = style;
  return save;
}

int x_set_color(Window w, int clr)
{ gg_window* win = win_list[w];
  int save = win->COLOR;
  //win->COLOR = clr;
  win->COLOR = clr & 0xffffff;
  return save;
 }

drawing_mode x_set_mode(Window w, drawing_mode mode)
{ gg_window* win = win_list[w];
  drawing_mode save = win->MODE;
  win->MODE = mode;
  return save;
}

int x_text_height(Window w, const char*)
{ gg_window* win = win_list[w];
  int h = int(0.5f + win->FONT_HEIGHT * win->FONT_SCALE);
  if (win->FONT_WIDTH == FIXED_FONT_WIDTH) h += 5;
  return h;
}

int x_text_width(Window w, const char* s) { 
  return x_text_width(w,s,(int)strlen(s));
}

int  x_text_width(Window w, const char* s, int len)
{ gg_window* win = win_list[w];
  if (len > int(strlen(s))) len = int(strlen(s));
  float f = win->FONT_SCALE;
  int width = 0;
  for(int i = 0; i < len; i++) { 
    int c = s[i] & 127;
    width += int(0.5 + f*win->FONT_WIDTH[c]);
  }
  return width;
}


void x_get_rgb(int col, int& r, int& g, int& b)
{ if (col == invisible)
  { r = -1;
    g = -1;
    b = -1;
   }
  else
  { b = col & 0xff;
    g = (col >> 8) & 0xff;
    r = (col >> 16) & 0xff;
  }
 }


int x_new_color(int r, int g, int b) {
  if (r == -1 && g == -1 && b == -1) return invisible;
  return (r<<16) + (g<<8) + b;
}



inline void bigpix(gg_window* win, int x, int y)
{ x_image* canvas = win->canvas;
  pixel clr = win->COLOR;
  int mode = win->MODE;
  int lw = win->LINEWIDTH;

  for(int i=0; i<lw; i++)
  { int dx = (i%2) ? -(i+1)/2 : +(i+1)/2;
    for(int j=0; j<lw; j++)
    { int dy = (j%2) ? -(j+1)/2 : +(j+1)/2;
      canvas->set_pixel(x+dx,y+dy,clr,mode);
     }
   }
}

inline void bigpix1(gg_window* win, int x, int y)
{ x_image* canvas = win->canvas;

  int lw = win->LINEWIDTH;
  int clr = win->COLOR;
  int mode = win->MODE;

  int w = lw + (1 - lw % 2);

  int x0 = x - w/2; 
  int y0 = y - w/2; 

  for(int i=0; i<w; i++)
  { for(int j=0; j<w; j++)
    { int alpha = 255;
      if (w > lw) {
/*
        if (i == 0 || i == w-1) alpha /= 2;
        if (j == 0 || j == w-1) alpha /= 2;
*/
        if (i == 0 || i == w-1) alpha /= 4;
        if (j == 0 || j == w-1) alpha /= 4;
      }
      clr |= (alpha << 24);
      canvas->blend_pixel(x0+i,y0+j,clr,mode);
    }
   }
}




static void insert_bit_line(gg_window* win, int x, int y, unsigned char* lp,
                                                      int len)
{ unsigned char* stop = lp+len;
  while (lp < stop)
  { unsigned char c = *lp++;
    for(int i=7; i>=0; i--)
    { if (c & (1<<i)) win->canvas->set_pixel(x,y,win->COLOR,win->MODE);
      x++;
     }
   }
}


static void hline(gg_window* win, int x0, int x1, int y)
{ 
  // 1-pixel horizontal line

  if (x0 > x1) SWAP(x0,x1);

  x_image* canv = win->canvas;
  pixel clr = win->COLOR;
  int mode = win->MODE;

  // clip

  if (y  < canv->clip_y0 || y  > canv->clip_y1 || 
      x1 < canv->clip_x0 || x0 > canv->clip_x1) return;

  if (x0 < canv->clip_x0) x0 = canv->clip_x0;
  if (x1 > canv->clip_x1) x1 = canv->clip_x1;

  for(int x = x0; x<=x1; x++) canv->set_pixel(x,y,clr,mode);

  //for(int x = x0; x<=x1; x++) canv->blend_pixel(x,y,clr|0x33000000,mode);
}


static void vline(gg_window* win,int x, int y0, int y1)
{ 
  // 1-pixel vertical line

  if (y0 > y1) SWAP(y0,y1);

  x_image* canv = win->canvas;
  pixel clr = win->COLOR;
  int mode = win->MODE;

  // clip

  if (x < canv->clip_x0 || x > canv->clip_x1 || 
      y1 < canv->clip_y0 || y0 > canv->clip_y1) return;

  if (y0 < canv->clip_y0) y0 = canv->clip_y0;
  if (y1 > canv->clip_y1) y1 = canv->clip_y1;

  for(int y = y0; y<=y1; y++) canv->set_pixel(x,y,clr,mode);
}


//-----------------------------------------------------------------------------
// Wu's Line Algorithm (anti-aliasing)
//-----------------------------------------------------------------------------

inline void plot(gg_window* win, int x, int y, pixel clr, double bright) {

bigpix(win,x,y);

/*
  int a = int(bright*255);
  clr |= (a << 24);
  win->canvas->blend_pixel(x,y,clr,src_mode);
*/
}


inline int ipart(double x) {
  // integer part of x
  return floor(x);
}

inline double fpart(double x) {
  // fractional part of x
  return x - floor(x);
}

inline double rfpart(double x) {
  // 1 - fractional part
  return 1 - fpart(x);
}

void WuDrawLine(gg_window* win, int xx0, int yy0, int xx1, int yy1) 
{
  x_image* canv = win->canvas;

  pixel clr = win->COLOR;

  double x0 = xx0;
  double y0 = yy0;
  double x1 = xx1;
  double y1 = yy1;

  bool steep = abs(y1-y0) > abs(x1-x0);
  
  if (steep)  {
    SWAP(x0, y0);
    SWAP(x1, y1);
  }

  if (x0 > x1){
    SWAP(x0, x1);
    SWAP(y0, y1);
  }
  
  double dx = x1 - x0;
  double dy = y1 - y0;

  double gradient = (dx == 0) ? 1.0 : dy/dx;

  // handle first endpoint

  double xend = round(x0);
  double yend = y0 + gradient * (xend - x0);
  double xgap = rfpart(x0 + 0.5);
  double xpxl1 = xend; // this will be used in the main loop
  double ypxl1 = ipart(yend);

  if (steep) {
     plot(win,ypxl1,   xpxl1, clr, rfpart(yend) * xgap);
     plot(win,ypxl1+1, xpxl1, clr,  fpart(yend) * xgap);
  }
  else {
      plot(win,xpxl1, ypxl1  ,clr,  rfpart(yend) * xgap);
      plot(win,xpxl1, ypxl1+1,clr,   fpart(yend) * xgap);
  }

  double intery = yend + gradient; // first y-intersection for the main loop
  
  // handle second endpoint

  xend = round(x1);
  yend = y1 + gradient * (xend - x1);
  xgap = fpart(x1 + 0.5);

  double xpxl2 = xend; //this will be used in the main loop
  double ypxl2 = ipart(yend);

  if (steep) {
      plot(win,ypxl2  , xpxl2,clr,  rfpart(yend) * xgap);
      plot(win,ypxl2+1, xpxl2,clr,   fpart(yend) * xgap);
  }
  else {
      plot(win,xpxl2, ypxl2,  clr, rfpart(yend) * xgap);
      plot(win,xpxl2, ypxl2+1,clr,  fpart(yend) * xgap);
  }
  
  // main loop

  if (steep) 
    for (int x = xpxl1 + 1; x < xpxl2 - 1; x++)
    { plot(win,ipart(intery)  , x, clr, rfpart(intery));
      plot(win,ipart(intery)+1, x, clr,  fpart(intery));
      intery = intery + gradient;
     }
  else
    for (int x = xpxl1 + 1; x < xpxl2 - 1; x++)
    { plot(win,x, ipart(intery),  clr, rfpart(intery));
      plot(win,x, ipart(intery)+1,clr,  fpart(intery));
      intery = intery + gradient;
    }

  int lw = win->LINEWIDTH;
  if (xx0 > xx1) SWAP(xx0,xx1);
  if (yy0 > yy1) SWAP(yy0,yy1);
  FlushDisplay(win,xx0-lw,yy0-lw,xx1+lw,yy1+lw);
}




static void extend_line(int s, int& x1, int& y1, int& x2, int& y2)
{ // extend line by s pixels

  int dx = x2 - x1;
  int dy = y2 - y1;
  if (dx == 0 && dy == 0) return;

  int xoff = s;
  int yoff = s;

  if (dx < 0) { dx = -dx; xoff = -s; }
  if (dy < 0) { dy = -dy; yoff = -s; }

  if ( dx >= dy) x2 += xoff;
  if ( dx <= dy) y2 += yoff;
 }



void x_line0(Window w, int x1, int y1, int x2, int y2) { 
   x_line(w,x1,y1,x2,y2); 
}



void x_line(Window w, int x1, int y1, int x2, int y2)
{ 
  gg_window* win = set_draw_window(w);

  if (x1 > x2) { SWAP(x1,x2); SWAP(y1,y2); }

  int lw = win->LINEWIDTH;

/*
  int lw2 = win->LINEWIDTH/2;
  extend_line(lw2,x1,y1,x2,y2);
  extend_line(lw2,x2,y2,x1,y1);
*/

  int sx = 1;
  int sy = 1;

  int dx = x2 - x1;
  int dy = y2 - y1;

  if (dy < 0)
  { dy = -dy;
    sy = -1;
   }
 
  int x = x1;
  int y = y1;

  if (dx > dy)
  { int c = dx / 2;
    bigpix(win,x,y);
    while(x != x2)
    { x += sx;
      if ((c+=dy) >= dx)
      { c -= dx;
        y += sy;
       }
      bigpix(win,x,y);
    }
  }
  else  
  { // dy >= dx
    int c = dy / 2;
    bigpix(win,x,y);
    while(y != y2)
    { y += sy;
      if ((c+=dx) >= dy)
      { c -= dy;
        x += sx;
       }
      bigpix(win,x,y);
    }
  }

  if (y1 > y2) SWAP(y1,y2);

  FlushDisplay(win,x1-lw,y1-lw,x2+lw,y2+lw);
}



void x_line_old(Window w, int x1, int y1, int x2, int y2)
{ 
  gg_window* win = set_draw_window(w);

  x_image* canv = win->canvas;

  int xmin = (x1 < x2) ? x1 : x2;
  int xmax = (x1 < x2) ? x2 : x1;
  int ymin = (y1 < y2) ? y1 : y2;
  int ymax = (y1 < y2) ? y2 : y1;

  int lw = win->LINEWIDTH;

/*
  int lw2 = win->LINEWIDTH/2;
  extend_line(lw2,x1,y1,x2,y2);
  extend_line(lw2,x2,y2,x1,y1);
*/

/*

  if (x1 == x2)
  { for(int i=0; i<lw; i++)
    { int d = (i%2) ? -(i+1)/2 : +(i+1)/2;
      vline(win,x1+d,y1,y2);
     }
    FlushDisplay(win,xmin-lw,ymin-lw,xmax+lw,ymax+lw);
    return;
   }

  if (y1 == y2)
  { for(int i=0; i<lw; i++)
    { int d = (i%2) ? -(i+1)/2 : +(i+1)/2;
      hline(win,x1,x2,y1+d);
     }
    FlushDisplay(win,xmin-lw,ymin-lw,xmax+lw,ymax+lw);
    return;
   }
*/

  // draw line

  int sy = 1;
  int dx = x2 - x1;
  int dy = y2 - y1;

  if (dx < 0)
  { dx = -dx;
    SWAP(x1,x2);
    dy = -dy;
    SWAP(y1,y2);
   }

  if (dy < 0)
  { dy = -dy;
    sy = -1;
   }


/*
  // clip line into window

  // clipping done by x_image::setpixel

 // if (dx > dy)
    { if (x1 < canv->clip_x0)
      { int d = canv->clip_x0 - x1;
        x1 += d;
        y1 += (d*sy*dy)/dx;
       }
      if (x2 > canv->clip_x1)
      { int d = x2 - canv->clip_x1;
        x2 -= d;
        y2 -= (d*sy*dy)/dx;
       }
     }
 // else  // dy > dx
    {
      if (y1 < canv->clip_y0)
      { int d = canv->clip_y0 - y1;
        y1 += d;
        x1 += (d*dx)/(sy*dy);
       }

      if (y2 < canv->clip_y0)
      { int d = canv->clip_y0 - y2;
        y2 += d;
        x2 += (d*dx)/(sy*dy);
       }

      if (y1 > canv->clip_y1)
      { int d = y1 - canv->clip_y1;
        y1 -= d;
        x1 -= (d*dx)/(sy*dy);
       }

      if (y2 > canv->clip_y1)
      { int d = y2 - canv->clip_y1;
        y2 -= d;
        x2 -= (d*dx)/(sy*dy);
       }
    
     }

  dx = x2 - x1;
  dy = sy*(y2 - y1);
  if (dx < 0 || dy < 0) return;
*/


  if (dx > dy)
  { int c = dx / 2;
    bigpix(win,x1,y1);
    while(x1 != x2)
    { x1++;
      if ((c+=dy) > dx)
      { c  -= dx;
        y1 += sy;
       }
      bigpix(win,x1,y1);
    }
  }
  else  // dy >= dx
  { int c = dy / 2;
    bigpix(win,x1,y1);
    while(y1 != y2)
    { y1 += sy;
      if ((c+=dx) >= dy)
      { c -= dy;
        x1++;
       }
      bigpix(win,x1,y1);
    }
  }

  FlushDisplay(win,xmin-lw,ymin-lw,xmax+lw,ymax+lw);
}


void x_lines(Window w, int n, int* x1, int* y1, int* x2, int* y2) 
{ gg_window* win = set_draw_window(w);

  if (n <= 0) return;

  bool win_flush  = win->flush;

  win->flush = false;

  int x_min = x1[0];
  int x_max = x1[0];
  int y_min = y1[0];
  int y_max = y1[0];

  for (int i=0; i<n; i++)
  { x_line(w,x1[i],y1[i],x2[i],y2[i]); 

   if (x1[i] < x_min) x_min = x1[i];
   if (x1[i] > x_max) x_max = x1[i];
   if (x2[i] < x_min) x_min = x2[i];
   if (x2[i] > x_max) x_max = x2[i];

   if (y1[i] < y_min) y_min = y1[i];
   if (y1[i] > y_max) y_max = y1[i];
   if (y2[i] < y_min) y_min = y2[i];
   if (y2[i] > y_max) y_max = y2[i];
  }

  int lw = win->LINEWIDTH;
  FlushDisplay(win,x_min-lw,y_min-lw,x_max+lw,y_max+lw);

  win->flush = win_flush;
}


static void clip_bitmap(x_image* canv, int x, int y, unsigned char* data,
                                                int width, int height)
{ int y1 = y + height - 1;
  int x1 = x + 8*width - 1;

  int left_margin = -1;
  int top_margin  = -1;
  int right_margin  = width;
  int bottom_margin = height;

  unsigned char mask1 = 0xFF;
  unsigned char mask2 = 0xFF;

  if (y  < canv->clip_y0) top_margin    += canv->clip_y0 - y;
  if (y1 > canv->clip_y1) bottom_margin -= y1 - canv->clip_y1;

  if (x  < canv->clip_x0)  
  { left_margin  = canv->clip_x0 - x;
    mask1 = (0xFF >> (left_margin%8));
    left_margin /= 8;
   }

  if (x1 > canv->clip_x1) 
  { right_margin = (x1 - canv->clip_x1);
    mask2 = ~(0xFF >> (right_margin%8));
    right_margin = width - right_margin/8 - 1;
   }

  
  for (int i = 0; i < height; i++)
  { unsigned char* p = data + i*width;
    if (i < top_margin || i > bottom_margin)
       { unsigned char* stop = p + width;
         while (p < stop) *p++ = 0;
        }
    else
      for(int j=0; j < width; j++)
      { if (j < left_margin || j > right_margin) *p=0;
        if (j == left_margin) *p &= mask1;
        if (j == right_margin) *p &= mask2;
        p++;
       }
   }
}

void x_polyline(Window w, int n, int* xcoord, int* ycoord, int adjust)
{ int x = xcoord[0];
  int y = ycoord[0];
  for(int i = 1; i<n; i++)
  { int x1 = xcoord[i];
    int y1 = ycoord[i];
    x_line(w,x,y,x1,y1);
    x = x1;
    y = y1;
   }
}


void x_polygon(Window w, int n, int* xcoord, int* ycoord)
{ int x = xcoord[n-1];
  int y = ycoord[n-1];
  for(int i = 0; i<n; i++)
  { int x1 = xcoord[i];
    int y1 = ycoord[i];
    x_line(w,x,y,x1,y1);
    x = x1;
    y = y1;
   }
}
  


static void flood_fill(x_image* canv, int x, int y, int x_min, int y_min, 
                       int x_max, int y_max, pixel clr)
{ 

  int width = canv->w;

  if (x_min < canv->clip_x0+1) x_min = canv->clip_x0+1;
  if (x_max > canv->clip_x1-1) x_max = canv->clip_x1-1;
  if (y_min < canv->clip_y0+1) y_min = canv->clip_y0+1;
  if (y_max > canv->clip_y1-1) y_max = canv->clip_y1-1;

  if (x < x_min || x > x_max || y < y_min || y > y_max) return;


  pixel* buf = canv->buf;
  pixel* pos = buf + y*width + x;

  b_queue<pixel*> queue(16384);

  //if (*pos != clr)
  if ((*pos & 0x80000000) == 0)
  { *pos = clr;
     queue.append(pos);
   }

  int offset[4] = { -1, +1, -width, +width };

  int max_qsize = 0;

  while (!queue.empty())
  { int sz = queue.size();
    if (sz > max_qsize) max_qsize = sz;
    pixel* p = queue.pop();

    if (sz > queue.max_size() - 4) continue;

    for (int i=0; i<4; i++)
    { pixel* q = p + offset[i];
      int xx = (q-buf) % width;
      int yy = (q-buf) / width;
      if (xx < x_min || xx > x_max || yy < y_min || yy > y_max) continue;
      //if (*q != clr)

      if ((*q & 0x80000000) == 0) queue.append(q); 

      if (clr & 0x01000000) // clip mask bit
        *q = (*q & 0x00ffffff) | (clr & 0xff000000);
      else
        *q = clr; 
    }
  }

  if (max_qsize > queue.max_size()-10) { 
    cout << "max_qsize = " << max_qsize << endl;
  }
}



void x_fill_polygon(Window w, int n, int* xcoord, int* ycoord)
{
  gg_window* win = set_draw_window(w);

  //double A_eps = 1.0;
  double A_eps = 10;

  // bounding box (x_min,y_min,x_max,y_max)

  int x_min = xcoord[0];
  int x_max = xcoord[0];
  int y_min = ycoord[0];
  int y_max = ycoord[0];

  for(int i=1;i<n;i++)
  { if (xcoord[i] < x_min) x_min = xcoord[i];
    if (xcoord[i] > x_max) x_max = xcoord[i];
    if (ycoord[i] < y_min) y_min = ycoord[i];
    if (ycoord[i] > y_max) y_max = ycoord[i];
   }

  if (n < 3 || x_min == x_max || y_min == y_max) { 
    // degenerate
    return;
   }


  // check convexity or find convex corner

  int last_a = 0;
  double A = 0;
  for(int i=1;i<n-1;i++) {
    // total area (sign gives orientation)
    A += AREA(xcoord[0],ycoord[0],xcoord[i],ycoord[i],xcoord[i+1],ycoord[i+1]);
  }

  //cout << "AREA = " << A << endl;

  if (fabs(A) < A_eps) 
  { // draw boundary
    for(int i=0; i<n; i++)
      x_line(w,xcoord[i],ycoord[i],xcoord[(i+1)%n],ycoord[(i+1)%n]);
    return;
  }


  // draw outline of polygon with 0x80 bit set
  // and flood fill from an interior point

  pixel win_clr = win->COLOR;
  int win_lw = win->LINEWIDTH;
  drawing_mode win_mode = win->MODE;

  win->LINEWIDTH = 1;
  win->COLOR = 0x80000000;
  win->MODE = or_mode;

  for(int i=0; i<n; i++)
    x_line(w,xcoord[i],ycoord[i],xcoord[(i+1)%n],ycoord[(i+1)%n]);


  bool convex = true;
  int convex_i = 0;

  if (n > 3)
  { // check convexity / find a convex corner with maximal triangle area
    double triang_A = 0;
    for(int i=0;i<n;i++)
    { int j = (i+1) % n; 
      int k = (j+1) % n; 
      double a=AREA(xcoord[i],ycoord[i],xcoord[j],ycoord[j],xcoord[k],ycoord[k]);
      if (fabs(a) < A_eps) continue;

      //if ((a < 0 && A < 0) || (a > 0 && A > 0))
      if (TEST_SIGN(a,A))
      { // convex corner
        if (fabs(a) > triang_A) {
          convex_i = i;
          triang_A = fabs(a);
        }
      }
      else
        convex = false;
    }
  }


  // interior point
  int int_x = xcoord[0];
  int int_y = ycoord[0];

  if (convex)
  { //cout << "CONVEX POLYGON" << endl;
    // use center of gravity
    double sum_x = 0;
    double sum_y = 0;
    for(int i=0; i<n; i++)
    { sum_x += xcoord[i];
      sum_y += ycoord[i];
     }
    int_x = int(0.5 + sum_x/n);
    int_y = int(0.5 + sum_y/n);

/*
    // check interior point
    for(int i=1;i<n-1;i++)
    { int j = (i+1) % n; 
      double a = AREA(xcoord[i],ycoord[i],xcoord[j],ycoord[j],int_x,int_y);
      if (!TEST_SIGN(a,A)) {
        cout << "a = " << a << " A = " << A << endl;
        return;
      }
    }
*/

  }
  else
  {  //cout << "NON-CONVEX POLYGON" << endl;
    int i = convex_i;
    int j = (i+1) % n; 
    int k = (j+1) % n; 

    // find interior point in triangle (i,j,k) close to j
/*
    double sum_x = xcoord[i] + 2*xcoord[j] + xcoord[k];
    double sum_y = ycoord[i] + 2*ycoord[j] + ycoord[k];
    int_x = int(0.5 + sum_x/4);
    int_y = int(0.5 + sum_y/4);
*/

    double gx = (xcoord[i] + xcoord[j] + xcoord[k])/3;
    double gy = (ycoord[i] + ycoord[j] + ycoord[k])/3;

    // vector (dx,dy) from p_j towards center of gravity (gx,gy)

    double dx = (gx - xcoord[j])/100;
    double dy = (gy - ycoord[j])/100;

    for(int t=1; t<100; t++)
    { int_x = int(0.5 + xcoord[j] + t*dx);
      int_y = int(0.5 + ycoord[j] + t*dy);
      pixel c = win->canvas->get_pixel(int_x,int_y);
      if ((c & 0x80000000) == 0) break;
     }
  }


  // flood fill from (int_x,int_y)

  pixel fill_clr = win_clr | 0x80000000;
  flood_fill(win->canvas,int_x,int_y,x_min,y_min,x_max,y_max,fill_clr);

  // reset fill poly bit
  SetAlpha(win->canvas, x_min,y_min,x_max,y_max,~0x80,and_mode);

  win->LINEWIDTH = win_lw;
  win->COLOR = win_clr;
  win->MODE = win_mode;


  FlushDisplay(win,x_min,y_min,x_max,y_max);
}


void x_box(Window w, int x0, int y0, int x1, int y1)
{ gg_window* win = set_draw_window(w);

  if (x0 > x1) SWAP(x0,x1);
  if (y0 > y1) SWAP(y0,y1);

  x_image* canv = win->canvas;

  // clipping
  if (x0 < canv->clip_x0) x0 = canv->clip_x0;
  if (x1 > canv->clip_x1) x1 = canv->clip_x1;
  if (y0 < canv->clip_y0) y0 = canv->clip_y0;
  if (y1 > canv->clip_y1) y1 = canv->clip_y1;

  if (x0 <= x1 && y0 <= y1) 
    for(int y=y0; y<=y1; y++) hline(win,x0,x1,y);

  FlushDisplay(win,x0,y0,x1,y1);
}

void x_clear_window(Window w, int x0, int y0, int x1, int y1, int xorig, 
                                                               int yorig)
{ gg_window* win = set_draw_window(w);

  //if (win->mapped == 0) return;
  if (win->canvas == 0) return; // not displayed
  

  char* pm = win->bg_pixmap;

  if (pm == 0)
  { int c = x_set_color(w,win->bg_clr);
    x_box(w,x0,y0,x1,y1);
    x_set_color(w,c);
    return;
   }

  x_image* im = (x_image*)pm;
  int wi = im->w;
  int he = im->h;

  if (xorig > 0)
     while (xorig > 0) xorig -= wi;
  else
     while (xorig+wi < 0) xorig += wi;

  if (yorig > 0)
     while (yorig > 0) yorig -= he;
  else
     while (yorig+he < 0) yorig += he;

  int xmax = win->canvas->w;
  int ymax = win->canvas->h;

  for(int y = yorig;  y < ymax; y += he)
    for(int x = xorig; x < xmax; x += wi)
      if (x < x1 && x+wi > x0 && y < y1 && y+he >y0)
        CopyArea(im,0,0,win->canvas,x,y,wi,he);

  FlushDisplay(win,0,0,xmax-1,ymax-1);
}




void x_rect(Window w, int x0, int y0, int x1, int y1)
{ gg_window* win = set_draw_window(w);

  if (x0 > x1) SWAP(x0,x1);
  if (y0 > y1) SWAP(y0,y1);

  int lw = win->LINEWIDTH;

  for(int i=0; i<lw; i++)
  { int d = (i%2) ? -(i+1)/2 : +(i+1)/2;
    hline(win,x0+d,x1-d,y0+d);
    hline(win,x0+d,x1-d,y1-d);
    vline(win,x0+d,y0+d,y1-d);
    vline(win,x1-d,y0+d,y1-d);
  }

  FlushDisplay(win,x0-lw,y0-lw,x0+lw,y1+lw);
  FlushDisplay(win,x1-lw,y0-lw,x1+lw,y1+lw);
  FlushDisplay(win,x0-lw,y0-lw,x1+lw,y0+lw);
  FlushDisplay(win,x0-lw,y1-lw,x1+lw,y1+lw);
}


/*
void x_circle0(Window w, int x0,int y0,int r0)
{ 
  gg_window* win = set_draw_window(w);
  x_image* canvas = win->canvas;
  pixel clr = win->COLOR;
  int mode = win->MODE;

  int r;
  for (r = r0-win->LINEWIDTH/2; r <= r0+win->LINEWIDTH/2; r++)
  { int y = r;
    int x = 0;
    int e = 3-2*y;

    canvas->set_pixel(x0,y0+r,clr,mode);
    canvas->set_pixel(x0,y0-r,clr,mode);
    canvas->set_pixel(x0+r,y0,clr,mode);
    canvas->set_pixel(x0-r,y0,clr,mode);

    for (x=1;x<y;)
    { canvas->set_pixel(x0+x,y0+y,clr,mode);
      canvas->set_pixel(x0+x,y0-y,clr,mode);
      canvas->set_pixel(x0-x,y0+y,clr,mode);
      canvas->set_pixel(x0-x,y0-y,clr,mode);
      canvas->set_pixel(x0+y,y0+x,clr,mode);
      canvas->set_pixel(x0+y,y0-x,clr,mode);
      canvas->set_pixel(x0-y,y0+x,clr,mode);
      canvas->set_pixel(x0-y,y0-x,clr,mode);
      x++;
      if (e>=0) { y--; e = e - 4*y; }
      e = e + 4*x + 2;
     }

    if (x == y)
    { canvas->set_pixel(x0+x,y0+y,clr,mode);
      canvas->set_pixel(x0+x,y0-y,clr,mode);
      canvas->set_pixel(x0-x,y0+y,clr,mode);
      canvas->set_pixel(x0-x,y0-y,clr,mode);
     }
  }

  FlushDisplay(win,x0-r0,y0-r0,x0+r0,y0+r0);
}
*/



#define SETARCPIX(x,y)\
{ if (orient >= 0)\
 { if (ax*y >= ay*x && x*by >= y*bx) canvas->set_pixel(x0+x,y0+y,clr,mode); }\
 else  if (ax*y >= ay*x || x*by >= y*bx) canvas->set_pixel(x0+x,y0+y,clr,mode);\
}
 

void x_arc(Window w, int x0, int y0, int r0, int r1, double start, double angle)
{ gg_window* win = set_draw_window(w);
  x_image* canvas = win->canvas;
  pixel clr = win->COLOR;
  int mode = win->MODE;

  double ax =  cos(start);
  double ay = -sin(start);
  double bx =  cos(start+angle);
  double by = -sin(start+angle);
  double orient = ax*by - ay*bx;

  for (int r = r0-win->LINEWIDTH/2; r <= r0+win->LINEWIDTH/2; r++)
  { int y = r;
    int x = 0;
    int e = 3-2*y;

    SETARCPIX( 0, r);
    SETARCPIX( 0,-r);
    SETARCPIX( r, 0);
    SETARCPIX(-r, 0);

    for (x=1;x<y;)
      { SETARCPIX( x, y);
        SETARCPIX( x,-y);
        SETARCPIX(-x, y);
        SETARCPIX(-x,-y);
        SETARCPIX( y, x);
        SETARCPIX( y,-x);
        SETARCPIX(-y, x);
        SETARCPIX(-y,-x);
        x++;
        if (e>=0) { y--; e = e - 4*y; }
        e = e + 4*x + 2;
       }

    if (x == y)
    { SETARCPIX( x, y);
      SETARCPIX( x,-y);
      SETARCPIX(-x, y);
      SETARCPIX(-x,-y);
     }
  }

  if (r0 < r1) r0 = r1;
  FlushDisplay(win,x0-r0,y0-r0,x0+r1,y0+r1);
}




/*
void x_fill_circle0(Window w, int x0, int y0, int r)
{ 
  gg_window* win = set_draw_window(w);
  int y = 1;
  int x = r;
  int e = 3-2*r;

  hline(win,x0-x,x0+x,y0);

  while (y<=x)
  { hline(win,x0-x,x0+x,y0+y);
    hline(win,x0-x,x0+x,y0-y);

    if (y<x && e>=0)
    { hline(win,x0-y,x0+y,y0+x);
      hline(win,x0-y,x0+y,y0-x);
      x--;
      e = e - 4*x;
     }
    y++;
    e = e + 4*y + 2;
   }
  FlushDisplay(win,x0-r,y0-r,x0+r,y0+r);
}
*/


void x_pixel(Window w, int x, int y) 
{ gg_window* win =set_draw_window(w);
  win->canvas->set_pixel(x,y,win->COLOR,win->MODE);
  FlushDisplay(win,x,y,x,y);
}


int x_get_pixel(Window w, int x, int y) 
{ gg_window* win =set_draw_window(w);
  return (int)win->canvas->get_pixel(x,y);
}

void x_pixels(Window w, int n, int* x, int* y)
{ if (n <=0 ) return;

  gg_window* win = set_draw_window(w);
  x_image* canvas = win->canvas;
  int clr = win->COLOR;
  int mode = win->MODE;

  int xmin = *x;
  int xmax = *x;
  int ymin = *y;
  int ymax = *y;
  while(n--) 
  { if (xmin > *x) xmin = *x;
    if (xmax < *x) xmax = *x;
    if (ymin > *y) ymin = *y;
    if (ymax < *y) ymax = *y;
    canvas->set_pixel(*x++,*y++,clr,mode); 
   }
  FlushDisplay(win,xmin,ymin,xmax,ymax);
}


static void draw_ellipse(Window w, int x0, int y0, int a, int b, bool fill)
{
  /* Foley, van Dam, Feiner, Huges: Computer Graphics, page 90 */

  gg_window* win = set_draw_window(w);

  double xc = x0;
  double yc = y0;
  double rx = a;
  double ry = b;

  double rx_2 = rx*rx;
  double ry_2 = ry*ry;

/*
  bigpix(win, xc, yc-ry);
  bigpix(win, xc, yc+ry);
  bigpix(win, xc-rx, yc);
  bigpix(win, xc+rx, yc);
*/

  double x = 0;
  double y = ry;

  double d1 = ry_2 + rx_2*(0.25 - b); 
   
  while (rx_2*(y - 0.5) > ry_2*(x+1))
  { if (d1 < 0)
      d1 += ry_2*(2*x + 3);
    else
      { d1 += ry_2*(2*x + 3) + rx_2*(2 - 2*y);
        y--;
       }

    x++;

    if (fill)
    { hline(win,int(xc-x),int(xc+x),int(yc+y));
      hline(win,int(xc-x),int(xc+x),int(yc-y));
     }
    else
    { bigpix(win,int(xc+x),int(yc+y));
      bigpix(win,int(xc-x),int(yc+y));
      bigpix(win,int(xc+x),int(yc-y));
      bigpix(win,int(xc-x),int(yc-y));
     }
  }

  double d2 = ry_2*(x+0.5)*(x+0.5) + rx_2*(y - 1)*(y - 1) - rx_2*ry_2;

  while (y > 0)
  { if (d2 < 0)
      { d2 += ry_2*(2*x+2)+rx_2*(3-2*y);
        x++;
       }
    else
      d2 += rx_2*(3-2*y);

    y--;

    if (fill)
    { hline(win,int(xc-x),int(xc+x),int(yc+y));
      hline(win,int(xc-x),int(xc+x),int(yc-y));
     }
    else
    { bigpix(win,int(xc+x),int(yc+y));
      bigpix(win,int(xc-x),int(yc+y));
      bigpix(win,int(xc+x),int(yc-y));
      bigpix(win,int(xc-x),int(yc-y));
     }
  }

  int lw = win->LINEWIDTH;
  FlushDisplay(win,int(xc-rx-lw),int(yc-ry-lw),int(xc+rx+lw),int(yc+ry+lw));
}


void x_ellipse(Window w, int x0, int y0, int a, int b)
{ draw_ellipse(w,x0,y0,a,b,false); }

void x_fill_ellipse(Window w, int x0, int y0, int a, int b)
{ draw_ellipse(w,x0,y0,a,b,true); }


//------------------------------------------------------------------------------
// bitmaps & pixrects
//------------------------------------------------------------------------------

static unsigned char rev_byte(unsigned char c)
{ unsigned char c1 = 0x00;
   for(int i=0; i<8; i++)
   { c1 <<= 1;
     if (c&1) c1 |= 1;
     c >>= 1;
    }
  return c1;
 }

void x_set_pixrect_alpha(char* p, int) {}

void x_get_pixrect_anchor(char* p, int& x, int& y) {
    x = ((x_image*)p)->anchor_x;
    y = ((x_image*)p)->anchor_y;
}

void x_set_pixrect_anchor(char* p, int x, int y) {
    ((x_image*)p)->anchor_x = x;
    ((x_image*)p)->anchor_y = y;
}

void x_set_pixrect_phi(char* p, float phi) {
  ((x_image*)p)->phi = phi;
}



char* x_create_pixrect_from_bits(Window w, int wi, int he, unsigned char* data,
                                                                        int fg, 
                                                                        int bg)
{ //gg_window* win = set_draw_window(w);
  x_image* img = new x_image((pixel*)NULL,wi,he);
  return (char*)img;
 }


char* x_create_pixrect(Window w, int x0, int y0, int x1, int y1)
{ gg_window* win = set_draw_window(w);

  if (x0 > x1) SWAP(x0,x1);
  if (y0 > y1) SWAP(y0,y1);

  int wi = x1-x0+1;
  int he = y1-y0+1;


  x_image* img = new x_image(wi,he);

  pixel* p = img->buf;

  //for(int y = y0; y < y1; y++)
  for(int y = y0; y <= y1; y++)
  { pixel* q = win->canvas->buf + x0 + y*win->canvas->w;
    for(int x = x0; x <= x1; x++) *p++ = *q++;
   }

  return (char*)img;
 }

char* x_root_pixrect(int x0, int y0, int x1, int y1) 
{ return x_create_pixrect(0,x0,y0,x1,y1); }


void x_get_pixrect_size(char* p, int& w, int& h)
{ x_image* img = (x_image*)p;
  if (img)
  { w = img->w;
    h = img->h;
   }
  else w = h = 0;
}


void x_insert_pixrect(Window w, int x, int y, char* p)
{ gg_window* win = set_draw_window(w);
  if (p == NULL) return; 
  x_image* img = (x_image*)p;
  y -= (img->h -1);
  CopyArea(img,0,0,win->canvas,x,y,img->w,img->h,img->use_alpha);
  FlushDisplay(win,x,y,x+img->w,y+img->h);
}


void x_insert_pixrect(Window w, char* p)
{ gg_window* win = set_draw_window(w);
  if (p == NULL) return; 
  x_image* img = (x_image*)p;
  CopyArea(img,0,0,win->canvas,0,0,img->w,img->h,img->use_alpha);
  FlushDisplay(win,0,0,img->w,img->h);
}


void x_insert_pixrect(Window w, int x, int y, char* p, int x0, int y0, 
                                                       int width, int height)
{ gg_window* win = set_draw_window(w);
  assert(p != NULL);
  x_image* img = (x_image*)p;
  y -= (height-1);
  CopyArea(img,x0,y0,win->canvas,x,y,width,height,img->use_alpha);
  FlushDisplay(win,x,y,width,height);
}

void  x_insert_pixrect(Window w, int x, int y, int wi, int he, char* p)
{
  // scale pixrect into box (x,y,x+wi,y-he) (upper left corner)

  gg_window* win = set_draw_window(w);
  x_image* im = (x_image*)p;

  double scale_x = double(wi)/im->w;
  double scale_y = double(he)/im->h;

  x -= int(0.5+scale_x*im->anchor_x);
  y += int(0.5+scale_y*im->anchor_y);

  if (wi != im->w || he != im->h || im->phi != 0) 
  { //x_image* im1 = im->resize_rotate(wi,he,im->phi);

    int x0 = (x < 0) ? -x : 0;
    int x1 = x0 + x_window_width(w);

    int y0 = he - y;
    if (y0 < 0) y0 = 0; 
    int y1 = y0 + x_window_height(w);

    x_image* im1 = im->resize_rotate(wi,he,im->phi,x0,y0,x1,y1);

    x_insert_pixrect(w,x,y,(char*)im1);
    x_delete_pixrect((char*)im1);
    return;
  }

  x_insert_pixrect(w,x,y,p);
}


void x_delete_pixrect(char* p) 
{ x_image * img = (x_image*)p; 
  if (img->buf) delete[] img->buf;
  delete img;
}


void x_pixrect_to_matrix(Window w, char* prect, int* matrix) {}

void x_matrix_to_pixrect(Window w, int* matrix, int width, int height,
                                   char* prect) {}


unsigned char* scale_bitmap(int f, unsigned char* bits, int width, int height)
{
 int bw1 = (width+7)/8;

 int w2 = f*width;
 int h2 = f*height;
 int bw2 = (w2+7)/8;

 unsigned char* bits2 = new unsigned char[bw2*h2];

 for(int i=0; i<h2; i++)
   for(int j=0; j<bw2; j++)
   { char x = bits[(i/f)*bw1 + j/f]; // source byte
     int y = 0;
     for(int k = 0; k<8; k++) {
       if (x & (1<<k)) y |=  ((1<<f)-1) << f*k;
     }
     unsigned char* z = (unsigned char*)&y;
     bits2[i*bw2 + j] = z[j%f];
    }

  return bits2;
}


char* x_create_bitmap(Window w, int width, int height, unsigned char* p, int f)
{ //gg_window* win = set_draw_window(w);

  if (f > 1)
  { p = scale_bitmap(f,p,width,height);
    width *= f;
    height *= f;
  }


  x_image* im = new x_image(width,height,1); // bitmap
  char* q = (char*)im->bmp;
  char* stop = q + im->size;
  while (q < stop) *q++ = rev_byte(*p++);

  return (char*)im;
}

char* x_create_pixrect_from_bgra(Window w, int wi, int he, unsigned char* bgra, 
                                                           float f ) 
{ int sz = wi*he;
  pixel* p = new pixel[sz];
  memcpy(p,bgra,4*sz);
  x_image* im = new x_image(p,wi,he); 
  im->use_alpha = true;
  im = im->resize_rotate(int(0.5+f*im->w),int(0.5+f*im->h),0);
  return (char*)im; 
}


unsigned char* x_pixrect_to_bgra(Window w, char* prect) { 
  x_image* im = (x_image*)prect;
  int sz = im->w * im->h;
  unsigned int* bgra = new unsigned int[sz];
  memcpy(bgra,im->buf,4*sz);
  return (unsigned char*)bgra; 
}




void x_insert_bitmap(Window w, int x, int y, char* bm)
{ gg_window* win = set_draw_window(w);
  
  x_image* img = (x_image*)bm;

  int lbytes  = img->size/img->h;

  unsigned char* pm = img->bmp;

  y -= (img->h-1);

  clip_bitmap(win->canvas,x,y,pm,lbytes,img->h);

  for(int i=0; i<img->h; i++)
     insert_bit_line(win,x,y+i,pm+i*lbytes,lbytes);

  FlushDisplay(win,x,y,x+img->w,y+img->h);
}


void x_delete_bitmap(char* bm) { x_delete_pixrect(bm); }


void x_insert_bitmap(Window w, int x, int y, int width, int height, 
                                                        unsigned char* data)
{ gg_window* win = set_draw_window(w);

  int len =  width/8;
  if (width % 8) len++;

  unsigned char* buf = new unsigned char[len*height];
  unsigned char* q = buf;
  unsigned char* stop = q + len*height;

  while (q < stop) *q++ = rev_byte(*data++);

  clip_bitmap(win->canvas,x,y,buf,len,height);

  for(int i=0; i<height; i++)
    insert_bit_line(win,x,y+i,buf+i*len,len);

  FlushDisplay(win,x,y,x+len,y+height);

  delete[] buf;
}

void x_text(Window w, int x, int y, const char *txt)
{ gg_window* win = set_draw_window(w);

  x_image* canv = win->canvas;

  int x1 = x + x_text_width(w,txt);
  int y1 = y + x_text_height(w,txt);

/*
  x_rect(w,x,y,x1,y1);
*/

  if (win->FONT_WIDTH == FIXED_FONT_WIDTH) y += 5;

  if (win->TEXTMODE == opaque)  // clear background
  { int save_color = x_set_color(w,win->bg_clr);
    drawing_mode save_mode = x_set_mode(w,src_mode);
    x_box(w,x,y,x1,y1);
    x_set_mode(w,save_mode);
    x_set_color(w,save_color);
   }


  int xmin = canv->clip_x0;
  int xmax = canv->clip_x1 - 5;


/*
  int offset = 0;
  for(int i=0; i<128; i++) {
    assert(offset == FONT_OFFSET[i]);
    offset += win->FONT_WIDTH[i] * FONT_HEIGHT;
  }
*/

  int xx = x;
  for (unsigned int i = 0; i < strlen(txt); i++) 
  { if (xx < xmin || xx > xmax) continue;
    int c = txt[i] & 127;
    int tw  = win->FONT_WIDTH[c];
    int th  = win->FONT_HEIGHT;
    int off = win->FONT_OFFSET[c];
    int pix_num = tw*th;
    pixel* pixels = new pixel[pix_num];
    for(int j=0;j<pix_num;j++)
    { int alpha = win->FONT_ALPHA[off+j];
      pixels[j] = win->COLOR | (alpha << 24);
     }

    x_image* img = new x_image(pixels,tw,th);

    float f = win->FONT_SCALE;
    if (f != 1)
    { x_image* p = img->resize_rotate(int(0.5+f*img->w),int(0.5+f*img->h),0);
      x_delete_pixrect((char*)img);
      img = p;
     }

    CopyArea(img,0,0,win->canvas,xx,y,img->w,img->h,true);

    xx += int(0.5 + f*tw);

    x_delete_pixrect((char*)img);
  }

  FlushDisplay(win,x,y,x1,y1);
}


void x_text(Window w, int x, int y, const char *text, int l)
{ char* str = new char[strlen(text)+1];
  strcpy(str,text);
  if (unsigned(l) < strlen(text)) str[l] = '\0';
  x_text(w,x,y,str);
  delete[] str;
}

void x_ctext(Window w, int x, int y, const char* str)
{ x_text(w,x-(x_text_width(w,str)-1)/2, y-(x_text_height(w,str)-1)/2, str); }

void x_text_underline(Window w, int x, int y, const char* text, int, int)
{ x_text(w,x,y,text); }

void x_ctext_underline(Window w, int x, int y, const char* text, int, int)
{ x_ctext(w,x,y,text); }




void x_set_clip_rectangle(Window w, int x ,int y, int width, int height) 
{ 
/*
  printf("clip_rect(w = %d  x = %d y = %d  w = %d  h = %d)\n",
                    w,x,y,width,height);
*/

  gg_window* win = win_list[w];
  x_image* canv = win->canvas;

  assert(canv != 0);

  if (y < 0) y = 0;
  if (y >= canv->h) y = canv->h - 1;
  if (x < 0) x = 0;
  if (x >= canv->w) y = canv->w - 1;

  canv->clip_x0 = x;
  canv->clip_y0 = y;
  int dx = canv->w - (x+width);
  int dy = canv->h - (y+height);
  if (dx < 0) width  += dx;
  if (dy < 0) height += dy;
  canv->clip_x1 = x + width - 1;
  canv->clip_y1 = y + height - 1;
}


static void set_clipping_params(Window w, int clip_mode)
{ gg_window* win = win_list[w];
  if (clip_mode == 0) {
    win->COLOR = 0x01000000;
    win->MODE = or_mode;
  }
  else
  { // unset clip bit 
    win->COLOR = ~0x01000000;
    win->MODE = and_mode;
   }
}

void x_clip_mask_rectangle(Window w, int x0,int y0,int x1,int y1,int clip_mode)
{ gg_window* win = win_list[w];
  pixel clr = win->COLOR;
  drawing_mode mode = win->MODE;
  set_clipping_params(w,clip_mode);
  x_box(w,x0,y0,x1,y1);
  win->COLOR = clr;
  win->MODE = mode;
 }


void x_clip_mask_ellipse(Window w, int x, int y, int a, int b, int clip_mode) 
{ gg_window* win = win_list[w];
  pixel clr = win->COLOR;
  drawing_mode mode = win->MODE;
  set_clipping_params(w,clip_mode);
  draw_ellipse(w,x,y,a,b,true);
  win->COLOR = clr;
  win->MODE = mode;
}

void x_clip_mask_polygon(Window w, int n, int* xc, int* yc, int clip_mode) 
{ gg_window* win = win_list[w];

  if (n == 0) 
  { // empty polygon: entire canvas
    x_image* canv = win->canvas;
    x_clip_mask_rectangle(w, 0, 0, canv->w, canv->h,clip_mode);
    return;
  }

  pixel clr = win->COLOR;
  drawing_mode mode = win->MODE;
  set_clipping_params(w,clip_mode);
  x_fill_polygon(w,n,xc,yc);
  win->COLOR = clr;
  win->MODE = mode;

}


void x_set_read_gc(Window w)
{ gg_window* win = set_draw_window(w);
  win->save_clr  = x_set_color(w,black);
  win->save_mode = x_set_mode(w,xor_mode);
  win->save_ls   = x_set_line_style(w,solid);
  win->save_lw   = x_set_line_width(w,1);
 }


void x_reset_gc(Window w)
{ gg_window* win = set_draw_window(w);
  x_set_color(w,win->save_clr);
  x_set_mode(w,win->save_mode);
  x_set_line_style(w,win->save_ls);
  x_set_line_width(w,win->save_lw);
 }


void x_screen_to_window(Window w, int& x, int& y) 
{ gg_window* win = set_draw_window(w);
  x -= win->xpos;
  y -= win->ypos;
 }

void x_window_to_screen(Window w, int& x, int& y)
{ gg_window* win = set_draw_window(w);
  x += win->xpos;
  y += win->ypos;
 }

int   x_set_bg_color(Window w, int c) 
{ gg_window* win = set_draw_window(w);
  int old_c = win->bg_clr;
  win->bg_clr = c;
  return old_c;
}

char* x_set_bg_pixmap(Window w, char* pm) 
{ gg_window* win = set_draw_window(w);
  char* old_pm = win->bg_pixmap;
  win->bg_pixmap = pm;
  return old_pm;
}


void x_set_border_color(Window w, int c) 
{ gg_window* win = set_draw_window(w);
  win->border_clr = c;
}

void x_set_border_width(Window w, int width) 
{ gg_window* win = set_draw_window(w);
  win->border_w = width;
}



// buffering

int   x_create_buffer(Window w, int wi, int he) 
{ gg_window* win = set_draw_window(w);
  if (win->buffer == 0 || wi != win->buffer->w || he != win->buffer->h)
  { if (win->buffer) x_delete_buffer(w);
    win->buffer = new x_image(wi,he);
    if (win->canvas_save) win->canvas = win->buffer;
   }
  return 1;
 }

int   x_create_buffer(Window w) 
{ int wi = x_window_width(w);
  int he = x_window_height(w);
  x_create_buffer(w,wi,he);
  return 1;
 }


void x_delete_buffer(Window w)
{ gg_window* win = set_draw_window(w);
  x_stop_buffering(w);
  if (win->buffer) 
  { delete win->buffer;
    win->buffer = 0;
   }
}


int   x_start_buffering(Window w, int wi ,int he)
{ gg_window* win = set_draw_window(w);
  assert(win != root_win);
  x_stop_buffering(w);
  x_delete_buffer(w);
  x_create_buffer(w,wi,he);
  win->canvas_save = win->canvas;
  win->canvas = win->buffer;
  win->flush = false;
  return 1;
}

int   x_start_buffering(Window w) 
{ gg_window* win = set_draw_window(w);
  assert(win != root_win);
  if (win->canvas_save) return 0;
  int new_buf = (win->buffer == 0);
  if (new_buf) x_create_buffer(w);
  win->canvas_save = win->canvas;
  win->canvas = win->buffer;
  win->flush = false;
  return new_buf;
}


void x_flush_buffer(Window w,int x0,int y0,int x1,int y1, int xoff, int yoff)
{ gg_window* win = set_draw_window(w);

  assert(win != root_win);

  if (win->buffer == 0) return; 

  if (y0 > y1) SWAP(y0,y1);
  if (x0 > x1) SWAP(x0,x1);

  int wi = x1-x0+1;
  int he = y1-y0+1;

  if (win->canvas_save )
  { // buffering on
    CopyArea(win->buffer,x0+xoff,y0+yoff,win->canvas_save,x0,y0,wi,he);
   }
  else
  { // buffering has been finished before
    CopyArea(win->buffer,x0+xoff,y0+yoff,win->canvas,x0,y0,wi,he);
   }

  x0 += xoff;
  x1 += xoff;
  y0 += yoff;
  y1 += yoff;

  bool flush = win->flush;
  win->flush = true;
  FlushDisplay(win,x0,y0,x1,y1);
  win->flush = flush;

}


void x_flush_buffer(Window w ,int x0, int y0, int x1, int y1) 
{ x_flush_buffer(w,x0,y0,x1,y1,0,0); }


void  x_stop_buffering(Window w)
{ gg_window* win = set_draw_window(w);
  if (win->canvas_save)
  { win->canvas = win->canvas_save;
    win->canvas_save = 0;
   }
  win->flush = true;
}

void  x_stop_buffering(Window w, char** ppr)  
{ gg_window* win = set_draw_window(w);
  if (win->canvas_save)
  { win->canvas = win->canvas_save;
    win->canvas_save = 0;
    *ppr = (char*)win->buffer;
    win->buffer = 0;
  }
  else *ppr = 0;
  win->flush = true;
}

int   x_test_buffer(Window w) 
{ gg_window* win = set_draw_window(w);
  return win->canvas_save != 0; 
 }

void  x_set_buffer(int,char*) {}



char* x_create_pixrect_from_xpm(Window w, const char** xpm) 
{ 
  int width;
  int height;
  int colors;
  int chars;

  sscanf(*xpm,"%d %d %d %d",&width,&height,&colors, &chars);

  if (chars > 2)
  { fprintf(stderr,"xpm: sorry, chars_per_pixel > 2 not implemented.\n");
    return 0;
   }

  x_image* im = new x_image(width,height);

  im->use_alpha = true;

  pixel color_table[1<<16]; 
  int i = 1<<16; 
  while (i--) color_table[i] = white;

  for(i=0; i<colors; i++)
  { xpm++;
    char c1=0;
    char c2=0;
    char rgb_str[16];

    if (chars == 1) sscanf(*xpm,"%c c %s",&c1,rgb_str);
    if (chars == 2) sscanf(*xpm,"%c%c c %s",&c1,&c2,rgb_str);

    int len = strlen(rgb_str);

    if (strcmp(rgb_str,"None") == 0)
    {
      if (w == 0)
         color_table[c1+256*c2] = black; // icon
      else
         //color_table[c1+256*c2] = grey1;
        //color_table[c1+256*c2] = yellow;
        color_table[c1+256*c2] = 0x00ffffff;
    }
    else
      { if (len == 13)
        { rgb_str[3] = rgb_str[5];
          rgb_str[4] = rgb_str[6];
          rgb_str[5] = rgb_str[9];
          rgb_str[6] = rgb_str[10];
          rgb_str[7] = 0;
         }
    
        int x;
        sscanf(rgb_str+1,"%x",&x);
     
        int b = x % 256; x = x/256;
        int g = x % 256; x = x/256;
        int r = x % 256;
        int col = x_new_color(r,g,b);
      //printf("new color = %d\n",col);
        color_table[c1+256*c2] = col | 0xff000000;
       }
   }

   for(int y=0; y<height; y++)
   { const char* line = *++xpm;
     for(int x=0; x<width; x++)
     { int index = line[x*chars];
       if (chars == 2) index += 256*line[x*chars+1];
       im->set_pixel(x,y,color_table[index],src_mode); 
      }
    }

  float f = display_dpi/96.0;

  x_image* img = im->resize_rotate(int(0.5+f*im->w),int(0.5+f*im->h),0);

  x_delete_pixrect((char*)im);

  return (char*)img;
}


void x_window_frame(int w, int& x0, int& y0, int& x1, int& y1) 
{ gg_window* win = win_list[w];
  x0 = win->x0;
  y0 = win->y0;
  x1 = win->x1;
  y1 = win->y1;
}




//------------------------------------------------------------------------------
// window manager
//------------------------------------------------------------------------------

const int ROOT_COLOR = blue2;   
const int LABEL_COLOR = grey2;

static int mouse_x = 0;
static int mouse_y = 0;


static int handle_next_event(int* win, int* x, int* y, int* val1, int* val2, 
                                                    unsigned long* t, int msec)
{ 
  // msec > 0: blocking / msec timeout
  // msec = 0: blocking without timeout
  // msec < 0: non blocking

  *val2 = 0;

  // find next event

  int e = no_event;

  int cur_x = last_event.x;
  int cur_y = last_event.y;

  if (event_buf.ev != -1)
  { e = event_buf.ev;
    *val1 = event_buf.val;
    cur_x = event_buf.x;
    cur_y = event_buf.y;
    event_buf.ev = -1;
   }
  else
    e = CHECK_NEXT_EVENT(*val1,cur_x,cur_y,*t,msec);

  if (e == display_event)
  { //cout << string ("DISPLAY EVENT: %d %d %d",*val1,cur_x,cur_y) << endl;

    if (*val1 > 0) {
    // display size changed
    //cout << string ("DISPLAY CHANGED: %d x %d (%d)",cur_x,cur_y,*val1);
      redraw_root();
    }

  //cout << "DISPLAY CLOSED" << endl;

    *win = 0;
    return e;
  }

  if (e != no_event)
  { last_event.ev  = e;
    last_event.val = *val1;
    last_event.x = cur_x;
    last_event.y = cur_y;
    if (e != key_press_event && e != key_release_event)
    { mouse_x = cur_x;
      mouse_y = cur_y;
     }
  }

  gg_window* w = grab_win;

  if (w == 0)
  { for(int i = win_top; i >= 0; i--)
    { w = win_stack[i];
      if (mouse_x >= w->x0  && mouse_x <= w->x1 &&
          mouse_y >= w->y0  && mouse_y <= w->y1 )  break;
     }
   }

  *win = w->id;
  *x = mouse_x - w->xpos;
  *y = mouse_y - w->ypos;

/*
cout << event_name[e] << endl;
cout << "win = " << *win << " val = " << *val1 << " x = " << *x << " y = " << *y << "  t = " << *t << endl;
*/

  return e;
}



static void draw_window_button(gg_window* win, int i)
{
  if (win->minimized) return;

  int x1 = win->x1 - BUTTON_X[i]; 
  int x2 = x1 + BUTTON_W;
  int y1 = win->y0 + BUTTON_Y;
  int y2 = y1 + BUTTON_W;

  x_set_line_width(0,3);

  int clr = white;
  int bg_clr = white;

  if (active_win_button == i+1)
  { clr = black;
    bg_clr = grey1;
   }
  else
  { clr = white;
    bg_clr = win->label_clr;
   }
/*
  x_set_color(0,bg_clr);
  x_box(root_win->id,x1-10,y1-10,x2+10,y2+10);
*/

  x_set_color(0,clr);

  switch (i) {
 
   case 0: x_line(root_win->id,x1,(y1+y2)/2,x2,(y1+y2)/2);
           break;

   case 1: x_rect(root_win->id,x1,y1,x2,y2);
           break;

   case 2: x_line(root_win->id,x1,y1,x2,y2);
           x_line(root_win->id,x1,y2,x2,y1);
           break;
  }

}


static int check_window_button(gg_window* win)
{
  if (win->y0+BUTTON_Y-10 > mouse_y || win->y0+BUTTON_Y+BUTTON_W+10 < mouse_y)
    return 0;

  int but = 0;
  for(int i=0; i<3; i++) {
    if (win->x1-BUTTON_X[i]-10 <= mouse_x && 
        win->x1-BUTTON_X[i]+BUTTON_W+10 >= mouse_x) but = i+1;
  }

  return but;
}

static void set_window_label(Window w, const char* s)
{
  gg_window* win = win_list[w];

  int          clr = x_set_color(0,win->label_clr);
  text_mode    tm = x_set_text_mode(0,transparent);
  drawing_mode mode = x_set_mode(0,src_mode);

  //x_set_font(0,"T29"); 
  x_set_font(0,"T35"); 

  //int label_w = win->x1 - win->x0 - 220;
  int label_w = win->x1 - win->x0 - int(1.5*display_dpi);
  int label_h = x_text_height(0,"H");

  int label_x = win->x0 + 35;
  int label_y = win->y0 + int(0.5*(HEADER_W - label_h)) + 2;

  x_set_color(0,win->label_clr);
  x_box(root_win->id,label_x,label_y,label_x+label_w,label_y+label_h);

  int len = strlen(s);
  while (x_text_width(0,s,len) > label_w) len--;

  if (s != win->header) strcpy(win->header,s);

  x_set_color(0,black);
  x_text(root_win->id,label_x,label_y,win->header,len);

  x_set_text_mode(0,tm);
  x_set_color(0,clr);
  x_set_mode(0,mode);

}

void x_set_label(Window w, const char* s)
{ gg_window* win = win_list[w];
  root_win->flush = false;
  set_window_label(w,s);
  root_win->flush = true;
  FlushDisplay(root_win,win->x0,win->y0,win->x1,win->y0+HEADER_W);
}



static void draw_window(gg_window* win, int clear_win=1)
{
  if (win->minimized)
  { int icon_w, icon_h;
    x_get_pixrect_size(win->icon_pixmap,icon_w,icon_h);
    x_insert_pixrect(0,win->xpos,win->ypos+icon_h-1,win->icon_pixmap);
    x_set_color(0,black);
    x_rect(0,win->x0,win->y0,win->x1,win->y1);

    char str[64];
    strncpy(str,win->label,64);
    str[63] = '\0';

    int sz = strlen(str);
    while (x_text_width(0,str) > icon_w) str[--sz] = 0;

    int label_x = (win->x0 + win->x1)/2;
    int label_y = win->y1 + int(0.7*x_text_height(0,"H"));
    x_set_color(0,white);
    x_ctext(0,label_x,label_y,str);
    x_set_color(0,black);
    return;
   }

  int border_w  = win->xpos - win->x0 - 2;
  int label_h  = win->ypos - win->y0 - 2;

  int x0 = win->x0;
  int y0 = win->y0;

  int x1 = win->x1;
  int y1 = win->y1;

  drawing_mode save_mode = x_set_mode(0,src_mode);
  int save_line_width = x_set_line_width(0,1);

  if (clear_win)
  { x_set_color(0,win->bg_clr);
    //x_box(0,x0,y0,x1,y1);
    x_box(0,win->xpos, win->ypos,
                       win->xpos+win->width-1, win->ypos+win->height-1);
   }

  if (border_w > 1)
  { 
    //int bw = border_w;
    //int bw1 = 2*(border_w + 1);

    int bw = border_w + 1;
    int bw1 = int(2.5*border_w);

x_set_color(0,grey1);
x_box(root_win->id,x0,y0,x1,y1);

    x_set_color(0,win->label_clr);
    x_box(root_win->id,x0+bw,y0+bw,x1-bw,y0+label_h);
    x_box(root_win->id,x0,y0+bw1,x0+bw,y1-bw1);
    x_box(root_win->id,x1-bw,y0+bw1,x1,y1-bw1);
    x_box(root_win->id,x0+bw1,y1-bw,x1-bw1,y1);
    x_box(root_win->id,x0+bw1,y0,x1-bw1,y0+bw);

    x_set_color(0,black);
    x_rect(root_win->id,x0,y0,x1,y1);

    x_line(root_win->id,x0+bw,y0+bw,x0+bw1,y0+bw);
    x_line(root_win->id,x0+bw1,y0+bw,x0+bw1,y0+1);
    x_line(root_win->id,x0+bw1,y0+1,x1-bw1,y0+1);
    x_line(root_win->id,x1-bw1,y0+1,x1-bw1,y0+bw);
    x_line(root_win->id,x1-bw1,y0+bw,x1-bw,y0+bw);
    x_line(root_win->id,x1-bw,y0+bw,x1-bw,y0+bw1);
    x_line(root_win->id,x1-bw,y0+bw1,x1-1,y0+bw1);
    x_line(root_win->id,x1-1,y0+bw1,x1-1,y1-bw1);
    x_line(root_win->id,x1-1,y1-bw1,x1-bw,y1-bw1);
    x_line(root_win->id,x1-bw,y1-bw1,x1-bw,y1-bw);
    x_line(root_win->id,x1-bw,y1-bw,x1-bw1,y1-bw);
    x_line(root_win->id,x1-bw1,y1-bw,x1-bw1,y1-1);
    x_line(root_win->id,x1-bw1,y1-1,x0+bw1,y1-1);
    x_line(root_win->id,x0+bw1,y1-1,x0+bw1,y1-bw);
    x_line(root_win->id,x0+bw1,y1-bw,x0+bw,y1-bw);
    x_line(root_win->id,x0+bw,y1-bw,x0+bw,y1-bw1);
    x_line(root_win->id,x0+bw,y1-bw1,x0+1,y1-bw1);
    x_line(root_win->id,x0+1,y1-bw1,x0+1,y0+bw1);
    x_line(root_win->id,x0+1,y0+bw1,x0+bw,y0+bw1);
    x_line(root_win->id,x0+bw,y0+bw1,x0+bw,y0+bw);
  }

  if (win->border_w > 0)
  { x_set_color(0,win->border_clr);
    x_rect(root_win->id, win->xpos-1, win->ypos-1,
                         win->xpos+win->width, win->ypos+win->height);
   }

  if (label_h > 15) {
    for(int i=0; i<3; i++) draw_window_button(win,i);
    //x_set_label(win->id,win->header);
    set_window_label(win->id,win->header);
   }
  
  x_set_mode(0,save_mode);
  x_set_line_width(0,save_line_width);

/*
  if (win->redraw) win->redraw(win->inf,0,0,win->width,win->height,1);
*/

  x_image* canv = win->canvas;
  if (win->canvas_save) canv = win->canvas_save;

  CopyWindowAreaToRoot(win,0,0,canv->w-1,canv->h-1);
}



void redraw_root(int x0, int y0, int x1, int y1)
{
  root_win->flush = false;

  // clear root window
  FillArea(root_win->canvas,0,0,root_win->x1,root_win->y1,root_win->bg_clr);

  root_win->COLOR = black;

  // draw windows
  for(int i=1; i<=win_top; i++) 
  { gg_window* wp = win_stack[i];
    if (wp->pwin && wp->pwin->minimized) continue;
    draw_window(wp);
   }

  // flush to display

  x_image* root_canv = root_win->canvas;

  if (x0 < 0) x0 = 0;
  if (y0 < 0) y0 = 0;
  if (x1 >= root_canv->w) x1 = root_canv->w - 1;
  if (y1 >= root_canv->h) y1 = root_canv->h - 1;

  root_win->flush = true;
  UPDATE_SCREEN_AREA(root_canv->buf,root_canv->w,root_canv->h,x0,y0,x1,y1,x0,y0);

}


void redraw_root() 
{ redraw_root(root_win->x0,root_win->y0,root_win->x1,root_win->y1); }


static void minimize(gg_window* win)
{ 
  int width = win->save_width;
  int height = win->save_height;

  int x0 = win->save_x0;
  int y0 = win->save_y0;
  int x1 = win->save_x1;
  int y1 = win->save_y1;
  int xpos = win->save_xpos;
  int ypos = win->save_ypos;

  win->save_width  = win->width;
  win->save_height = win->height;
  win->save_x0 = win->x0;
  win->save_y0 = win->y0;
  win->save_x1 = win->x1;
  win->save_y1 = win->y1;
  win->save_xpos = win->xpos;
  win->save_ypos = win->ypos;

  win->width  = width;
  win->height = height;
  win->x0 = x0;
  win->y0 = y0;
  win->x1 = x1;
  win->y1 = y1;
  win->xpos = xpos;
  win->ypos = ypos;

  win->minimized = 1-win->minimized;

  redraw_root();
}



static void move_win(gg_window* win, int x, int y)
{ int xp0 = win->x0;
  int yp0 = win->y0;
  int xp1 = win->x1;
  int yp1 = win->y1;
  int wi  = xp1-xp0+1;
  int he  = yp1-yp0+1;

  int xc = mouse_x; /* absolute cursor coordinates */
  int yc = mouse_y; 
  int dx = xc-xp0;  /* relative to upper left corner */
  int dy = yc-yp0;

  Window w;
  int e,val1,val2;
  unsigned long t;

  root_win->MODE = xor_mode;
  root_win->COLOR = root_win->bg_clr;
  root_win->LINEWIDTH = 2;

  x_rect(0,xp0-1,yp0-1,xp1+1,yp1+1);

  do { e = handle_next_event(&w,&x,&y,&val1,&val2,&t,0);
      if (mouse_x != xc || mouse_y != yc)
       { int rx0 = mouse_x-dx-1;
         int ry0 = mouse_y-dy-1;
         int rx1 = mouse_x-dx+wi;
         int ry1 = mouse_y-dy+he;
         x_rect(0,rx0,ry0,rx1,ry1);
         x_rect(0,xc-dx-1,yc-dy-1,xc-dx+wi,yc-dy+he);
         xc = mouse_x;
         yc = mouse_y;
       }
     } while (e != button_release_event);

  xc -= dx;
  yc -= dy;

  x_rect(0,xc-1,yc-1,xc+wi,yc+he);

  root_win->MODE = src_mode;
  root_win->COLOR = black;
  root_win->LINEWIDTH = 1;

  dx = xc - win->x0; 
  dy = yc - win->y0; 

  win->x0 += dx;
  win->y0 += dy;
  win->x1 += dx;
  win->y1 += dy;
  win->xpos += dx;
  win->ypos += dy;

  // move child windows
  for(int i=1; i<=win_count; i++)
  { gg_window* wp = win_list[i];
    if (wp == 0 || wp->pwin != win) continue;
    wp->x0 += dx;
    wp->y0 += dy;
    wp->x1 += dx;
    wp->y1 += dy;
    wp->xpos += dx;
    wp->ypos += dy;
  }
  
  redraw_root();

}


static void resize_win(gg_window* win, int* x, int* y, int pos)
{ int xp0 = win->x0;
  int yp0 = win->y0;
  int xp1 = win->x1;
  int yp1 = win->y1;
  int xb = win->xpos - xp0;
  int yb = win->ypos - yp0;
  
  int xc  = mouse_x; /* absolute cursor coordinates */
  int yc  = mouse_y;

  Window w;
  int e,val1,val2,dx=0,dy=0;
  unsigned long t;

  root_win->MODE = xor_mode;
  root_win->COLOR = root_win->bg_clr;
  root_win->LINEWIDTH = 2;
  x_rect(0,xp0-1,yp0-1,xp1+1,yp1+1);

  switch(pos) {
   case 0: dx = xp0-xc; dy = yp0-yc;
           break;
   case 1: dx = xp1-xc; dy = yp0-yc;
           break;
   case 2: dx = xp1-xc; dy = yp1-yc;
           break;
   case 3: dx = xp0-xc; dy = yp1-yc;
           break;
   }

  do { e = handle_next_event(&w,x,y,&val1,&val2,&t,0);
       if (mouse_x != xc || mouse_y != yc)
        { 
          switch(pos) {
          case 0: x_rect(0,mouse_x+dx-1,mouse_y+dy-1,xp1+1,yp1+1);
                  x_rect(0,xc+dx-1,yc+dy-1,xp1+1,yp1+1);
                  break;
          case 1: x_rect(0,xp0-1,mouse_y+dy-1,mouse_x+dx+1,yp1+1);
                  x_rect(0,xp0-1,yc+dy-1,xc+dx+1,yp1+1);
                  break;
          case 2: x_rect(0,xp0-1,yp0-1,mouse_x+dx+1,mouse_y+dy+1);
                  x_rect(0,xp0-1,yp0-1,xc+dx+1,yc+dy+1);
                  break;
          case 3: x_rect(0,mouse_x+dx-1,yp0-1,xp1+1,mouse_y+dy+1);
                  x_rect(0,xc+dx-1,yp0-1,xp1+1,yc+dy+1);
                  break;
           }

          xc = mouse_x;
          yc = mouse_y;
        }
     } while (e != button_release_event);


     switch(pos) {
     case 0: x_rect(0,xc+dx-1,yc+dy-1,xp1+1,yp1+1);
             break;
     case 1: x_rect(0,xp0-1,yc+dy-1,xc+dx+1,yp1+1);
             break;
     case 2: x_rect(0,xp0-1,yp0-1,xc+dx+1,yc+dy+1);
             break;
     case 3: x_rect(0,xc+dx-1,yp0-1,xp1+1,yc+dy+1);
             break;
     }

   root_win->MODE = src_mode;
   root_win->COLOR = black;
   root_win->LINEWIDTH = 1;

   xc += dx;
   yc += dy;

   switch(pos) {
   case 0: win->x0 = xc; win->y0 = yc; 
           break;
   case 1: win->x1 = xc; win->y0 = yc; 
           break;
   case 2: win->x1 = xc; win->y1 = yc; 
           break;
   case 3: win->x0 = xc; win->y1 = yc; 
           break;
   }

  win->xpos = win->x0 + xb;
  win->ypos = win->y0 + yb;
  win->width  = win->x1 - win->x0 - 2*xb + 1; 
  win->height = win->y1 - win->y0 - xb - yb + 1; 

  *x = win->width;
  *y = win->height;

  if (win->canvas != 0) 
  { x_image* new_canv = new x_image(win->width,win->height);
    FillArea(new_canv,0,0,win->width,win->height,win->bg_clr);
    delete win->canvas;
    win->canvas = new_canv;
   }

  redraw_root();
}



//------------------------------------------------------------------------------


static int manage_next_event(Window* w, int* x, int* y, int* val1, int* val2,
                                                  unsigned long* t, int msec)

{
  // get next event and run a simple window manager 

  // msec > 0: blocking / timeout
  // msec = 0: blocking 
  // msec < 0: non-blocking 


  int e = handle_next_event(w,x,y,val1,val2,t,msec);

  //cout << "manage_next_event: " << event_name[e] << endl;

  if (e == destroy_event)  return e;

  if (e == display_event) return e;

  if (e == no_event) return e;

  if (*w == 0 || win_list[*w]->pwin) {
    // root or child window
    //if (active_win != 0) SET_MOUSE_CURSOR(-1);
    SET_MOUSE_CURSOR(-1);
    active_win = 0;

    if (active_win_button != 0) {
      int but = active_win_button;
      active_win_button = 0;
      for(int i=1; i<=win_count; i++) {
        if (win_list[i]->pwin != 0) continue;
        draw_window_button(win_list[i],but);
      }
    }

    return e;
   }

  gg_window* win = win_list[*w];

  int x0 = win->x0;
  int y0 = win->y0;
  int x1 = win->x1;
  int y1 = win->y1;

  int border_w = win->xpos - win->x0; 
  int label_h = win->ypos - win->y0;  


  if (win->minimized  && e == button_press_event) 
  { // click on icon
    active_win = 0;
    e = handle_next_event(w,x,y,val1,val2,t,msec);
    if (e == button_release_event)
     { minimize(win);
        *x = 0;
        *y = 0;
        *val1 = win->width;
        *val2 = win->height;
        return exposure_event;
       }
    else
      { if (win->pwin == 0) move_win(win,*x,*y);
        return no_event;
       }
   }

  if (win->minimized) return no_event; 



  if ((x0<=mouse_x && mouse_x<=x0+border_w) || 
      (x1>=mouse_x && mouse_x>=x1-border_w) ||  
      (y0<=mouse_y && mouse_y<=y0+label_h)  || 
      (y1>=mouse_y && mouse_y>=y1-border_w) )
  {
    // pointer on window frame: raise/move or resize window

    if (active_win != 0) SET_MOUSE_CURSOR(-1);

    active_win = 0;

    // check resize corners

    int cx[4];
    int cy[4];
    cx[0] = x0;          cy[0] = y0;
    cx[1] = x1-RESIZE_W; cy[1] = y0;
    cx[2] = x1-RESIZE_W; cy[2] = y1-RESIZE_W;
    cx[3] = x0;          cy[3] = y1-RESIZE_W;

    int CURSOR[] = { XC_top_left_corner,
                     XC_top_right_corner,
                     XC_bottom_right_corner,
                     XC_bottom_left_corner };
  
    for(int i=0; i<4; i++) {
      if (cy[i] <= mouse_y && mouse_y <= cy[i]+RESIZE_W &&
          cx[i] <= mouse_x && mouse_x <= cx[i]+RESIZE_W )
       { 
         SET_MOUSE_CURSOR(CURSOR[i]);

         if (e == button_press_event)
         { resize_win(win,x,y,i);
           x_set_color(win->id,win->bg_clr);
           x_box(win->id,0,0,*x,*y);

           *val1 = *x;
           *val2 = *y;
           *x = 0;
           *y = 0;
/*
           *x = win->xpos;
           *y = win->ypos;
           *val1 = win->width;
           *val2 = win->height;
*/
           return exposure_event;
          }

          return no_event;
        }
    }

   
    // check window buttons

    int but = 0;

    if (win->y0+BUTTON_Y-10 <= mouse_y && 
        win->y0+BUTTON_Y+BUTTON_W+10 >= mouse_y)
    { for(int i=0; i<3; i++) {
         if (win->x1-BUTTON_X[i]-10 <= mouse_x && 
             win->x1-BUTTON_X[i]+BUTTON_W+10 >= mouse_x) but = i+1;
      }
    }

    if (but != active_win_button) {
       active_win_button = but;
       for(int i=0; i<3; i++) draw_window_button(win,i);
     }

    if (e != button_press_event) {
      SET_MOUSE_CURSOR(-1);
      return no_event;
    }

    switch (but) {

      case 1: // iconize
              minimize(win);
              *x = 0;
              *y = 0;
              *val1 = win->width;
              *val2 = win->height;
              //return (win->minimized) ? no_event : configure_event;
              return no_event;

      case 2: { // maximize
                int xoff  = win->xpos - win->x0;
                int yoff  = win->ypos - win->y0;

                win->x0 = 0; 
                win->y0 = 0; 
                win->x1 = display_width;
                win->y1 = display_height;

                win->xpos = win->x0 + xoff;
                win->ypos = win->y0 + yoff;
                win->width  = win->x1 - win->x0 - 2*xoff + 1; 
                win->height = win->y1 - win->y0 - xoff - yoff + 1; 

                redraw_root();

                *x = 0;
                *y = 0;
                *val1 = win->width;
                *val2 = win->height;
                return exposure_event;
               }

      case 3: // close
              return destroy_event;
    }


    // button pressed on frame

    if (*val1 == 1)
    { // left button: raise and move window
      x_set_topmost(*w);
      move_win(win,*x,*y); 
      return configure_event; 
     }

    if (*val1 == 3)
    { // right button: write pixels
      char tmp[512];
      strcpy(tmp,win->header);
      x_set_label(win->id,"Saving Image to File (pixels.bgra)");
      int w = x1-x0+1;
      int h = y1-y0+1;
      string fname = string("pixels-%d-%d.bgra",w,h);
      root_win->canvas->write_bin(fname,x0,y0,x1,y1);
      x_set_label(win->id,tmp);
      return no_event;
     }


  }  // pointer on frame
  else
  { // pointer in client area
    if (active_win != *w)
    { active_win = *w;
      gg_window* wp = win_list[*w];
      if (wp) SET_MOUSE_CURSOR(wp->cursor);
     }
  }

  return e;
}


int x_get_next_event(Window& win, int& x, int& y, int& val1, int& val2,
                                                             unsigned long &t,
                                                             int msec)
{ // get next event 
  return  manage_next_event(&win,&x,&y,&val1,&val2,&t,msec);
}




void x_put_back_event()
{ event_buf.ev = last_event.ev;  
  event_buf.val = last_event.val; 
  event_buf.x = last_event.x; 
  event_buf.y = last_event.y; 
}


int x_open_display()
{ 
  if (do_not_open_display) return -1;

  if (root_win) return display_fd; // called before

  display_fd = 
   INIT_GRAPHICS(1,display_width,display_height,display_depth,display_dpi);

  if (display_fd < 0)
  { cout << "x_open_display failed" << endl;
    exit(1);
    return -1;
   }

  char* p = getenv("LEDA_DPI");
  if (p != 0) display_dpi = atoi(p);

/*
  HEADER_W = display_dpi/4;

  BORDER_W = HEADER_W/4;
  RESIZE_W = HEADER_W/3;
  BUTTON_W = HEADER_W/3;

  BUTTON_Y    = int(0.36 * HEADER_W);
  BUTTON_X[0] = int(3.7*HEADER_W);
  BUTTON_X[1] = int(2.5*HEADER_W);
  BUTTON_X[2] = int(1.3*HEADER_W);
*/

  HEADER_W = int(display_dpi/3.7);
  BORDER_W = int(HEADER_W/4.5);

  RESIZE_W = HEADER_W/3;
  BUTTON_W = HEADER_W/3;

  BUTTON_Y    = int(0.36 * HEADER_W);
  BUTTON_X[0] = int(3.7*HEADER_W);
  BUTTON_X[1] = int(2.5*HEADER_W);
  BUTTON_X[2] = int(1.3*HEADER_W);


/*
  cout << "display_width  = " << display_width << endl;
  cout << "display_height = " << display_height << endl;
  cout << "display_dpi    = " << display_dpi << endl;
*/

  mouse_x = 0;
  mouse_y = 0;

  grab_win = 0;

  // create root window (win_list[0])

  root_win = new gg_window;
  root_win->id = 0;
  root_win->flush = true;
  win_list[0] = root_win;
  win_stack[0] = root_win;
  win_count = 0;
  app_count = 0;
  win_top = 0;

  root_win->inf = 0;
  root_win->xpos = 0;
  root_win->ypos = 0;
  root_win->width  = display_width;
  root_win->height = display_height;

  root_win->x0 = 0;
  root_win->y0 = 0;
  root_win->x1 = root_win->width-1;
  root_win->y1 = root_win->height-1;

  root_win->bg_clr = ROOT_COLOR;

  root_win->border_clr = black;
  root_win->border_w = 1;
  root_win->cursor = -1;

  strcpy(root_win->header,"Root Window");

  root_win->COLOR = black;
  root_win->LINEWIDTH = 1;
  root_win->LINESTYLE = solid;
  root_win->MODE = src_mode;
  root_win->TEXTMODE = transparent;

/*
  strcpy(root_win->FONT_NAME,"T32");
  strcpy(root_win->FONT_NAME_TMP,"T32");
*/
  strcpy(root_win->FONT_NAME,"T38");
  strcpy(root_win->FONT_NAME_TMP,"T38");

  root_win->bg_pixmap = 0;

  root_win->icon_pixmap = x_create_pixrect_from_xpm(0,leda_icon);

  root_win->canvas = new x_image(display_width,display_height);
  root_win->canvas_save = 0;

  root_win->minimized = false;

  if (win_top==0) // no windows (other than root)
     x_clear_window(0,0,0,root_win->x1, root_win->y1,0,0);

  return display_fd;
}



void x_close_display()
{ delete (x_image*)root_win->icon_pixmap;
  delete root_win;
  root_win = 0;
  INIT_GRAPHICS(0,display_width,display_height,display_depth,display_dpi);
 }


int   x_window_width(Window win)  { return win_list[win]->width; }
int   x_window_height(Window win) { return win_list[win]->height; }
void* x_window_inf(Window win)    { return win_list[win]->inf; }
void* x_window_handle(Window w) { return win_list[w]; }


void x_window_position(Window win,int* x,int* y) 
{ *x = win_list[win]->x0; 
  *y = win_list[win]->y0; 
 }


Window x_create_window(void* inf, int width,int height,int bg,
                     const char* header, const char* label, int wp,
                     void (*func)(void*,int,int,int,int,int))
{ 
/*
  if (width > display_width) width = display_width;
  if (height > display_height) height = display_height;
*/

  gg_window* win = new gg_window;


  bool app_window = false;
  if (width > 0 && height > 0 && strlen(header) > 0) {
    app_window = true;
    app_count++;
  }

/*
  win_list[++win_count] = win;
  win->id = win_count;
*/

  // find first free index >= 1 in win_list

  int index = 1;
  while (index <= win_count && win_list[index] != 0) index++;
  if (index == win_count+1) win_count = index;

  win_list[index] = win;
  win->id = index;

  win->pwin = 0;

  strcpy(win->header,header);
  strcpy(win->label,label);

/*
  strcpy(win->FONT_NAME,"T32");
  strcpy(win->FONT_NAME_TMP,"T32");
*/
  strcpy(win->FONT_NAME,"T35");
  strcpy(win->FONT_NAME_TMP,"T35");

  win->bg_clr = bg;
  win->border_clr = black;
  win->border_w = 1;
  win->cursor = -1;

  win->label_clr = LABEL_COLOR;

  win->width = width;
  win->height = height;

  win->canvas = 0;
  win->canvas_save = 0;
  win->buffer = 0;

  if (width > 0) width  += 2*BORDER_W;
  if (height > 0) height += HEADER_W+BORDER_W;

  win->x0 = 0;
  win->y0 = 0;
  win->x1 = width-1;
  win->y1 = height-1;

  win->xpos = BORDER_W;
  win->ypos = HEADER_W;

  win->minimized = 0;

  int icon_w, icon_h;
  x_get_pixrect_size(root_win->icon_pixmap,icon_w,icon_h);

  win->save_x0 = BORDER_W;
  win->save_y0 = BORDER_W + (icon_h + 7*BORDER_W)*(app_count-1);

/*
if (app_window) {
cout << app_count << "  " << win->header << endl;
cout << "width = " << width << "  height = " << height <<endl;
cout << "icon: " << icon_w << " x " << icon_w << " ypos = " << win->save_y0 << endl;
}
*/

  win->save_x1 = win->save_x0 + icon_w-1;
  win->save_y1 = win->save_y0 + icon_h-1;
  win->save_xpos = win->save_x0;
  win->save_ypos = win->save_y0;
  win->save_width = icon_w;
  win->save_height = icon_h;

  win->redraw = func;
  win->inf = inf;

  win->COLOR = black;
  win->LINEWIDTH = 1;
  win->LINESTYLE = solid;
  win->MODE = src_mode;
  win->TEXTMODE = transparent;

  win->save_ls = solid;
  win->save_lw = 1;
  win->save_clr = black;
  win->save_mode = src_mode;

  win->icon_pixmap = root_win->icon_pixmap;
  win->bg_pixmap = 0;

  win->mapped = 0;

  win->flush = true;

  return win->id;
}


void x_open_window(int w, int x, int y, int width, int height, int pw, bool hidden)
{
  gg_window* win = win_list[w];

  if (width > display_width) width = display_width;
  if (height > display_height) height = display_height;

  if (win->mapped) return;

  if (width < 0)
  { width = -width;
    x -= (width-1);
    if (pw == 0) x -= 2*BORDER_W;
   }

  if (height < 0)
  { height = -height;
    y -= (height-1);
    if (pw == 0) y -= (BORDER_W+HEADER_W);
   }

  int bw = BORDER_W;
  int hw = (pw==0) ? HEADER_W : BORDER_W;


  if (pw)
  { bw = win->border_w;
    hw = win->border_w;
   }

  // push window on win_stack

  win_top++;

  if (pw <= 0)
  { // app window (not child): push on top
    win_stack[win_top] = win;
    win->mapped = win_top;
   }
  else
  { // child window: insert above pw and its children
    // find first position k above pw and its children
    gg_window* p_win = win_list[pw];
    int k = p_win->mapped + 1;
    while (k < win_top && win_stack[k]->pwin == p_win) k++;

    for(int i = win_top; i > k ; i--) 
    { win_stack[i] = win_stack[i-1];
      win_stack[i]->mapped = i;
     }

    win_stack[k] = win;
    win->mapped = k;
   }


  width  += 2*bw;
  height += hw+bw;

  if (pw > 0)
  { gg_window* p_win = win_list[pw];
    x += p_win->xpos;
    y += p_win->ypos;
    win->pwin = p_win;
   }

  win->x0 = x;
  win->y0 = y;
  win->x1 = x+width-1;
  win->y1 = y+height-1;
  win->xpos = x+bw;
  win->ypos = y+hw;
  win->width = width - 2*bw; 
  win->height = height - hw - bw;

  if (win->canvas != 0) delete win->canvas;

  win->canvas = new x_image(win->width,win->height);

  //draw_window(win);

  redraw_root();

/*
  // consume all pending events
  int e,val,px,py;
  unsigned long t;
  while ((e = CHECK_NEXT_EVENT(val,px,py,t,100)) != no_event) {
    cout << "SKIP EVENT: " << event_name[e] << endl;
  }
*/
}


int x_window_opened(Window w) 
{  return win_list[w] && win_list[w]->mapped; }


void x_close_window(Window w)
{ gg_window* win = win_list[w];

  if (grab_win == win) grab_win = 0;

  if (win->mapped == 0) return;

  //win->canvas->write_bin("pixels.bgra");

  int x0 = win->x0;
  int y0 = win->y0;
  int x1 = win->x1;
  int y1 = win->y1;

  for(int i=win->mapped; i < win_top; i++) 
  { gg_window* wx = win_stack[i+1];
    win_stack[i] = wx;
    wx->mapped = i;
   }

  win_top--;
  win->mapped = 0;

  redraw_root(x0,y0,x1,y1);

}


void x_destroy_window(Window w) 
{ gg_window* win = win_list[w];
  if (win->mapped) x_close_window(w);
  delete win; 
  win_list[w] = 0;
  if (w == win_count) 
  { while (win_list[w] == 0) w--;
    win_count = w;
  }
}

void x_set_icon_label(Window w, const char* s) 
{ gg_window* win = win_list[w];
  strcpy(win->label,s); 
  if (win->minimized) draw_window(win);
}

void  x_set_icon_pixmap(Window w, char* pm)
{ gg_window* win = win_list[w];
  win->icon_pixmap = pm;
  int icon_w, icon_h;
  x_get_pixrect_size(pm,icon_w,icon_h);
  if (!win->minimized)
  { win->save_x1 = win->save_x0 + icon_w-1;
    win->save_y1 = win->save_y0 + icon_h-1;
    win->save_width = icon_w;
    win->save_height = icon_h;
   }
  else
  { win->x1 = win->x0 + icon_w-1;
    win->y1 = win->y0 + icon_h-1;
    win->width = icon_w;
    win->height = icon_h;
    draw_window(win);
   }
}

void x_minimize_window(Window w)
{ gg_window* win = win_list[w];
  minimize(win);
}



void x_grab_pointer(Window w) { grab_win = (w > 0) ? win_list[w] : 0; }


void x_resize_window(Window wi, int x, int y, int w, int h, int p) 
{
  // x,y: position of upper left FRAME corner
  // w,h: width and height of CLIENT area

/*
  cout << string("resize:  wi = %d x = %d y = %d w = %d h = %d p = %d",
                                                         wi,x,y,w,h,p) << endl;  
*/
  x_delete_buffer(wi);

  gg_window* win = win_list[wi];

  gg_window* pw = win->pwin;

  if (pw) {
   x += pw->xpos;
   y += pw->ypos;
  }

  win->x0 = x;
  win->y0 = y;
  win->x1 = x + w;
  win->y1 = y + h;

  win->xpos = x;
  win->ypos = y;
  win->width = w;
  win->height = h;
  
  if (pw == 0)
  { win->xpos += BORDER_W;
    win->ypos += HEADER_W;
    win->x1 += 2*BORDER_W;
    win->y1 += HEADER_W + BORDER_W;
  }

  if (win->canvas != 0) 
  { x_image* new_canv = new x_image(win->width,win->height);
    FillArea(new_canv,0,0,win->width,win->height,win->bg_clr);
    delete win->canvas;
    win->canvas = new_canv;
   }

  redraw_root();

  if (win->redraw) win->redraw(win->inf,0,0,w,h,1);
}



void x_set_topmost(Window w)
{ gg_window* win = win_list[w];

  // copy window group (window and its children)
  gg_window* grp[win_max];
  int k = 0;
  int i = win->mapped;
  do { grp[k++] = win_stack[i++];
  } while (i <= win_top && win_stack[i]->pwin == win);

  // remove group
  win_top -= k;
  for(int i=win->mapped; i <= win_top; i++) {
     win_stack[i] = win_stack[i+k];
     win_stack[i]->mapped = i;
   }

  // copy back to top of stack
  for(int i=0; i<k; i++)
  { win_stack[++win_top] = grp[i];
    grp[i]->mapped = win_top;
  }

  redraw_root();
}


//------------------------------------------------------------------------------
// not implemented
//------------------------------------------------------------------------------

void x_flush_display() { }

void x_do_not_open_display(bool b) { do_not_open_display = b; }

void x_set_alpha(int, int) {}

void x_set_focus(Window) {}

void x_move_pointer(Window,int,int) {}

void x_set_stipple(Window,char*,int) {}

int  x_set_join_style(Window,int) { return 0; }

void x_fill_arc(Window,int,int,int,int,double,double) {}


//clipping
void x_clip_mask_chord(int, int, int, int, int, int, int, int, int) {}

//tooltips
void x_add_tooltip(int) {}
void x_set_tooltip(int, int, int, int, int, int, const char*) {}
void x_del_tooltip(int, int) {}


// ms-windows

void x_set_special_event_handler(int, 
                   void (*func)(void*,const char*,const char*,int,int), 
                   unsigned long data)  {}

void x_polyline(int w, int n, int *xcoord, int *ycoord, int*) {}

int x_window_minimized(int win) { return 0 ; }

int x_choose_file(int,int,const char*,const char*,const char*, char*,char*)  
{ return 0; }

void  x_text_to_clipboard(int, const char*) {}
char* x_text_from_clipboard(int) { return 0; }

void x_open_metafile(int, char const *) {}
void x_load_metafile(int, int, int, int, int, char const *) {}
void x_close_metafile(int) {}

void  x_metafile_to_clipboard(int) {}
void  x_pixrect_to_clipboard(int, char *) {}
char* x_pixrect_from_clipboard(int) { return 0; }

void x_start_timer(Window,int) {}
void x_stop_timer(Window) {}

int  x_choose_file(int,int,char const *,char const *,char *,char *) { return 0;}
int  x_choose_color(int, int) { return 0; }
void x_set_drop_handler(int, void (*func)(void*,const char*,int,int)) {}


// not defined
int x_get_resizing(int) { return 0; }


LEDA_END_NAMESPACE
